#!/usr/bin/env perl

#==============================================================================
#
#  Optimized Chacha routines for x86-64 by Andrew Moon
#
#  sse2: 64, 256 byte blocks
#  ssse3: 64, 256 byte blocks
#  avx: 64, 128, 256 byte blocks
#  xop: 64, 128, 256 byte blocks
#  avx2: 64, 128, 256, 512 byte blocks
#
#  License: Public Domain or MIT
#
#==============================================================================
#
# struct CHACHA_CTX {
#     unsigned char state[48] {
#         uint32_le key[8]
#         uint64_le counter
#         uint32_le iv[2]
#     }
#     size_t rounds;
# }
#
# chacha_blocks_ext(
#     CHACHA_CTX *ctx,
#     const unsigned char *in,
#     unsigned char *out,
#     size_t inl
# )
#
# If 'in' is NULL, bytestream is written directly to 'out'
#
# If 'inl' is not 0, then 'ctx->counter' will be treated as the lower and upper
# halves of a 64 bit integer and incremented by (inl + 63) / 64
#
#==============================================================================

$flavour = shift;
$output = shift;
if ($flavour =~ /\./) {
	$output = $flavour;
	undef $flavour;
}

$0 =~ m/(.*[\/\\])[^\/\\]+$/; 
$dir = $1;

( $xlate = "${dir}x86_64-xlate.pl" and -f $xlate ) or
( $xlate = "${dir}../../perlasm/x86_64-xlate.pl" and -f $xlate) or
die "can't locate x86_64-xlate.pl";

open OUT, "| \"$^X\" $xlate $flavour $output";

*STDOUT = *OUT;

# why isn't this already handled for implementors?
%ext = (
	x86 => 0,
	sse2 => 0,
	sse3 => 0,
	ssse3 => 0,
	sse4_1 => 0,
	sse4_2 => 0,
	avx => 0,
	xop => 0,
	avx2 => 0,
	avx512 => 0,
);

if (`$ENV{CC} -Wa,-v -c -o /dev/null -x assembler /dev/null 2>&1` =~ /GNU assembler version ([2-9]\.[0-9]+)/) {
	$ext{'x86'} = 1;
	$ext{'sse2'} = ($1 >= 2.16);
	$ext{'sse3'} = ($1 >= 2.17);
	$ext{'ssse3'} = ($1 >= 2.18);
	$ext{'sse4_1'} = ($1 >= 2.18);
	$ext{'sse4_2'} = ($1 >= 2.18);
	$ext{'avx'} = ($1 >= 2.19);
	$ext{'xop'} = ($1 >= 2.21);
	$ext{'avx2'} = ($1 >= 2.22);
	$ext{'avx512'} = ($1 >= 2.24);
}

if (`$ENV{CC} -v 2>&1` =~ /(^clang version|based on LLVM) ([3-9])\.([0-9]+)/) {
	my $ver = $2 + ($3 / 100.0); # 3.1->3.01, 3.10->3.10
	$ext{'x86'} = 1;
	$ext{'sse2'} = ($1 >= 2.01); # ?
	$ext{'sse3'} = ($1 >= 2.01); # ?
	$ext{'ssse3'} = ($1 >= 2.01);
	$ext{'sse4_1'} = ($1 >= 2.01); # ?
	$ext{'sse4_2'} = ($1 >= 2.01); # ?
	$ext{'avx'} = ($ver >= 3.0);
	$ext{'xop'} = ($ver >= 3.01);
	$ext{'avx2'} = ($ver >= 3.01);
	$ext{'avx512'} = ($ver >= 3.05); # ?
}

$win64 = ($flavour =~ /[nm]asm|mingw64/ || $output =~ /\.asm$/);

if ($win64 && ($flavour =~ /nasm/ || $ENV{ASM} =~ /nasm/) && `nasm -v 2>&1` =~ /NASM version ([2-9]\.[0-9]+)/) {
	$ext{'x86'} = 1;
	$ext{'sse2'} = ($1 >= 2.00);
	$ext{'sse3'} = ($1 >= 2.01);
	$ext{'ssse3'} = ($1 >= 2.01);
	$ext{'sse4_1'} = ($1 >= 2.01);
	$ext{'sss4_2'} = ($1 >= 2.01);
	$ext{'avx'} = ($1 >= 2.09);
	$ext{'xop'} = ($1 >= 2.09); # ?
	$ext{'avx2'} = ($1 >= 2.10);
	$ext{'avx512'} = ($1 >= 2.12); # ?
}

if ($win64 && ($flavour =~ /masm/ || $ENV{ASM} =~ /ml64/) && `ml64 2>&1` =~ /Version ([0-9]+)\./) {
	$ext{'x86'} = 1;
	$ext{'sse2'} = ($1 >= 7); # ?
	$ext{'sse3'} = ($1 >= 9); # ?
	$ext{'ssse3'} = ($1 >= 9);
	$ext{'sse4_1'} = ($1 >= 9);
	$ext{'sse4_2'} = ($1 >= 9);
	$ext{'avx'} = ($1 >= 10);
	$ext{'xop'} = ($1 >= 10);
	$ext{'avx2'} = ($1 >= 11);
	$ext{'avx512'} = ($1 >= 14); # ?
}

%last_valid_fallback = ();

sub add_function {
	my ($name, $args, $ext_xmm, $flag, $body) = @_;

	$fullname = "$name\_$flag";

	$code .=
<<___;
.globl $fullname
.type $fullname, \@abi-omnipotent
.align 16
$fullname:
___

	if ($ext{$flag}) {
		if ($win64) {
			$stack_space = 48;
			$rdi_rsi = 32;
			if ($ext_xmm) {
				$stack_space += 8 * 16;
				$rdi_rsi += 8 * 16;
			}

			$code .= 
<<___;
subq \$$stack_space, %rsp
movdqa %xmm6, 0(%rsp)
movdqa %xmm7, 16(%rsp)
___

			$code .= 
<<___ if ($ext_xmm);
movdqa %xmm8, 32(%rsp)
movdqa %xmm9, 48(%rsp)
movdqa %xmm10, 64(%rsp)
movdqa %xmm11, 80(%rsp)
movdqa %xmm12, 96(%rsp)
movdqa %xmm13, 112(%rsp)
movdqa %xmm14, 128(%rsp)
movdqa %xmm15, 144(%rsp)
___

			$code .=
<<___;
movq %rdi, $rdi_rsi(%rsp)
movq %rsi, $rdi_rsi + 8(%rsp)
movq %rcx, %rdi
movq %rdx, %rsi
movq %r8, %rdx
movq %r9, %rcx
___

			$code .= 'movq '.($stack_space + 40).'(%rsp), %r8\n' if ($args >= 5);
			$code .= 'movq '.($stack_space + 48).'(%rsp), %r9\n' if ($args >= 6);

			$code .= "call $fullname_thunk\n";

			$code = 
<<___;
movdqa 0(%rsp), %xmm6
movdqa 16(%rsp), %xmm7
___

			$code .=
<<___ if ($ext_xmm);
movdqa 32(%rsp), %xmm8
movdqa 48(%rsp), %xmm9
movdqa 64(%rsp), %xmm10
movdqa 80(%rsp), %xmm11
movdqa 96(%rsp), %xmm12
movdqa 112(%rsp), %xmm13
movdqa 128(%rsp), %xmm14
movdqa 144(%rsp), %xmm15
___

			$code =
<<___;
movq $rdi_rsi(%rsp), %rdi
movq $rdi_rsi + 8(%rsp), %rsi
addq \$$stack_space, %rsp
ret
___
		}
		$code .= "$fullname\_local:\n";
		$code .= $body;
		$last_valid_fallback{$name} = "$fullname\_local" if ($ext{$flag});
	} else {
		$code .= "jmp $last_valid_fallback{$name}\n";
	}
	$code .= ".size $fullname, .-$fullname\n\n";
}

#
#add_function(
#	'function', 4, false, 'ext',
#<<___
#___
#);
#


if ($ext{'x86'}) {

$code .=
<<___;
.text
.align 16
chacha_constants:
.long 0x61707865,0x3320646e,0x79622d32,0x6b206574
.byte 2,3,0,1,6,7,4,5,10,11,8,9,14,15,12,13
.byte 3,0,1,2,7,4,5,6,11,8,9,10,15,12,13,14

___

# add an 'always exists' version first, and add functions in order
# of most common extension to least so the fallback order is correct
add_function(
	'chacha_blocks', 4, true, 'sse2',
<<___
pushq %rbx
pushq %rbp
movq %rsp, %rbp
andq \$~63, %rsp
subq \$512, %rsp
movq \$0x3320646e61707865, %rax
movq \$0x6b20657479622d32, %r8
movd %rax, %xmm8
movd %r8, %xmm14
punpcklqdq %xmm14, %xmm8
movdqu 0(%rdi), %xmm9
movdqu 16(%rdi), %xmm10
movdqu 32(%rdi), %xmm11
movq 48(%rdi), %rax
movq \$1, %r9
movdqa %xmm8, 0(%rsp)
movdqa %xmm9, 16(%rsp)
movdqa %xmm10, 32(%rsp)
movdqa %xmm11, 48(%rsp)
movq %rax, 64(%rsp)
cmpq \$256, %rcx
jb chacha_blocks_sse2_below256
pshufd \$0x00, %xmm8, %xmm0
pshufd \$0x55, %xmm8, %xmm1
pshufd \$0xaa, %xmm8, %xmm2
pshufd \$0xff, %xmm8, %xmm3
movdqa %xmm0, 128(%rsp)
movdqa %xmm1, 144(%rsp)
movdqa %xmm2, 160(%rsp)
movdqa %xmm3, 176(%rsp)
pshufd \$0x00, %xmm9, %xmm0
pshufd \$0x55, %xmm9, %xmm1
pshufd \$0xaa, %xmm9, %xmm2
pshufd \$0xff, %xmm9, %xmm3
movdqa %xmm0, 192(%rsp)
movdqa %xmm1, 208(%rsp)
movdqa %xmm2, 224(%rsp)
movdqa %xmm3, 240(%rsp)
pshufd \$0x00, %xmm10, %xmm0
pshufd \$0x55, %xmm10, %xmm1
pshufd \$0xaa, %xmm10, %xmm2
pshufd \$0xff, %xmm10, %xmm3
movdqa %xmm0, 256(%rsp)
movdqa %xmm1, 272(%rsp)
movdqa %xmm2, 288(%rsp)
movdqa %xmm3, 304(%rsp)
pshufd \$0xaa, %xmm11, %xmm0
pshufd \$0xff, %xmm11, %xmm1
movdqa %xmm0, 352(%rsp)
movdqa %xmm1, 368(%rsp)
jmp chacha_blocks_sse2_atleast256
.align 64
chacha_blocks_sse2_atleast256:
movq 48(%rsp), %rax
leaq 1(%rax), %r8
leaq 2(%rax), %r9
leaq 3(%rax), %r10
leaq 4(%rax), %rbx
movl %eax, 320(%rsp)
movl %r8d, 4+320(%rsp)
movl %r9d, 8+320(%rsp)
movl %r10d, 12+320(%rsp)
shrq \$32, %rax
shrq \$32, %r8
shrq \$32, %r9
shrq \$32, %r10
movl %eax, 336(%rsp)
movl %r8d, 4+336(%rsp)
movl %r9d, 8+336(%rsp)
movl %r10d, 12+336(%rsp)
movq %rbx, 48(%rsp)
movq 64(%rsp), %rax
movdqa 128(%rsp), %xmm0
movdqa 144(%rsp), %xmm1
movdqa 160(%rsp), %xmm2
movdqa 176(%rsp), %xmm3
movdqa 192(%rsp), %xmm4
movdqa 208(%rsp), %xmm5
movdqa 224(%rsp), %xmm6
movdqa 240(%rsp), %xmm7
movdqa 256(%rsp), %xmm8
movdqa 272(%rsp), %xmm9
movdqa 288(%rsp), %xmm10
movdqa 304(%rsp), %xmm11
movdqa 320(%rsp), %xmm12
movdqa 336(%rsp), %xmm13
movdqa 352(%rsp), %xmm14
movdqa 368(%rsp), %xmm15
chacha_blocks_sse2_mainloop1:
paddd %xmm4, %xmm0
paddd %xmm5, %xmm1
pxor %xmm0, %xmm12
pxor %xmm1, %xmm13
paddd %xmm6, %xmm2
paddd %xmm7, %xmm3
movdqa %xmm6, 96(%rsp)
pxor %xmm2, %xmm14
pxor %xmm3, %xmm15
pshuflw \$0xb1,%xmm12,%xmm12
pshufhw \$0xb1,%xmm12,%xmm12
pshuflw \$0xb1,%xmm13,%xmm13
pshufhw \$0xb1,%xmm13,%xmm13
pshuflw \$0xb1,%xmm14,%xmm14
pshufhw \$0xb1,%xmm14,%xmm14
pshuflw \$0xb1,%xmm15,%xmm15
pshufhw \$0xb1,%xmm15,%xmm15
paddd %xmm12, %xmm8
paddd %xmm13, %xmm9
paddd %xmm14, %xmm10
paddd %xmm15, %xmm11
movdqa %xmm12, 112(%rsp)
pxor %xmm8, %xmm4
pxor %xmm9, %xmm5
movdqa 96(%rsp), %xmm6
movdqa %xmm4, %xmm12
pslld \$ 12, %xmm4
psrld \$20, %xmm12
pxor %xmm12, %xmm4
movdqa %xmm5, %xmm12
pslld \$ 12, %xmm5
psrld \$20, %xmm12
pxor %xmm12, %xmm5
pxor %xmm10, %xmm6
pxor %xmm11, %xmm7
movdqa %xmm6, %xmm12
pslld \$ 12, %xmm6
psrld \$20, %xmm12
pxor %xmm12, %xmm6
movdqa %xmm7, %xmm12
pslld \$ 12, %xmm7
psrld \$20, %xmm12
pxor %xmm12, %xmm7
movdqa 112(%rsp), %xmm12
paddd %xmm4, %xmm0
paddd %xmm5, %xmm1
pxor %xmm0, %xmm12
pxor %xmm1, %xmm13
paddd %xmm6, %xmm2
paddd %xmm7, %xmm3
movdqa %xmm6, 96(%rsp)
pxor %xmm2, %xmm14
pxor %xmm3, %xmm15
movdqa %xmm12, %xmm6
pslld \$ 8, %xmm12
psrld \$24, %xmm6
pxor %xmm6, %xmm12
movdqa %xmm13, %xmm6
pslld \$ 8, %xmm13
psrld \$24, %xmm6
pxor %xmm6, %xmm13
paddd %xmm12, %xmm8
paddd %xmm13, %xmm9
movdqa %xmm14, %xmm6
pslld \$ 8, %xmm14
psrld \$24, %xmm6
pxor %xmm6, %xmm14
movdqa %xmm15, %xmm6
pslld \$ 8, %xmm15
psrld \$24, %xmm6
pxor %xmm6, %xmm15
paddd %xmm14, %xmm10
paddd %xmm15, %xmm11
movdqa %xmm12, 112(%rsp)
pxor %xmm8, %xmm4
pxor %xmm9, %xmm5
movdqa 96(%rsp), %xmm6
movdqa %xmm4, %xmm12
pslld \$ 7, %xmm4
psrld \$25, %xmm12
pxor %xmm12, %xmm4
movdqa %xmm5, %xmm12
pslld \$ 7, %xmm5
psrld \$25, %xmm12
pxor %xmm12, %xmm5
pxor %xmm10, %xmm6
pxor %xmm11, %xmm7
movdqa %xmm6, %xmm12
pslld \$ 7, %xmm6
psrld \$25, %xmm12
pxor %xmm12, %xmm6
movdqa %xmm7, %xmm12
pslld \$ 7, %xmm7
psrld \$25, %xmm12
pxor %xmm12, %xmm7
movdqa 112(%rsp), %xmm12
paddd %xmm5, %xmm0
paddd %xmm6, %xmm1
pxor %xmm0, %xmm15
pxor %xmm1, %xmm12
paddd %xmm7, %xmm2
paddd %xmm4, %xmm3
movdqa %xmm7, 96(%rsp)
pxor %xmm2, %xmm13
pxor %xmm3, %xmm14
pshuflw \$0xb1,%xmm15,%xmm15
pshufhw \$0xb1,%xmm15,%xmm15
pshuflw \$0xb1,%xmm12,%xmm12
pshufhw \$0xb1,%xmm12,%xmm12
pshuflw \$0xb1,%xmm13,%xmm13
pshufhw \$0xb1,%xmm13,%xmm13
pshuflw \$0xb1,%xmm14,%xmm14
pshufhw \$0xb1,%xmm14,%xmm14
paddd %xmm15, %xmm10
paddd %xmm12, %xmm11
paddd %xmm13, %xmm8
paddd %xmm14, %xmm9
movdqa %xmm15, 112(%rsp)
pxor %xmm10, %xmm5
pxor %xmm11, %xmm6
movdqa 96(%rsp), %xmm7
movdqa %xmm5, %xmm15
pslld \$ 12, %xmm5
psrld \$20, %xmm15
pxor %xmm15, %xmm5
movdqa %xmm6, %xmm15
pslld \$ 12, %xmm6
psrld \$20, %xmm15
pxor %xmm15, %xmm6
pxor %xmm8, %xmm7
pxor %xmm9, %xmm4
movdqa %xmm7, %xmm15
pslld \$ 12, %xmm7
psrld \$20, %xmm15
pxor %xmm15, %xmm7
movdqa %xmm4, %xmm15
pslld \$ 12, %xmm4
psrld \$20, %xmm15
pxor %xmm15, %xmm4
movdqa 112(%rsp), %xmm15
paddd %xmm5, %xmm0
paddd %xmm6, %xmm1
pxor %xmm0, %xmm15
pxor %xmm1, %xmm12
paddd %xmm7, %xmm2
paddd %xmm4, %xmm3
movdqa %xmm7, 96(%rsp)
pxor %xmm2, %xmm13
pxor %xmm3, %xmm14
movdqa %xmm15, %xmm7
pslld \$ 8, %xmm15
psrld \$24, %xmm7
pxor %xmm7, %xmm15
movdqa %xmm12, %xmm7
pslld \$ 8, %xmm12
psrld \$24, %xmm7
pxor %xmm7, %xmm12
paddd %xmm15, %xmm10
paddd %xmm12, %xmm11
movdqa %xmm13, %xmm7
pslld \$ 8, %xmm13
psrld \$24, %xmm7
pxor %xmm7, %xmm13
movdqa %xmm14, %xmm7
pslld \$ 8, %xmm14
psrld \$24, %xmm7
pxor %xmm7, %xmm14
paddd %xmm13, %xmm8
paddd %xmm14, %xmm9
movdqa %xmm15, 112(%rsp)
pxor %xmm10, %xmm5
pxor %xmm11, %xmm6
movdqa 96(%rsp), %xmm7
movdqa %xmm5, %xmm15
pslld \$ 7, %xmm5
psrld \$25, %xmm15
pxor %xmm15, %xmm5
movdqa %xmm6, %xmm15
pslld \$ 7, %xmm6
psrld \$25, %xmm15
pxor %xmm15, %xmm6
pxor %xmm8, %xmm7
pxor %xmm9, %xmm4
movdqa %xmm7, %xmm15
pslld \$ 7, %xmm7
psrld \$25, %xmm15
pxor %xmm15, %xmm7
movdqa %xmm4, %xmm15
pslld \$ 7, %xmm4
psrld \$25, %xmm15
pxor %xmm15, %xmm4
movdqa 112(%rsp), %xmm15
subq \$2, %rax
jnz chacha_blocks_sse2_mainloop1
paddd 128(%rsp), %xmm0
paddd 144(%rsp), %xmm1
paddd 160(%rsp), %xmm2
paddd 176(%rsp), %xmm3
paddd 192(%rsp), %xmm4
paddd 208(%rsp), %xmm5
paddd 224(%rsp), %xmm6
paddd 240(%rsp), %xmm7
paddd 256(%rsp), %xmm8
paddd 272(%rsp), %xmm9
paddd 288(%rsp), %xmm10
paddd 304(%rsp), %xmm11
paddd 320(%rsp), %xmm12
paddd 336(%rsp), %xmm13
paddd 352(%rsp), %xmm14
paddd 368(%rsp), %xmm15
movdqa %xmm8, 384(%rsp)
movdqa %xmm9, 400(%rsp)
movdqa %xmm10, 416(%rsp)
movdqa %xmm11, 432(%rsp)
movdqa %xmm12, 448(%rsp)
movdqa %xmm13, 464(%rsp)
movdqa %xmm14, 480(%rsp)
movdqa %xmm15, 496(%rsp)
movdqa %xmm0, %xmm8
movdqa %xmm2, %xmm9
movdqa %xmm4, %xmm10
movdqa %xmm6, %xmm11
punpckhdq %xmm1, %xmm0
punpckhdq %xmm3, %xmm2
punpckhdq %xmm5, %xmm4
punpckhdq %xmm7, %xmm6
punpckldq %xmm1, %xmm8
punpckldq %xmm3, %xmm9
punpckldq %xmm5, %xmm10
punpckldq %xmm7, %xmm11
movdqa %xmm0, %xmm1
movdqa %xmm4, %xmm3
movdqa %xmm8, %xmm5
movdqa %xmm10, %xmm7
punpckhqdq %xmm2, %xmm0
punpckhqdq %xmm6, %xmm4
punpckhqdq %xmm9, %xmm8
punpckhqdq %xmm11, %xmm10
punpcklqdq %xmm2, %xmm1
punpcklqdq %xmm6, %xmm3
punpcklqdq %xmm9, %xmm5
punpcklqdq %xmm11, %xmm7
andq %rsi, %rsi
jz chacha_blocks_sse2_noinput1
movdqu 0(%rsi), %xmm2
movdqu 16(%rsi), %xmm6
movdqu 64(%rsi), %xmm9
movdqu 80(%rsi), %xmm11
movdqu 128(%rsi), %xmm12
movdqu 144(%rsi), %xmm13
movdqu 192(%rsi), %xmm14
movdqu 208(%rsi), %xmm15
pxor %xmm2, %xmm5
pxor %xmm6, %xmm7
pxor %xmm9, %xmm8
pxor %xmm11, %xmm10
pxor %xmm12, %xmm1
pxor %xmm13, %xmm3
pxor %xmm14, %xmm0
pxor %xmm15, %xmm4
movdqu %xmm5, 0(%rdx)
movdqu %xmm7, 16(%rdx)
movdqu %xmm8, 64(%rdx)
movdqu %xmm10, 80(%rdx)
movdqu %xmm1, 128(%rdx)
movdqu %xmm3, 144(%rdx)
movdqu %xmm0, 192(%rdx)
movdqu %xmm4, 208(%rdx)
movdqa 384(%rsp), %xmm0
movdqa 400(%rsp), %xmm1
movdqa 416(%rsp), %xmm2
movdqa 432(%rsp), %xmm3
movdqa 448(%rsp), %xmm4
movdqa 464(%rsp), %xmm5
movdqa 480(%rsp), %xmm6
movdqa 496(%rsp), %xmm7
movdqa %xmm0, %xmm8
movdqa %xmm2, %xmm9
movdqa %xmm4, %xmm10
movdqa %xmm6, %xmm11
punpckldq %xmm1, %xmm8
punpckldq %xmm3, %xmm9
punpckhdq %xmm1, %xmm0
punpckhdq %xmm3, %xmm2
punpckldq %xmm5, %xmm10
punpckldq %xmm7, %xmm11
punpckhdq %xmm5, %xmm4
punpckhdq %xmm7, %xmm6
movdqa %xmm8, %xmm1
movdqa %xmm0, %xmm3
movdqa %xmm10, %xmm5
movdqa %xmm4, %xmm7
punpcklqdq %xmm9, %xmm1
punpcklqdq %xmm11, %xmm5
punpckhqdq %xmm9, %xmm8
punpckhqdq %xmm11, %xmm10
punpcklqdq %xmm2, %xmm3
punpcklqdq %xmm6, %xmm7
punpckhqdq %xmm2, %xmm0
punpckhqdq %xmm6, %xmm4
movdqu 32(%rsi), %xmm2
movdqu 48(%rsi), %xmm6
movdqu 96(%rsi), %xmm9
movdqu 112(%rsi), %xmm11
movdqu 160(%rsi), %xmm12
movdqu 176(%rsi), %xmm13
movdqu 224(%rsi), %xmm14
movdqu 240(%rsi), %xmm15
pxor %xmm2, %xmm1
pxor %xmm6, %xmm5
pxor %xmm9, %xmm8
pxor %xmm11, %xmm10
pxor %xmm12, %xmm3
pxor %xmm13, %xmm7
pxor %xmm14, %xmm0
pxor %xmm15, %xmm4
movdqu %xmm1, 32(%rdx)
movdqu %xmm5, 48(%rdx)
movdqu %xmm8, 96(%rdx)
movdqu %xmm10, 112(%rdx)
movdqu %xmm3, 160(%rdx)
movdqu %xmm7, 176(%rdx)
movdqu %xmm0, 224(%rdx)
movdqu %xmm4, 240(%rdx)
addq \$256, %rsi
jmp chacha_blocks_sse2_mainloop_cont
chacha_blocks_sse2_noinput1:
movdqu %xmm5, 0(%rdx)
movdqu %xmm7, 16(%rdx)
movdqu %xmm8, 64(%rdx)
movdqu %xmm10, 80(%rdx)
movdqu %xmm1, 128(%rdx)
movdqu %xmm3, 144(%rdx)
movdqu %xmm0, 192(%rdx)
movdqu %xmm4, 208(%rdx)
movdqa 384(%rsp), %xmm0
movdqa 400(%rsp), %xmm1
movdqa 416(%rsp), %xmm2
movdqa 432(%rsp), %xmm3
movdqa 448(%rsp), %xmm4
movdqa 464(%rsp), %xmm5
movdqa 480(%rsp), %xmm6
movdqa 496(%rsp), %xmm7
movdqa %xmm0, %xmm8
movdqa %xmm2, %xmm9
movdqa %xmm4, %xmm10
movdqa %xmm6, %xmm11
punpckldq %xmm1, %xmm8
punpckldq %xmm3, %xmm9
punpckhdq %xmm1, %xmm0
punpckhdq %xmm3, %xmm2
punpckldq %xmm5, %xmm10
punpckldq %xmm7, %xmm11
punpckhdq %xmm5, %xmm4
punpckhdq %xmm7, %xmm6
movdqa %xmm8, %xmm1
movdqa %xmm0, %xmm3
movdqa %xmm10, %xmm5
movdqa %xmm4, %xmm7
punpcklqdq %xmm9, %xmm1
punpcklqdq %xmm11, %xmm5
punpckhqdq %xmm9, %xmm8
punpckhqdq %xmm11, %xmm10
punpcklqdq %xmm2, %xmm3
punpcklqdq %xmm6, %xmm7
punpckhqdq %xmm2, %xmm0
punpckhqdq %xmm6, %xmm4
movdqu %xmm1, 32(%rdx)
movdqu %xmm5, 48(%rdx)
movdqu %xmm8, 96(%rdx)
movdqu %xmm10, 112(%rdx)
movdqu %xmm3, 160(%rdx)
movdqu %xmm7, 176(%rdx)
movdqu %xmm0, 224(%rdx)
movdqu %xmm4, 240(%rdx)
chacha_blocks_sse2_mainloop_cont:
addq \$256, %rdx
subq \$256, %rcx
cmp \$256, %rcx
jae chacha_blocks_sse2_atleast256
movdqa 0(%rsp), %xmm8
movdqa 16(%rsp), %xmm9
movdqa 32(%rsp), %xmm10
movdqa 48(%rsp), %xmm11
movq \$1, %r9
chacha_blocks_sse2_below256:
movq %r9, %xmm5
andq %rcx, %rcx
jz chacha_blocks_sse2_done
cmpq \$64, %rcx
jae chacha_blocks_sse2_above63
movq %rdx, %r9
andq %rsi, %rsi
jz chacha_blocks_sse2_noinput2
movq %rcx, %r10
movq %rsp, %rdx
addq %r10, %rsi
addq %r10, %rdx
negq %r10
chacha_blocks_sse2_copyinput:
movb (%rsi, %r10), %al
movb %al, (%rdx, %r10)
incq %r10
jnz chacha_blocks_sse2_copyinput
movq %rsp, %rsi
chacha_blocks_sse2_noinput2:
movq %rsp, %rdx
chacha_blocks_sse2_above63:
movdqa %xmm8, %xmm0
movdqa %xmm9, %xmm1
movdqa %xmm10, %xmm2
movdqa %xmm11, %xmm3
movq 64(%rsp), %rax
chacha_blocks_sse2_mainloop2:
paddd %xmm1, %xmm0
pxor %xmm0, %xmm3
pshuflw \$0xb1,%xmm3,%xmm3
pshufhw \$0xb1,%xmm3,%xmm3
paddd %xmm3, %xmm2
pxor %xmm2, %xmm1
movdqa %xmm1,%xmm4
pslld \$12, %xmm1
psrld \$20, %xmm4
pxor %xmm4, %xmm1
paddd %xmm1, %xmm0
pxor %xmm0, %xmm3
movdqa %xmm3,%xmm4
pslld \$8, %xmm3
psrld \$24, %xmm4
pshufd \$0x93,%xmm0,%xmm0
pxor %xmm4, %xmm3
paddd %xmm3, %xmm2
pshufd \$0x4e,%xmm3,%xmm3
pxor %xmm2, %xmm1
pshufd \$0x39,%xmm2,%xmm2
movdqa %xmm1,%xmm4
pslld \$7, %xmm1
psrld \$25, %xmm4
pxor %xmm4, %xmm1
subq \$2, %rax
paddd %xmm1, %xmm0
pxor %xmm0, %xmm3
pshuflw \$0xb1,%xmm3,%xmm3
pshufhw \$0xb1,%xmm3,%xmm3
paddd %xmm3, %xmm2
pxor %xmm2, %xmm1
movdqa %xmm1,%xmm4
pslld \$12, %xmm1
psrld \$20, %xmm4
pxor %xmm4, %xmm1
paddd %xmm1, %xmm0
pxor %xmm0, %xmm3
movdqa %xmm3,%xmm4
pslld \$8, %xmm3
psrld \$24, %xmm4
pshufd \$0x39,%xmm0,%xmm0
pxor %xmm4, %xmm3
paddd %xmm3, %xmm2
pshufd \$0x4e,%xmm3,%xmm3
pxor %xmm2, %xmm1
pshufd \$0x93,%xmm2,%xmm2
movdqa %xmm1,%xmm4
pslld \$7, %xmm1
psrld \$25, %xmm4
pxor %xmm4, %xmm1
jnz chacha_blocks_sse2_mainloop2
paddd %xmm8, %xmm0
paddd %xmm9, %xmm1
paddd %xmm10, %xmm2
paddd %xmm11, %xmm3
andq %rsi, %rsi
jz chacha_blocks_sse2_noinput3
movdqu 0(%rsi), %xmm12
movdqu 16(%rsi), %xmm13
movdqu 32(%rsi), %xmm14
movdqu 48(%rsi), %xmm15
pxor %xmm12, %xmm0
pxor %xmm13, %xmm1
pxor %xmm14, %xmm2
pxor %xmm15, %xmm3
addq \$64, %rsi
chacha_blocks_sse2_noinput3:
movdqu %xmm0, 0(%rdx)
movdqu %xmm1, 16(%rdx)
movdqu %xmm2, 32(%rdx)
movdqu %xmm3, 48(%rdx)
paddq %xmm5, %xmm11
cmpq \$64, %rcx
jbe chacha_blocks_sse2_mainloop2_finishup
addq \$64, %rdx
subq \$64, %rcx
jmp chacha_blocks_sse2_below256
chacha_blocks_sse2_mainloop2_finishup:
cmpq \$64, %rcx
je chacha_blocks_sse2_done
addq %rcx, %r9
addq %rcx, %rdx
negq %rcx
chacha_blocks_sse2_copyoutput:
movb (%rdx, %rcx), %al
movb %al, (%r9, %rcx)
incq %rcx
jnz chacha_blocks_sse2_copyoutput
chacha_blocks_sse2_done:
movdqu %xmm11, 32(%rdi)
movq %rbp, %rsp
popq %rbp
popq %rbx
ret
___
);

add_function(
	'chacha_blocks', 4, true, 'ssse3',
<<___
pushq %rbx
pushq %rbp
movq %rsp, %rbp
andq \$~63, %rsp
subq \$512, %rsp
leaq chacha_constants(%rip), %rax
movdqa 0(%rax), %xmm8
movdqa 16(%rax), %xmm6
movdqa 32(%rax), %xmm7
movdqu 0(%rdi), %xmm9
movdqu 16(%rdi), %xmm10
movdqu 32(%rdi), %xmm11
movq 48(%rdi), %rax
movq \$1, %r9
movdqa %xmm8, 0(%rsp)
movdqa %xmm9, 16(%rsp)
movdqa %xmm10, 32(%rsp)
movdqa %xmm11, 48(%rsp)
movdqa %xmm6, 80(%rsp)
movdqa %xmm7, 96(%rsp)
movq %rax, 64(%rsp)
cmpq \$256, %rcx
jb chacha_blocks_ssse3_below256
pshufd \$0x00, %xmm8, %xmm0
pshufd \$0x55, %xmm8, %xmm1
pshufd \$0xaa, %xmm8, %xmm2
pshufd \$0xff, %xmm8, %xmm3
movdqa %xmm0, 128(%rsp)
movdqa %xmm1, 144(%rsp)
movdqa %xmm2, 160(%rsp)
movdqa %xmm3, 176(%rsp)
pshufd \$0x00, %xmm9, %xmm0
pshufd \$0x55, %xmm9, %xmm1
pshufd \$0xaa, %xmm9, %xmm2
pshufd \$0xff, %xmm9, %xmm3
movdqa %xmm0, 192(%rsp)
movdqa %xmm1, 208(%rsp)
movdqa %xmm2, 224(%rsp)
movdqa %xmm3, 240(%rsp)
pshufd \$0x00, %xmm10, %xmm0
pshufd \$0x55, %xmm10, %xmm1
pshufd \$0xaa, %xmm10, %xmm2
pshufd \$0xff, %xmm10, %xmm3
movdqa %xmm0, 256(%rsp)
movdqa %xmm1, 272(%rsp)
movdqa %xmm2, 288(%rsp)
movdqa %xmm3, 304(%rsp)
pshufd \$0xaa, %xmm11, %xmm0
pshufd \$0xff, %xmm11, %xmm1
movdqa %xmm0, 352(%rsp)
movdqa %xmm1, 368(%rsp)
jmp chacha_blocks_ssse3_atleast256
.align 64
nop
nop
nop
nop
chacha_blocks_ssse3_atleast256:
movq 48(%rsp), %rax
leaq 1(%rax), %r8
leaq 2(%rax), %r9
leaq 3(%rax), %r10
leaq 4(%rax), %rbx
movl %eax, 320(%rsp)
movl %r8d, 4+320(%rsp)
movl %r9d, 8+320(%rsp)
movl %r10d, 12+320(%rsp)
shrq \$32, %rax
shrq \$32, %r8
shrq \$32, %r9
shrq \$32, %r10
movl %eax, 336(%rsp)
movl %r8d, 4+336(%rsp)
movl %r9d, 8+336(%rsp)
movl %r10d, 12+336(%rsp)
movq %rbx, 48(%rsp)
movq 64(%rsp), %rax
movdqa 128(%rsp), %xmm0
movdqa 144(%rsp), %xmm1
movdqa 160(%rsp), %xmm2
movdqa 176(%rsp), %xmm3
movdqa 192(%rsp), %xmm4
movdqa 208(%rsp), %xmm5
movdqa 224(%rsp), %xmm6
movdqa 240(%rsp), %xmm7
movdqa 256(%rsp), %xmm8
movdqa 272(%rsp), %xmm9
movdqa 288(%rsp), %xmm10
movdqa 304(%rsp), %xmm11
movdqa 320(%rsp), %xmm12
movdqa 336(%rsp), %xmm13
movdqa 352(%rsp), %xmm14
movdqa 368(%rsp), %xmm15
chacha_blocks_ssse3_mainloop1:
paddd %xmm4, %xmm0
paddd %xmm5, %xmm1
pxor %xmm0, %xmm12
pxor %xmm1, %xmm13
paddd %xmm6, %xmm2
paddd %xmm7, %xmm3
pxor %xmm2, %xmm14
pxor %xmm3, %xmm15
pshufb 80(%rsp), %xmm12
pshufb 80(%rsp), %xmm13
paddd %xmm12, %xmm8
paddd %xmm13, %xmm9
pshufb 80(%rsp), %xmm14
pshufb 80(%rsp), %xmm15
paddd %xmm14, %xmm10
paddd %xmm15, %xmm11
movdqa %xmm12, 112(%rsp)
pxor %xmm8, %xmm4
pxor %xmm9, %xmm5
movdqa %xmm4, %xmm12
pslld \$ 12, %xmm4
psrld \$20, %xmm12
pxor %xmm12, %xmm4
movdqa %xmm5, %xmm12
pslld \$ 12, %xmm5
psrld \$20, %xmm12
pxor %xmm12, %xmm5
pxor %xmm10, %xmm6
pxor %xmm11, %xmm7
movdqa %xmm6, %xmm12
pslld \$ 12, %xmm6
psrld \$20, %xmm12
pxor %xmm12, %xmm6
movdqa %xmm7, %xmm12
pslld \$ 12, %xmm7
psrld \$20, %xmm12
pxor %xmm12, %xmm7
movdqa 112(%rsp), %xmm12
paddd %xmm4, %xmm0
paddd %xmm5, %xmm1
pxor %xmm0, %xmm12
pxor %xmm1, %xmm13
paddd %xmm6, %xmm2
paddd %xmm7, %xmm3
pxor %xmm2, %xmm14
pxor %xmm3, %xmm15
pshufb 96(%rsp), %xmm12
pshufb 96(%rsp), %xmm13
paddd %xmm12, %xmm8
paddd %xmm13, %xmm9
pshufb 96(%rsp), %xmm14
pshufb 96(%rsp), %xmm15
paddd %xmm14, %xmm10
paddd %xmm15, %xmm11
movdqa %xmm12, 112(%rsp)
pxor %xmm8, %xmm4
pxor %xmm9, %xmm5
movdqa %xmm4, %xmm12
pslld \$ 7, %xmm4
psrld \$25, %xmm12
pxor %xmm12, %xmm4
movdqa %xmm5, %xmm12
pslld \$ 7, %xmm5
psrld \$25, %xmm12
pxor %xmm12, %xmm5
pxor %xmm10, %xmm6
pxor %xmm11, %xmm7
movdqa %xmm6, %xmm12
pslld \$ 7, %xmm6
psrld \$25, %xmm12
pxor %xmm12, %xmm6
movdqa %xmm7, %xmm12
pslld \$ 7, %xmm7
psrld \$25, %xmm12
pxor %xmm12, %xmm7
movdqa 112(%rsp), %xmm12
paddd %xmm5, %xmm0
paddd %xmm6, %xmm1
pxor %xmm0, %xmm15
pxor %xmm1, %xmm12
paddd %xmm7, %xmm2
paddd %xmm4, %xmm3
pxor %xmm2, %xmm13
pxor %xmm3, %xmm14
pshufb 80(%rsp), %xmm15
pshufb 80(%rsp), %xmm12
paddd %xmm15, %xmm10
paddd %xmm12, %xmm11
pshufb 80(%rsp), %xmm13
pshufb 80(%rsp), %xmm14
paddd %xmm13, %xmm8
paddd %xmm14, %xmm9
movdqa %xmm15, 112(%rsp)
pxor %xmm10, %xmm5
pxor %xmm11, %xmm6
movdqa %xmm5, %xmm15
pslld \$ 12, %xmm5
psrld \$20, %xmm15
pxor %xmm15, %xmm5
movdqa %xmm6, %xmm15
pslld \$ 12, %xmm6
psrld \$20, %xmm15
pxor %xmm15, %xmm6
pxor %xmm8, %xmm7
pxor %xmm9, %xmm4
movdqa %xmm7, %xmm15
pslld \$ 12, %xmm7
psrld \$20, %xmm15
pxor %xmm15, %xmm7
movdqa %xmm4, %xmm15
pslld \$ 12, %xmm4
psrld \$20, %xmm15
pxor %xmm15, %xmm4
movdqa 112(%rsp), %xmm15
paddd %xmm5, %xmm0
paddd %xmm6, %xmm1
pxor %xmm0, %xmm15
pxor %xmm1, %xmm12
paddd %xmm7, %xmm2
paddd %xmm4, %xmm3
pxor %xmm2, %xmm13
pxor %xmm3, %xmm14
pshufb 96(%rsp), %xmm15
pshufb 96(%rsp), %xmm12
paddd %xmm15, %xmm10
paddd %xmm12, %xmm11
pshufb 96(%rsp), %xmm13
pshufb 96(%rsp), %xmm14
paddd %xmm13, %xmm8
paddd %xmm14, %xmm9
movdqa %xmm15, 112(%rsp)
pxor %xmm10, %xmm5
pxor %xmm11, %xmm6
movdqa %xmm5, %xmm15
pslld \$ 7, %xmm5
psrld \$25, %xmm15
pxor %xmm15, %xmm5
movdqa %xmm6, %xmm15
pslld \$ 7, %xmm6
psrld \$25, %xmm15
pxor %xmm15, %xmm6
pxor %xmm8, %xmm7
pxor %xmm9, %xmm4
movdqa %xmm7, %xmm15
pslld \$ 7, %xmm7
psrld \$25, %xmm15
pxor %xmm15, %xmm7
movdqa %xmm4, %xmm15
pslld \$ 7, %xmm4
psrld \$25, %xmm15
pxor %xmm15, %xmm4
subq \$2, %rax
movdqa 112(%rsp), %xmm15
jnz chacha_blocks_ssse3_mainloop1
paddd 128(%rsp), %xmm0
paddd 144(%rsp), %xmm1
paddd 160(%rsp), %xmm2
paddd 176(%rsp), %xmm3
paddd 192(%rsp), %xmm4
paddd 208(%rsp), %xmm5
paddd 224(%rsp), %xmm6
paddd 240(%rsp), %xmm7
paddd 256(%rsp), %xmm8
paddd 272(%rsp), %xmm9
paddd 288(%rsp), %xmm10
paddd 304(%rsp), %xmm11
paddd 320(%rsp), %xmm12
paddd 336(%rsp), %xmm13
paddd 352(%rsp), %xmm14
paddd 368(%rsp), %xmm15
movdqa %xmm8, 384(%rsp)
movdqa %xmm9, 400(%rsp)
movdqa %xmm10, 416(%rsp)
movdqa %xmm11, 432(%rsp)
movdqa %xmm12, 448(%rsp)
movdqa %xmm13, 464(%rsp)
movdqa %xmm14, 480(%rsp)
movdqa %xmm15, 496(%rsp)
movdqa %xmm0, %xmm8
movdqa %xmm2, %xmm9
movdqa %xmm4, %xmm10
movdqa %xmm6, %xmm11
punpckhdq %xmm1, %xmm0
punpckhdq %xmm3, %xmm2
punpckhdq %xmm5, %xmm4
punpckhdq %xmm7, %xmm6
punpckldq %xmm1, %xmm8
punpckldq %xmm3, %xmm9
punpckldq %xmm5, %xmm10
punpckldq %xmm7, %xmm11
movdqa %xmm0, %xmm1
movdqa %xmm4, %xmm3
movdqa %xmm8, %xmm5
movdqa %xmm10, %xmm7
punpckhqdq %xmm2, %xmm0
punpckhqdq %xmm6, %xmm4
punpckhqdq %xmm9, %xmm8
punpckhqdq %xmm11, %xmm10
punpcklqdq %xmm2, %xmm1
punpcklqdq %xmm6, %xmm3
punpcklqdq %xmm9, %xmm5
punpcklqdq %xmm11, %xmm7
andq %rsi, %rsi
jz chacha_blocks_ssse3_noinput1
movdqu 0(%rsi), %xmm2
movdqu 16(%rsi), %xmm6
movdqu 64(%rsi), %xmm9
movdqu 80(%rsi), %xmm11
movdqu 128(%rsi), %xmm12
movdqu 144(%rsi), %xmm13
movdqu 192(%rsi), %xmm14
movdqu 208(%rsi), %xmm15
pxor %xmm2, %xmm5
pxor %xmm6, %xmm7
pxor %xmm9, %xmm8
pxor %xmm11, %xmm10
pxor %xmm12, %xmm1
pxor %xmm13, %xmm3
pxor %xmm14, %xmm0
pxor %xmm15, %xmm4
movdqu %xmm5, 0(%rdx)
movdqu %xmm7, 16(%rdx)
movdqu %xmm8, 64(%rdx)
movdqu %xmm10, 80(%rdx)
movdqu %xmm1, 128(%rdx)
movdqu %xmm3, 144(%rdx)
movdqu %xmm0, 192(%rdx)
movdqu %xmm4, 208(%rdx)
movdqa 384(%rsp), %xmm0
movdqa 400(%rsp), %xmm1
movdqa 416(%rsp), %xmm2
movdqa 432(%rsp), %xmm3
movdqa 448(%rsp), %xmm4
movdqa 464(%rsp), %xmm5
movdqa 480(%rsp), %xmm6
movdqa 496(%rsp), %xmm7
movdqa %xmm0, %xmm8
movdqa %xmm2, %xmm9
movdqa %xmm4, %xmm10
movdqa %xmm6, %xmm11
punpckldq %xmm1, %xmm8
punpckldq %xmm3, %xmm9
punpckhdq %xmm1, %xmm0
punpckhdq %xmm3, %xmm2
punpckldq %xmm5, %xmm10
punpckldq %xmm7, %xmm11
punpckhdq %xmm5, %xmm4
punpckhdq %xmm7, %xmm6
movdqa %xmm8, %xmm1
movdqa %xmm0, %xmm3
movdqa %xmm10, %xmm5
movdqa %xmm4, %xmm7
punpcklqdq %xmm9, %xmm1
punpcklqdq %xmm11, %xmm5
punpckhqdq %xmm9, %xmm8
punpckhqdq %xmm11, %xmm10
punpcklqdq %xmm2, %xmm3
punpcklqdq %xmm6, %xmm7
punpckhqdq %xmm2, %xmm0
punpckhqdq %xmm6, %xmm4
movdqu 32(%rsi), %xmm2
movdqu 48(%rsi), %xmm6
movdqu 96(%rsi), %xmm9
movdqu 112(%rsi), %xmm11
movdqu 160(%rsi), %xmm12
movdqu 176(%rsi), %xmm13
movdqu 224(%rsi), %xmm14
movdqu 240(%rsi), %xmm15
pxor %xmm2, %xmm1
pxor %xmm6, %xmm5
pxor %xmm9, %xmm8
pxor %xmm11, %xmm10
pxor %xmm12, %xmm3
pxor %xmm13, %xmm7
pxor %xmm14, %xmm0
pxor %xmm15, %xmm4
movdqu %xmm1, 32(%rdx)
movdqu %xmm5, 48(%rdx)
movdqu %xmm8, 96(%rdx)
movdqu %xmm10, 112(%rdx)
movdqu %xmm3, 160(%rdx)
movdqu %xmm7, 176(%rdx)
movdqu %xmm0, 224(%rdx)
movdqu %xmm4, 240(%rdx)
addq \$256, %rsi
jmp chacha_blocks_ssse3_mainloop_cont
chacha_blocks_ssse3_noinput1:
movdqu %xmm5, 0(%rdx)
movdqu %xmm7, 16(%rdx)
movdqu %xmm8, 64(%rdx)
movdqu %xmm10, 80(%rdx)
movdqu %xmm1, 128(%rdx)
movdqu %xmm3, 144(%rdx)
movdqu %xmm0, 192(%rdx)
movdqu %xmm4, 208(%rdx)
movdqa 384(%rsp), %xmm0
movdqa 400(%rsp), %xmm1
movdqa 416(%rsp), %xmm2
movdqa 432(%rsp), %xmm3
movdqa 448(%rsp), %xmm4
movdqa 464(%rsp), %xmm5
movdqa 480(%rsp), %xmm6
movdqa 496(%rsp), %xmm7
movdqa %xmm0, %xmm8
movdqa %xmm2, %xmm9
movdqa %xmm4, %xmm10
movdqa %xmm6, %xmm11
punpckldq %xmm1, %xmm8
punpckldq %xmm3, %xmm9
punpckhdq %xmm1, %xmm0
punpckhdq %xmm3, %xmm2
punpckldq %xmm5, %xmm10
punpckldq %xmm7, %xmm11
punpckhdq %xmm5, %xmm4
punpckhdq %xmm7, %xmm6
movdqa %xmm8, %xmm1
movdqa %xmm0, %xmm3
movdqa %xmm10, %xmm5
movdqa %xmm4, %xmm7
punpcklqdq %xmm9, %xmm1
punpcklqdq %xmm11, %xmm5
punpckhqdq %xmm9, %xmm8
punpckhqdq %xmm11, %xmm10
punpcklqdq %xmm2, %xmm3
punpcklqdq %xmm6, %xmm7
punpckhqdq %xmm2, %xmm0
punpckhqdq %xmm6, %xmm4
movdqu %xmm1, 32(%rdx)
movdqu %xmm5, 48(%rdx)
movdqu %xmm8, 96(%rdx)
movdqu %xmm10, 112(%rdx)
movdqu %xmm3, 160(%rdx)
movdqu %xmm7, 176(%rdx)
movdqu %xmm0, 224(%rdx)
movdqu %xmm4, 240(%rdx)
chacha_blocks_ssse3_mainloop_cont:
addq \$256, %rdx
subq \$256, %rcx
cmp \$256, %rcx
jae chacha_blocks_ssse3_atleast256
movdqa 80(%rsp), %xmm6
movdqa 96(%rsp), %xmm7
movdqa 0(%rsp), %xmm8
movdqa 16(%rsp), %xmm9
movdqa 32(%rsp), %xmm10
movdqa 48(%rsp), %xmm11
movq \$1, %r9
chacha_blocks_ssse3_below256:
movq %r9, %xmm5
andq %rcx, %rcx
jz chacha_blocks_ssse3_done
cmpq \$64, %rcx
jae chacha_blocks_ssse3_above63
movq %rdx, %r9
andq %rsi, %rsi
jz chacha_blocks_ssse3_noinput2
movq %rcx, %r10
movq %rsp, %rdx
addq %r10, %rsi
addq %r10, %rdx
negq %r10
chacha_blocks_ssse3_copyinput:
movb (%rsi, %r10), %al
movb %al, (%rdx, %r10)
incq %r10
jnz chacha_blocks_ssse3_copyinput
movq %rsp, %rsi
chacha_blocks_ssse3_noinput2:
movq %rsp, %rdx
chacha_blocks_ssse3_above63:
movdqa %xmm8, %xmm0
movdqa %xmm9, %xmm1
movdqa %xmm10, %xmm2
movdqa %xmm11, %xmm3
movq 64(%rsp), %rax
chacha_blocks_ssse3_mainloop2:
paddd %xmm1, %xmm0
pxor %xmm0, %xmm3
pshufb %xmm6, %xmm3
paddd %xmm3, %xmm2
pxor %xmm2, %xmm1
movdqa %xmm1, %xmm4
pslld \$12, %xmm4
psrld \$20, %xmm1
pxor %xmm4, %xmm1
paddd %xmm1, %xmm0
pxor %xmm0, %xmm3
pshufb %xmm7, %xmm3
pshufd \$0x93, %xmm0, %xmm0
paddd %xmm3, %xmm2
pshufd \$0x4e, %xmm3, %xmm3
pxor %xmm2, %xmm1
pshufd \$0x39, %xmm2, %xmm2
movdqa %xmm1, %xmm4
pslld \$7, %xmm4
psrld \$25, %xmm1
pxor %xmm4, %xmm1
paddd %xmm1, %xmm0
pxor %xmm0, %xmm3
pshufb %xmm6, %xmm3
paddd %xmm3, %xmm2
pxor %xmm2, %xmm1
movdqa %xmm1, %xmm4
pslld \$12, %xmm4
psrld \$20, %xmm1
pxor %xmm4, %xmm1
paddd %xmm1, %xmm0
pxor %xmm0, %xmm3
pshufb %xmm7, %xmm3
pshufd \$0x39, %xmm0, %xmm0
paddd %xmm3, %xmm2
pshufd \$0x4e, %xmm3, %xmm3
pxor %xmm2, %xmm1
pshufd \$0x93, %xmm2, %xmm2
movdqa %xmm1, %xmm4
pslld \$7, %xmm4
psrld \$25, %xmm1
pxor %xmm4, %xmm1
subq \$2, %rax
jnz chacha_blocks_ssse3_mainloop2
paddd %xmm8, %xmm0
paddd %xmm9, %xmm1
paddd %xmm10, %xmm2
paddd %xmm11, %xmm3
andq %rsi, %rsi
jz chacha_blocks_ssse3_noinput3
movdqu 0(%rsi), %xmm12
movdqu 16(%rsi), %xmm13
movdqu 32(%rsi), %xmm14
movdqu 48(%rsi), %xmm15
pxor %xmm12, %xmm0
pxor %xmm13, %xmm1
pxor %xmm14, %xmm2
pxor %xmm15, %xmm3
addq \$64, %rsi
chacha_blocks_ssse3_noinput3:
movdqu %xmm0, 0(%rdx)
movdqu %xmm1, 16(%rdx)
movdqu %xmm2, 32(%rdx)
movdqu %xmm3, 48(%rdx)
paddq %xmm5, %xmm11
cmpq \$64, %rcx
jbe chacha_blocks_ssse3_mainloop2_finishup
addq \$64, %rdx
subq \$64, %rcx
jmp chacha_blocks_ssse3_below256
chacha_blocks_ssse3_mainloop2_finishup:
cmpq \$64, %rcx
je chacha_blocks_ssse3_done
addq %rcx, %r9
addq %rcx, %rdx
negq %rcx
chacha_blocks_ssse3_copyoutput:
movb (%rdx, %rcx), %al
movb %al, (%r9, %rcx)
incq %rcx
jnz chacha_blocks_ssse3_copyoutput
chacha_blocks_ssse3_done:
movdqu %xmm11, 32(%rdi)
movq %rbp, %rsp
popq %rbp
popq %rbx
ret
___
);


add_function(
	'chacha_blocks', 4, true, 'avx',
<<___
pushq %rbx
pushq %rbp
movq %rsp, %rbp
andq \$~63, %rsp
subq \$512, %rsp
leaq chacha_constants(%rip), %rax
vmovdqa 0(%rax), %xmm8
vmovdqa 16(%rax), %xmm6
vmovdqa 32(%rax), %xmm7
vmovdqa 0(%rdi), %xmm9
vmovdqa 16(%rdi), %xmm10
vmovdqa 32(%rdi), %xmm11
movq 32(%rdi), %rbx
movq 48(%rdi), %rax
movq \$1, %r9
vmovdqa %xmm8, 0(%rsp)
vmovdqa %xmm9, 16(%rsp)
vmovdqa %xmm10, 32(%rsp)
vmovdqa %xmm11, 48(%rsp)
vmovdqa %xmm6, 80(%rsp)
vmovdqa %xmm7, 96(%rsp)
movq %rax, 64(%rsp)
cmp \$128, %rcx
jb chacha_blocks_avx_below128_fixup
cmp \$256, %rcx
jb chacha_blocks_avx_atleast128
chacha_blocks_avx_atleast256_fixup:
vpshufd \$0x00, %xmm8, %xmm0
vpshufd \$0x55, %xmm8, %xmm1
vpshufd \$0xaa, %xmm8, %xmm2
vpshufd \$0xff, %xmm8, %xmm3
vmovdqa %xmm0, 128(%rsp)
vmovdqa %xmm1, 144(%rsp)
vmovdqa %xmm2, 160(%rsp)
vmovdqa %xmm3, 176(%rsp)
vpshufd \$0x00, %xmm9, %xmm0
vpshufd \$0x55, %xmm9, %xmm1
vpshufd \$0xaa, %xmm9, %xmm2
vpshufd \$0xff, %xmm9, %xmm3
vmovdqa %xmm0, 192(%rsp)
vmovdqa %xmm1, 208(%rsp)
vmovdqa %xmm2, 224(%rsp)
vmovdqa %xmm3, 240(%rsp)
vpshufd \$0x00, %xmm10, %xmm0
vpshufd \$0x55, %xmm10, %xmm1
vpshufd \$0xaa, %xmm10, %xmm2
vpshufd \$0xff, %xmm10, %xmm3
vmovdqa %xmm0, 256(%rsp)
vmovdqa %xmm1, 272(%rsp)
vmovdqa %xmm2, 288(%rsp)
vmovdqa %xmm3, 304(%rsp)
vpshufd \$0xaa, %xmm11, %xmm0
vpshufd \$0xff, %xmm11, %xmm1
vmovdqa %xmm0, 352(%rsp)
vmovdqa %xmm1, 368(%rsp)
jmp chacha_blocks_avx_atleast256
.align 64
nop
nop
nop
nop
nop
chacha_blocks_avx_atleast256:
movq %rbx, %rax
leaq 1(%rbx), %r8
leaq 2(%rbx), %r9
leaq 3(%rbx), %r10
movl %eax, 320(%rsp)
movl %r8d, 4+320(%rsp)
movl %r9d, 8+320(%rsp)
movl %r10d, 12+320(%rsp)
shrq \$32, %rax
shrq \$32, %r8
shrq \$32, %r9
shrq \$32, %r10
addq \$4, %rbx
movl %eax, 336(%rsp)
movl %r8d, 4+336(%rsp)
movl %r9d, 8+336(%rsp)
movl %r10d, 12+336(%rsp)
movq 64(%rsp), %rax
vmovdqa 128(%rsp), %xmm0
vmovdqa 144(%rsp), %xmm1
vmovdqa 160(%rsp), %xmm2
vmovdqa 176(%rsp), %xmm3
vmovdqa 192(%rsp), %xmm4
vmovdqa 208(%rsp), %xmm5
vmovdqa 224(%rsp), %xmm6
vmovdqa 240(%rsp), %xmm7
vmovdqa 256(%rsp), %xmm8
vmovdqa 272(%rsp), %xmm9
vmovdqa 288(%rsp), %xmm10
vmovdqa 304(%rsp), %xmm11
vmovdqa 320(%rsp), %xmm12
vmovdqa 336(%rsp), %xmm13
vmovdqa 352(%rsp), %xmm14
vmovdqa 368(%rsp), %xmm15
chacha_blocks_avx_mainloop1:
vpaddd %xmm0, %xmm4, %xmm0
vpaddd %xmm1, %xmm5, %xmm1
vpxor %xmm12, %xmm0, %xmm12
vpxor %xmm13, %xmm1, %xmm13
vpaddd %xmm2, %xmm6, %xmm2
vpaddd %xmm3, %xmm7, %xmm3
vpxor %xmm14, %xmm2, %xmm14
vpxor %xmm15, %xmm3, %xmm15
vpshufb 80(%rsp), %xmm12, %xmm12
vpshufb 80(%rsp), %xmm13, %xmm13
vpaddd %xmm8, %xmm12, %xmm8
vpaddd %xmm9, %xmm13, %xmm9
vpshufb 80(%rsp), %xmm14, %xmm14
vpshufb 80(%rsp), %xmm15, %xmm15
vpaddd %xmm10, %xmm14, %xmm10
vpaddd %xmm11, %xmm15, %xmm11
vmovdqa %xmm12, 112(%rsp)
vpxor %xmm4, %xmm8, %xmm4
vpxor %xmm5, %xmm9, %xmm5
vpslld \$ 12, %xmm4, %xmm12
vpsrld \$20, %xmm4, %xmm4
vpxor %xmm4, %xmm12, %xmm4
vpslld \$ 12, %xmm5, %xmm12
vpsrld \$20, %xmm5, %xmm5
vpxor %xmm5, %xmm12, %xmm5
vpxor %xmm6, %xmm10, %xmm6
vpxor %xmm7, %xmm11, %xmm7
vpslld \$ 12, %xmm6, %xmm12
vpsrld \$20, %xmm6, %xmm6
vpxor %xmm6, %xmm12, %xmm6
vpslld \$ 12, %xmm7, %xmm12
vpsrld \$20, %xmm7, %xmm7
vpxor %xmm7, %xmm12, %xmm7
vpaddd %xmm0, %xmm4, %xmm0
vpaddd %xmm1, %xmm5, %xmm1
vpxor 112(%rsp), %xmm0, %xmm12
vpxor %xmm13, %xmm1, %xmm13
vpaddd %xmm2, %xmm6, %xmm2
vpaddd %xmm3, %xmm7, %xmm3
vpxor %xmm14, %xmm2, %xmm14
vpxor %xmm15, %xmm3, %xmm15
vpshufb 96(%rsp), %xmm12, %xmm12
vpshufb 96(%rsp), %xmm13, %xmm13
vpaddd %xmm8, %xmm12, %xmm8
vpaddd %xmm9, %xmm13, %xmm9
vpshufb 96(%rsp), %xmm14, %xmm14
vpshufb 96(%rsp), %xmm15, %xmm15
vpaddd %xmm10, %xmm14, %xmm10
vpaddd %xmm11, %xmm15, %xmm11
vmovdqa %xmm12, 112(%rsp)
vpxor %xmm4, %xmm8, %xmm4
vpxor %xmm5, %xmm9, %xmm5
vpslld \$ 7, %xmm4, %xmm12
vpsrld \$25, %xmm4, %xmm4
vpxor %xmm4, %xmm12, %xmm4
vpslld \$ 7, %xmm5, %xmm12
vpsrld \$25, %xmm5, %xmm5
vpxor %xmm5, %xmm12, %xmm5
vpxor %xmm6, %xmm10, %xmm6
vpxor %xmm7, %xmm11, %xmm7
vpslld \$ 7, %xmm6, %xmm12
vpsrld \$25, %xmm6, %xmm6
vpxor %xmm6, %xmm12, %xmm6
vpslld \$ 7, %xmm7, %xmm12
vpsrld \$25, %xmm7, %xmm7
vpxor %xmm7, %xmm12, %xmm7
vpaddd %xmm0, %xmm5, %xmm0
vpaddd %xmm1, %xmm6, %xmm1
vpxor %xmm15, %xmm0, %xmm15
vpxor 112(%rsp), %xmm1, %xmm12
vpaddd %xmm2, %xmm7, %xmm2
vpaddd %xmm3, %xmm4, %xmm3
vpxor %xmm13, %xmm2, %xmm13
vpxor %xmm14, %xmm3, %xmm14
vpshufb 80(%rsp), %xmm15, %xmm15
vpshufb 80(%rsp), %xmm12, %xmm12
vpaddd %xmm10, %xmm15, %xmm10
vpaddd %xmm11, %xmm12, %xmm11
vpshufb 80(%rsp), %xmm13, %xmm13
vpshufb 80(%rsp), %xmm14, %xmm14
vpaddd %xmm8, %xmm13, %xmm8
vpaddd %xmm9, %xmm14, %xmm9
vmovdqa %xmm15, 112(%rsp)
vpxor %xmm5, %xmm10, %xmm5
vpxor %xmm6, %xmm11, %xmm6
vpslld \$ 12, %xmm5, %xmm15
vpsrld \$20, %xmm5, %xmm5
vpxor %xmm5, %xmm15, %xmm5
vpslld \$ 12, %xmm6, %xmm15
vpsrld \$20, %xmm6, %xmm6
vpxor %xmm6, %xmm15, %xmm6
vpxor %xmm7, %xmm8, %xmm7
vpxor %xmm4, %xmm9, %xmm4
vpslld \$ 12, %xmm7, %xmm15
vpsrld \$20, %xmm7, %xmm7
vpxor %xmm7, %xmm15, %xmm7
vpslld \$ 12, %xmm4, %xmm15
vpsrld \$20, %xmm4, %xmm4
vpxor %xmm4, %xmm15, %xmm4
vpaddd %xmm0, %xmm5, %xmm0
vpaddd %xmm1, %xmm6, %xmm1
vpxor 112(%rsp), %xmm0, %xmm15
vpxor %xmm12, %xmm1, %xmm12
vpaddd %xmm2, %xmm7, %xmm2
vpaddd %xmm3, %xmm4, %xmm3
vpxor %xmm13, %xmm2, %xmm13
vpxor %xmm14, %xmm3, %xmm14
vpshufb 96(%rsp), %xmm15, %xmm15
vpshufb 96(%rsp), %xmm12, %xmm12
vpaddd %xmm10, %xmm15, %xmm10
vpaddd %xmm11, %xmm12, %xmm11
vpshufb 96(%rsp), %xmm13, %xmm13
vpshufb 96(%rsp), %xmm14, %xmm14
vpaddd %xmm8, %xmm13, %xmm8
vpaddd %xmm9, %xmm14, %xmm9
vmovdqa %xmm15, 112(%rsp)
vpxor %xmm5, %xmm10, %xmm5
vpxor %xmm6, %xmm11, %xmm6
vpslld \$ 7, %xmm5, %xmm15
vpsrld \$25, %xmm5, %xmm5
vpxor %xmm5, %xmm15, %xmm5
vpslld \$ 7, %xmm6, %xmm15
vpsrld \$25, %xmm6, %xmm6
vpxor %xmm6, %xmm15, %xmm6
vpxor %xmm7, %xmm8, %xmm7
vpxor %xmm4, %xmm9, %xmm4
vpslld \$ 7, %xmm7, %xmm15
vpsrld \$25, %xmm7, %xmm7
vpxor %xmm7, %xmm15, %xmm7
vpslld \$ 7, %xmm4, %xmm15
vpsrld \$25, %xmm4, %xmm4
vpxor %xmm4, %xmm15, %xmm4
vmovdqa 112(%rsp), %xmm15
subq \$2, %rax
jnz chacha_blocks_avx_mainloop1
vpaddd 128(%rsp), %xmm0, %xmm0
vpaddd 144(%rsp), %xmm1, %xmm1
vpaddd 160(%rsp), %xmm2, %xmm2
vpaddd 176(%rsp), %xmm3, %xmm3
vpaddd 192(%rsp), %xmm4, %xmm4
vpaddd 208(%rsp), %xmm5, %xmm5
vpaddd 224(%rsp), %xmm6, %xmm6
vpaddd 240(%rsp), %xmm7, %xmm7
vpaddd 256(%rsp), %xmm8, %xmm8
vpaddd 272(%rsp), %xmm9, %xmm9
vpaddd 288(%rsp), %xmm10, %xmm10
vpaddd 304(%rsp), %xmm11, %xmm11
vpaddd 320(%rsp), %xmm12, %xmm12
vpaddd 336(%rsp), %xmm13, %xmm13
vpaddd 352(%rsp), %xmm14, %xmm14
vpaddd 368(%rsp), %xmm15, %xmm15
vmovdqa %xmm8, 384(%rsp)
vmovdqa %xmm9, 400(%rsp)
vmovdqa %xmm10, 416(%rsp)
vmovdqa %xmm11, 432(%rsp)
vmovdqa %xmm12, 448(%rsp)
vmovdqa %xmm13, 464(%rsp)
vmovdqa %xmm14, 480(%rsp)
vmovdqa %xmm15, 496(%rsp)
vpunpckldq %xmm1, %xmm0, %xmm8
vpunpckldq %xmm3, %xmm2, %xmm9
vpunpckhdq %xmm1, %xmm0, %xmm12
vpunpckhdq %xmm3, %xmm2, %xmm13
vpunpckldq %xmm5, %xmm4, %xmm10
vpunpckldq %xmm7, %xmm6, %xmm11
vpunpckhdq %xmm5, %xmm4, %xmm14
vpunpckhdq %xmm7, %xmm6, %xmm15
vpunpcklqdq %xmm9, %xmm8, %xmm0
vpunpcklqdq %xmm11, %xmm10, %xmm1
vpunpckhqdq %xmm9, %xmm8, %xmm2
vpunpckhqdq %xmm11, %xmm10, %xmm3
vpunpcklqdq %xmm13, %xmm12, %xmm4
vpunpcklqdq %xmm15, %xmm14, %xmm5
vpunpckhqdq %xmm13, %xmm12, %xmm6
vpunpckhqdq %xmm15, %xmm14, %xmm7
andq %rsi, %rsi
jz chacha_blocks_avx_noinput1
vpxor 0(%rsi), %xmm0, %xmm0
vpxor 16(%rsi), %xmm1, %xmm1
vpxor 64(%rsi), %xmm2, %xmm2
vpxor 80(%rsi), %xmm3, %xmm3
vpxor 128(%rsi), %xmm4, %xmm4
vpxor 144(%rsi), %xmm5, %xmm5
vpxor 192(%rsi), %xmm6, %xmm6
vpxor 208(%rsi), %xmm7, %xmm7
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 64(%rdx)
vmovdqu %xmm3, 80(%rdx)
vmovdqu %xmm4, 128(%rdx)
vmovdqu %xmm5, 144(%rdx)
vmovdqu %xmm6, 192(%rdx)
vmovdqu %xmm7, 208(%rdx)
vmovdqa 384(%rsp), %xmm0
vmovdqa 400(%rsp), %xmm1
vmovdqa 416(%rsp), %xmm2
vmovdqa 432(%rsp), %xmm3
vmovdqa 448(%rsp), %xmm4
vmovdqa 464(%rsp), %xmm5
vmovdqa 480(%rsp), %xmm6
vmovdqa 496(%rsp), %xmm7
vpunpckldq %xmm1, %xmm0, %xmm8
vpunpckldq %xmm3, %xmm2, %xmm9
vpunpckhdq %xmm1, %xmm0, %xmm12
vpunpckhdq %xmm3, %xmm2, %xmm13
vpunpckldq %xmm5, %xmm4, %xmm10
vpunpckldq %xmm7, %xmm6, %xmm11
vpunpckhdq %xmm5, %xmm4, %xmm14
vpunpckhdq %xmm7, %xmm6, %xmm15
vpunpcklqdq %xmm9, %xmm8, %xmm0
vpunpcklqdq %xmm11, %xmm10, %xmm1
vpunpckhqdq %xmm9, %xmm8, %xmm2
vpunpckhqdq %xmm11, %xmm10, %xmm3
vpunpcklqdq %xmm13, %xmm12, %xmm4
vpunpcklqdq %xmm15, %xmm14, %xmm5
vpunpckhqdq %xmm13, %xmm12, %xmm6
vpunpckhqdq %xmm15, %xmm14, %xmm7
vpxor 32(%rsi), %xmm0, %xmm0
vpxor 48(%rsi), %xmm1, %xmm1
vpxor 96(%rsi), %xmm2, %xmm2
vpxor 112(%rsi), %xmm3, %xmm3
vpxor 160(%rsi), %xmm4, %xmm4
vpxor 176(%rsi), %xmm5, %xmm5
vpxor 224(%rsi), %xmm6, %xmm6
vpxor 240(%rsi), %xmm7, %xmm7
vmovdqu %xmm0, 32(%rdx)
vmovdqu %xmm1, 48(%rdx)
vmovdqu %xmm2, 96(%rdx)
vmovdqu %xmm3, 112(%rdx)
vmovdqu %xmm4, 160(%rdx)
vmovdqu %xmm5, 176(%rdx)
vmovdqu %xmm6, 224(%rdx)
vmovdqu %xmm7, 240(%rdx)
addq \$256, %rsi
jmp chacha_blocks_avx_mainloop1_cont
chacha_blocks_avx_noinput1:
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 64(%rdx)
vmovdqu %xmm3, 80(%rdx)
vmovdqu %xmm4, 128(%rdx)
vmovdqu %xmm5, 144(%rdx)
vmovdqu %xmm6, 192(%rdx)
vmovdqu %xmm7, 208(%rdx)
vmovdqa 384(%rsp), %xmm0
vmovdqa 400(%rsp), %xmm1
vmovdqa 416(%rsp), %xmm2
vmovdqa 432(%rsp), %xmm3
vmovdqa 448(%rsp), %xmm4
vmovdqa 464(%rsp), %xmm5
vmovdqa 480(%rsp), %xmm6
vmovdqa 496(%rsp), %xmm7
vpunpckldq %xmm1, %xmm0, %xmm8
vpunpckldq %xmm3, %xmm2, %xmm9
vpunpckhdq %xmm1, %xmm0, %xmm12
vpunpckhdq %xmm3, %xmm2, %xmm13
vpunpckldq %xmm5, %xmm4, %xmm10
vpunpckldq %xmm7, %xmm6, %xmm11
vpunpckhdq %xmm5, %xmm4, %xmm14
vpunpckhdq %xmm7, %xmm6, %xmm15
vpunpcklqdq %xmm9, %xmm8, %xmm0
vpunpcklqdq %xmm11, %xmm10, %xmm1
vpunpckhqdq %xmm9, %xmm8, %xmm2
vpunpckhqdq %xmm11, %xmm10, %xmm3
vpunpcklqdq %xmm13, %xmm12, %xmm4
vpunpcklqdq %xmm15, %xmm14, %xmm5
vpunpckhqdq %xmm13, %xmm12, %xmm6
vpunpckhqdq %xmm15, %xmm14, %xmm7
vmovdqu %xmm0, 32(%rdx)
vmovdqu %xmm1, 48(%rdx)
vmovdqu %xmm2, 96(%rdx)
vmovdqu %xmm3, 112(%rdx)
vmovdqu %xmm4, 160(%rdx)
vmovdqu %xmm5, 176(%rdx)
vmovdqu %xmm6, 224(%rdx)
vmovdqu %xmm7, 240(%rdx)
chacha_blocks_avx_mainloop1_cont:
addq \$256, %rdx
subq \$256, %rcx
cmp \$256, %rcx
jae chacha_blocks_avx_atleast256
movq %rbx, 48(%rsp)
vmovdqa 80(%rsp), %xmm6
vmovdqa 96(%rsp), %xmm7
vmovdqa 0(%rsp), %xmm8
vmovdqa 16(%rsp), %xmm9
vmovdqa 32(%rsp), %xmm10
vmovdqa 48(%rsp), %xmm11
movq \$1, %r9
cmp \$128, %rcx
jb chacha_blocks_avx_below128_fixup
chacha_blocks_avx_atleast128:
vmovd %r9, %xmm5
vmovdqa %xmm8, %xmm0
vmovdqa %xmm9, %xmm1
vmovdqa %xmm10, %xmm2
vmovdqa %xmm11, %xmm3
vpaddq %xmm5, %xmm11, %xmm11
vmovdqa %xmm8, 384(%rsp)
vmovdqa %xmm9, 400(%rsp)
vmovdqa %xmm10, 416(%rsp)
vmovdqa %xmm3, 432(%rsp)
vmovdqa %xmm11, 448(%rsp)
movq 64(%rsp), %rax
chacha_blocks_avx_mainloop2:
vpaddd %xmm0, %xmm1, %xmm0
vpaddd %xmm8, %xmm9, %xmm8
vpxor %xmm3, %xmm0, %xmm3
vpxor %xmm11, %xmm8, %xmm11
vpshufb %xmm6, %xmm3, %xmm3
vpshufb %xmm6, %xmm11, %xmm11
vpaddd %xmm2, %xmm3, %xmm2
vpaddd %xmm10, %xmm11, %xmm10
vpxor %xmm1, %xmm2, %xmm1
vpxor %xmm9, %xmm10, %xmm9
vpslld \$ 12, %xmm1, %xmm4
vpsrld \$20, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpslld \$ 12, %xmm9, %xmm5
vpsrld \$20, %xmm9, %xmm9
vpxor %xmm9, %xmm5, %xmm9
vpaddd %xmm0, %xmm1, %xmm0
vpaddd %xmm8, %xmm9, %xmm8
vpxor %xmm3, %xmm0, %xmm3
vpxor %xmm11, %xmm8, %xmm11
vpshufb %xmm7, %xmm3, %xmm3
vpshufb %xmm7, %xmm11, %xmm11
vpshufd \$0x93, %xmm0, %xmm0
vpshufd \$0x93, %xmm8, %xmm8
vpaddd %xmm2, %xmm3, %xmm2
vpaddd %xmm10, %xmm11, %xmm10
vpshufd \$0x4e, %xmm3, %xmm3
vpshufd \$0x4e, %xmm11, %xmm11
vpxor %xmm1, %xmm2, %xmm1
vpxor %xmm9, %xmm10, %xmm9
vpshufd \$0x39, %xmm2, %xmm2
vpshufd \$0x39, %xmm10, %xmm10
vpslld \$ 7, %xmm1, %xmm4
vpsrld \$25, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpslld \$ 7, %xmm9, %xmm5
vpsrld \$25, %xmm9, %xmm9
vpxor %xmm9, %xmm5, %xmm9
vpaddd %xmm0, %xmm1, %xmm0
vpaddd %xmm8, %xmm9, %xmm8
vpxor %xmm3, %xmm0, %xmm3
vpxor %xmm11, %xmm8, %xmm11
vpshufb %xmm6, %xmm3, %xmm3
vpshufb %xmm6, %xmm11, %xmm11
vpaddd %xmm2, %xmm3, %xmm2
vpaddd %xmm10, %xmm11, %xmm10
vpxor %xmm1, %xmm2, %xmm1
vpxor %xmm9, %xmm10, %xmm9
vpslld \$ 12, %xmm1, %xmm4
vpsrld \$20, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpslld \$ 12, %xmm9, %xmm5
vpsrld \$20, %xmm9, %xmm9
vpxor %xmm9, %xmm5, %xmm9
vpaddd %xmm0, %xmm1, %xmm0
vpaddd %xmm8, %xmm9, %xmm8
vpxor %xmm3, %xmm0, %xmm3
vpxor %xmm11, %xmm8, %xmm11
vpshufb %xmm7, %xmm3, %xmm3
vpshufb %xmm7, %xmm11, %xmm11
vpshufd \$0x39, %xmm0, %xmm0
vpshufd \$0x39, %xmm8, %xmm8
vpaddd %xmm2, %xmm3, %xmm2
vpaddd %xmm10, %xmm11, %xmm10
vpshufd \$0x4e, %xmm3, %xmm3
vpshufd \$0x4e, %xmm11, %xmm11
vpxor %xmm1, %xmm2, %xmm1
vpxor %xmm9, %xmm10, %xmm9
vpshufd \$0x93, %xmm2, %xmm2
vpshufd \$0x93, %xmm10, %xmm10
vpslld \$ 7, %xmm1, %xmm4
vpsrld \$25, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpslld \$ 7, %xmm9, %xmm5
vpsrld \$25, %xmm9, %xmm9
vpxor %xmm9, %xmm5, %xmm9
subq \$2, %rax
jnz chacha_blocks_avx_mainloop2
vmovdqa 384(%rsp), %xmm12
vmovdqa 400(%rsp), %xmm13
vmovdqa 416(%rsp), %xmm14
vmovdqa 432(%rsp), %xmm15
vmovdqa 448(%rsp), %xmm5
vpaddd %xmm12, %xmm0, %xmm0
vpaddd %xmm13, %xmm1, %xmm1
vpaddd %xmm14, %xmm2, %xmm2
vpaddd %xmm15, %xmm3, %xmm3
vpaddd %xmm12, %xmm8, %xmm8
vpaddd %xmm13, %xmm9, %xmm9
vpaddd %xmm14, %xmm10, %xmm10
vpaddd %xmm5, %xmm11, %xmm11
andq %rsi, %rsi
jz chacha_blocks_avx_noinput2
vpxor 0(%rsi), %xmm0, %xmm0
vpxor 16(%rsi), %xmm1, %xmm1
vpxor 32(%rsi), %xmm2, %xmm2
vpxor 48(%rsi), %xmm3, %xmm3
vpxor 64(%rsi), %xmm8, %xmm8
vpxor 80(%rsi), %xmm9, %xmm9
vpxor 96(%rsi), %xmm10, %xmm10
vpxor 112(%rsi), %xmm11, %xmm11
addq \$128, %rsi
chacha_blocks_avx_noinput2:
addq \$2, %rbx
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 32(%rdx)
vmovdqu %xmm3, 48(%rdx)
vmovdqu %xmm8, 64(%rdx)
vmovdqu %xmm9, 80(%rdx)
vmovdqu %xmm10, 96(%rdx)
vmovdqu %xmm11, 112(%rdx)
addq \$128, %rdx
subq \$128, %rcx
movq %rbx, 48(%rsp)
vmovdqa 80(%rsp), %xmm6
vmovdqa 96(%rsp), %xmm7
vmovdqa 0(%rsp), %xmm8
vmovdqa 16(%rsp), %xmm9
vmovdqa 32(%rsp), %xmm10
vmovdqa 48(%rsp), %xmm11
chacha_blocks_avx_below128_fixup:
movq \$1, %r9
vmovd %r9, %xmm5
chacha_blocks_avx_below128:
andq %rcx, %rcx
jz chacha_blocks_avx_done
cmpq \$64, %rcx
jae chacha_blocks_avx_above63
movq %rdx, %r9
andq %rsi, %rsi
jz chacha_blocks_avx_noinput3
movq %rcx, %r10
movq %rsp, %rdx
addq %r10, %rsi
addq %r10, %rdx
negq %r10
chacha_blocks_avx_copyinput:
movb (%rsi, %r10), %al
movb %al, (%rdx, %r10)
incq %r10
jnz chacha_blocks_avx_copyinput
movq %rsp, %rsi
chacha_blocks_avx_noinput3:
movq %rsp, %rdx
chacha_blocks_avx_above63:
vmovdqa %xmm8, %xmm0
vmovdqa %xmm9, %xmm1
vmovdqa %xmm10, %xmm2
vmovdqa %xmm11, %xmm3
movq 64(%rsp), %rax
chacha_blocks_avx_mainloop3:
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vpshufb %xmm6, %xmm3, %xmm3
vpaddd %xmm2, %xmm3, %xmm2
vpxor %xmm1, %xmm2, %xmm1
vpslld \$12, %xmm1, %xmm4
vpsrld \$20, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vpshufb %xmm7, %xmm3, %xmm3
vpshufd \$0x93, %xmm0, %xmm0
vpaddd %xmm2, %xmm3, %xmm2
vpshufd \$0x4e, %xmm3, %xmm3
vpxor %xmm1, %xmm2, %xmm1
vpshufd \$0x39, %xmm2, %xmm2
vpslld \$7, %xmm1, %xmm4
vpsrld \$25, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vpshufb %xmm6, %xmm3, %xmm3
vpaddd %xmm2, %xmm3, %xmm2
vpxor %xmm1, %xmm2, %xmm1
vpslld \$12, %xmm1, %xmm4
vpsrld \$20, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vpshufb %xmm7, %xmm3, %xmm3
vpshufd \$0x39, %xmm0, %xmm0
vpaddd %xmm2, %xmm3, %xmm2
vpshufd \$0x4e, %xmm3, %xmm3
vpxor %xmm1, %xmm2, %xmm1
vpshufd \$0x93, %xmm2, %xmm2
vpslld \$7, %xmm1, %xmm4
vpsrld \$25, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
subq \$2, %rax
jnz chacha_blocks_avx_mainloop3
vpaddd %xmm0, %xmm8, %xmm0
vpaddd %xmm1, %xmm9, %xmm1
vpaddd %xmm2, %xmm10, %xmm2
vpaddd %xmm3, %xmm11, %xmm3
andq %rsi, %rsi
jz chacha_blocks_avx_noinput4
vpxor 0(%rsi), %xmm0, %xmm0
vpxor 16(%rsi), %xmm1, %xmm1
vpxor 32(%rsi), %xmm2, %xmm2
vpxor 48(%rsi), %xmm3, %xmm3
addq \$64, %rsi
chacha_blocks_avx_noinput4:
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 32(%rdx)
vmovdqu %xmm3, 48(%rdx)
vpaddq %xmm11, %xmm5, %xmm11
addq \$1, %rbx
cmpq \$64, %rcx
jbe chacha_blocks_avx_mainloop3_finishup
addq \$64, %rdx
subq \$64, %rcx
jmp chacha_blocks_avx_below128
chacha_blocks_avx_mainloop3_finishup:
cmpq \$64, %rcx
je chacha_blocks_avx_done
addq %rcx, %r9
addq %rcx, %rdx
negq %rcx
chacha_blocks_avx_copyoutput:
movb (%rdx, %rcx), %al
movb %al, (%r9, %rcx)
incq %rcx
jnz chacha_blocks_avx_copyoutput
chacha_blocks_avx_done:
movq %rbx, 32(%rdi)
movq %rbp, %rsp
popq %rbp
popq %rbx
ret
___
);



add_function(
	'chacha_blocks', 4, true, 'xop',
<<___
pushq %rbx
pushq %rbp
movq %rsp, %rbp
andq \$~63, %rsp
subq \$512, %rsp
leaq chacha_constants(%rip), %rax
vmovdqa 0(%rax), %xmm8
vmovdqu 0(%rdi), %xmm9
vmovdqu 16(%rdi), %xmm10
vmovdqu 32(%rdi), %xmm11
mov 32(%rdi), %rbx
movq 48(%rdi), %rax
movq \$1, %r9
vmovdqa %xmm8, 0(%rsp)
vmovdqa %xmm9, 16(%rsp)
vmovdqa %xmm10, 32(%rsp)
vmovdqa %xmm11, 48(%rsp)
movq %rax, 64(%rsp)
cmpq \$128, %rcx
jb chacha_blocks_xop_below128_fixup
cmpq \$256, %rcx
jb chacha_blocks_xop_atleast128
chacha_blocks_xop_atleast256_fixup:
vpshufd \$0x00, %xmm8, %xmm0
vpshufd \$0x55, %xmm8, %xmm1
vpshufd \$0xaa, %xmm8, %xmm2
vpshufd \$0xff, %xmm8, %xmm3
vmovdqa %xmm0, 128(%rsp)
vmovdqa %xmm1, 144(%rsp)
vmovdqa %xmm2, 160(%rsp)
vmovdqa %xmm3, 176(%rsp)
vpshufd \$0x00, %xmm9, %xmm0
vpshufd \$0x55, %xmm9, %xmm1
vpshufd \$0xaa, %xmm9, %xmm2
vpshufd \$0xff, %xmm9, %xmm3
vmovdqa %xmm0, 192(%rsp)
vmovdqa %xmm1, 208(%rsp)
vmovdqa %xmm2, 224(%rsp)
vmovdqa %xmm3, 240(%rsp)
vpshufd \$0x00, %xmm10, %xmm0
vpshufd \$0x55, %xmm10, %xmm1
vpshufd \$0xaa, %xmm10, %xmm2
vpshufd \$0xff, %xmm10, %xmm3
vmovdqa %xmm0, 256(%rsp)
vmovdqa %xmm1, 272(%rsp)
vmovdqa %xmm2, 288(%rsp)
vmovdqa %xmm3, 304(%rsp)
vpshufd \$0xaa, %xmm11, %xmm0
vpshufd \$0xff, %xmm11, %xmm1
vmovdqa %xmm0, 352(%rsp)
vmovdqa %xmm1, 368(%rsp)
jmp chacha_blocks_xop_atleast256
.align 64
chacha_blocks_xop_atleast256:
movq %rbx, %rax
leaq 1(%rbx), %r8
leaq 2(%rbx), %r9
leaq 3(%rbx), %r10
movl %eax, 320(%rsp)
movl %r8d, 4+320(%rsp)
movl %r9d, 8+320(%rsp)
movl %r10d, 12+320(%rsp)
shrq \$32, %rax
shrq \$32, %r8
shrq \$32, %r9
shrq \$32, %r10
addq \$4, %rbx
movl %eax, 336(%rsp)
movl %r8d, 4+336(%rsp)
movl %r9d, 8+336(%rsp)
movl %r10d, 12+336(%rsp)
movq 64(%rsp), %rax
vmovdqa 128(%rsp), %xmm0
vmovdqa 144(%rsp), %xmm1
vmovdqa 160(%rsp), %xmm2
vmovdqa 176(%rsp), %xmm3
vmovdqa 192(%rsp), %xmm4
vmovdqa 208(%rsp), %xmm5
vmovdqa 224(%rsp), %xmm6
vmovdqa 240(%rsp), %xmm7
vmovdqa 256(%rsp), %xmm8
vmovdqa 272(%rsp), %xmm9
vmovdqa 288(%rsp), %xmm10
vmovdqa 304(%rsp), %xmm11
vmovdqa 320(%rsp), %xmm12
vmovdqa 336(%rsp), %xmm13
vmovdqa 352(%rsp), %xmm14
vmovdqa 368(%rsp), %xmm15
chacha_blocks_xop_mainloop1:
vpaddd %xmm0, %xmm4, %xmm0
vpaddd %xmm1, %xmm5, %xmm1
vpaddd %xmm2, %xmm6, %xmm2
vpaddd %xmm3, %xmm7, %xmm3
vpxor %xmm12, %xmm0, %xmm12
vpxor %xmm13, %xmm1, %xmm13
vpxor %xmm14, %xmm2, %xmm14
vpxor %xmm15, %xmm3, %xmm15
vprotd \$16, %xmm12, %xmm12
vprotd \$16, %xmm13, %xmm13
vprotd \$16, %xmm14, %xmm14
vprotd \$16, %xmm15, %xmm15
vpaddd %xmm8, %xmm12, %xmm8
vpaddd %xmm9, %xmm13, %xmm9
vpaddd %xmm10, %xmm14, %xmm10
vpaddd %xmm11, %xmm15, %xmm11
vpxor %xmm4, %xmm8, %xmm4
vpxor %xmm5, %xmm9, %xmm5
vpxor %xmm6, %xmm10, %xmm6
vpxor %xmm7, %xmm11, %xmm7
vprotd \$12, %xmm4, %xmm4
vprotd \$12, %xmm5, %xmm5
vprotd \$12, %xmm6, %xmm6
vprotd \$12, %xmm7, %xmm7
vpaddd %xmm0, %xmm4, %xmm0
vpaddd %xmm1, %xmm5, %xmm1
vpaddd %xmm2, %xmm6, %xmm2
vpaddd %xmm3, %xmm7, %xmm3
vpxor %xmm12, %xmm0, %xmm12
vpxor %xmm13, %xmm1, %xmm13
vpxor %xmm14, %xmm2, %xmm14
vpxor %xmm15, %xmm3, %xmm15
vprotd \$8, %xmm12, %xmm12
vprotd \$8, %xmm13, %xmm13
vprotd \$8, %xmm14, %xmm14
vprotd \$8, %xmm15, %xmm15
vpaddd %xmm8, %xmm12, %xmm8
vpaddd %xmm9, %xmm13, %xmm9
vpaddd %xmm10, %xmm14, %xmm10
vpaddd %xmm11, %xmm15, %xmm11
vpxor %xmm4, %xmm8, %xmm4
vpxor %xmm5, %xmm9, %xmm5
vpxor %xmm6, %xmm10, %xmm6
vpxor %xmm7, %xmm11, %xmm7
vprotd \$7, %xmm4, %xmm4
vprotd \$7, %xmm5, %xmm5
vprotd \$7, %xmm6, %xmm6
vprotd \$7, %xmm7, %xmm7
vpaddd %xmm0, %xmm5, %xmm0
vpaddd %xmm1, %xmm6, %xmm1
vpaddd %xmm2, %xmm7, %xmm2
vpaddd %xmm3, %xmm4, %xmm3
vpxor %xmm15, %xmm0, %xmm15
vpxor %xmm12, %xmm1, %xmm12
vpxor %xmm13, %xmm2, %xmm13
vpxor %xmm14, %xmm3, %xmm14
vprotd \$16, %xmm15, %xmm15
vprotd \$16, %xmm12, %xmm12
vprotd \$16, %xmm13, %xmm13
vprotd \$16, %xmm14, %xmm14
vpaddd %xmm10, %xmm15, %xmm10
vpaddd %xmm11, %xmm12, %xmm11
vpaddd %xmm8, %xmm13, %xmm8
vpaddd %xmm9, %xmm14, %xmm9
vpxor %xmm5, %xmm10, %xmm5
vpxor %xmm6, %xmm11, %xmm6
vpxor %xmm7, %xmm8, %xmm7
vpxor %xmm4, %xmm9, %xmm4
vprotd \$12, %xmm5, %xmm5
vprotd \$12, %xmm6, %xmm6
vprotd \$12, %xmm7, %xmm7
vprotd \$12, %xmm4, %xmm4
vpaddd %xmm0, %xmm5, %xmm0
vpaddd %xmm1, %xmm6, %xmm1
vpaddd %xmm2, %xmm7, %xmm2
vpaddd %xmm3, %xmm4, %xmm3
vpxor %xmm15, %xmm0, %xmm15
vpxor %xmm12, %xmm1, %xmm12
vpxor %xmm13, %xmm2, %xmm13
vpxor %xmm14, %xmm3, %xmm14
vprotd \$8, %xmm15, %xmm15
vprotd \$8, %xmm12, %xmm12
vprotd \$8, %xmm13, %xmm13
vprotd \$8, %xmm14, %xmm14
vpaddd %xmm10, %xmm15, %xmm10
vpaddd %xmm11, %xmm12, %xmm11
vpaddd %xmm8, %xmm13, %xmm8
vpaddd %xmm9, %xmm14, %xmm9
vpxor %xmm5, %xmm10, %xmm5
vpxor %xmm6, %xmm11, %xmm6
vpxor %xmm7, %xmm8, %xmm7
vpxor %xmm4, %xmm9, %xmm4
vprotd \$7, %xmm5, %xmm5
vprotd \$7, %xmm6, %xmm6
vprotd \$7, %xmm7, %xmm7
vprotd \$7, %xmm4, %xmm4
subq \$2, %rax
jnz chacha_blocks_xop_mainloop1
vpaddd 128(%rsp), %xmm0, %xmm0
vpaddd 144(%rsp), %xmm1, %xmm1
vpaddd 160(%rsp), %xmm2, %xmm2
vpaddd 176(%rsp), %xmm3, %xmm3
vpaddd 192(%rsp), %xmm4, %xmm4
vpaddd 208(%rsp), %xmm5, %xmm5
vpaddd 224(%rsp), %xmm6, %xmm6
vpaddd 240(%rsp), %xmm7, %xmm7
vpaddd 256(%rsp), %xmm8, %xmm8
vpaddd 272(%rsp), %xmm9, %xmm9
vpaddd 288(%rsp), %xmm10, %xmm10
vpaddd 304(%rsp), %xmm11, %xmm11
vpaddd 320(%rsp), %xmm12, %xmm12
vpaddd 336(%rsp), %xmm13, %xmm13
vpaddd 352(%rsp), %xmm14, %xmm14
vpaddd 368(%rsp), %xmm15, %xmm15
vmovdqa %xmm8, 384(%rsp)
vmovdqa %xmm9, 400(%rsp)
vmovdqa %xmm10, 416(%rsp)
vmovdqa %xmm11, 432(%rsp)
vmovdqa %xmm12, 448(%rsp)
vmovdqa %xmm13, 464(%rsp)
vmovdqa %xmm14, 480(%rsp)
vmovdqa %xmm15, 496(%rsp)
vpunpckldq %xmm1, %xmm0, %xmm8
vpunpckldq %xmm3, %xmm2, %xmm9
vpunpckhdq %xmm1, %xmm0, %xmm12
vpunpckhdq %xmm3, %xmm2, %xmm13
vpunpckldq %xmm5, %xmm4, %xmm10
vpunpckldq %xmm7, %xmm6, %xmm11
vpunpckhdq %xmm5, %xmm4, %xmm14
vpunpckhdq %xmm7, %xmm6, %xmm15
vpunpcklqdq %xmm9, %xmm8, %xmm0
vpunpcklqdq %xmm11, %xmm10, %xmm1
vpunpckhqdq %xmm9, %xmm8, %xmm2
vpunpckhqdq %xmm11, %xmm10, %xmm3
vpunpcklqdq %xmm13, %xmm12, %xmm4
vpunpcklqdq %xmm15, %xmm14, %xmm5
vpunpckhqdq %xmm13, %xmm12, %xmm6
vpunpckhqdq %xmm15, %xmm14, %xmm7
andq %rsi, %rsi
jz chacha_blocks_xop_noinput1
vpxor 0(%rsi), %xmm0, %xmm0
vpxor 16(%rsi), %xmm1, %xmm1
vpxor 64(%rsi), %xmm2, %xmm2
vpxor 80(%rsi), %xmm3, %xmm3
vpxor 128(%rsi), %xmm4, %xmm4
vpxor 144(%rsi), %xmm5, %xmm5
vpxor 192(%rsi), %xmm6, %xmm6
vpxor 208(%rsi), %xmm7, %xmm7
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 64(%rdx)
vmovdqu %xmm3, 80(%rdx)
vmovdqu %xmm4, 128(%rdx)
vmovdqu %xmm5, 144(%rdx)
vmovdqu %xmm6, 192(%rdx)
vmovdqu %xmm7, 208(%rdx)
vmovdqa 384(%rsp), %xmm0
vmovdqa 400(%rsp), %xmm1
vmovdqa 416(%rsp), %xmm2
vmovdqa 432(%rsp), %xmm3
vmovdqa 448(%rsp), %xmm4
vmovdqa 464(%rsp), %xmm5
vmovdqa 480(%rsp), %xmm6
vmovdqa 496(%rsp), %xmm7
vpunpckldq %xmm1, %xmm0, %xmm8
vpunpckldq %xmm3, %xmm2, %xmm9
vpunpckhdq %xmm1, %xmm0, %xmm12
vpunpckhdq %xmm3, %xmm2, %xmm13
vpunpckldq %xmm5, %xmm4, %xmm10
vpunpckldq %xmm7, %xmm6, %xmm11
vpunpckhdq %xmm5, %xmm4, %xmm14
vpunpckhdq %xmm7, %xmm6, %xmm15
vpunpcklqdq %xmm9, %xmm8, %xmm0
vpunpcklqdq %xmm11, %xmm10, %xmm1
vpunpckhqdq %xmm9, %xmm8, %xmm2
vpunpckhqdq %xmm11, %xmm10, %xmm3
vpunpcklqdq %xmm13, %xmm12, %xmm4
vpunpcklqdq %xmm15, %xmm14, %xmm5
vpunpckhqdq %xmm13, %xmm12, %xmm6
vpunpckhqdq %xmm15, %xmm14, %xmm7
vpxor 32(%rsi), %xmm0, %xmm0
vpxor 48(%rsi), %xmm1, %xmm1
vpxor 96(%rsi), %xmm2, %xmm2
vpxor 112(%rsi), %xmm3, %xmm3
vpxor 160(%rsi), %xmm4, %xmm4
vpxor 176(%rsi), %xmm5, %xmm5
vpxor 224(%rsi), %xmm6, %xmm6
vpxor 240(%rsi), %xmm7, %xmm7
vmovdqu %xmm0, 32(%rdx)
vmovdqu %xmm1, 48(%rdx)
vmovdqu %xmm2, 96(%rdx)
vmovdqu %xmm3, 112(%rdx)
vmovdqu %xmm4, 160(%rdx)
vmovdqu %xmm5, 176(%rdx)
vmovdqu %xmm6, 224(%rdx)
vmovdqu %xmm7, 240(%rdx)
addq \$256, %rsi
jmp chacha_blocks_xop_mainloop_cont
chacha_blocks_xop_noinput1:
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 64(%rdx)
vmovdqu %xmm3, 80(%rdx)
vmovdqu %xmm4, 128(%rdx)
vmovdqu %xmm5, 144(%rdx)
vmovdqu %xmm6, 192(%rdx)
vmovdqu %xmm7, 208(%rdx)
vmovdqa 384(%rsp), %xmm0
vmovdqa 400(%rsp), %xmm1
vmovdqa 416(%rsp), %xmm2
vmovdqa 432(%rsp), %xmm3
vmovdqa 448(%rsp), %xmm4
vmovdqa 464(%rsp), %xmm5
vmovdqa 480(%rsp), %xmm6
vmovdqa 496(%rsp), %xmm7
vpunpckldq %xmm1, %xmm0, %xmm8
vpunpckldq %xmm3, %xmm2, %xmm9
vpunpckhdq %xmm1, %xmm0, %xmm12
vpunpckhdq %xmm3, %xmm2, %xmm13
vpunpckldq %xmm5, %xmm4, %xmm10
vpunpckldq %xmm7, %xmm6, %xmm11
vpunpckhdq %xmm5, %xmm4, %xmm14
vpunpckhdq %xmm7, %xmm6, %xmm15
vpunpcklqdq %xmm9, %xmm8, %xmm0
vpunpcklqdq %xmm11, %xmm10, %xmm1
vpunpckhqdq %xmm9, %xmm8, %xmm2
vpunpckhqdq %xmm11, %xmm10, %xmm3
vpunpcklqdq %xmm13, %xmm12, %xmm4
vpunpcklqdq %xmm15, %xmm14, %xmm5
vpunpckhqdq %xmm13, %xmm12, %xmm6
vpunpckhqdq %xmm15, %xmm14, %xmm7
vmovdqu %xmm0, 32(%rdx)
vmovdqu %xmm1, 48(%rdx)
vmovdqu %xmm2, 96(%rdx)
vmovdqu %xmm3, 112(%rdx)
vmovdqu %xmm4, 160(%rdx)
vmovdqu %xmm5, 176(%rdx)
vmovdqu %xmm6, 224(%rdx)
vmovdqu %xmm7, 240(%rdx)
chacha_blocks_xop_mainloop_cont:
addq \$256, %rdx
subq \$256, %rcx
cmp \$256, %rcx
jae chacha_blocks_xop_atleast256
movq %rbx, 48(%rsp)
vmovdqa 0(%rsp), %xmm8
vmovdqa 16(%rsp), %xmm9
vmovdqa 32(%rsp), %xmm10
vmovdqa 48(%rsp), %xmm11
movq \$1, %r9
cmp \$128, %rcx
jb chacha_blocks_xop_below128_fixup
chacha_blocks_xop_atleast128:
vmovd %r9, %xmm5
vmovdqa %xmm8, %xmm0
vmovdqa %xmm9, %xmm1
vmovdqa %xmm10, %xmm2
vmovdqa %xmm11, %xmm3
vpaddq %xmm5, %xmm11, %xmm11
vmovdqa %xmm8, 384(%rsp)
vmovdqa %xmm9, 400(%rsp)
vmovdqa %xmm10, 416(%rsp)
vmovdqa %xmm3, 432(%rsp)
vmovdqa %xmm11, 448(%rsp)
movq 64(%rsp), %rax
chacha_blocks_xop_mainloop2:
vpaddd %xmm0, %xmm1, %xmm0
vpaddd %xmm8, %xmm9, %xmm8
vpxor %xmm3, %xmm0, %xmm3
vpxor %xmm11, %xmm8, %xmm11
vprotd \$16, %xmm3, %xmm3
vprotd \$16, %xmm11, %xmm11
vpaddd %xmm2, %xmm3, %xmm2
vpaddd %xmm10, %xmm11, %xmm10
vpxor %xmm1, %xmm2, %xmm1
vpxor %xmm9, %xmm10, %xmm9
vprotd \$12, %xmm1, %xmm1
vprotd \$12, %xmm9, %xmm9
vpaddd %xmm0, %xmm1, %xmm0
vpaddd %xmm8, %xmm9, %xmm8
vpxor %xmm3, %xmm0, %xmm3
vpxor %xmm11, %xmm8, %xmm11
vprotd \$8, %xmm3, %xmm3
vprotd \$8, %xmm11, %xmm11
vpshufd \$0x93, %xmm0, %xmm0
vpshufd \$0x93, %xmm8, %xmm8
vpaddd %xmm2, %xmm3, %xmm2
vpaddd %xmm10, %xmm11, %xmm10
vpshufd \$0x4e, %xmm3, %xmm3
vpshufd \$0x4e, %xmm11, %xmm11
vpxor %xmm1, %xmm2, %xmm1
vpxor %xmm9, %xmm10, %xmm9
vpshufd \$0x39, %xmm2, %xmm2
vpshufd \$0x39, %xmm10, %xmm10
vprotd \$7, %xmm1, %xmm1
vprotd \$7, %xmm9, %xmm9
vpaddd %xmm0, %xmm1, %xmm0
vpaddd %xmm8, %xmm9, %xmm8
vpxor %xmm3, %xmm0, %xmm3
vpxor %xmm11, %xmm8, %xmm11
vprotd \$16, %xmm3, %xmm3
vprotd \$16, %xmm11, %xmm11
vpaddd %xmm2, %xmm3, %xmm2
vpaddd %xmm10, %xmm11, %xmm10
vpxor %xmm1, %xmm2, %xmm1
vpxor %xmm9, %xmm10, %xmm9
vprotd \$12, %xmm1, %xmm1
vprotd \$12, %xmm9, %xmm9
vpaddd %xmm0, %xmm1, %xmm0
vpaddd %xmm8, %xmm9, %xmm8
vpxor %xmm3, %xmm0, %xmm3
vpxor %xmm11, %xmm8, %xmm11
vprotd \$8, %xmm3, %xmm3
vprotd \$8, %xmm11, %xmm11
vpshufd \$0x39, %xmm0, %xmm0
vpshufd \$0x39, %xmm8, %xmm8
vpaddd %xmm2, %xmm3, %xmm2
vpaddd %xmm10, %xmm11, %xmm10
vpshufd \$0x4e, %xmm3, %xmm3
vpshufd \$0x4e, %xmm11, %xmm11
vpxor %xmm1, %xmm2, %xmm1
vpxor %xmm9, %xmm10, %xmm9
vpshufd \$0x93, %xmm2, %xmm2
vpshufd \$0x93, %xmm10, %xmm10
vprotd \$7, %xmm1, %xmm1
vprotd \$7, %xmm9, %xmm9
subq \$2, %rax
jnz chacha_blocks_xop_mainloop2
vmovdqa 384(%rsp), %xmm12
vmovdqa 400(%rsp), %xmm13
vmovdqa 416(%rsp), %xmm14
vmovdqa 432(%rsp), %xmm15
vmovdqa 448(%rsp), %xmm5
vpaddd %xmm12, %xmm0, %xmm0
vpaddd %xmm13, %xmm1, %xmm1
vpaddd %xmm14, %xmm2, %xmm2
vpaddd %xmm15, %xmm3, %xmm3
vpaddd %xmm12, %xmm8, %xmm8
vpaddd %xmm13, %xmm9, %xmm9
vpaddd %xmm14, %xmm10, %xmm10
vpaddd %xmm5, %xmm11, %xmm11
andq %rsi, %rsi
jz chacha_blocks_xop_noinput2
vpxor 0(%rsi), %xmm0, %xmm0
vpxor 16(%rsi), %xmm1, %xmm1
vpxor 32(%rsi), %xmm2, %xmm2
vpxor 48(%rsi), %xmm3, %xmm3
vpxor 64(%rsi), %xmm8, %xmm8
vpxor 80(%rsi), %xmm9, %xmm9
vpxor 96(%rsi), %xmm10, %xmm10
vpxor 112(%rsi), %xmm11, %xmm11
addq \$128, %rsi
chacha_blocks_xop_noinput2:
addq \$2, %rbx
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 32(%rdx)
vmovdqu %xmm3, 48(%rdx)
vmovdqu %xmm8, 64(%rdx)
vmovdqu %xmm9, 80(%rdx)
vmovdqu %xmm10, 96(%rdx)
vmovdqu %xmm11, 112(%rdx)
addq \$128, %rdx
subq \$128, %rcx
movq %rbx, 48(%rsp)
vmovdqa 0(%rsp), %xmm8
vmovdqa 16(%rsp), %xmm9
vmovdqa 32(%rsp), %xmm10
vmovdqa 48(%rsp), %xmm11
movq \$1, %r9
chacha_blocks_xop_below128_fixup:
vmovd %r9, %xmm5
chacha_blocks_xop_below128:
andq %rcx, %rcx
jz chacha_blocks_xop_done
cmpq \$64, %rcx
jae chacha_blocks_xop_above63
movq %rdx, %r9
andq %rsi, %rsi
jz chacha_blocks_xop_noinput3
movq %rcx, %r10
movq %rsp, %rdx
addq %r10, %rsi
addq %r10, %rdx
negq %r10
chacha_blocks_xop_copyinput:
movb (%rsi, %r10), %al
movb %al, (%rdx, %r10)
incq %r10
jnz chacha_blocks_xop_copyinput
movq %rsp, %rsi
chacha_blocks_xop_noinput3:
movq %rsp, %rdx
chacha_blocks_xop_above63:
vmovdqa %xmm8, %xmm0
vmovdqa %xmm9, %xmm1
vmovdqa %xmm10, %xmm2
vmovdqa %xmm11, %xmm3
movq 64(%rsp), %rax
chacha_blocks_xop_mainloop3:
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vprotd \$16, %xmm3, %xmm3
vpaddd %xmm2, %xmm3, %xmm2
vpxor %xmm1, %xmm2, %xmm1
vprotd \$12, %xmm1, %xmm1
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vprotd \$8, %xmm3, %xmm3
vpaddd %xmm2, %xmm3, %xmm2
vpshufd \$0x93, %xmm0, %xmm0
vpxor %xmm1, %xmm2, %xmm1
vprotd \$7, %xmm1, %xmm1
vpshufd \$0x4e, %xmm3, %xmm3
vpaddd %xmm0, %xmm1, %xmm0
vpshufd \$0x39, %xmm2, %xmm2
vpxor %xmm3, %xmm0, %xmm3
vprotd \$16, %xmm3, %xmm3
vpaddd %xmm2, %xmm3, %xmm2
vpxor %xmm1, %xmm2, %xmm1
vprotd \$12, %xmm1, %xmm1
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vprotd \$8, %xmm3, %xmm3
vpaddd %xmm2, %xmm3, %xmm2
vpxor %xmm1, %xmm2, %xmm1
vpshufd \$0x39, %xmm0, %xmm0
vprotd \$7, %xmm1, %xmm1
vpshufd \$0x4e, %xmm3, %xmm3
vpshufd \$0x93, %xmm2, %xmm2
subq \$2, %rax
jnz chacha_blocks_xop_mainloop3
vpaddd %xmm0, %xmm8, %xmm0
vpaddd %xmm1, %xmm9, %xmm1
vpaddd %xmm2, %xmm10, %xmm2
vpaddd %xmm3, %xmm11, %xmm3
andq %rsi, %rsi
jz chacha_blocks_xop_noinput4
vpxor 0(%rsi), %xmm0, %xmm0
vpxor 16(%rsi), %xmm1, %xmm1
vpxor 32(%rsi), %xmm2, %xmm2
vpxor 48(%rsi), %xmm3, %xmm3
addq \$64, %rsi
chacha_blocks_xop_noinput4:
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 32(%rdx)
vmovdqu %xmm3, 48(%rdx)
vpaddq %xmm11, %xmm5, %xmm11
addq \$1, %rbx
cmpq \$64, %rcx
jbe chacha_blocks_xop_mainloop3_finishup
addq \$64, %rdx
subq \$64, %rcx
jmp chacha_blocks_xop_below128
chacha_blocks_xop_mainloop3_finishup:
cmpq \$64, %rcx
je chacha_blocks_xop_done
addq %rcx, %r9
addq %rcx, %rdx
negq %rcx
chacha_blocks_xop_copyoutput:
movb (%rdx, %rcx), %al
movb %al, (%r9, %rcx)
incq %rcx
jnz chacha_blocks_xop_copyoutput
chacha_blocks_xop_done:
movq %rbx, 32(%rdi)
movq %rbp, %rsp
popq %rbp
popq %rbx
ret
___
);





add_function(
	'chacha_blocks', 4, true, 'avx2',
<<___
vzeroupper
pushq %rbx
pushq %rbp
pushq %r12
pushq %r13
pushq %r14
movq %rsp, %rbp
andq \$~63, %rsp
subq \$512, %rsp
leaq chacha_constants(%rip), %rax
vbroadcasti128 0(%rax), %ymm8
vbroadcasti128 16(%rax), %ymm6
vbroadcasti128 32(%rax), %ymm7
vbroadcasti128 0(%rdi), %ymm9
vbroadcasti128 16(%rdi), %ymm10
vbroadcasti128 32(%rdi), %ymm11
movq 32(%rdi), %r14
movq 48(%rdi), %rax
vmovdqa %xmm8, 0(%rsp)
vmovdqa %xmm9, 16(%rsp)
vmovdqa %xmm10, 32(%rsp)
vmovdqa %xmm11, 48(%rsp)
movq %rax, 64(%rsp)
vmovdqa %ymm6, 448(%rsp)
vmovdqa %ymm7, 480(%rsp)
cmp \$128, %rcx
jb chacha_blocks_avx2_below128_fixup
cmp \$256, %rcx
jb chacha_blocks_avx2_atleast128
cmp \$512, %rcx
jb chacha_blocks_avx2_atleast256
.align 64
chacha_blocks_avx2_atleast512:
movq %r14, %rax
leaq 1(%r14), %r8
leaq 2(%r14), %r9
leaq 3(%r14), %r10
leaq 4(%r14), %rbx
leaq 5(%r14), %r11
leaq 6(%r14), %r12
leaq 7(%r14), %r13
movl %eax, 128(%rsp)
movl %r8d, 4+128(%rsp)
movl %r9d, 8+128(%rsp)
movl %r10d, 12+128(%rsp)
movl %ebx, 16+128(%rsp)
movl %r11d, 20+128(%rsp)
movl %r12d, 24+128(%rsp)
movl %r13d, 28+128(%rsp)
addq \$8, %r14
shrq \$32, %rax
shrq \$32, %r8
shrq \$32, %r9
shrq \$32, %r10
shrq \$32, %rbx
shrq \$32, %r11
shrq \$32, %r12
shrq \$32, %r13
movl %eax, 160(%rsp)
movl %r8d, 4+160(%rsp)
movl %r9d, 8+160(%rsp)
movl %r10d, 12+160(%rsp)
movl %ebx, 16+160(%rsp)
movl %r11d, 20+160(%rsp)
movl %r12d, 24+160(%rsp)
movl %r13d, 28+160(%rsp)
movq 64(%rsp), %rax
vpbroadcastd 0(%rsp), %ymm0
vpbroadcastd 4+0(%rsp), %ymm1
vpbroadcastd 8+0(%rsp), %ymm2
vpbroadcastd 12+0(%rsp), %ymm3
vpbroadcastd 16(%rsp), %ymm4
vpbroadcastd 4+16(%rsp), %ymm5
vpbroadcastd 8+16(%rsp), %ymm6
vpbroadcastd 12+16(%rsp), %ymm7
vpbroadcastd 32(%rsp), %ymm8
vpbroadcastd 4+32(%rsp), %ymm9
vpbroadcastd 8+32(%rsp), %ymm10
vpbroadcastd 12+32(%rsp), %ymm11
vpbroadcastd 8+48(%rsp), %ymm14
vpbroadcastd 12+48(%rsp), %ymm15
vmovdqa 128(%rsp), %ymm12
vmovdqa 160(%rsp), %ymm13
chacha_blocks_avx2_mainloop1:
vpaddd %ymm0, %ymm4, %ymm0
vpaddd %ymm1, %ymm5, %ymm1
vpxor %ymm12, %ymm0, %ymm12
vpxor %ymm13, %ymm1, %ymm13
vpaddd %ymm2, %ymm6, %ymm2
vpaddd %ymm3, %ymm7, %ymm3
vpxor %ymm14, %ymm2, %ymm14
vpxor %ymm15, %ymm3, %ymm15
vpshufb 448(%rsp), %ymm12, %ymm12
vpshufb 448(%rsp), %ymm13, %ymm13
vpaddd %ymm8, %ymm12, %ymm8
vpaddd %ymm9, %ymm13, %ymm9
vpshufb 448(%rsp), %ymm14, %ymm14
vpshufb 448(%rsp), %ymm15, %ymm15
vpaddd %ymm10, %ymm14, %ymm10
vpaddd %ymm11, %ymm15, %ymm11
vmovdqa %ymm12, 96(%rsp)
vpxor %ymm4, %ymm8, %ymm4
vpxor %ymm5, %ymm9, %ymm5
vpslld \$ 12, %ymm4, %ymm12
vpsrld \$20, %ymm4, %ymm4
vpxor %ymm4, %ymm12, %ymm4
vpslld \$ 12, %ymm5, %ymm12
vpsrld \$20, %ymm5, %ymm5
vpxor %ymm5, %ymm12, %ymm5
vpxor %ymm6, %ymm10, %ymm6
vpxor %ymm7, %ymm11, %ymm7
vpslld \$ 12, %ymm6, %ymm12
vpsrld \$20, %ymm6, %ymm6
vpxor %ymm6, %ymm12, %ymm6
vpslld \$ 12, %ymm7, %ymm12
vpsrld \$20, %ymm7, %ymm7
vpxor %ymm7, %ymm12, %ymm7
vpaddd %ymm0, %ymm4, %ymm0
vpaddd %ymm1, %ymm5, %ymm1
vpxor 96(%rsp), %ymm0, %ymm12
vpxor %ymm13, %ymm1, %ymm13
vpaddd %ymm2, %ymm6, %ymm2
vpaddd %ymm3, %ymm7, %ymm3
vpxor %ymm14, %ymm2, %ymm14
vpxor %ymm15, %ymm3, %ymm15
vpshufb 480(%rsp), %ymm12, %ymm12
vpshufb 480(%rsp), %ymm13, %ymm13
vpaddd %ymm8, %ymm12, %ymm8
vpaddd %ymm9, %ymm13, %ymm9
vpshufb 480(%rsp), %ymm14, %ymm14
vpshufb 480(%rsp), %ymm15, %ymm15
vpaddd %ymm10, %ymm14, %ymm10
vpaddd %ymm11, %ymm15, %ymm11
vmovdqa %ymm12, 96(%rsp)
vpxor %ymm4, %ymm8, %ymm4
vpxor %ymm5, %ymm9, %ymm5
vpslld \$ 7, %ymm4, %ymm12
vpsrld \$25, %ymm4, %ymm4
vpxor %ymm4, %ymm12, %ymm4
vpslld \$ 7, %ymm5, %ymm12
vpsrld \$25, %ymm5, %ymm5
vpxor %ymm5, %ymm12, %ymm5
vpxor %ymm6, %ymm10, %ymm6
vpxor %ymm7, %ymm11, %ymm7
vpslld \$ 7, %ymm6, %ymm12
vpsrld \$25, %ymm6, %ymm6
vpxor %ymm6, %ymm12, %ymm6
vpslld \$ 7, %ymm7, %ymm12
vpsrld \$25, %ymm7, %ymm7
vpxor %ymm7, %ymm12, %ymm7
vpaddd %ymm0, %ymm5, %ymm0
vpaddd %ymm1, %ymm6, %ymm1
vpxor %ymm15, %ymm0, %ymm15
vpxor 96(%rsp), %ymm1, %ymm12
vpaddd %ymm2, %ymm7, %ymm2
vpaddd %ymm3, %ymm4, %ymm3
vpxor %ymm13, %ymm2, %ymm13
vpxor %ymm14, %ymm3, %ymm14
vpshufb 448(%rsp), %ymm15, %ymm15
vpshufb 448(%rsp), %ymm12, %ymm12
vpaddd %ymm10, %ymm15, %ymm10
vpaddd %ymm11, %ymm12, %ymm11
vpshufb 448(%rsp), %ymm13, %ymm13
vpshufb 448(%rsp), %ymm14, %ymm14
vpaddd %ymm8, %ymm13, %ymm8
vpaddd %ymm9, %ymm14, %ymm9
vmovdqa %ymm15, 96(%rsp)
vpxor %ymm5, %ymm10, %ymm5
vpxor %ymm6, %ymm11, %ymm6
vpslld \$ 12, %ymm5, %ymm15
vpsrld \$20, %ymm5, %ymm5
vpxor %ymm5, %ymm15, %ymm5
vpslld \$ 12, %ymm6, %ymm15
vpsrld \$20, %ymm6, %ymm6
vpxor %ymm6, %ymm15, %ymm6
vpxor %ymm7, %ymm8, %ymm7
vpxor %ymm4, %ymm9, %ymm4
vpslld \$ 12, %ymm7, %ymm15
vpsrld \$20, %ymm7, %ymm7
vpxor %ymm7, %ymm15, %ymm7
vpslld \$ 12, %ymm4, %ymm15
vpsrld \$20, %ymm4, %ymm4
vpxor %ymm4, %ymm15, %ymm4
vpaddd %ymm0, %ymm5, %ymm0
vpaddd %ymm1, %ymm6, %ymm1
vpxor 96(%rsp), %ymm0, %ymm15
vpxor %ymm12, %ymm1, %ymm12
vpaddd %ymm2, %ymm7, %ymm2
vpaddd %ymm3, %ymm4, %ymm3
vpxor %ymm13, %ymm2, %ymm13
vpxor %ymm14, %ymm3, %ymm14
vpshufb 480(%rsp), %ymm15, %ymm15
vpshufb 480(%rsp), %ymm12, %ymm12
vpaddd %ymm10, %ymm15, %ymm10
vpaddd %ymm11, %ymm12, %ymm11
vpshufb 480(%rsp), %ymm13, %ymm13
vpshufb 480(%rsp), %ymm14, %ymm14
vpaddd %ymm8, %ymm13, %ymm8
vpaddd %ymm9, %ymm14, %ymm9
vmovdqa %ymm15, 96(%rsp)
vpxor %ymm5, %ymm10, %ymm5
vpxor %ymm6, %ymm11, %ymm6
vpslld \$ 7, %ymm5, %ymm15
vpsrld \$25, %ymm5, %ymm5
vpxor %ymm5, %ymm15, %ymm5
vpslld \$ 7, %ymm6, %ymm15
vpsrld \$25, %ymm6, %ymm6
vpxor %ymm6, %ymm15, %ymm6
vpxor %ymm7, %ymm8, %ymm7
vpxor %ymm4, %ymm9, %ymm4
vpslld \$ 7, %ymm7, %ymm15
vpsrld \$25, %ymm7, %ymm7
vpxor %ymm7, %ymm15, %ymm7
vpslld \$ 7, %ymm4, %ymm15
vpsrld \$25, %ymm4, %ymm4
vpxor %ymm4, %ymm15, %ymm4
vmovdqa 96(%rsp), %ymm15
subq \$2, %rax
jnz chacha_blocks_avx2_mainloop1
vmovdqa %ymm8, 192(%rsp)
vmovdqa %ymm9, 224(%rsp)
vmovdqa %ymm10, 256(%rsp)
vmovdqa %ymm11, 288(%rsp)
vmovdqa %ymm12, 320(%rsp)
vmovdqa %ymm13, 352(%rsp)
vmovdqa %ymm14, 384(%rsp)
vmovdqa %ymm15, 416(%rsp)
vpbroadcastd 0(%rsp), %ymm8
vpbroadcastd 4+0(%rsp), %ymm9
vpbroadcastd 8+0(%rsp), %ymm10
vpbroadcastd 12+0(%rsp), %ymm11
vpbroadcastd 16(%rsp), %ymm12
vpbroadcastd 4+16(%rsp), %ymm13
vpbroadcastd 8+16(%rsp), %ymm14
vpbroadcastd 12+16(%rsp), %ymm15
vpaddd %ymm8, %ymm0, %ymm0
vpaddd %ymm9, %ymm1, %ymm1
vpaddd %ymm10, %ymm2, %ymm2
vpaddd %ymm11, %ymm3, %ymm3
vpaddd %ymm12, %ymm4, %ymm4
vpaddd %ymm13, %ymm5, %ymm5
vpaddd %ymm14, %ymm6, %ymm6
vpaddd %ymm15, %ymm7, %ymm7
vpunpckldq %ymm1, %ymm0, %ymm8
vpunpckldq %ymm3, %ymm2, %ymm9
vpunpckhdq %ymm1, %ymm0, %ymm12
vpunpckhdq %ymm3, %ymm2, %ymm13
vpunpckldq %ymm5, %ymm4, %ymm10
vpunpckldq %ymm7, %ymm6, %ymm11
vpunpckhdq %ymm5, %ymm4, %ymm14
vpunpckhdq %ymm7, %ymm6, %ymm15
vpunpcklqdq %ymm9, %ymm8, %ymm0
vpunpcklqdq %ymm11, %ymm10, %ymm1
vpunpckhqdq %ymm9, %ymm8, %ymm2
vpunpckhqdq %ymm11, %ymm10, %ymm3
vpunpcklqdq %ymm13, %ymm12, %ymm4
vpunpcklqdq %ymm15, %ymm14, %ymm5
vpunpckhqdq %ymm13, %ymm12, %ymm6
vpunpckhqdq %ymm15, %ymm14, %ymm7
vperm2i128 \$0x20, %ymm1, %ymm0, %ymm8
vperm2i128 \$0x20, %ymm3, %ymm2, %ymm9
vperm2i128 \$0x31, %ymm1, %ymm0, %ymm12
vperm2i128 \$0x31, %ymm3, %ymm2, %ymm13
vperm2i128 \$0x20, %ymm5, %ymm4, %ymm10
vperm2i128 \$0x20, %ymm7, %ymm6, %ymm11
vperm2i128 \$0x31, %ymm5, %ymm4, %ymm14
vperm2i128 \$0x31, %ymm7, %ymm6, %ymm15
andq %rsi, %rsi
jz chacha_blocks_avx2_noinput1
vpxor 0(%rsi), %ymm8, %ymm8
vpxor 64(%rsi), %ymm9, %ymm9
vpxor 128(%rsi), %ymm10, %ymm10
vpxor 192(%rsi), %ymm11, %ymm11
vpxor 256(%rsi), %ymm12, %ymm12
vpxor 320(%rsi), %ymm13, %ymm13
vpxor 384(%rsi), %ymm14, %ymm14
vpxor 448(%rsi), %ymm15, %ymm15
vmovdqu %ymm8, 0(%rdx)
vmovdqu %ymm9, 64(%rdx)
vmovdqu %ymm10, 128(%rdx)
vmovdqu %ymm11, 192(%rdx)
vmovdqu %ymm12, 256(%rdx)
vmovdqu %ymm13, 320(%rdx)
vmovdqu %ymm14, 384(%rdx)
vmovdqu %ymm15, 448(%rdx)
vmovdqa 192(%rsp), %ymm0
vmovdqa 224(%rsp), %ymm1
vmovdqa 256(%rsp), %ymm2
vmovdqa 288(%rsp), %ymm3
vmovdqa 320(%rsp), %ymm4
vmovdqa 352(%rsp), %ymm5
vmovdqa 384(%rsp), %ymm6
vmovdqa 416(%rsp), %ymm7
vpbroadcastd 32(%rsp), %ymm8
vpbroadcastd 4+32(%rsp), %ymm9
vpbroadcastd 8+32(%rsp), %ymm10
vpbroadcastd 12+32(%rsp), %ymm11
vmovdqa 128(%rsp), %ymm12
vmovdqa 160(%rsp), %ymm13
vpbroadcastd 8+48(%rsp), %ymm14
vpbroadcastd 12+48(%rsp), %ymm15
vpaddd %ymm8, %ymm0, %ymm0
vpaddd %ymm9, %ymm1, %ymm1
vpaddd %ymm10, %ymm2, %ymm2
vpaddd %ymm11, %ymm3, %ymm3
vpaddd %ymm12, %ymm4, %ymm4
vpaddd %ymm13, %ymm5, %ymm5
vpaddd %ymm14, %ymm6, %ymm6
vpaddd %ymm15, %ymm7, %ymm7
vpunpckldq %ymm1, %ymm0, %ymm8
vpunpckldq %ymm3, %ymm2, %ymm9
vpunpckhdq %ymm1, %ymm0, %ymm12
vpunpckhdq %ymm3, %ymm2, %ymm13
vpunpckldq %ymm5, %ymm4, %ymm10
vpunpckldq %ymm7, %ymm6, %ymm11
vpunpckhdq %ymm5, %ymm4, %ymm14
vpunpckhdq %ymm7, %ymm6, %ymm15
vpunpcklqdq %ymm9, %ymm8, %ymm0
vpunpcklqdq %ymm11, %ymm10, %ymm1
vpunpckhqdq %ymm9, %ymm8, %ymm2
vpunpckhqdq %ymm11, %ymm10, %ymm3
vpunpcklqdq %ymm13, %ymm12, %ymm4
vpunpcklqdq %ymm15, %ymm14, %ymm5
vpunpckhqdq %ymm13, %ymm12, %ymm6
vpunpckhqdq %ymm15, %ymm14, %ymm7
vperm2i128 \$0x20, %ymm1, %ymm0, %ymm8
vperm2i128 \$0x20, %ymm3, %ymm2, %ymm9
vperm2i128 \$0x31, %ymm1, %ymm0, %ymm12
vperm2i128 \$0x31, %ymm3, %ymm2, %ymm13
vperm2i128 \$0x20, %ymm5, %ymm4, %ymm10
vperm2i128 \$0x20, %ymm7, %ymm6, %ymm11
vperm2i128 \$0x31, %ymm5, %ymm4, %ymm14
vperm2i128 \$0x31, %ymm7, %ymm6, %ymm15
vpxor 32(%rsi), %ymm8, %ymm8
vpxor 96(%rsi), %ymm9, %ymm9
vpxor 160(%rsi), %ymm10, %ymm10
vpxor 224(%rsi), %ymm11, %ymm11
vpxor 288(%rsi), %ymm12, %ymm12
vpxor 352(%rsi), %ymm13, %ymm13
vpxor 416(%rsi), %ymm14, %ymm14
vpxor 480(%rsi), %ymm15, %ymm15
vmovdqu %ymm8, 32(%rdx)
vmovdqu %ymm9, 96(%rdx)
vmovdqu %ymm10, 160(%rdx)
vmovdqu %ymm11, 224(%rdx)
vmovdqu %ymm12, 288(%rdx)
vmovdqu %ymm13, 352(%rdx)
vmovdqu %ymm14, 416(%rdx)
vmovdqu %ymm15, 480(%rdx)
addq \$512, %rsi
jmp chacha_blocks_avx2_mainloop1_cont
chacha_blocks_avx2_noinput1:
vmovdqu %ymm8, 0(%rdx)
vmovdqu %ymm9, 64(%rdx)
vmovdqu %ymm10, 128(%rdx)
vmovdqu %ymm11, 192(%rdx)
vmovdqu %ymm12, 256(%rdx)
vmovdqu %ymm13, 320(%rdx)
vmovdqu %ymm14, 384(%rdx)
vmovdqu %ymm15, 448(%rdx)
vmovdqa 192(%rsp), %ymm0
vmovdqa 224(%rsp), %ymm1
vmovdqa 256(%rsp), %ymm2
vmovdqa 288(%rsp), %ymm3
vmovdqa 320(%rsp), %ymm4
vmovdqa 352(%rsp), %ymm5
vmovdqa 384(%rsp), %ymm6
vmovdqa 416(%rsp), %ymm7
vpbroadcastd 32(%rsp), %ymm8
vpbroadcastd 4+32(%rsp), %ymm9
vpbroadcastd 8+32(%rsp), %ymm10
vpbroadcastd 12+32(%rsp), %ymm11
vmovdqa 128(%rsp), %ymm12
vmovdqa 160(%rsp), %ymm13
vpbroadcastd 8+48(%rsp), %ymm14
vpbroadcastd 12+48(%rsp), %ymm15
vpaddd %ymm8, %ymm0, %ymm0
vpaddd %ymm9, %ymm1, %ymm1
vpaddd %ymm10, %ymm2, %ymm2
vpaddd %ymm11, %ymm3, %ymm3
vpaddd %ymm12, %ymm4, %ymm4
vpaddd %ymm13, %ymm5, %ymm5
vpaddd %ymm14, %ymm6, %ymm6
vpaddd %ymm15, %ymm7, %ymm7
vpunpckldq %ymm1, %ymm0, %ymm8
vpunpckldq %ymm3, %ymm2, %ymm9
vpunpckhdq %ymm1, %ymm0, %ymm12
vpunpckhdq %ymm3, %ymm2, %ymm13
vpunpckldq %ymm5, %ymm4, %ymm10
vpunpckldq %ymm7, %ymm6, %ymm11
vpunpckhdq %ymm5, %ymm4, %ymm14
vpunpckhdq %ymm7, %ymm6, %ymm15
vpunpcklqdq %ymm9, %ymm8, %ymm0
vpunpcklqdq %ymm11, %ymm10, %ymm1
vpunpckhqdq %ymm9, %ymm8, %ymm2
vpunpckhqdq %ymm11, %ymm10, %ymm3
vpunpcklqdq %ymm13, %ymm12, %ymm4
vpunpcklqdq %ymm15, %ymm14, %ymm5
vpunpckhqdq %ymm13, %ymm12, %ymm6
vpunpckhqdq %ymm15, %ymm14, %ymm7
vperm2i128 \$0x20, %ymm1, %ymm0, %ymm8
vperm2i128 \$0x20, %ymm3, %ymm2, %ymm9
vperm2i128 \$0x31, %ymm1, %ymm0, %ymm12
vperm2i128 \$0x31, %ymm3, %ymm2, %ymm13
vperm2i128 \$0x20, %ymm5, %ymm4, %ymm10
vperm2i128 \$0x20, %ymm7, %ymm6, %ymm11
vperm2i128 \$0x31, %ymm5, %ymm4, %ymm14
vperm2i128 \$0x31, %ymm7, %ymm6, %ymm15
vmovdqu %ymm8, 32(%rdx)
vmovdqu %ymm9, 96(%rdx)
vmovdqu %ymm10, 160(%rdx)
vmovdqu %ymm11, 224(%rdx)
vmovdqu %ymm12, 288(%rdx)
vmovdqu %ymm13, 352(%rdx)
vmovdqu %ymm14, 416(%rdx)
vmovdqu %ymm15, 480(%rdx)
chacha_blocks_avx2_mainloop1_cont:
addq \$512, %rdx
subq \$512, %rcx
cmp \$512, %rcx
jae chacha_blocks_avx2_atleast512
movq %r14, 48(%rsp)
vbroadcasti128 0(%rsp), %ymm8
vbroadcasti128 16(%rsp), %ymm9
vbroadcasti128 32(%rsp), %ymm10
vbroadcasti128 48(%rsp), %ymm11
vbroadcasti128 448(%rsp), %ymm6
vbroadcasti128 480(%rsp), %ymm7
cmp \$128, %rcx
jb chacha_blocks_avx2_below128_fixup
cmp \$256, %rcx
jb chacha_blocks_avx2_atleast128
chacha_blocks_avx2_atleast256:
movq \$1, %r9
movq \$2, %r10
vpxor %ymm13, %ymm13, %ymm13
vmovd %r9, %xmm14
vmovd %r10, %xmm15
vinserti128 \$1, %xmm14, %ymm13, %ymm4
vinserti128 \$1, %xmm15, %ymm15, %ymm5
vmovdqa %ymm8, %ymm0
vmovdqa %ymm9, %ymm1
vmovdqa %ymm10, %ymm2
vpaddq %ymm4, %ymm11, %ymm3
vpaddq %ymm5, %ymm3, %ymm11
vmovdqa %ymm8, 192(%rsp)
vmovdqa %ymm9, 224(%rsp)
vmovdqa %ymm10, 256(%rsp)
vmovdqa %ymm3, 288(%rsp)
vmovdqa %ymm11, 320(%rsp)
movq 64(%rsp), %rax
chacha_blocks_avx2_mainloop2:
vpaddd %ymm0, %ymm1, %ymm0
vpaddd %ymm8, %ymm9, %ymm8
vpxor %ymm3, %ymm0, %ymm3
vpxor %ymm11, %ymm8, %ymm11
vpshufb %ymm6, %ymm3, %ymm3
vpshufb %ymm6, %ymm11, %ymm11
vpaddd %ymm2, %ymm3, %ymm2
vpaddd %ymm10, %ymm11, %ymm10
vpxor %ymm1, %ymm2, %ymm1
vpxor %ymm9, %ymm10, %ymm9
vpslld \$ 12, %ymm1, %ymm4
vpsrld \$20, %ymm1, %ymm1
vpxor %ymm1, %ymm4, %ymm1
vpslld \$ 12, %ymm9, %ymm5
vpsrld \$20, %ymm9, %ymm9
vpxor %ymm9, %ymm5, %ymm9
vpaddd %ymm0, %ymm1, %ymm0
vpaddd %ymm8, %ymm9, %ymm8
vpxor %ymm3, %ymm0, %ymm3
vpxor %ymm11, %ymm8, %ymm11
vpshufb %ymm7, %ymm3, %ymm3
vpshufb %ymm7, %ymm11, %ymm11
vpshufd \$0x93, %ymm0, %ymm0
vpshufd \$0x93, %ymm8, %ymm8
vpaddd %ymm2, %ymm3, %ymm2
vpaddd %ymm10, %ymm11, %ymm10
vpshufd \$0x4e, %ymm3, %ymm3
vpshufd \$0x4e, %ymm11, %ymm11
vpxor %ymm1, %ymm2, %ymm1
vpxor %ymm9, %ymm10, %ymm9
vpshufd \$0x39, %ymm2, %ymm2
vpshufd \$0x39, %ymm10, %ymm10
vpslld \$ 7, %ymm1, %ymm4
vpsrld \$25, %ymm1, %ymm1
vpxor %ymm1, %ymm4, %ymm1
vpslld \$ 7, %ymm9, %ymm5
vpsrld \$25, %ymm9, %ymm9
vpxor %ymm9, %ymm5, %ymm9
vpaddd %ymm0, %ymm1, %ymm0
vpaddd %ymm8, %ymm9, %ymm8
vpxor %ymm3, %ymm0, %ymm3
vpxor %ymm11, %ymm8, %ymm11
vpshufb %ymm6, %ymm3, %ymm3
vpshufb %ymm6, %ymm11, %ymm11
vpaddd %ymm2, %ymm3, %ymm2
vpaddd %ymm10, %ymm11, %ymm10
vpxor %ymm1, %ymm2, %ymm1
vpxor %ymm9, %ymm10, %ymm9
vpslld \$ 12, %ymm1, %ymm4
vpsrld \$20, %ymm1, %ymm1
vpxor %ymm1, %ymm4, %ymm1
vpslld \$ 12, %ymm9, %ymm5
vpsrld \$20, %ymm9, %ymm9
vpxor %ymm9, %ymm5, %ymm9
vpaddd %ymm0, %ymm1, %ymm0
vpaddd %ymm8, %ymm9, %ymm8
vpxor %ymm3, %ymm0, %ymm3
vpxor %ymm11, %ymm8, %ymm11
vpshufb %ymm7, %ymm3, %ymm3
vpshufb %ymm7, %ymm11, %ymm11
vpshufd \$0x39, %ymm0, %ymm0
vpshufd \$0x39, %ymm8, %ymm8
vpaddd %ymm2, %ymm3, %ymm2
vpaddd %ymm10, %ymm11, %ymm10
vpshufd \$0x4e, %ymm3, %ymm3
vpshufd \$0x4e, %ymm11, %ymm11
vpxor %ymm1, %ymm2, %ymm1
vpxor %ymm9, %ymm10, %ymm9
vpshufd \$0x93, %ymm2, %ymm2
vpshufd \$0x93, %ymm10, %ymm10
vpslld \$ 7, %ymm1, %ymm4
vpsrld \$25, %ymm1, %ymm1
vpxor %ymm1, %ymm4, %ymm1
vpslld \$ 7, %ymm9, %ymm5
vpsrld \$25, %ymm9, %ymm9
vpxor %ymm9, %ymm5, %ymm9
subq \$2, %rax
jnz chacha_blocks_avx2_mainloop2
vmovdqa 192(%rsp), %ymm12
vmovdqa 224(%rsp), %ymm13
vmovdqa 256(%rsp), %ymm14
vmovdqa 288(%rsp), %ymm15
vmovdqa 320(%rsp), %ymm5
vpaddd %ymm12, %ymm0, %ymm0
vpaddd %ymm12, %ymm8, %ymm8
vpaddd %ymm13, %ymm1, %ymm1
vpaddd %ymm13, %ymm9, %ymm9
vpaddd %ymm14, %ymm2, %ymm2
vpaddd %ymm14, %ymm10, %ymm10
vpaddd %ymm15, %ymm3, %ymm3
vpaddd %ymm5, %ymm11, %ymm11
vperm2i128 \$2, %ymm0, %ymm1, %ymm12
vperm2i128 \$2, %ymm2, %ymm3, %ymm13
vperm2i128 \$19, %ymm0, %ymm1, %ymm14
vperm2i128 \$19, %ymm2, %ymm3, %ymm15
vperm2i128 \$2, %ymm8, %ymm9, %ymm0
vperm2i128 \$2, %ymm10, %ymm11, %ymm1
vperm2i128 \$19, %ymm8, %ymm9, %ymm2
vperm2i128 \$19, %ymm10, %ymm11, %ymm3
andq %rsi, %rsi
jz chacha_blocks_avx2_noinput2
vpxor 0(%rsi), %ymm12, %ymm12
vpxor 32(%rsi), %ymm13, %ymm13
vpxor 64(%rsi), %ymm14, %ymm14
vpxor 96(%rsi), %ymm15, %ymm15
vpxor 128(%rsi), %ymm0, %ymm0
vpxor 160(%rsi), %ymm1, %ymm1
vpxor 192(%rsi), %ymm2, %ymm2
vpxor 224(%rsi), %ymm3, %ymm3
addq \$256, %rsi
chacha_blocks_avx2_noinput2:
addq \$4, %r14
vmovdqu %ymm12, 0(%rdx)
vmovdqu %ymm13, 32(%rdx)
vmovdqu %ymm14, 64(%rdx)
vmovdqu %ymm15, 96(%rdx)
vmovdqu %ymm0, 128(%rdx)
vmovdqu %ymm1, 160(%rdx)
vmovdqu %ymm2, 192(%rdx)
vmovdqu %ymm3, 224(%rdx)
movq %r14, 48(%rsp)
addq \$256, %rdx
subq \$256, %rcx
vbroadcasti128 0(%rsp), %ymm8
vbroadcasti128 16(%rsp), %ymm9
vbroadcasti128 32(%rsp), %ymm10
vbroadcasti128 48(%rsp), %ymm11
cmp \$128, %rcx
jb chacha_blocks_avx2_below128_fixup
chacha_blocks_avx2_atleast128:
movq \$1, %r9
vpxor %ymm13, %ymm13, %ymm13
vmovd %r9, %xmm14
vinserti128 \$1, %xmm14, %ymm13, %ymm13
vpaddq %ymm13, %ymm11, %ymm11
vmovdqa %ymm8, %ymm0
vmovdqa %ymm9, %ymm1
vmovdqa %ymm10, %ymm2
vmovdqa %ymm11, %ymm3
movq 64(%rsp), %rax
chacha_blocks_avx2_mainloop3:
vpaddd %ymm0, %ymm1, %ymm0
vpxor %ymm3, %ymm0, %ymm3
vpshufb %ymm6, %ymm3, %ymm3
vpaddd %ymm2, %ymm3, %ymm2
vpxor %ymm1, %ymm2, %ymm1
vpslld \$ 12, %ymm1, %ymm4
vpsrld \$20, %ymm1, %ymm1
vpxor %ymm1, %ymm4, %ymm1
vpaddd %ymm0, %ymm1, %ymm0
vpxor %ymm3, %ymm0, %ymm3
vpshufb %ymm7, %ymm3, %ymm3
vpshufd \$0x93, %ymm0, %ymm0
vpaddd %ymm2, %ymm3, %ymm2
vpshufd \$0x4e, %ymm3, %ymm3
vpxor %ymm1, %ymm2, %ymm1
vpshufd \$0x39, %ymm2, %ymm2
vpslld \$ 7, %ymm1, %ymm4
vpsrld \$25, %ymm1, %ymm1
vpxor %ymm1, %ymm4, %ymm1
vpaddd %ymm0, %ymm1, %ymm0
vpxor %ymm3, %ymm0, %ymm3
vpshufb %ymm6, %ymm3, %ymm3
vpaddd %ymm2, %ymm3, %ymm2
vpxor %ymm1, %ymm2, %ymm1
vpslld \$ 12, %ymm1, %ymm4
vpsrld \$20, %ymm1, %ymm1
vpxor %ymm1, %ymm4, %ymm1
vpaddd %ymm0, %ymm1, %ymm0
vpxor %ymm3, %ymm0, %ymm3
vpshufb %ymm7, %ymm3, %ymm3
vpshufd \$0x39, %ymm0, %ymm0
vpaddd %ymm2, %ymm3, %ymm2
vpshufd \$0x4e, %ymm3, %ymm3
vpxor %ymm1, %ymm2, %ymm1
vpshufd \$0x93, %ymm2, %ymm2
vpslld \$ 7, %ymm1, %ymm4
vpsrld \$25, %ymm1, %ymm1
vpxor %ymm1, %ymm4, %ymm1
subq \$2, %rax
jnz chacha_blocks_avx2_mainloop3
vpaddd %ymm8, %ymm0, %ymm0
vpaddd %ymm9, %ymm1, %ymm1
vpaddd %ymm10, %ymm2, %ymm2
vpaddd %ymm11, %ymm3, %ymm3
vperm2i128 \$2, %ymm0, %ymm1, %ymm12
vperm2i128 \$2, %ymm2, %ymm3, %ymm13
vperm2i128 \$19, %ymm0, %ymm1, %ymm14
vperm2i128 \$19, %ymm2, %ymm3, %ymm15
andq %rsi, %rsi
jz chacha_blocks_avx2_noinput3
vpxor 0(%rsi), %ymm12, %ymm12
vpxor 32(%rsi), %ymm13, %ymm13
vpxor 64(%rsi), %ymm14, %ymm14
vpxor 96(%rsi), %ymm15, %ymm15
addq \$128, %rsi
chacha_blocks_avx2_noinput3:
addq \$2, %r14
vmovdqu %ymm12, 0(%rdx)
vmovdqu %ymm13, 32(%rdx)
vmovdqu %ymm14, 64(%rdx)
vmovdqu %ymm15, 96(%rdx)
movq \$2, %r9
vmovd %r9, %xmm4
addq \$128, %rdx
vpaddq %xmm4, %xmm11, %xmm11
subq \$128, %rcx
movq %r14, 48(%rsp)
chacha_blocks_avx2_below128_fixup:
movq \$1, %r9
vmovd %r9, %xmm5
chacha_blocks_avx2_below128:
andq %rcx, %rcx
jz chacha_blocks_avx2_done
cmpq \$64, %rcx
jae chacha_blocks_avx2_above63
movq %rdx, %r9
andq %rsi, %rsi
jz chacha_blocks_avx2_noinput4
movq %rcx, %r10
movq %rsp, %rdx
addq %r10, %rsi
addq %r10, %rdx
negq %r10
chacha_blocks_avx2_copyinput:
movb (%rsi, %r10), %al
movb %al, (%rdx, %r10)
incq %r10
jnz chacha_blocks_avx2_copyinput
movq %rsp, %rsi
chacha_blocks_avx2_noinput4:
movq %rsp, %rdx
chacha_blocks_avx2_above63:
vmovdqa %xmm8, %xmm0
vmovdqa %xmm9, %xmm1
vmovdqa %xmm10, %xmm2
vmovdqa %xmm11, %xmm3
movq 64(%rsp), %rax
chacha_blocks_avx2_mainloop4:
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vpshufb %xmm6, %xmm3, %xmm3
vpaddd %xmm2, %xmm3, %xmm2
vpxor %xmm1, %xmm2, %xmm1
vpslld \$12, %xmm1, %xmm4
vpsrld \$20, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vpshufb %xmm7, %xmm3, %xmm3
vpshufd \$0x93, %xmm0, %xmm0
vpaddd %xmm2, %xmm3, %xmm2
vpshufd \$0x4e, %xmm3, %xmm3
vpxor %xmm1, %xmm2, %xmm1
vpshufd \$0x39, %xmm2, %xmm2
vpslld \$7, %xmm1, %xmm4
vpsrld \$25, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vpshufb %xmm6, %xmm3, %xmm3
vpaddd %xmm2, %xmm3, %xmm2
vpxor %xmm1, %xmm2, %xmm1
vpslld \$12, %xmm1, %xmm4
vpsrld \$20, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
vpaddd %xmm0, %xmm1, %xmm0
vpxor %xmm3, %xmm0, %xmm3
vpshufb %xmm7, %xmm3, %xmm3
vpshufd \$0x39, %xmm0, %xmm0
vpaddd %xmm2, %xmm3, %xmm2
vpshufd \$0x4e, %xmm3, %xmm3
vpxor %xmm1, %xmm2, %xmm1
vpshufd \$0x93, %xmm2, %xmm2
vpslld \$7, %xmm1, %xmm4
vpsrld \$25, %xmm1, %xmm1
vpxor %xmm1, %xmm4, %xmm1
subq \$2, %rax
jnz chacha_blocks_avx2_mainloop4
vpaddd %xmm0, %xmm8, %xmm0
vpaddd %xmm1, %xmm9, %xmm1
vpaddd %xmm2, %xmm10, %xmm2
vpaddd %xmm3, %xmm11, %xmm3
andq %rsi, %rsi
jz chacha_blocks_avx2_noinput5
vpxor 0(%rsi), %xmm0, %xmm0
vpxor 16(%rsi), %xmm1, %xmm1
vpxor 32(%rsi), %xmm2, %xmm2
vpxor 48(%rsi), %xmm3, %xmm3
addq \$64, %rsi
chacha_blocks_avx2_noinput5:
vmovdqu %xmm0, 0(%rdx)
vmovdqu %xmm1, 16(%rdx)
vmovdqu %xmm2, 32(%rdx)
vmovdqu %xmm3, 48(%rdx)
vpaddq %xmm11, %xmm5, %xmm11
addq \$1, %r14
cmpq \$64, %rcx
jbe chacha_blocks_avx2_mainloop4_finishup
addq \$64, %rdx
subq \$64, %rcx
jmp chacha_blocks_avx2_below128
chacha_blocks_avx2_mainloop4_finishup:
cmpq \$64, %rcx
je chacha_blocks_avx2_done
addq %rcx, %r9
addq %rcx, %rdx
negq %rcx
chacha_blocks_avx2_copyoutput:
movb (%rdx, %rcx), %al
movb %al, (%r9, %rcx)
incq %rcx
jnz chacha_blocks_avx2_copyoutput
chacha_blocks_avx2_done:
movq %r14, 32(%rdi)
movq %rbp, %rsp
popq %r14
popq %r13
popq %r12
popq %rbp
popq %rbx
vzeroupper
ret
___
);


} # $ext{'x86'}


print $code;

close STDOUT;

