/*
	Chacha reference implementation and driver by Andrew Moon

	Public Domain or MIT
*/

#include <openssl/bn.h> /* 'abuse' SIXTY_FOUR_BIT_LONG etc */
#include <openssl/crypto.h> /* OPENSSL_assert */
#include <string.h>
#include "chacha.h"

#include <stdio.h>

/* why isn't openssl detecting exact sizes automatically? */
#if defined(SIXTY_FOUR_BIT_LONG)
typedef unsigned int chacha_uint32;
#define MASK32(x) (x)
#elif defined(SIXTY_FOUR_BIT) || defined(THIRTY_TWO_BIT)
typedef unsigned long chacha_uint32;
#define MASK32(x) (x)
#else
typedef unsigned long chacha_uint32;
#define MASK32(x) ((x) & 0xfffffffful)
#endif

enum chacha_constants {
	CHACHA_BLOCKBYTES_REF = 64
};

/* interpret four 8 bit unsigned integers as a 32 bit unsigned integer in little endian */
static chacha_uint32
U8TO32(const unsigned char *p) {
	return 
	(((chacha_uint32)(p[0])      ) |
	 ((chacha_uint32)(p[1]) <<  8) |
	 ((chacha_uint32)(p[2]) << 16) |
	 ((chacha_uint32)(p[3]) << 24));
}

/* store a 32 bit unsigned integer as four 8 bit unsigned integers in little endian */
static void
U32TO8(unsigned char *p, chacha_uint32 v) {
	p[0] = (v      ) & 0xff;
	p[1] = (v >>  8) & 0xff;
	p[2] = (v >> 16) & 0xff;
	p[3] = (v >> 24) & 0xff;
}

/* 32 bit left rotate */
static chacha_uint32
ROTL32(chacha_uint32 x, int k) {
	return MASK32((x << k) | (x >> (32 - k)));
}

static void
chacha_init_ref(CHACHA_CTX *ctx, const unsigned char *key, const unsigned char *iv, size_t ivlen, size_t rounds) {
	OPENSSL_assert((ivlen > 0) && (ivlen < 16));
	memcpy(ctx->state + 0, key, 32); /* key */
	memset(ctx->state + 32, 0, 16 - ivlen); /* ctr */
	memcpy(ctx->state + 48 - ivlen, iv, ivlen); /* iv */
	ctx->rounds = rounds;
}

static void
chacha_init_iv_ref(CHACHA_CTX *ctx, const unsigned char *iv, size_t ivlen) {
	OPENSSL_assert((ivlen > 0) && (ivlen < 16));
	memset(ctx->state + 32, 0, 16 - ivlen); /* ctr */
	memcpy(ctx->state + 48 - ivlen, iv, ivlen); /* iv */
}

static void
chacha_blocks_ref(CHACHA_CTX *ctx, const unsigned char *in, unsigned char *out, size_t inl) {
	chacha_uint32 x[16], j[16];
	chacha_uint32 t;
	unsigned char *out_actual = out, tmp[CHACHA_BLOCKBYTES_REF];
	size_t i, r;

	if (!inl) return;

	/* "expand 32-byte k", as 4 little endian 32-bit unsigned integers */
	j[0] = 0x61707865;
	j[1] = 0x3320646e;
	j[2] = 0x79622d32;
	j[3] = 0x6b206574;
	for (i = 0; i < 12; i++)
		j[i + 4] = U8TO32(ctx->state + (i * 4));

	r = ctx->rounds;

	for (;;) {
		if (inl < CHACHA_BLOCKBYTES_REF) {
			if (in) {
				for (i = 0; i < inl; i++) tmp[i] = in[i];
				in = tmp;
			}
			out_actual = out;
			out = tmp;
		}

		for (i = 0; i < 16; i++)
			x[i] = j[i];

		#define quarter(a,b,c,d) \
			a += b; t = d^a; d = ROTL32(t,16); \
			c += d; t = b^c; b = ROTL32(t,12); \
			a += b; t = d^a; d = ROTL32(t, 8); \
			c += d; t = b^c; b = ROTL32(t, 7);

		i = r;
		do {
			quarter( x[0], x[4], x[8],x[12])
			quarter( x[1], x[5], x[9],x[13])
			quarter( x[2], x[6],x[10],x[14])
			quarter( x[3], x[7],x[11],x[15])
			quarter( x[0], x[5],x[10],x[15])
			quarter( x[1], x[6],x[11],x[12])
			quarter( x[2], x[7], x[8],x[13])
			quarter( x[3], x[4], x[9],x[14])

			i -= 2;
		} while (i);

		for (i = 0; i < 16; i++)
			x[i] += j[i];

		if (in) {
			for (i = 0; i < 16; i++)
				U32TO8(out + (i * 4), x[i] ^ U8TO32(in + (i * 4)));
			in += CHACHA_BLOCKBYTES_REF;
		} else {
			for (i = 0; i < 16; i++)
				U32TO8(out + (i * 4), x[i]);
		}

		/* increment the 64 bit counter, split in to two 32 bit halves */
		j[12]++;
		if (!j[12])
			j[13]++;

		if (inl <= CHACHA_BLOCKBYTES_REF) {
			if (inl < CHACHA_BLOCKBYTES_REF) for (i = 0; i < inl; i++) out_actual[i] = out[i];

			/* store the counter back to the ctx */
			U32TO8(ctx->state + 32, j[12]);
			U32TO8(ctx->state + 36, j[13]);
			return;
		}
		inl -= CHACHA_BLOCKBYTES_REF;
		out += CHACHA_BLOCKBYTES_REF;
	}
}

typedef struct chacha_impl_s {
	int cpuflag1, cpuflag2;
	size_t parallel_blocks;
	int require_aligned;
	void (*chacha_blocks)(CHACHA_CTX *ctx, const unsigned char *in, unsigned char *out, size_t inl);
} chacha_impl;


#define CHACHA_DECLARE(ext) \
	void chacha_blocks_##ext(CHACHA_CTX *ctx, const unsigned char *in, unsigned char *out, size_t inl);

#define CHACHA_IMPL(cpuflag1, ext, parallel_blocks, aligned) \
	{cpuflag1, -1, parallel_blocks, aligned, chacha_blocks_##ext}

#define CHACHA_IMPL2(cpuflag1, cpuflag2, ext, parallel_blocks, aligned) \
	{cpuflag1, cpuflag2, parallel_blocks, aligned, chacha_blocks_##ext}

#if defined(CHACHA_ASM_X86_64)
	CHACHA_DECLARE(avx2)
	CHACHA_DECLARE(xop)
	CHACHA_DECLARE(avx)
	CHACHA_DECLARE(ssse3)
	CHACHA_DECLARE(sse2)

	extern unsigned int OPENSSL_ia32cap_P[];
#endif

#if defined(CHACHA_ASM_X86)
	CHACHA_DECLARE(avx2)
	CHACHA_DECLARE(xop)
	CHACHA_DECLARE(avx)
	CHACHA_DECLARE(ssse3)
	CHACHA_DECLARE(sse2)
	CHACHA_DECLARE(x86)

	extern unsigned int OPENSSL_ia32cap_P[];
#endif

static const chacha_impl chacha_impl_list[] = {
#if defined(CHACHA_ASM_X86_64)
	CHACHA_IMPL2(59, 69,  avx2,  8, 0),
	CHACHA_IMPL2(24, 43,   xop,  4, 0),
	CHACHA_IMPL2(24, 60,   avx,  4, 0),
	CHACHA_IMPL2(24, 41, ssse3,  4, 0),
	CHACHA_IMPL2(24, 26,  sse2,  4, 0),
#endif

#if defined(CHACHA_ASM_X86)
	CHACHA_IMPL2(59, 69,  avx2,  8, 0),
	CHACHA_IMPL2(24, 43,   xop,  4, 0),
	CHACHA_IMPL2(24, 60,   avx,  4, 0),
	CHACHA_IMPL2(24, 41, ssse3,  4, 0),
	CHACHA_IMPL2(24, 26,  sse2,  4, 0),
	CHACHA_IMPL( -1,       x86,  1, 0),
#endif

	CHACHA_IMPL(-1, ref, 1, 0)
};


static const chacha_impl *
chacha_get_first_available_impl(const chacha_impl *opt) {
	for (; opt->chacha_blocks != chacha_blocks_ref; opt++) {
#if defined(CHACHA_ASM_X86) || defined(CHACHA_ASM_X86_64)
			if ((OPENSSL_ia32cap_P[opt->cpuflag1 / 32] & (1 << (opt->cpuflag1 & 31))) == 0)
				continue;
			if ((opt->cpuflag2 != -1) && ((OPENSSL_ia32cap_P[opt->cpuflag2 / 32] & (1 << (opt->cpuflag2 & 31))) == 0))
				continue;
			break;
#endif
	}
	return opt;
}

static const chacha_impl *chacha_opt = NULL;

static const chacha_impl *
chacha_get_opt(void) {
	const chacha_impl *opt = chacha_opt;

	if (!opt) {
		opt = chacha_get_first_available_impl(chacha_impl_list);
		chacha_opt = opt;
	}

	return opt;
}

int
CHACHA_test_next_opt(void) {
	const chacha_impl *opt = chacha_get_opt();
	if (opt->chacha_blocks == chacha_blocks_ref)
		return 0;
	chacha_opt = chacha_get_first_available_impl(opt + 1);
	return 1;
}

/* is the pointer aligned on a word boundary? */
static int
chacha_is_aligned(const void *p) {
	return ((size_t)p & (sizeof(size_t) - 1)) == 0;
}

/* processes inl bytes (can do partial blocks), handling input/ouput alignment */
static void
chacha_consume(const chacha_impl *opt, CHACHA_CTX *ctx, const unsigned char *in, unsigned char *out, size_t inl) {
	int in_aligned, out_aligned;

	/* it's ok to call with 0 bytes */
	if (!inl)
		return;

	/* if everything is aligned, handle directly */
	in_aligned = chacha_is_aligned(in);
	out_aligned = chacha_is_aligned(out);
	if (!opt->require_aligned || (in_aligned && out_aligned)) {
		opt->chacha_blocks(ctx, in, out, inl);
		return;
	}

	/* copy the unaligned data to an aligned buffer and process in chunks */
	while (inl) {
		unsigned char buffer[16 * CHACHA_BLOCKBYTES_REF];
		const size_t bytes = (inl > sizeof(buffer)) ? sizeof(buffer) : inl;
		const unsigned char *src = in;
		unsigned char *dst = (out_aligned) ? out : buffer;
		if (!in_aligned) {
			memcpy(buffer, in, bytes);
			src = buffer;
		}
		opt->chacha_blocks(ctx, src, dst, bytes);
		if (!out_aligned)
			memcpy(out, buffer, bytes);
		if (in)
			in += bytes;
		out += bytes;
		inl -= bytes;
	}
}

size_t
CHACHA_get_parallel_blocks(void) {
	const chacha_impl *opt = chacha_get_opt();
	return opt->parallel_blocks;
}

void
CHACHA_init(CHACHA_CTX *ctx, const unsigned char *key, const unsigned char *iv, size_t ivlen, size_t rounds) {
	OPENSSL_assert(((rounds & 1) == 0) && (rounds >= 2)); /* rounds must be even and >= 2 */
	chacha_init_ref(ctx, key, iv, ivlen, rounds);
	ctx->leftover = 0;
}

void
CHACHA_init_iv(CHACHA_CTX *ctx, const unsigned char *iv, size_t ivlen) {
	chacha_init_iv_ref(ctx, iv, ivlen);
	ctx->leftover = 0;
}

void
CHACHA_update(CHACHA_CTX *ctx, const unsigned char *in, size_t inl, unsigned char *out, size_t *outl) {
	const chacha_impl *opt = chacha_get_opt();
	unsigned char *out_start = out;
	size_t bytes;

	/* enough for at least one block? */
	if ((ctx->leftover + inl) >= CHACHA_BLOCKBYTES_REF) {
		/* handle the previous data */
		if (ctx->leftover) {
			bytes = (CHACHA_BLOCKBYTES_REF - ctx->leftover);
			if (in) {
				memcpy(ctx->buffer + ctx->leftover, in, bytes);
				in += bytes;
			}
			chacha_consume(opt, ctx, (in) ? ctx->buffer : NULL, out, CHACHA_BLOCKBYTES_REF);
			inl -= bytes;
			out += CHACHA_BLOCKBYTES_REF;
			ctx->leftover = 0;
		}

		/* handle the direct data */
		bytes = (inl & ~(CHACHA_BLOCKBYTES_REF - 1));
		if (bytes) {
			chacha_consume(opt, ctx, in, out, bytes);
			inl -= bytes;
			if (in)
				in += bytes;
			out += bytes;
		}
	}

	/* handle leftover data */
	if (inl) {
		if (in)
			memcpy(ctx->buffer + ctx->leftover, in, inl);
		else
			memset(ctx->buffer + ctx->leftover, 0, inl);
		ctx->leftover += inl;
	}

	*outl = out - out_start;
}

void
CHACHA_update_and_finish(CHACHA_CTX *ctx, const unsigned char *in, size_t inl, unsigned char *out, size_t *outl) {
	const chacha_impl *opt = chacha_get_opt();
	unsigned char *out_start = out;
	size_t bytes;

	/* handle the previous data */
	if (ctx->leftover) {
		bytes = (CHACHA_BLOCKBYTES_REF - ctx->leftover);
		if (inl < bytes)
			bytes = inl;
		if (in) {
			memcpy(ctx->buffer + ctx->leftover, in, bytes);
			in += bytes;
		}
		chacha_consume(opt, ctx, (in) ? ctx->buffer : NULL, out, CHACHA_BLOCKBYTES_REF);
		inl -= bytes;
		out += CHACHA_BLOCKBYTES_REF;
		ctx->leftover = 0;
	}

	/* handle the rest */
	if (inl) {
		chacha_consume(opt, ctx, in, out, inl);
		out += inl;
	}

	*outl = out - out_start;
}

void
CHACHA_finish(CHACHA_CTX *ctx, unsigned char *out, size_t *outl) {
	const chacha_impl *opt = chacha_get_opt();
	size_t leftover = ctx->leftover;
	if (leftover) {
		if (!opt->require_aligned || chacha_is_aligned(out)) {
			opt->chacha_blocks(ctx, ctx->buffer, out, leftover);
		} else {
			opt->chacha_blocks(ctx, ctx->buffer, ctx->buffer, leftover);
			memcpy(out, ctx->buffer, leftover);
		}
	}
	*outl = leftover;
}

size_t
CHACHA20_get_parallel_blocks(void) {
	return CHACHA_get_parallel_blocks();
}

void
CHACHA20_init(CHACHA_CTX *ctx, const unsigned char *key, const unsigned char *iv, size_t ivlen) {
	CHACHA_init(ctx, key, iv, ivlen, 20);
}

void
CHACHA20_init_iv(CHACHA_CTX *ctx, const unsigned char *iv, size_t ivlen) {
	CHACHA_init_iv(ctx, iv, ivlen);
}

void
CHACHA20_update(CHACHA_CTX *ctx, const unsigned char *in, size_t inl, unsigned char *out, size_t *outl) {
	CHACHA_update(ctx, in, inl, out, outl);
}

void
CHACHA20_update_and_finish(CHACHA_CTX *ctx, const unsigned char *in, size_t inl, unsigned char *out, size_t *outl) {
	CHACHA_update_and_finish(ctx, in, inl, out, outl);
}

void CHACHA20_finish(CHACHA_CTX *ctx, unsigned char *out, size_t *outl) {
	CHACHA_finish(ctx, out, outl);
}
