/*
	Chacha20/Poly1305 Chrome compatible optimized TLS implementation
*/

#include <openssl/opensslconf.h>

#if !defined(OPENSSL_NO_CHACHA) && !defined(OPENSSL_NO_POLY1305)
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/chacha.h>
#include <openssl/poly1305.h>
#include <string.h>

#define CHACHA20_POLY1305_IVLEN 8

typedef struct evp_chacha20_poly1305_s {
	CHACHA_CTX cipher_ctx;
	size_t valid;
	unsigned char aad[13 + 8];
} EVP_CHACHA20_POLY1305_CTX;

static void
POLY1305_update_with_64bit_le_length(POLY1305_CTX *ctx, const unsigned char *in, size_t inl) {
	unsigned char length_bytes[8];
	size_t i, v = inl;
	for (i = 0; i < 8; i++) {
		length_bytes[i] = (unsigned char)v;
		v >>= 8;
	}
	POLY1305_update(ctx, in, inl);
	POLY1305_update(ctx, length_bytes, 8);
}

static int
EVP_chacha20_poly1305_init(EVP_CIPHER_CTX *ctx, const unsigned char *key, const unsigned char *iv, int enc) {
	EVP_CHACHA20_POLY1305_CTX *aead = (EVP_CHACHA20_POLY1305_CTX *)ctx->cipher_data;
	if (key) {
		static const unsigned char null_iv[CHACHA20_POLY1305_IVLEN] = {0};
		CHACHA20_init(&aead->cipher_ctx, key, null_iv, CHACHA20_POLY1305_IVLEN);
	}
	aead->valid = 0;
	return 1;
}

static int
EVP_chacha20_poly1305_cipher(EVP_CIPHER_CTX *ctx, unsigned char *out, const unsigned char *in, size_t inl) {
	EVP_CHACHA20_POLY1305_CTX *aead = (EVP_CHACHA20_POLY1305_CTX *)ctx->cipher_data;
	POLY1305_CTX auth_ctx;
	size_t written;
	size_t next64;

	OPENSSL_assert(out && in);

	if (!aead->valid)
		return -1;
	aead->valid = 0;

	/* adjust for mac */
	inl -= POLY1305_TAG_LEN;

	/* set the correct length in the aad */
	aead->aad[11] = (unsigned char)(inl >> 8);
	aead->aad[12] = (unsigned char)(inl     );

	/* set the iv to the sequence number */
	CHACHA20_init_iv(&aead->cipher_ctx, aead->aad, CHACHA20_POLY1305_IVLEN);

	next64 = (inl + 64 + 63) & ~63;

	/* see if it's more efficient to do everything in one chacha call */
	if (next64 <= (CHACHA20_get_parallel_blocks() * 64)) {
		unsigned char buffer[1024];
		unsigned char *poly1305_key = buffer;
		unsigned char *ciphered_data = buffer + 64;

		memset(poly1305_key, 0, POLY1305_KEY_LEN); /* clear the buffer receiving the poly1305 key */
		memcpy(ciphered_data, in, inl); /* copy in the data to en/decrypt after the key block */
		CHACHA20_update_and_finish(&aead->cipher_ctx, buffer, next64, buffer, &written);

		POLY1305_init_ex(&auth_ctx, poly1305_key, sizeof(aead->aad) + inl + 8);
		POLY1305_update(&auth_ctx, aead->aad, sizeof(aead->aad));
		POLY1305_update_with_64bit_le_length(&auth_ctx, (ctx->encrypt) ? ciphered_data : in, inl);

		if (ctx->encrypt) {
			POLY1305_finish(&auth_ctx, out + inl);
		} else {
			unsigned char mac[POLY1305_TAG_LEN];

			/* unfortunately can't auth the data before decrypting, oh well */
			POLY1305_finish(&auth_ctx, mac);
			if (POLY1305_verify(in + inl, mac))
				return -1;
		}

		/* copy the en/decrypted data to out */
		memcpy(out, ciphered_data, inl);
	} else {
		/* message was too long for underlying implementation to short-message accelerate */
		unsigned char poly1305_key[64];

		CHACHA20_update(&aead->cipher_ctx, NULL, 64, poly1305_key, &written);
		POLY1305_init_ex(&auth_ctx, poly1305_key, sizeof(aead->aad) + inl + 8);
		POLY1305_update(&auth_ctx, aead->aad, sizeof(aead->aad));

		if (ctx->encrypt) {
			CHACHA20_update_and_finish(&aead->cipher_ctx, in, inl, out, &written);
			POLY1305_update_with_64bit_le_length(&auth_ctx, out, inl);
			POLY1305_finish(&auth_ctx, out + inl);
		} else {
			unsigned char mac[POLY1305_TAG_LEN];

			POLY1305_update_with_64bit_le_length(&auth_ctx, in, inl);
			POLY1305_finish(&auth_ctx, mac);
			if (POLY1305_verify(in + inl, mac))
				return -1;

			CHACHA20_update_and_finish(&aead->cipher_ctx, in, inl, out, &written);
		}
	}

	return inl + POLY1305_TAG_LEN;
}

static int
EVP_chacha20_poly1305_cleanup(EVP_CIPHER_CTX *ctx) {
	OPENSSL_cleanse(ctx->cipher_data, sizeof(EVP_CHACHA20_POLY1305_CTX));
	return 1;
}

static int
EVP_chacha20_poly1305_ctrl(EVP_CIPHER_CTX *ctx, int type, int arg, void *ptr) {
	EVP_CHACHA20_POLY1305_CTX *aead = (EVP_CHACHA20_POLY1305_CTX *)ctx->cipher_data;

	switch (type) {
		case EVP_CTRL_AEAD_TLS1_AAD: {
			if (arg != 13)
				return 0;

			memcpy(aead->aad, ptr, 13);
			memset(aead->aad + 13, 0, 8);
			aead->aad[13] = 13;
			aead->valid = 1;
			return POLY1305_TAG_LEN;
		}

		default:
			break;
	}

	return 0;
}


static const EVP_CIPHER chacha20_poly1305 = {
	NID_undef,                         /* nid */
	1,                                 /* block size */
	CHACHA_KEY_LEN,                    /* key len */
	0,                                 /* iv len, handled internally */

	EVP_CIPH_STREAM_CIPHER |           /* this is a stream cipher */
	EVP_CIPH_FLAG_CUSTOM_CIPHER |      /* return value for _cipher is the number of bytes written, -1 for err */
	EVP_CIPH_FLAG_AEAD_CIPHER,         /* only tls cares about this */

	EVP_chacha20_poly1305_init,        /* init */
	EVP_chacha20_poly1305_cipher,      /* encrypt/decrypt */
	EVP_chacha20_poly1305_cleanup,     /* cleanup */
	sizeof(EVP_CHACHA20_POLY1305_CTX), /* ctx size */
	NULL,                              /* set asn1 parameters */
	NULL,                              /* get asn1 parameters */
	EVP_chacha20_poly1305_ctrl,        /* control function */
	NULL                               /* app data */
};

const EVP_CIPHER *
EVP_chacha20_poly1305(void) {
	return &chacha20_poly1305;
}

#endif /* !defined(OPENSSL_NO_CHACHA) && !defined(OPENSSL_NO_POLY1305) */
