/*
	Poly1305 reference implementation and driver by Andrew Moon

	Public Domain or MIT
*/

#include <openssl/bn.h> /* 'abuse' SIXTY_FOUR_BIT_LONG etc */
#include <string.h>
#include "poly.h"

/* why isn't openssl detecting exact sizes automatically? */
#if defined(SIXTY_FOUR_BIT_LONG)
typedef unsigned int poly1305_uint32;
typedef unsigned long poly1305_uint64;
#define MASK32(x) (x)
#elif defined(SIXTY_FOUR_BIT) || defined(THIRTY_TWO_BIT)
typedef unsigned long poly1305_uint32;
typedef unsigned long long poly1305_uint64;
#define MASK32(x) (x)
#else
typedef unsigned long poly1305_uint32;
typedef unsigned long long poly1305_uint64;
#define MASK32(x) ((x) & 0xfffffffful)
#endif

enum {
	POLY1305_BLOCK_SIZE_REF = 16
};

typedef struct poly1305_state_ref_s {
	poly1305_uint32 r[5];
	poly1305_uint32 h[5];
	poly1305_uint32 pad[4];
	unsigned char final;
} POLY1305_STATE_REF;

/* interpret four 8 bit unsigned integers as a 32 bit unsigned integer in little endian */
static poly1305_uint32
U8TO32(const unsigned char *p) {
	return 
	(((poly1305_uint32)(p[0])      ) |
	 ((poly1305_uint32)(p[1]) <<  8) |
	 ((poly1305_uint32)(p[2]) << 16) |
	 ((poly1305_uint32)(p[3]) << 24));
}

/* store a 32 bit unsigned integer as four 8 bit unsigned integers in little endian */
static void
U32TO8(unsigned char *p, poly1305_uint32 v) {
	p[0] = (v      ) & 0xff;
	p[1] = (v >>  8) & 0xff;
	p[2] = (v >> 16) & 0xff;
	p[3] = (v >> 24) & 0xff;
}

static void
poly1305_init_ex_ref(POLY1305_CTX *ctx, const unsigned char *key, size_t bytes_hint) {
	POLY1305_STATE_REF *st = (POLY1305_STATE_REF *)ctx->state;

	/* bytes_hint not used by ref */
	(void)bytes_hint;

	/* r &= 0xffffffc0ffffffc0ffffffc0fffffff */
	st->r[0] = (U8TO32(key +  0)     ) & 0x3ffffff;
	st->r[1] = (U8TO32(key +  3) >> 2) & 0x3ffff03;
	st->r[2] = (U8TO32(key +  6) >> 4) & 0x3ffc0ff;
	st->r[3] = (U8TO32(key +  9) >> 6) & 0x3f03fff;
	st->r[4] = (U8TO32(key + 12) >> 8) & 0x00fffff;

	/* h = 0 */
	st->h[0] = 0;
	st->h[1] = 0;
	st->h[2] = 0;
	st->h[3] = 0;
	st->h[4] = 0;

	/* save pad for later */
	st->pad[0] = U8TO32(key + 16);
	st->pad[1] = U8TO32(key + 20);
	st->pad[2] = U8TO32(key + 24);
	st->pad[3] = U8TO32(key + 28);

	st->final = 0;
}

static void
poly1305_blocks_ref(POLY1305_CTX *ctx, const unsigned char *in, size_t bytes) {
	POLY1305_STATE_REF *st = (POLY1305_STATE_REF *)ctx->state;
	const poly1305_uint32 hibit = (st->final) ? 0 : ((poly1305_uint32)1 << 24); /* 1 << 128 */
	poly1305_uint32 r0,r1,r2,r3,r4;
	poly1305_uint32 s1,s2,s3,s4;
	poly1305_uint32 h0,h1,h2,h3,h4;
	poly1305_uint64 d0,d1,d2,d3,d4;
	poly1305_uint32 c;

	r0 = st->r[0];
	r1 = st->r[1];
	r2 = st->r[2];
	r3 = st->r[3];
	r4 = st->r[4];

	s1 = r1 * 5;
	s2 = r2 * 5;
	s3 = r3 * 5;
	s4 = r4 * 5;

	h0 = st->h[0];
	h1 = st->h[1];
	h2 = st->h[2];
	h3 = st->h[3];
	h4 = st->h[4];

	while (bytes >= POLY1305_BLOCK_SIZE_REF) {
		/* h += m[i] */
		h0 += (U8TO32(in +  0)     ) & 0x3ffffff;
		h1 += (U8TO32(in +  3) >> 2) & 0x3ffffff;
		h2 += (U8TO32(in +  6) >> 4) & 0x3ffffff;
		h3 += (U8TO32(in +  9) >> 6) & 0x3ffffff;
		h4 += (U8TO32(in + 12) >> 8) | hibit;

		/* h *= r */
		d0 = ((poly1305_uint64)h0 * r0) + ((poly1305_uint64)h1 * s4) + ((poly1305_uint64)h2 * s3) + ((poly1305_uint64)h3 * s2) + ((poly1305_uint64)h4 * s1);
		d1 = ((poly1305_uint64)h0 * r1) + ((poly1305_uint64)h1 * r0) + ((poly1305_uint64)h2 * s4) + ((poly1305_uint64)h3 * s3) + ((poly1305_uint64)h4 * s2);
		d2 = ((poly1305_uint64)h0 * r2) + ((poly1305_uint64)h1 * r1) + ((poly1305_uint64)h2 * r0) + ((poly1305_uint64)h3 * s4) + ((poly1305_uint64)h4 * s3);
		d3 = ((poly1305_uint64)h0 * r3) + ((poly1305_uint64)h1 * r2) + ((poly1305_uint64)h2 * r1) + ((poly1305_uint64)h3 * r0) + ((poly1305_uint64)h4 * s4);
		d4 = ((poly1305_uint64)h0 * r4) + ((poly1305_uint64)h1 * r3) + ((poly1305_uint64)h2 * r2) + ((poly1305_uint64)h3 * r1) + ((poly1305_uint64)h4 * r0);

		/* (partial) h %= p */
		c = (poly1305_uint32)(d0 >> 26); h0 = (poly1305_uint32)d0 & 0x3ffffff; d1 += c;
		c = (poly1305_uint32)(d1 >> 26); h1 = (poly1305_uint32)d1 & 0x3ffffff; d2 += c;
		c = (poly1305_uint32)(d2 >> 26); h2 = (poly1305_uint32)d2 & 0x3ffffff; d3 += c;
		c = (poly1305_uint32)(d3 >> 26); h3 = (poly1305_uint32)d3 & 0x3ffffff; d4 += c;
		c = (poly1305_uint32)(d4 >> 26); h4 = (poly1305_uint32)d4 & 0x3ffffff; h0 += c * 5;
		c =                  (h0 >> 26); h0 =                  h0 & 0x3ffffff; h1 += c;

		in += POLY1305_BLOCK_SIZE_REF;
		bytes -= POLY1305_BLOCK_SIZE_REF;
	}

	st->h[0] = h0;
	st->h[1] = h1;
	st->h[2] = h2;
	st->h[3] = h3;
	st->h[4] = h4;
}

static void
poly1305_finish_ex_ref(POLY1305_CTX *ctx, const unsigned char *in, size_t remaining, unsigned char mac[16]) {
	POLY1305_STATE_REF *st = (POLY1305_STATE_REF *)ctx->state;
	poly1305_uint32 h0,h1,h2,h3,h4,c;
	poly1305_uint32 g0,g1,g2,g3,g4;
	poly1305_uint64 f;
	poly1305_uint32 mask;

	/* process the remaining block */
	if (remaining) {
		unsigned char final[POLY1305_BLOCK_SIZE_REF] = {0};
		size_t i;
		for (i = 0; i < remaining; i++)
			final[i] = in[i];
		final[remaining] = 1;
		st->final = 1;
		poly1305_blocks_ref(ctx, final, POLY1305_BLOCK_SIZE_REF);
	}

	/* fully carry h */
	h0 = st->h[0];
	h1 = st->h[1];
	h2 = st->h[2];
	h3 = st->h[3];
	h4 = st->h[4];

	c = h1 >> 26; h1 = h1 & 0x3ffffff; h2 +=     c; 
	c = h2 >> 26; h2 = h2 & 0x3ffffff; h3 +=     c;
	c = h3 >> 26; h3 = h3 & 0x3ffffff; h4 +=     c;
	c = h4 >> 26; h4 = h4 & 0x3ffffff; h0 += c * 5;
	c = h0 >> 26; h0 = h0 & 0x3ffffff; h1 +=     c;

	/* compute h + -p */
	g0 = h0 + 5; c = g0 >> 26; g0 &= 0x3ffffff;
	g1 = h1 + c; c = g1 >> 26; g1 &= 0x3ffffff;
	g2 = h2 + c; c = g2 >> 26; g2 &= 0x3ffffff;
	g3 = h3 + c; c = g3 >> 26; g3 &= 0x3ffffff;
	g4 = h4 + c; c = g4 >> 26; g4 &= 0x3ffffff;

	/* select h if h < p, or h + -p if h >= p */
	mask = (c ^ 1) - 1;
	g0 &= mask;
	g1 &= mask;
	g2 &= mask;
	g3 &= mask;
	g4 &= mask;
	mask = ~mask;
	h0 = (h0 & mask) | g0;
	h1 = (h1 & mask) | g1;
	h2 = (h2 & mask) | g2;
	h3 = (h3 & mask) | g3;
	h4 = (h4 & mask) | g4;

	/* h = h % (2^128) */
	h0 = MASK32((h0      ) | (h1 << 26));
	h1 = MASK32((h1 >>  6) | (h2 << 20));
	h2 = MASK32((h2 >> 12) | (h3 << 14));
	h3 = MASK32((h3 >> 18) | (h4 <<  8));

	/* mac = (h + pad) % (2^128) */
	f = (poly1305_uint64)h0 + st->pad[0]            ; h0 = (poly1305_uint32)f;
	f = (poly1305_uint64)h1 + st->pad[1] + (f >> 32); h1 = (poly1305_uint32)f;
	f = (poly1305_uint64)h2 + st->pad[2] + (f >> 32); h2 = (poly1305_uint32)f;
	f = (poly1305_uint64)h3 + st->pad[3] + (f >> 32); h3 = (poly1305_uint32)f;

	U32TO8(mac +  0, h0);
	U32TO8(mac +  4, h1);
	U32TO8(mac +  8, h2);
	U32TO8(mac + 12, h3);
}

static void
poly1305_ref(unsigned char *mac, const unsigned char *in, size_t inl, const unsigned char *key) {
	POLY1305_CTX ctx;
	size_t blocks;
	poly1305_init_ex_ref(&ctx, key, inl);
	blocks = (inl & ~(POLY1305_BLOCK_SIZE_REF - 1));
	if (blocks) {
		poly1305_blocks_ref(&ctx, in, blocks);
		in += blocks;
		inl -= blocks;
	}
	poly1305_finish_ex_ref(&ctx, in, inl, mac);
}


/* optimized support */

typedef struct poly1305_impl_s {
	int cpuflag1, cpuflag2;
	int require_aligned;
	size_t block_size;
	void (*init_ex)(POLY1305_CTX *ctx, const unsigned char *key, size_t bytes_hint);
	void (*blocks)(POLY1305_CTX *ctx, const unsigned char *in, size_t inl);
	void (*finish_ex)(POLY1305_CTX *ctx, const unsigned char *in, size_t remaining, unsigned char *mac);
	void (*poly1305)(unsigned char *mac, const unsigned char *in, size_t inl, const unsigned char *key);
} poly1305_impl;


#define POLY1305_DECLARE(ext) \
	void poly1305_init_ex_##ext(POLY1305_CTX *ctx, const unsigned char *key, size_t bytes_hint); \
	void poly1305_blocks_##ext(POLY1305_CTX *ctx, const unsigned char *in, size_t inl); \
	void poly1305_finish_ex_##ext(POLY1305_CTX *ctx, const unsigned char *in, size_t remaining, unsigned char *mac); \
	void poly1305_##ext(unsigned char *mac, const unsigned char *m, size_t inl, const unsigned char *key);

#define POLY1305_IMPL(cpuflag1, ext, aligned, block_size) \
	{cpuflag1, -1, aligned, block_size, poly1305_init_ex_##ext, poly1305_blocks_##ext, poly1305_finish_ex_##ext, poly1305_##ext}

#define POLY1305_IMPL2(cpuflag1, cpuflag2, ext, aligned, block_size) \
	{cpuflag1, cpuflag2, aligned, block_size, poly1305_init_ex_##ext, poly1305_blocks_##ext, poly1305_finish_ex_##ext, poly1305_##ext}

#if defined(POLY1305_ASM_X86_64)
	POLY1305_DECLARE(avx2)
	POLY1305_DECLARE(avx)
	POLY1305_DECLARE(sse2)

	extern unsigned int OPENSSL_ia32cap_P[];
#endif

#if defined(POLY1305_ASM_X86)
	POLY1305_DECLARE(avx2)
	POLY1305_DECLARE(avx)
	POLY1305_DECLARE(sse2)
	POLY1305_DECLARE(x86)

	extern unsigned int OPENSSL_ia32cap_P[];
#endif

static const poly1305_impl poly1305_impl_list[] = {
#if defined(POLY1305_ASM_X86_64)
	POLY1305_IMPL2(59, 69, avx2, 0, 64),
	POLY1305_IMPL2(24, 60, avx, 0, 32),
	POLY1305_IMPL2(24, 26, sse2, 0, 32),
#endif

#if defined(POLY1305_ASM_X86)
	POLY1305_IMPL2(59, 69, avx2, 0, 64),
	POLY1305_IMPL2(24, 60, avx, 0, 32),
	POLY1305_IMPL2(24, 26, sse2, 0, 32),
	POLY1305_IMPL(-1, x86, 0, 16),
#endif

	POLY1305_IMPL(-1, ref, 0, POLY1305_BLOCK_SIZE_REF)
};


static const poly1305_impl *
poly1305_get_first_available_impl(const poly1305_impl *opt) {
	for (; opt->blocks != poly1305_blocks_ref; opt++) {
#if defined(POLY1305_ASM_X86) || defined(POLY1305_ASM_X86_64)
			if ((OPENSSL_ia32cap_P[opt->cpuflag1 / 32] & (1 << (opt->cpuflag1 & 31))) == 0)
				continue;
			if ((opt->cpuflag2 != -1) && ((OPENSSL_ia32cap_P[opt->cpuflag2 / 32] & (1 << (opt->cpuflag2 & 31))) == 0))
				continue;
			break;
#endif
	}
	return opt;
}

static const poly1305_impl *poly1305_opt = NULL;

static const poly1305_impl *
poly1305_get_opt(void) {
	const poly1305_impl *opt = poly1305_opt;

	if (!opt) {
		opt = poly1305_get_first_available_impl(poly1305_impl_list);
		poly1305_opt = opt;
	}

	return opt;
}

int
POLY1305_test_next_opt(void) {
	const poly1305_impl *opt = poly1305_get_opt();
	if (opt->blocks == poly1305_blocks_ref)
		return 0;
	poly1305_opt = poly1305_get_first_available_impl(opt + 1);
	return 1;
}


size_t
POLY1305_get_block_size(void) {
	const poly1305_impl *opt = poly1305_get_opt();
	return opt->block_size;
}


/* is the pointer aligned on a word boundary? */
static int
poly1305_is_aligned(const void *p) {
	return ((size_t)p & (sizeof(size_t) - 1)) == 0;
}

/* processes inl bytes (full blocks only), handling input alignment */
static void
poly1305_consume(const poly1305_impl *opt, POLY1305_CTX *ctx, const unsigned char *in, size_t inl) {
	int in_aligned;

	/* it's ok to call with 0 bytes */
	if (!inl)
		return;

	/* if everything is aligned, handle directly */
	in_aligned = poly1305_is_aligned(in);
	if (!opt->require_aligned || in_aligned) {
		opt->blocks(ctx, in, inl);
		return;
	}

	/* copy the unaligned data to an aligned buffer and process in chunks */
	while (inl) {
		unsigned char buffer[1024];
		const size_t bytes = (inl > sizeof(buffer)) ? sizeof(buffer) : inl;
		memcpy(buffer, in, bytes);
		opt->blocks(ctx, buffer, bytes);
		in += bytes;
		inl -= bytes;
	}
}

void
POLY1305_init_ex(POLY1305_CTX *ctx, const unsigned char *key, size_t bytes_hint) {
	const poly1305_impl *opt = poly1305_get_opt();
	opt->init_ex(ctx, key, bytes_hint);
	ctx->block_size = opt->block_size;
	ctx->leftover = 0;
}

void
POLY1305_init(POLY1305_CTX *ctx, const unsigned char *key) {
	POLY1305_init_ex(ctx, key, ~(size_t)0);
}

void
POLY1305_update(POLY1305_CTX *ctx, const unsigned char *in, size_t inl) {
	const poly1305_impl *opt = poly1305_get_opt();

	/* handle leftover */
	if (ctx->leftover) {
		size_t want = (ctx->block_size - ctx->leftover);
		if (want > inl)
			want = inl;
		memcpy(ctx->buffer + ctx->leftover, in, want);
		inl -= want;
		in += want;
		ctx->leftover += want;
		if (ctx->leftover < ctx->block_size)
			return;
		opt->blocks(ctx, ctx->buffer, ctx->block_size);
		ctx->leftover = 0;
	}

	/* process full blocks */
	if (inl >= ctx->block_size) {
		size_t want = (inl & ~(ctx->block_size - 1));
		poly1305_consume(opt, ctx, in, want);
		in += want;
		inl -= want;
	}

	/* store leftover */
	if (inl) {
		memcpy(ctx->buffer + ctx->leftover, in, inl);
		ctx->leftover += inl;
	}
}

void
POLY1305_finish(POLY1305_CTX *ctx, unsigned char *mac) {
	const poly1305_impl *opt = poly1305_get_opt();
	opt->finish_ex(ctx, ctx->buffer, ctx->leftover, mac);
}

void
POLY1305(unsigned char *mac, const unsigned char *in, size_t inl, const unsigned char *key) {
	const poly1305_impl *opt = poly1305_get_opt();
	opt->poly1305(mac, in, inl, key);
}

int
POLY1305_verify(const unsigned char *mac1, const unsigned char *mac2) {
	size_t i;
	unsigned char sum = 0;
	for (i = 0; i < POLY1305_TAG_LEN; i++)
		sum |= (mac1[i] ^ mac2[i]);
	return sum ? 1 : 0;
}
