#!/usr/bin/env perl

#==============================================================================
#
#  Optimized Chacha routines for x86 by Andrew Moon
#
#  x86: 64 byte blocks
#  sse2: 64, 256 byte blocks
#  ssse3: 64, 256 byte blocks
#  avx: 64, 128, 256 byte blocks
#  xop: 64, 128, 256 byte blocks
#  avx2: 64, 128, 256, 512 byte blocks
#
#  License: Public Domain or MIT
#
#==============================================================================
#
# struct CHACHA_CTX {
#     unsigned char state[48] {
#         uint32_le key[8]
#         uint32_le counter[2]
#         uint32_le iv[2]
#     }
#     size_t rounds;
# }
#
# chacha_blocks_ext(
#     CHACHA_CTX *ctx,
#     const unsigned char *in,
#     unsigned char *out,
#     size_t inl
# )
#
# If 'in' is NULL, bytestream is written directly to 'out'
#
# If 'inl' is not 0, then 'ctx->counter' will be treated as the lower and upper
# halves of a 64 bit integer and incremented by (inl + 63) / 64
#
#==============================================================================

$0 =~ m/(.*[\/\\])[^\/\\]+$/; 
my $dir = $1;
push(@INC,"${dir}","${dir}../../perlasm");
require "x86asm.pl";

&asm_init($ARGV[0],"chacha_x86.pl",$ARGV[$#ARGV] eq "386");

# why isn't this already handled for implementors?
my %ext = (
	x86 => 0,
	sse2 => 0,
	sse3 => 0,
	ssse3 => 0,
	sse4_1 => 0,
	sse4_2 => 0,
	avx => 0,
	xop => 0,
	avx2 => 0,
	avx512 => 0,
);

if (exists $INC{'x86gas.pl'}) {
	if (`$ENV{CC} -Wa,-v -c -o /dev/null -x assembler /dev/null 2>&1` =~ /GNU assembler version ([2-9]\.[0-9]+)/) {
		$ext{'x86'} = 1;
		$ext{'sse2'} = ($1 >= 2.16);
		$ext{'sse3'} = ($1 >= 2.17);
		$ext{'ssse3'} = ($1 >= 2.18);
		$ext{'sse4_1'} = ($1 >= 2.18);
		$ext{'sse4_2'} = ($1 >= 2.18);
		$ext{'avx'} = ($1 >= 2.19);
		$ext{'xop'} = ($1 >= 2.21);
		$ext{'avx2'} = ($1 >= 2.22);
		$ext{'avx512'} = ($1 >= 2.24);
	}

	if ((exists $INC{'x86gas.pl'}) && `$ENV{CC} -v 2>&1` =~ /(^clang version|based on LLVM) ([3-9])\.([0-9]+)/) {
		my $ver = $2 + ($3 / 100.0); # 3.1->3.01, 3.10->3.10
		$ext{'x86'} = 1;
		$ext{'sse2'} = ($1 >= 2.01); # ?
		$ext{'sse3'} = ($1 >= 2.01); # ?
		$ext{'ssse3'} = ($1 >= 2.01);
		$ext{'sse4_1'} = ($1 >= 2.01); # ?
		$ext{'sse4_2'} = ($1 >= 2.01); # ?
		$ext{'avx'} = ($ver >= 3.0);
		$ext{'xop'} = ($ver >= 3.01);
		$ext{'avx2'} = ($ver >= 3.01);
		$ext{'avx512'} = ($ver >= 3.05); # ?
	}

	*{x86gas::opsize} = sub () {
		my $reg=shift;
    		if    ($reg =~ m/^%e/o)          { "l"; }
    		elsif ($reg =~ m/^%[a-d][hl]$/o) { "b"; }
    		elsif ($reg =~ m/^%[yxm]/o)      { undef; }
    		else                             { "w"; }
	}
}

if (exists $INC{'x86nasm.pl'}) {
	if (`nasm -v 2>&1` =~ /NASM version ([2-9]\.[0-9]+)/) {
		$ext{'x86'} = 1;
		$ext{'sse2'} = ($1 >= 2.00);
		$ext{'sse3'} = ($1 >= 2.01);
		$ext{'ssse3'} = ($1 >= 2.01);
		$ext{'sse4_1'} = ($1 >= 2.01);
		$ext{'sss4_2'} = ($1 >= 2.01);
		$ext{'avx'} = ($1 >= 2.09);
		$ext{'xop'} = ($1 >= 2.09); # ?
		$ext{'avx2'} = ($1 >= 2.10);
		$ext{'avx512'} = ($1 >= 2.12); # ?
	}
}

if (exists $INC{'x86masm.pl'}) {
	if (`ml64 2>&1` =~ /Version ([0-9]+)\./) {
		$ext{'x86'} = 1;
		$ext{'sse2'} = ($1 >= 7); # ?
		$ext{'sse3'} = ($1 >= 9); # ?
		$ext{'ssse3'} = ($1 >= 9);
		$ext{'sse4_1'} = ($1 >= 9);
		$ext{'sse4_2'} = ($1 >= 9);
		$ext{'avx'} = ($1 >= 10);
		$ext{'xop'} = ($1 >= 10);
		$ext{'avx2'} = ($1 >= 11);
		$ext{'avx512'} = ($1 >= 14); # ?
	}
}

my %last_valid_fallback = ();

sub add_function {
	my ($name, $flag, $fn) = @_;

	my $fullname = "$name\_$flag";
	&function_begin_B($fullname);
	&set_label("$fullname\_local");
	if ($ext{$flag}) {
		$fn->();
		$last_valid_fallback{$name} = "$fullname\_local";
	} else {
		&jmp($last_valid_fallback{$name});
	}
	&function_end_B($fullname);
}

if ($ext{'x86'}) {

sub rep_movsb {
	&data_byte(0xf3,0xa4);
}

sub label_to_reg {
	my ($label, $reg) = @_;
	&call(&label("get_pic"));
	&set_label("get_pic");
	&pop($reg);
	&lea($reg,&DWP(&label($label)."-".&label("get_pic"),$reg));
}

&align(16);
&static_label("chacha_constants");
&set_label("chacha_constants");
&data_word(0x61707865,0x3320646e,0x79622d32,0x6b206574); # "expand 32-byte k"
&data_byte(2,3,0,1,6,7,4,5,10,11,8,9,14,15,12,13);       # pshufb 32bit rotate by 16
&data_byte(3,0,1,2,7,4,5,6,11,8,9,10,15,12,13,14);       # pshufb 32bit rotate by 8

sub chacha_blocks_x86 {
	&push("esi");
	&push("edi");
	&push("ebx");
	&push("ebp");
	&sub("esp", 204);
	&mov("eax", &DWP(232,"esp"));
	&mov("edi", &DWP(236,"esp"));
	&test("edi", "edi");
	&mov("esi", &DWP(228,"esp"));
	&mov(&DWP(76,"esp"), "eax");
	&mov(&DWP(0,"esp"), "eax");
	&jne(&label("chacha_blocks_x86_3"));
	&add("esp", 204);
	&pop("ebp");
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
	&set_label("chacha_blocks_x86_3");
	&mov("ebx", &DWP(224,"esp"));
	&mov("ebp", &DWP(24,"ebx"));
	&mov(&DWP(120,"esp"), "ebp");
	&mov("ebp", &DWP(28,"ebx"));
	&mov(&DWP(108,"esp"), "ebp");
	&mov("ebp", &DWP(32,"ebx"));
	&mov(&DWP(80,"esp"), "ebp");
	&mov("eax", &DWP(0,"ebx"));
	&mov("edx", &DWP(4,"ebx"));
	&mov("ecx", &DWP(8,"ebx"));
	&mov("ebp", &DWP(36,"ebx"));
	&mov(&DWP(112,"esp"), "eax");
	&mov(&DWP(116,"esp"), "edx");
	&mov(&DWP(124,"esp"), "ecx");
	&mov(&DWP(84,"esp"), "ebp");
	&mov("ecx", &DWP(12,"ebx"));
	&mov("edx", &DWP(16,"ebx"));
	&mov("eax", &DWP(20,"ebx"));
	&mov("ebp", &DWP(40,"ebx"));
	&mov("ebx", &DWP(44,"ebx"));
	&mov(&DWP(104,"esp"), "ebp");
	&mov(&DWP(100,"esp"), "ebx");
	&mov(&DWP(96,"esp"), "eax");
	&mov(&DWP(92,"esp"), "edx");
	&mov(&DWP(88,"esp"), "ecx");
	&jmp(&label("chacha_blocks_x86_4"));
	&set_label("chacha_blocks_x86_22");
	&mov("eax", &DWP(76,"esp"));
	&add("edi", -64);
	&add("eax", 64);
	&mov(&DWP(76,"esp"), "eax");
	&set_label("chacha_blocks_x86_4");
	&cmp("edi", 64);
	&jae(&label("chacha_blocks_x86_15"));
	&test("esi", "esi");
	&je(&label("chacha_blocks_x86_14"));
	&test("edi", "edi");
	&jbe(&label("chacha_blocks_x86_13"));
	&mov("ecx", "edi");
	&shr("ecx", 1);
	&test("ecx", "ecx");
	&jbe(&label("chacha_blocks_x86_23"));
	&xor("ebx", "ebx");
	&set_label("chacha_blocks_x86_9");
	&movzb("edx", &DWP(0,"esi,%ebx,2"));
	&mov(&DWP(4,"esp,%ebx,2"), "dl");
	&movzb("eax", &DWP(1,"esi,%ebx,2"));
	&mov(&DWP(5,"esp,%ebx,2"), "al");
	&inc("ebx");
	&cmp("ebx", "ecx");
	&jb(&label("chacha_blocks_x86_9"));
	&lea("edx", &DWP(1,"ebx,%ebx"));
	&set_label("chacha_blocks_x86_11");
	&lea("eax", &DWP(-1,"edx"));
	&cmp("eax", "edi");
	&jae(&label("chacha_blocks_x86_13"));
	&movzb("eax", &DWP(0,"eax,%esi"));
	&mov(&DWP(3,"esp,%edx"), "al");
	&set_label("chacha_blocks_x86_13");
	&lea("esi", &DWP(4,"esp"));
	&set_label("chacha_blocks_x86_14");
	&mov("eax", &DWP(76,"esp"));
	&lea("edx", &DWP(4,"esp"));
	&mov(&DWP(0,"esp"), "eax");
	&mov(&DWP(76,"esp"), "edx");
	&set_label("chacha_blocks_x86_15");
	&mov("ebp", &DWP(96,"esp"));
	&mov(&DWP(168,"esp"), "ebp");
	&mov("ebp", &DWP(80,"esp"));
	&mov("edx", &DWP(124,"esp"));
	&mov("ebx", &DWP(88,"esp"));
	&mov(&DWP(180,"esp"), "ebp");
	&mov("ebp", &DWP(224,"esp"));
	&mov(&DWP(148,"esp"), "edx");
	&mov(&DWP(152,"esp"), "ebx");
	&mov("edx", &DWP(120,"esp"));
	&mov("ebx", &DWP(108,"esp"));
	&mov(&DWP(136,"esp"), "edx");
	&mov("eax", &DWP(116,"esp"));
	&mov("edx", &DWP(84,"esp"));
	&mov(&DWP(140,"esp"), "ebx");
	&mov("ebx", &DWP(100,"esp"));
	&mov("ebp", &DWP(48,"ebp"));
	&test("ebp", "ebp");
	&mov(&DWP(144,"esp"), "eax");
	&mov(&DWP(176,"esp"), "edx");
	&mov("ecx", &DWP(112,"esp"));
	&mov("eax", &DWP(92,"esp"));
	&mov("edx", &DWP(104,"esp"));
	&mov(&DWP(132,"esp"), "ebx");
	&mov("ebx", 1634760805);
	&mov(&DWP(160,"esp"), 857760878);
	&mov(&DWP(164,"esp"), 2036477234);
	&mov(&DWP(172,"esp"), 1797285236);
	&jbe(&label("chacha_blocks_x86_19"));
	&mov(&DWP(156,"esp"), "ebx");
	&mov(&DWP(128,"esp"), "ebp");
	&mov(&DWP(72,"esp"), "edi");
	&mov(&DWP(68,"esp"), "esi");
	&set_label("chacha_blocks_x86_17");
	&mov("ebp", &DWP(156,"esp"));
	&add("ebp", "ecx");
	&mov("edi", &DWP(180,"esp"));
	&xor("edi", "ebp");
	&rol("edi", 16);
	&add("eax", "edi");
	&xor("ecx", "eax");
	&rol("ecx", 12);
	&add("ebp", "ecx");
	&xor("edi", "ebp");
	&rol("edi", 8);
	&add("eax", "edi");
	&xor("ecx", "eax");
	&mov(&DWP(184,"esp"), "eax");
	&mov("esi", &DWP(144,"esp"));
	&mov("eax", &DWP(160,"esp"));
	&add("eax", "esi");
	&rol("ecx", 7);
	&mov(&DWP(188,"esp"), "ecx");
	&mov("ecx", &DWP(176,"esp"));
	&xor("ecx", "eax");
	&rol("ecx", 16);
	&mov("ebx", &DWP(168,"esp"));
	&add("ebx", "ecx");
	&xor("esi", "ebx");
	&rol("esi", 12);
	&add("eax", "esi");
	&mov(&DWP(180,"esp"), "edi");
	&xor("ecx", "eax");
	&mov(&DWP(160,"esp"), "eax");
	&mov("edi", &DWP(148,"esp"));
	&mov("eax", &DWP(164,"esp"));
	&add("eax", "edi");
	&xor("edx", "eax");
	&rol("ecx", 8);
	&rol("edx", 16);
	&add("ebx", "ecx");
	&mov(&DWP(176,"esp"), "ecx");
	&xor("esi", "ebx");
	&mov("ecx", &DWP(136,"esp"));
	&add("ecx", "edx");
	&xor("edi", "ecx");
	&rol("edi", 12);
	&add("eax", "edi");
	&xor("edx", "eax");
	&rol("edx", 8);
	&add("ecx", "edx");
	&xor("edi", "ecx");
	&mov(&DWP(168,"esp"), "ebx");
	&mov(&DWP(192,"esp"), "edx");
	&mov("edx", &DWP(152,"esp"));
	&mov("ebx", &DWP(172,"esp"));
	&add("ebx", "edx");
	&rol("edi", 7);
	&mov(&DWP(148,"esp"), "edi");
	&mov("edi", &DWP(132,"esp"));
	&xor("edi", "ebx");
	&rol("edi", 16);
	&mov(&DWP(164,"esp"), "eax");
	&mov("eax", &DWP(140,"esp"));
	&add("eax", "edi");
	&xor("edx", "eax");
	&rol("edx", 12);
	&add("ebx", "edx");
	&rol("esi", 7);
	&xor("edi", "ebx");
	&rol("edi", 8);
	&add("ebp", "esi");
	&add("eax", "edi");
	&xor("edi", "ebp");
	&rol("edi", 16);
	&xor("edx", "eax");
	&add("ecx", "edi");
	&xor("esi", "ecx");
	&rol("esi", 12);
	&add("ebp", "esi");
	&xor("edi", "ebp");
	&rol("edi", 8);
	&add("ecx", "edi");
	&xor("esi", "ecx");
	&mov(&DWP(156,"esp"), "ebp");
	&mov(&DWP(136,"esp"), "ecx");
	&mov("ecx", &DWP(148,"esp"));
	&mov("ebp", &DWP(160,"esp"));
	&add("ebp", "ecx");
	&rol("esi", 7);
	&mov(&DWP(144,"esp"), "esi");
	&mov("esi", &DWP(180,"esp"));
	&xor("esi", "ebp");
	&rol("esi", 16);
	&add("eax", "esi");
	&xor("ecx", "eax");
	&rol("ecx", 12);
	&add("ebp", "ecx");
	&rol("edx", 7);
	&xor("esi", "ebp");
	&mov(&DWP(160,"esp"), "ebp");
	&mov("ebp", &DWP(164,"esp"));
	&add("ebp", "edx");
	&mov(&DWP(132,"esp"), "edi");
	&rol("esi", 8);
	&mov("edi", &DWP(176,"esp"));
	&add("eax", "esi");
	&xor("edi", "ebp");
	&xor("ecx", "eax");
	&rol("edi", 16);
	&mov(&DWP(140,"esp"), "eax");
	&mov("eax", &DWP(184,"esp"));
	&add("eax", "edi");
	&xor("edx", "eax");
	&rol("edx", 12);
	&add("ebp", "edx");
	&xor("edi", "ebp");
	&rol("edi", 8);
	&add("eax", "edi");
	&rol("ecx", 7);
	&xor("edx", "eax");
	&mov(&DWP(148,"esp"), "ecx");
	&mov("ecx", &DWP(188,"esp"));
	&add("ebx", "ecx");
	&rol("edx", 7);
	&mov(&DWP(152,"esp"), "edx");
	&mov("edx", &DWP(192,"esp"));
	&xor("edx", "ebx");
	&rol("edx", 16);
	&mov(&DWP(164,"esp"), "ebp");
	&mov("ebp", &DWP(168,"esp"));
	&add("ebp", "edx");
	&xor("ecx", "ebp");
	&rol("ecx", 12);
	&add("ebx", "ecx");
	&xor("edx", "ebx");
	&rol("edx", 8);
	&add("ebp", "edx");
	&xor("ecx", "ebp");
	&mov(&DWP(172,"esp"), "ebx");
	&rol("ecx", 7);
	&mov("ebx", &DWP(128,"esp"));
	&mov(&DWP(180,"esp"), "esi");
	&add("ebx", -2);
	&mov(&DWP(176,"esp"), "edi");
	&mov(&DWP(168,"esp"), "ebp");
	&mov(&DWP(128,"esp"), "ebx");
	&jne(&label("chacha_blocks_x86_17"));
	&mov("ebx", &DWP(156,"esp"));
	&mov("edi", &DWP(72,"esp"));
	&mov("esi", &DWP(68,"esp"));
	&set_label("chacha_blocks_x86_19");
	&add("ebx", 1634760805);
	&test("esi", "esi");
	&je(&label("chacha_blocks_x86_30"));
	&mov("ebp", &DWP(0,"esi"));
	&xor("ebp", "ebx");
	&mov("ebx", &DWP(76,"esp"));
	&add("ecx", &DWP(112,"esp"));
	&add("eax", &DWP(92,"esp"));
	&mov(&DWP(0,"ebx"), "ebp");
	&mov("ebp", &DWP(160,"esp"));
	&add("ebp", 857760878);
	&xor("ebp", &DWP(4,"esi"));
	&mov(&DWP(4,"ebx"), "ebp");
	&mov("ebp", &DWP(164,"esp"));
	&add("ebp", 2036477234);
	&xor("ebp", &DWP(8,"esi"));
	&mov(&DWP(8,"ebx"), "ebp");
	&mov("ebp", &DWP(172,"esp"));
	&add("ebp", 1797285236);
	&xor("ebp", &DWP(12,"esi"));
	&mov(&DWP(12,"ebx"), "ebp");
	&xor("ecx", &DWP(16,"esi"));
	&mov(&DWP(16,"ebx"), "ecx");
	&mov("ecx", &DWP(144,"esp"));
	&add("ecx", &DWP(116,"esp"));
	&mov("ebp", &DWP(148,"esp"));
	&xor("ecx", &DWP(20,"esi"));
	&mov(&DWP(20,"ebx"), "ecx");
	&add("ebp", &DWP(124,"esp"));
	&mov("ecx", &DWP(152,"esp"));
	&xor("ebp", &DWP(24,"esi"));
	&mov(&DWP(24,"ebx"), "ebp");
	&add("ecx", &DWP(88,"esp"));
	&xor("ecx", &DWP(28,"esi"));
	&mov(&DWP(28,"ebx"), "ecx");
	&xor("eax", &DWP(32,"esi"));
	&mov(&DWP(32,"ebx"), "eax");
	&mov("eax", &DWP(168,"esp"));
	&add("eax", &DWP(96,"esp"));
	&mov("ecx", &DWP(136,"esp"));
	&xor("eax", &DWP(36,"esi"));
	&mov(&DWP(36,"ebx"), "eax");
	&add("ecx", &DWP(120,"esp"));
	&mov("ebp", &DWP(140,"esp"));
	&xor("ecx", &DWP(40,"esi"));
	&mov(&DWP(40,"ebx"), "ecx");
	&add("ebp", &DWP(108,"esp"));
	&mov("eax", &DWP(180,"esp"));
	&xor("ebp", &DWP(44,"esi"));
	&mov(&DWP(44,"ebx"), "ebp");
	&add("eax", &DWP(80,"esp"));
	&mov("ecx", &DWP(176,"esp"));
	&xor("eax", &DWP(48,"esi"));
	&mov(&DWP(48,"ebx"), "eax");
	&add("ecx", &DWP(84,"esp"));
	&xor("ecx", &DWP(52,"esi"));
	&mov(&DWP(52,"ebx"), "ecx");
	&add("edx", &DWP(104,"esp"));
	&xor("edx", &DWP(56,"esi"));
	&mov(&DWP(56,"ebx"), "edx");
	&mov("edx", &DWP(132,"esp"));
	&add("edx", &DWP(100,"esp"));
	&xor("edx", &DWP(60,"esi"));
	&add("esi", 64);
	&mov(&DWP(60,"ebx"), "edx");
	&set_label("chacha_blocks_x86_21");
	&mov("edx", &DWP(84,"esp"));
	&inc(&DWP(80,"esp"));
	&lea("eax", &DWP(1,"edx"));
	&jne(&label("..L2"));
	&mov("edx", "eax");
	&set_label("..L2");
	&mov(&DWP(84,"esp"), "edx");
	&cmp("edi", 64);
	&jbe(&label("chacha_blocks_x86_24"));
	&jmp(&label("chacha_blocks_x86_22"));
	&set_label("chacha_blocks_x86_23");
	&mov("edx", 1);
	&jmp(&label("chacha_blocks_x86_11"));
	&set_label("chacha_blocks_x86_24");
	&jae(&label("chacha_blocks_x86_29"));
	&test("edi", "edi");
	&jbe(&label("chacha_blocks_x86_29"));
	&mov("ecx", &DWP(0,"esp"));
	&xor("edx", "edx");
	&mov("ebx", &DWP(76,"esp"));
	&set_label("chacha_blocks_x86_27");
	&movzb("eax", &DWP(0,"edx,%ebx"));
	&mov(&DWP(0,"edx,%ecx"), "al");
	&inc("edx");
	&cmp("edx", "edi");
	&jb(&label("chacha_blocks_x86_27"));
	&set_label("chacha_blocks_x86_29");
	&mov("ecx", &DWP(224,"esp"));
	&mov("eax", &DWP(80,"esp"));
	&mov("edx", &DWP(84,"esp"));
	&mov(&DWP(32,"ecx"), "eax");
	&mov(&DWP(36,"ecx"), "edx");
	&add("esp", 204);
	&pop("ebp");
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
	&set_label("chacha_blocks_x86_30");
	&mov("ebp", &DWP(76,"esp"));
	&add("ecx", &DWP(112,"esp"));
	&add("eax", &DWP(92,"esp"));
	&mov(&DWP(16,"ebp"), "ecx");
	&mov("ecx", &DWP(144,"esp"));
	&add("ecx", &DWP(116,"esp"));
	&mov(&DWP(20,"ebp"), "ecx");
	&mov("ecx", &DWP(152,"esp"));
	&mov(&DWP(0,"ebp"), "ebx");
	&mov("ebx", &DWP(160,"esp"));
	&add("ebx", 857760878);
	&mov(&DWP(4,"ebp"), "ebx");
	&mov("ebx", &DWP(164,"esp"));
	&add("ecx", &DWP(88,"esp"));
	&add("ebx", 2036477234);
	&mov(&DWP(8,"ebp"), "ebx");
	&mov("ebx", &DWP(172,"esp"));
	&add("ebx", 1797285236);
	&mov(&DWP(12,"ebp"), "ebx");
	&mov(&DWP(28,"ebp"), "ecx");
	&mov(&DWP(32,"ebp"), "eax");
	&mov("ebx", &DWP(148,"esp"));
	&mov("eax", &DWP(168,"esp"));
	&mov("ecx", &DWP(136,"esp"));
	&add("ebx", &DWP(124,"esp"));
	&add("eax", &DWP(96,"esp"));
	&add("ecx", &DWP(120,"esp"));
	&add("edx", &DWP(104,"esp"));
	&mov(&DWP(24,"ebp"), "ebx");
	&mov(&DWP(36,"ebp"), "eax");
	&mov(&DWP(40,"ebp"), "ecx");
	&mov(&DWP(56,"ebp"), "edx");
	&mov("ebx", &DWP(140,"esp"));
	&mov("eax", &DWP(180,"esp"));
	&mov("ecx", &DWP(176,"esp"));
	&mov("edx", &DWP(132,"esp"));
	&add("ebx", &DWP(108,"esp"));
	&add("eax", &DWP(80,"esp"));
	&add("ecx", &DWP(84,"esp"));
	&add("edx", &DWP(100,"esp"));
	&mov(&DWP(44,"ebp"), "ebx");
	&mov(&DWP(48,"ebp"), "eax");
	&mov(&DWP(52,"ebp"), "ecx");
	&mov(&DWP(60,"ebp"), "edx");
	&jmp(&label("chacha_blocks_x86_21"));
}

sub chacha_blocks_sse2 {
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", ~63);
	&sub("esp", 704);
	&mov(&DWP(68,"esp"), "ebx");
	&mov(&DWP(72,"esp"), "esi");
	&mov(&DWP(76,"esp"), "edi");
	&mov("ecx", &DWP(8,"ebp"));
	&mov(&DWP(84,"esp"), "ecx");
	&mov("esi", &DWP(12,"ebp"));
	&mov("edx", &DWP(16,"ebp"));
	&mov("eax", &DWP(20,"ebp"));
	&label_to_reg("chacha_constants", "ebx");
	&movdqa("xmm0", &DWP(0,"ebx"));
	&movdqu("xmm1", &DWP(0,"ecx"));
	&movdqu("xmm2", &DWP(16,"ecx"));
	&movdqu("xmm3", &DWP(32,"ecx"));
	&movdqa(&DWP(0,"esp"), "xmm0");
	&movdqa(&DWP(16,"esp"), "xmm1");
	&movdqa(&DWP(32,"esp"), "xmm2");
	&movdqa(&DWP(48,"esp"), "xmm3");
	&mov("ecx", &DWP(48,"ecx"));
	&mov(&DWP(88,"esp"), "ecx");
	&cmp("eax", 0);
	&jbe(&label("chacha_blocks_sse2_done"));
	&cmp("eax", 256);
	&jb(&label("chacha_blocks_sse2_bytesbetween1and255"));
	&pshufd("xmm4", "xmm0", 0x00);
	&pshufd("xmm5", "xmm0", 0x55);
	&pshufd("xmm6", "xmm0", 0xaa);
	&pshufd("xmm0", "xmm0", 0xff);
	&movdqa(&DWP(128,"esp"), "xmm4");
	&movdqa(&DWP(144,"esp"), "xmm5");
	&movdqa(&DWP(160,"esp"), "xmm6");
	&movdqa(&DWP(176,"esp"), "xmm0");
	&pshufd("xmm0", "xmm1", 0x00);
	&pshufd("xmm4", "xmm1", 0x55);
	&pshufd("xmm5", "xmm1", 0xaa);
	&pshufd("xmm1", "xmm1", 0xff);
	&movdqa(&DWP(192,"esp"), "xmm0");
	&movdqa(&DWP(208,"esp"), "xmm4");
	&movdqa(&DWP(224,"esp"), "xmm5");
	&movdqa(&DWP(240,"esp"), "xmm1");
	&pshufd("xmm0", "xmm2", 0x00);
	&pshufd("xmm1", "xmm2", 0x55);
	&pshufd("xmm4", "xmm2", 0xaa);
	&pshufd("xmm2", "xmm2", 0xff);
	&movdqa(&DWP(256,"esp"), "xmm0");
	&movdqa(&DWP(272,"esp"), "xmm1");
	&movdqa(&DWP(288,"esp"), "xmm4");
	&movdqa(&DWP(304,"esp"), "xmm2");
	&pshufd("xmm0", "xmm3", 0xaa);
	&pshufd("xmm1", "xmm3", 0xff);
	&movdqa(&DWP(352,"esp"), "xmm0");
	&movdqa(&DWP(368,"esp"), "xmm1");
	&set_label("chacha_blocks_sse2_bytesatleast256");
	&mov("ecx", &DWP(48,"esp"));
	&mov("ebx", &DWP(4+48,"esp"));
	&mov(&DWP(320,"esp"), "ecx");
	&mov(&DWP(336,"esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(4+320,"esp"), "ecx");
	&mov(&DWP(4+336,"esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(8+320,"esp"), "ecx");
	&mov(&DWP(8+336,"esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(12+320,"esp"), "ecx");
	&mov(&DWP(12+336,"esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(48,"esp"), "ecx");
	&mov(&DWP(4+48,"esp"), "ebx");
	&mov(&DWP(92,"esp"), "eax");
	&mov("eax", &DWP(88,"esp"));
	&movdqa("xmm0", &DWP(160,"esp"));
	&movdqa("xmm1", &DWP(224,"esp"));
	&movdqa("xmm2", &DWP(288,"esp"));
	&movdqa("xmm3", &DWP(352,"esp"));
	&movdqa("xmm4", &DWP(176,"esp"));
	&movdqa("xmm5", &DWP(240,"esp"));
	&movdqa("xmm6", &DWP(304,"esp"));
	&movdqa("xmm7", &DWP(368,"esp"));
	&movdqa(&DWP(480,"esp"), "xmm0");
	&movdqa(&DWP(544,"esp"), "xmm1");
	&movdqa(&DWP(608,"esp"), "xmm2");
	&movdqa(&DWP(672,"esp"), "xmm3");
	&movdqa(&DWP(496,"esp"), "xmm4");
	&movdqa(&DWP(560,"esp"), "xmm5");
	&movdqa(&DWP(624,"esp"), "xmm6");
	&movdqa(&DWP(688,"esp"), "xmm7");
	&movdqa("xmm0", &DWP(128,"esp"));
	&movdqa("xmm1", &DWP(192,"esp"));
	&movdqa("xmm2", &DWP(256,"esp"));
	&movdqa("xmm3", &DWP(320,"esp"));
	&movdqa("xmm4", &DWP(144,"esp"));
	&movdqa("xmm5", &DWP(208,"esp"));
	&movdqa("xmm6", &DWP(272,"esp"));
	&movdqa("xmm7", &DWP(336,"esp"));
	&jmp(&label("chacha_blocks_sse2_mainloop1"));
	&set_label("chacha_blocks_sse2_mainloop1", 64);
	&paddd("xmm0", "xmm1");
	&paddd("xmm4", "xmm5");
	&pxor("xmm3", "xmm0");
	&pxor("xmm7", "xmm4");
	&pshuflw("xmm3", "xmm3", 0xb1);
	&pshuflw("xmm7", "xmm7", 0xb1);
	&pshufhw("xmm3", "xmm3", 0xb1);
	&pshufhw("xmm7", "xmm7", 0xb1);
	&movdqa(&DWP(464,"esp"), "xmm4");
	&paddd("xmm2", "xmm3");
	&paddd("xmm6", "xmm7");
	&pxor("xmm1", "xmm2");
	&pxor("xmm5", "xmm6");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&por("xmm1", "xmm4");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 12);
	&psrld("xmm4", 20);
	&por("xmm5", "xmm4");
	&movdqa("xmm4", &DWP(464,"esp"));
	&paddd("xmm0", "xmm1");
	&paddd("xmm4", "xmm5");
	&pxor("xmm3", "xmm0");
	&pxor("xmm7", "xmm4");
	&movdqa(&DWP(448,"esp"), "xmm0");
	&movdqa(&DWP(464,"esp"), "xmm4");
	&movdqa("xmm4", "xmm3");
	&pslld("xmm3", 8);
	&psrld("xmm4", 24);
	&por("xmm3", "xmm4");
	&movdqa("xmm4", "xmm7");
	&pslld("xmm7", 8);
	&psrld("xmm4", 24);
	&por("xmm7", "xmm4");
	&paddd("xmm2", "xmm3");
	&paddd("xmm6", "xmm7");
	&movdqa(&DWP(640,"esp"), "xmm3");
	&movdqa(&DWP(656,"esp"), "xmm7");
	&pxor("xmm1", "xmm2");
	&pxor("xmm5", "xmm6");
	&movdqa(&DWP(576,"esp"), "xmm2");
	&movdqa(&DWP(592,"esp"), "xmm6");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&movdqa("xmm3", &DWP(672,"esp"));
	&psrld("xmm4", 25);
	&movdqa("xmm7", &DWP(688,"esp"));
	&por("xmm1", "xmm4");
	&movdqa(&DWP(512,"esp"), "xmm1");
	&movdqa("xmm0", "xmm5");
	&pslld("xmm5", 7);
	&movdqa("xmm2", &DWP(608,"esp"));
	&psrld("xmm0", 25);
	&movdqa("xmm6", &DWP(624,"esp"));
	&por("xmm5", "xmm0");
	&movdqa(&DWP(528,"esp"), "xmm5");
	&movdqa("xmm0", &DWP(480,"esp"));
	&movdqa("xmm4", &DWP(496,"esp"));
	&movdqa("xmm1", &DWP(544,"esp"));
	&movdqa("xmm5", &DWP(560,"esp"));
	&paddd("xmm0", "xmm1");
	&paddd("xmm4", "xmm5");
	&pxor("xmm3", "xmm0");
	&pxor("xmm7", "xmm4");
	&pshuflw("xmm3", "xmm3", 0xb1);
	&pshuflw("xmm7", "xmm7", 0xb1);
	&pshufhw("xmm3", "xmm3", 0xb1);
	&pshufhw("xmm7", "xmm7", 0xb1);
	&movdqa(&DWP(496,"esp"), "xmm4");
	&paddd("xmm2", "xmm3");
	&paddd("xmm6", "xmm7");
	&pxor("xmm1", "xmm2");
	&pxor("xmm5", "xmm6");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&por("xmm1", "xmm4");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 12);
	&psrld("xmm4", 20);
	&por("xmm5", "xmm4");
	&movdqa("xmm4", &DWP(496,"esp"));
	&paddd("xmm0", "xmm1");
	&paddd("xmm4", "xmm5");
	&pxor("xmm3", "xmm0");
	&pxor("xmm7", "xmm4");
	&movdqa(&DWP(480,"esp"), "xmm0");
	&movdqa(&DWP(496,"esp"), "xmm4");
	&movdqa("xmm4", "xmm3");
	&pslld("xmm3", 8);
	&psrld("xmm4", 24);
	&por("xmm3", "xmm4");
	&movdqa("xmm4", "xmm7");
	&pslld("xmm7", 8);
	&psrld("xmm4", 24);
	&por("xmm7", "xmm4");
	&paddd("xmm2", "xmm3");
	&paddd("xmm6", "xmm7");
	&movdqa(&DWP(672,"esp"), "xmm3");
	&pxor("xmm1", "xmm2");
	&pxor("xmm5", "xmm6");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&psrld("xmm4", 25);
	&movdqa("xmm3", &DWP(640,"esp"));
	&por("xmm1", "xmm4");
	&movdqa("xmm0", "xmm5");
	&pslld("xmm5", 7);
	&psrld("xmm0", 25);
	&por("xmm5", "xmm0");
	&movdqa(&DWP(560,"esp"), "xmm5");
	&movdqa("xmm0", &DWP(448,"esp"));
	&movdqa("xmm5", &DWP(528,"esp"));
	&movdqa("xmm4", &DWP(464,"esp"));
	&paddd("xmm0", "xmm5");
	&paddd("xmm4", "xmm1");
	&pxor("xmm7", "xmm0");
	&pxor("xmm3", "xmm4");
	&pshuflw("xmm7", "xmm7", 0xb1);
	&pshuflw("xmm3", "xmm3", 0xb1);
	&pshufhw("xmm7", "xmm7", 0xb1);
	&pshufhw("xmm3", "xmm3", 0xb1);
	&movdqa(&DWP(464,"esp"), "xmm4");
	&paddd("xmm2", "xmm7");
	&paddd("xmm6", "xmm3");
	&pxor("xmm5", "xmm2");
	&pxor("xmm1", "xmm6");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 12);
	&psrld("xmm4", 20);
	&por("xmm5", "xmm4");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&por("xmm1", "xmm4");
	&movdqa("xmm4", &DWP(464,"esp"));
	&paddd("xmm0", "xmm5");
	&paddd("xmm4", "xmm1");
	&pxor("xmm7", "xmm0");
	&pxor("xmm3", "xmm4");
	&movdqa(&DWP(448,"esp"), "xmm0");
	&movdqa(&DWP(464,"esp"), "xmm4");
	&movdqa("xmm4", "xmm7");
	&pslld("xmm7", 8);
	&psrld("xmm4", 24);
	&por("xmm7", "xmm4");
	&movdqa("xmm4", "xmm3");
	&pslld("xmm3", 8);
	&psrld("xmm4", 24);
	&por("xmm3", "xmm4");
	&paddd("xmm2", "xmm7");
	&paddd("xmm6", "xmm3");
	&movdqa(&DWP(688,"esp"), "xmm7");
	&movdqa(&DWP(640,"esp"), "xmm3");
	&pxor("xmm5", "xmm2");
	&pxor("xmm1", "xmm6");
	&movdqa(&DWP(608,"esp"), "xmm2");
	&movdqa(&DWP(624,"esp"), "xmm6");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 7);
	&movdqa("xmm7", &DWP(656,"esp"));
	&psrld("xmm4", 25);
	&movdqa("xmm3", &DWP(672,"esp"));
	&por("xmm5", "xmm4");
	&movdqa(&DWP(528,"esp"), "xmm5");
	&movdqa("xmm0", "xmm1");
	&pslld("xmm1", 7);
	&movdqa("xmm2", &DWP(576,"esp"));
	&psrld("xmm0", 25);
	&movdqa("xmm6", &DWP(592,"esp"));
	&por("xmm1", "xmm0");
	&movdqa(&DWP(544,"esp"), "xmm1");
	&movdqa("xmm0", &DWP(480,"esp"));
	&movdqa("xmm4", &DWP(496,"esp"));
	&movdqa("xmm5", &DWP(560,"esp"));
	&movdqa("xmm1", &DWP(512,"esp"));
	&paddd("xmm0", "xmm5");
	&paddd("xmm4", "xmm1");
	&pxor("xmm7", "xmm0");
	&pxor("xmm3", "xmm4");
	&pshuflw("xmm7", "xmm7", 0xb1);
	&pshuflw("xmm3", "xmm3", 0xb1);
	&pshufhw("xmm7", "xmm7", 0xb1);
	&pshufhw("xmm3", "xmm3", 0xb1);
	&movdqa(&DWP(496,"esp"), "xmm4");
	&paddd("xmm2", "xmm7");
	&paddd("xmm6", "xmm3");
	&pxor("xmm5", "xmm2");
	&pxor("xmm1", "xmm6");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 12);
	&psrld("xmm4", 20);
	&por("xmm5", "xmm4");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&por("xmm1", "xmm4");
	&movdqa("xmm4", &DWP(496,"esp"));
	&paddd("xmm0", "xmm5");
	&paddd("xmm4", "xmm1");
	&pxor("xmm7", "xmm0");
	&pxor("xmm3", "xmm4");
	&movdqa(&DWP(480,"esp"), "xmm0");
	&movdqa(&DWP(496,"esp"), "xmm4");
	&movdqa("xmm4", "xmm7");
	&pslld("xmm7", 8);
	&psrld("xmm4", 24);
	&por("xmm7", "xmm4");
	&movdqa("xmm4", "xmm3");
	&pslld("xmm3", 8);
	&psrld("xmm4", 24);
	&por("xmm3", "xmm4");
	&paddd("xmm2", "xmm7");
	&paddd("xmm6", "xmm3");
	&movdqa(&DWP(672,"esp"), "xmm3");
	&pxor("xmm5", "xmm2");
	&pxor("xmm1", "xmm6");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 7);
	&psrld("xmm4", 25);
	&movdqa("xmm3", &DWP(640,"esp"));
	&por("xmm5", "xmm4");
	&movdqa(&DWP(560,"esp"), "xmm5");
	&movdqa("xmm0", "xmm1");
	&pslld("xmm1", 7);
	&psrld("xmm0", 25);
	&por("xmm1", "xmm0");
	&movdqa("xmm0", &DWP(448,"esp"));
	&movdqa("xmm4", &DWP(464,"esp"));
	&movdqa("xmm5", &DWP(528,"esp"));
	&sub("eax", 2);
	&ja(&label("chacha_blocks_sse2_mainloop1"));
	&movdqa(&DWP(448,"esp"), "xmm0");
	&movdqa(&DWP(512,"esp"), "xmm1");
	&movdqa(&DWP(576,"esp"), "xmm2");
	&movdqa(&DWP(640,"esp"), "xmm3");
	&movdqa(&DWP(464,"esp"), "xmm4");
	&movdqa(&DWP(528,"esp"), "xmm5");
	&movdqa(&DWP(592,"esp"), "xmm6");
	&movdqa(&DWP(656,"esp"), "xmm7");
	&cmp("esi", 0);
	&movdqa("xmm0", &DWP(448,"esp"));
	&movdqa("xmm1", &DWP(464,"esp"));
	&movdqa("xmm2", &DWP(480,"esp"));
	&movdqa("xmm3", &DWP(496,"esp"));
	&paddd("xmm0", &DWP(128,"esp"));
	&paddd("xmm1", &DWP(144,"esp"));
	&paddd("xmm2", &DWP(160,"esp"));
	&paddd("xmm3", &DWP(176,"esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&jbe(&label("chacha_blocks_sse2_noinput1"));
	&movdqu("xmm2", &DWP(0,"esi"));
	&movdqu("xmm5", &DWP(64,"esi"));
	&movdqu("xmm6", &DWP(128,"esi"));
	&movdqu("xmm7", &DWP(192,"esi"));
	&pxor("xmm4", "xmm2");
	&pxor("xmm1", "xmm5");
	&pxor("xmm0", "xmm6");
	&pxor("xmm3", "xmm7");
	&movdqu(&DWP(0,"edx"), "xmm4");
	&movdqu(&DWP(64,"edx"), "xmm1");
	&movdqu(&DWP(128,"edx"), "xmm0");
	&movdqu(&DWP(192,"edx"), "xmm3");
	&movdqa("xmm0", &DWP(512,"esp"));
	&movdqa("xmm1", &DWP(528,"esp"));
	&movdqa("xmm2", &DWP(544,"esp"));
	&movdqa("xmm3", &DWP(560,"esp"));
	&paddd("xmm0", &DWP(192,"esp"));
	&paddd("xmm1", &DWP(208,"esp"));
	&paddd("xmm2", &DWP(224,"esp"));
	&paddd("xmm3", &DWP(240,"esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu("xmm2", &DWP(16+0,"esi"));
	&movdqu("xmm5", &DWP(16+64,"esi"));
	&movdqu("xmm6", &DWP(16+128,"esi"));
	&movdqu("xmm7", &DWP(16+192,"esi"));
	&pxor("xmm4", "xmm2");
	&pxor("xmm1", "xmm5");
	&pxor("xmm0", "xmm6");
	&pxor("xmm3", "xmm7");
	&movdqu(&DWP(16+0,"edx"), "xmm4");
	&movdqu(&DWP(16+64,"edx"), "xmm1");
	&movdqu(&DWP(16+128,"edx"), "xmm0");
	&movdqu(&DWP(16+192,"edx"), "xmm3");
	&movdqa("xmm0", &DWP(576,"esp"));
	&movdqa("xmm1", &DWP(592,"esp"));
	&movdqa("xmm2", &DWP(608,"esp"));
	&movdqa("xmm3", &DWP(624,"esp"));
	&paddd("xmm0", &DWP(256,"esp"));
	&paddd("xmm1", &DWP(272,"esp"));
	&paddd("xmm2", &DWP(288,"esp"));
	&paddd("xmm3", &DWP(304,"esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu("xmm2", &DWP(32+0,"esi"));
	&movdqu("xmm5", &DWP(32+64,"esi"));
	&movdqu("xmm6", &DWP(32+128,"esi"));
	&movdqu("xmm7", &DWP(32+192,"esi"));
	&pxor("xmm4", "xmm2");
	&pxor("xmm1", "xmm5");
	&pxor("xmm0", "xmm6");
	&pxor("xmm3", "xmm7");
	&movdqu(&DWP(32+0,"edx"), "xmm4");
	&movdqu(&DWP(32+64,"edx"), "xmm1");
	&movdqu(&DWP(32+128,"edx"), "xmm0");
	&movdqu(&DWP(32+192,"edx"), "xmm3");
	&movdqa("xmm0", &DWP(640,"esp"));
	&movdqa("xmm1", &DWP(656,"esp"));
	&movdqa("xmm2", &DWP(672,"esp"));
	&movdqa("xmm3", &DWP(688,"esp"));
	&paddd("xmm0", &DWP(320,"esp"));
	&paddd("xmm1", &DWP(336,"esp"));
	&paddd("xmm2", &DWP(352,"esp"));
	&paddd("xmm3", &DWP(368,"esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu("xmm2", &DWP(48+0,"esi"));
	&movdqu("xmm5", &DWP(48+64,"esi"));
	&movdqu("xmm6", &DWP(48+128,"esi"));
	&movdqu("xmm7", &DWP(48+192,"esi"));
	&pxor("xmm4", "xmm2");
	&pxor("xmm1", "xmm5");
	&pxor("xmm0", "xmm6");
	&pxor("xmm3", "xmm7");
	&movdqu(&DWP(48+0,"edx"), "xmm4");
	&movdqu(&DWP(48+64,"edx"), "xmm1");
	&movdqu(&DWP(48+128,"edx"), "xmm0");
	&movdqu(&DWP(48+192,"edx"), "xmm3");
	&add("esi", 256);
	&jmp(&label("chacha_blocks_sse2_mainloop1_cont"));
	&set_label("chacha_blocks_sse2_noinput1");
	&movdqu(&DWP(0,"edx"), "xmm4");
	&movdqu(&DWP(64,"edx"), "xmm1");
	&movdqu(&DWP(128,"edx"), "xmm0");
	&movdqu(&DWP(192,"edx"), "xmm3");
	&movdqa("xmm0", &DWP(512,"esp"));
	&movdqa("xmm1", &DWP(528,"esp"));
	&movdqa("xmm2", &DWP(544,"esp"));
	&movdqa("xmm3", &DWP(560,"esp"));
	&paddd("xmm0", &DWP(192,"esp"));
	&paddd("xmm1", &DWP(208,"esp"));
	&paddd("xmm2", &DWP(224,"esp"));
	&paddd("xmm3", &DWP(240,"esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu(&DWP(16+0,"edx"), "xmm4");
	&movdqu(&DWP(16+64,"edx"), "xmm1");
	&movdqu(&DWP(16+128,"edx"), "xmm0");
	&movdqu(&DWP(16+192,"edx"), "xmm3");
	&movdqa("xmm0", &DWP(576,"esp"));
	&movdqa("xmm1", &DWP(592,"esp"));
	&movdqa("xmm2", &DWP(608,"esp"));
	&movdqa("xmm3", &DWP(624,"esp"));
	&paddd("xmm0", &DWP(256,"esp"));
	&paddd("xmm1", &DWP(272,"esp"));
	&paddd("xmm2", &DWP(288,"esp"));
	&paddd("xmm3", &DWP(304,"esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu(&DWP(32+0,"edx"), "xmm4");
	&movdqu(&DWP(32+64,"edx"), "xmm1");
	&movdqu(&DWP(32+128,"edx"), "xmm0");
	&movdqu(&DWP(32+192,"edx"), "xmm3");
	&movdqa("xmm0", &DWP(640,"esp"));
	&movdqa("xmm1", &DWP(656,"esp"));
	&movdqa("xmm2", &DWP(672,"esp"));
	&movdqa("xmm3", &DWP(688,"esp"));
	&paddd("xmm0", &DWP(320,"esp"));
	&paddd("xmm1", &DWP(336,"esp"));
	&paddd("xmm2", &DWP(352,"esp"));
	&paddd("xmm3", &DWP(368,"esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu(&DWP(48+0,"edx"), "xmm4");
	&movdqu(&DWP(48+64,"edx"), "xmm1");
	&movdqu(&DWP(48+128,"edx"), "xmm0");
	&movdqu(&DWP(48+192,"edx"), "xmm3");
	&set_label("chacha_blocks_sse2_mainloop1_cont");
	&mov("eax", &DWP(92,"esp"));
	&sub("eax", 256);
	&add("edx", 256);
	&cmp("eax", 256);
	&jae(&label("chacha_blocks_sse2_bytesatleast256"));
	&cmp("eax", 0);
	&jbe(&label("chacha_blocks_sse2_done"));
	&set_label("chacha_blocks_sse2_bytesbetween1and255");
	&cmp("eax", 64);
	&jae(&label("chacha_blocks_sse2_nocopy"));
	&mov(&DWP(92,"esp"), "edx");
	&cmp("esi", 0);
	&jbe(&label("chacha_blocks_sse2_noinput2"));
	&lea("edi", &DWP(128,"esp"));
	&mov("ecx", "eax");
	&rep_movsb();
	&lea("esi", &DWP(128,"esp"));
	&set_label("chacha_blocks_sse2_noinput2");
	&lea("edx", &DWP(128,"esp"));
	&set_label("chacha_blocks_sse2_nocopy");
	&mov(&DWP(80,"esp"), "eax");
	&movdqa("xmm0", &DWP(0,"esp"));
	&movdqa("xmm1", &DWP(16,"esp"));
	&movdqa("xmm2", &DWP(32,"esp"));
	&movdqa("xmm3", &DWP(48,"esp"));
	&mov("eax", &DWP(88,"esp"));
	&set_label("chacha_blocks_sse2_mainloop2");
	&paddd("xmm0", "xmm1");
	&pxor("xmm3", "xmm0");
	&pshuflw("xmm3", "xmm3", 0xb1);
	&pshufhw("xmm3", "xmm3", 0xb1);
	&paddd("xmm2", "xmm3");
	&pxor("xmm1", "xmm2");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&pxor("xmm1", "xmm4");
	&paddd("xmm0", "xmm1");
	&pxor("xmm3", "xmm0");
	&movdqa("xmm4", "xmm3");
	&pslld("xmm3", 8);
	&psrld("xmm4", 24);
	&pshufd("xmm0", "xmm0", 0x93);
	&pxor("xmm3", "xmm4");
	&paddd("xmm2", "xmm3");
	&pshufd("xmm3", "xmm3", 0x4e);
	&pxor("xmm1", "xmm2");
	&pshufd("xmm2", "xmm2", 0x39);
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&psrld("xmm4", 25);
	&pxor("xmm1", "xmm4");
	&sub("eax", 2);
	&paddd("xmm0", "xmm1");
	&pxor("xmm3", "xmm0");
	&pshuflw("xmm3", "xmm3", 0xb1);
	&pshufhw("xmm3", "xmm3", 0xb1);
	&paddd("xmm2", "xmm3");
	&pxor("xmm1", "xmm2");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&pxor("xmm1", "xmm4");
	&paddd("xmm0", "xmm1");
	&pxor("xmm3", "xmm0");
	&movdqa("xmm4", "xmm3");
	&pslld("xmm3", 8);
	&psrld("xmm4", 24);
	&pshufd("xmm0", "xmm0", 0x39);
	&pxor("xmm3", "xmm4");
	&paddd("xmm2", "xmm3");
	&pshufd("xmm3", "xmm3", 0x4e);
	&pxor("xmm1", "xmm2");
	&pshufd("xmm2", "xmm2", 0x93);
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&psrld("xmm4", 25);
	&pxor("xmm1", "xmm4");
	&ja(&label("chacha_blocks_sse2_mainloop2"));
	&paddd("xmm0", &DWP(0,"esp"));
	&paddd("xmm1", &DWP(16,"esp"));
	&paddd("xmm2", &DWP(32,"esp"));
	&paddd("xmm3", &DWP(48,"esp"));
	&cmp("esi", 0);
	&jbe(&label("chacha_blocks_sse2_noinput3"));
	&movdqu("xmm4", &DWP(0,"esi"));
	&movdqu("xmm5", &DWP(16,"esi"));
	&movdqu("xmm6", &DWP(32,"esi"));
	&movdqu("xmm7", &DWP(48,"esi"));
	&pxor("xmm0", "xmm4");
	&pxor("xmm1", "xmm5");
	&pxor("xmm2", "xmm6");
	&pxor("xmm3", "xmm7");
	&add("esi", 64);
	&set_label("chacha_blocks_sse2_noinput3");
	&movdqu(&DWP(0,"edx"), "xmm0");
	&movdqu(&DWP(16,"edx"), "xmm1");
	&movdqu(&DWP(32,"edx"), "xmm2");
	&movdqu(&DWP(48,"edx"), "xmm3");
	&mov("eax", &DWP(80,"esp"));
	&mov("ecx", &DWP(48,"esp"));
	&mov("ebx", &DWP(4+48,"esp"));
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(48,"esp"), "ecx");
	&mov(&DWP(4+48,"esp"), "ebx");
	&cmp("eax", 64);
	&ja(&label("chacha_blocks_sse2_bytesatleast65"));
	&jae(&label("chacha_blocks_sse2_bytesatleast64"));
	&mov("esi", "edx");
	&mov("edi", &DWP(92,"esp"));
	&mov("ecx", "eax");
	&rep_movsb();
	&set_label("chacha_blocks_sse2_bytesatleast64");
	&set_label("chacha_blocks_sse2_done");
	&mov("eax", &DWP(84,"esp"));
	&movdqa("xmm0", &DWP(48,"esp"));
	&movdqu(&DWP(32,"eax"), "xmm0");
	&mov("eax", &DWP(64,"esp"));
	&mov("ebx", &DWP(68,"esp"));
	&mov("esi", &DWP(72,"esp"));
	&mov("edi", &DWP(76,"esp"));
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
	&set_label("chacha_blocks_sse2_bytesatleast65");
	&sub("eax", 64);
	&add("edx", 64);
	&jmp(&label("chacha_blocks_sse2_bytesbetween1and255"));
}

sub chacha_blocks_ssse3 {
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", ~63);
	&sub("esp", 704);
	&mov(&DWP(68, "esp"), "ebx");
	&mov(&DWP(72, "esp"), "esi");
	&mov(&DWP(76, "esp"), "edi");
	&mov("ecx", &DWP(8, "ebp"));
	&mov(&DWP(84, "esp"), "ecx");
	&mov("esi", &DWP(12, "ebp"));
	&mov("edx", &DWP(16, "ebp"));
	&mov("eax", &DWP(20, "ebp"));
	&label_to_reg("chacha_constants", "ebx");
	&movdqa("xmm0", &DWP(0, "ebx"));
	&movdqa("xmm4", &DWP(16, "ebx"));
	&movdqa("xmm5", &DWP(32, "ebx"));
	&movdqu("xmm1", &DWP(0, "ecx"));
	&movdqu("xmm2", &DWP(16, "ecx"));
	&movdqu("xmm3", &DWP(32, "ecx"));
	&movdqa(&DWP(0, "esp"), "xmm0");
	&movdqa(&DWP(16, "esp"), "xmm1");
	&movdqa(&DWP(32, "esp"), "xmm2");
	&movdqa(&DWP(48, "esp"), "xmm3");
	&movdqa(&DWP(96, "esp"), "xmm4");
	&movdqa(&DWP(112, "esp"), "xmm5");
	&mov("ecx", &DWP(48, "ecx"));
	&mov(&DWP(88, "esp"), "ecx");
	&cmp("eax", 0);
	&jbe(&label("chacha_blocks_ssse3_done"));
	&cmp("eax", 256);
	&jb(&label("chacha_blocks_ssse3_bytesbetween1and255"));
	&pshufd("xmm4", "xmm0", 0x00);
	&pshufd("xmm5", "xmm0", 0x55);
	&pshufd("xmm6", "xmm0", 0xaa);
	&pshufd("xmm0", "xmm0", 0xff);
	&movdqa(&DWP(128, "esp"), "xmm4");
	&movdqa(&DWP(144, "esp"), "xmm5");
	&movdqa(&DWP(160, "esp"), "xmm6");
	&movdqa(&DWP(176, "esp"), "xmm0");
	&pshufd("xmm0", "xmm1", 0x00);
	&pshufd("xmm4", "xmm1", 0x55);
	&pshufd("xmm5", "xmm1", 0xaa);
	&pshufd("xmm1", "xmm1", 0xff);
	&movdqa(&DWP(192, "esp"), "xmm0");
	&movdqa(&DWP(208, "esp"), "xmm4");
	&movdqa(&DWP(224, "esp"), "xmm5");
	&movdqa(&DWP(240, "esp"), "xmm1");
	&pshufd("xmm0", "xmm2", 0x00);
	&pshufd("xmm1", "xmm2", 0x55);
	&pshufd("xmm4", "xmm2", 0xaa);
	&pshufd("xmm2", "xmm2", 0xff);
	&movdqa(&DWP(256, "esp"), "xmm0");
	&movdqa(&DWP(272, "esp"), "xmm1");
	&movdqa(&DWP(288, "esp"), "xmm4");
	&movdqa(&DWP(304, "esp"), "xmm2");
	&pshufd("xmm0", "xmm3", 0xaa);
	&pshufd("xmm1", "xmm3", 0xff);
	&movdqa(&DWP(352, "esp"), "xmm0");
	&movdqa(&DWP(368, "esp"), "xmm1");
	&set_label("chacha_blocks_ssse3_bytesatleast256");
	&mov("ecx", &DWP(48, "esp"));
	&mov("ebx", &DWP(4+48, "esp"));
	&mov(&DWP(320, "esp"), "ecx");
	&mov(&DWP(336, "esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(4+320, "esp"), "ecx");
	&mov(&DWP(4+336, "esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(8+320, "esp"), "ecx");
	&mov(&DWP(8+336, "esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(12+320, "esp"), "ecx");
	&mov(&DWP(12+336, "esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(48, "esp"), "ecx");
	&mov(&DWP(4+48, "esp"), "ebx");
	&mov(&DWP(92, "esp"), "eax");
	&mov("eax", &DWP(88, "esp"));
	&movdqa("xmm0", &DWP(160, "esp"));
	&movdqa("xmm1", &DWP(224, "esp"));
	&movdqa("xmm2", &DWP(288, "esp"));
	&movdqa("xmm3", &DWP(352, "esp"));
	&movdqa("xmm4", &DWP(176, "esp"));
	&movdqa("xmm5", &DWP(240, "esp"));
	&movdqa("xmm6", &DWP(304, "esp"));
	&movdqa("xmm7", &DWP(368, "esp"));
	&movdqa(&DWP(480, "esp"), "xmm0");
	&movdqa(&DWP(544, "esp"), "xmm1");
	&movdqa(&DWP(608, "esp"), "xmm2");
	&movdqa(&DWP(672, "esp"), "xmm3");
	&movdqa(&DWP(496, "esp"), "xmm4");
	&movdqa(&DWP(560, "esp"), "xmm5");
	&movdqa(&DWP(624, "esp"), "xmm6");
	&movdqa(&DWP(688, "esp"), "xmm7");
	&movdqa("xmm0", &DWP(128, "esp"));
	&movdqa("xmm1", &DWP(192, "esp"));
	&movdqa("xmm2", &DWP(256, "esp"));
	&movdqa("xmm3", &DWP(320, "esp"));
	&movdqa("xmm4", &DWP(144, "esp"));
	&movdqa("xmm5", &DWP(208, "esp"));
	&movdqa("xmm6", &DWP(272, "esp"));
	&movdqa("xmm7", &DWP(336, "esp"));
	&jmp(&label("chacha_blocks_ssse3_mainloop1"));
	&set_label("chacha_blocks_ssse3_mainloop1", 64);
	&paddd("xmm0", "xmm1");
	&paddd("xmm4", "xmm5");
	&pxor("xmm3", "xmm0");
	&pshufb("xmm3", &DWP(96, "esp"));
	&pxor("xmm7", "xmm4");
	&pshufb("xmm7", &DWP(96, "esp"));
	&paddd("xmm2", "xmm3");
	&paddd("xmm6", "xmm7");
	&movdqa(&DWP(464, "esp"), "xmm4");
	&pxor("xmm1", "xmm2");
	&pxor("xmm5", "xmm6");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&por("xmm1", "xmm4");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 12);
	&psrld("xmm4", 20);
	&por("xmm5", "xmm4");
	&movdqa("xmm4", &DWP(464, "esp"));
	&paddd("xmm0", "xmm1");
	&paddd("xmm4", "xmm5");
	&pxor("xmm3", "xmm0");
	&pshufb("xmm3", &DWP(112, "esp"));
	&pxor("xmm7", "xmm4");
	&pshufb("xmm7", &DWP(112, "esp"));
	&movdqa(&DWP(448, "esp"), "xmm0");
	&paddd("xmm2", "xmm3");
	&movdqa(&DWP(464, "esp"), "xmm4");
	&paddd("xmm6", "xmm7");
	&movdqa(&DWP(640, "esp"), "xmm3");
	&movdqa(&DWP(656, "esp"), "xmm7");
	&pxor("xmm1", "xmm2");
	&pxor("xmm5", "xmm6");
	&movdqa(&DWP(576, "esp"), "xmm2");
	&movdqa(&DWP(592, "esp"), "xmm6");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&movdqa("xmm3", &DWP(672, "esp"));
	&psrld("xmm4", 25);
	&movdqa("xmm7", &DWP(688, "esp"));
	&por("xmm1", "xmm4");
	&movdqa(&DWP(512, "esp"), "xmm1");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 7);
	&movdqa("xmm2", &DWP(608, "esp"));
	&psrld("xmm4", 25);
	&movdqa("xmm6", &DWP(624, "esp"));
	&por("xmm5", "xmm4");
	&movdqa(&DWP(528, "esp"), "xmm5");
	&movdqa("xmm0", &DWP(480, "esp"));
	&movdqa("xmm4", &DWP(496, "esp"));
	&movdqa("xmm1", &DWP(544, "esp"));
	&movdqa("xmm5", &DWP(560, "esp"));
	&paddd("xmm0", "xmm1");
	&paddd("xmm4", "xmm5");
	&pxor("xmm3", "xmm0");
	&pshufb("xmm3", &DWP(96, "esp"));
	&pxor("xmm7", "xmm4");
	&pshufb("xmm7", &DWP(96, "esp"));
	&paddd("xmm2", "xmm3");
	&paddd("xmm6", "xmm7");
	&movdqa(&DWP(496, "esp"), "xmm4");
	&pxor("xmm1", "xmm2");
	&pxor("xmm5", "xmm6");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&por("xmm1", "xmm4");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 12);
	&psrld("xmm4", 20);
	&por("xmm5", "xmm4");
	&movdqa("xmm4", &DWP(496, "esp"));
	&paddd("xmm0", "xmm1");
	&paddd("xmm4", "xmm5");
	&pxor("xmm3", "xmm0");
	&pshufb("xmm3", &DWP(112, "esp"));
	&pxor("xmm7", "xmm4");
	&pshufb("xmm7", &DWP(112, "esp"));
	&movdqa(&DWP(480, "esp"), "xmm0");
	&paddd("xmm2", "xmm3");
	&movdqa(&DWP(496, "esp"), "xmm4");
	&paddd("xmm6", "xmm7");
	&movdqa(&DWP(672, "esp"), "xmm3");
	&pxor("xmm1", "xmm2");
	&pxor("xmm5", "xmm6");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&psrld("xmm4", 25);
	&movdqa("xmm3", &DWP(640, "esp"));
	&por("xmm1", "xmm4");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 7);
	&psrld("xmm4", 25);
	&por("xmm5", "xmm4");
	&movdqa(&DWP(560, "esp"), "xmm5");
	&movdqa("xmm0", &DWP(448, "esp"));
	&movdqa("xmm5", &DWP(528, "esp"));
	&movdqa("xmm4", &DWP(464, "esp"));
	&paddd("xmm0", "xmm5");
	&paddd("xmm4", "xmm1");
	&pxor("xmm7", "xmm0");
	&pshufb("xmm7", &DWP(96, "esp"));
	&pxor("xmm3", "xmm4");
	&pshufb("xmm3", &DWP(96, "esp"));
	&paddd("xmm2", "xmm7");
	&paddd("xmm6", "xmm3");
	&movdqa(&DWP(464, "esp"), "xmm4");
	&pxor("xmm5", "xmm2");
	&pxor("xmm1", "xmm6");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 12);
	&psrld("xmm4", 20);
	&por("xmm5", "xmm4");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&por("xmm1", "xmm4");
	&movdqa("xmm4", &DWP(464, "esp"));
	&paddd("xmm0", "xmm5");
	&paddd("xmm4", "xmm1");
	&pxor("xmm7", "xmm0");
	&pshufb("xmm7", &DWP(112, "esp"));
	&pxor("xmm3", "xmm4");
	&pshufb("xmm3", &DWP(112, "esp"));
	&movdqa(&DWP(448, "esp"), "xmm0");
	&paddd("xmm2", "xmm7");
	&movdqa(&DWP(464, "esp"), "xmm4");
	&paddd("xmm6", "xmm3");
	&movdqa(&DWP(688, "esp"), "xmm7");
	&movdqa(&DWP(640, "esp"), "xmm3");
	&pxor("xmm5", "xmm2");
	&pxor("xmm1", "xmm6");
	&movdqa(&DWP(608, "esp"), "xmm2");
	&movdqa(&DWP(624, "esp"), "xmm6");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 7);
	&movdqa("xmm7", &DWP(656, "esp"));
	&psrld("xmm4", 25);
	&movdqa("xmm3", &DWP(672, "esp"));
	&por("xmm5", "xmm4");
	&movdqa(&DWP(528, "esp"), "xmm5");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&movdqa("xmm2", &DWP(576, "esp"));
	&psrld("xmm4", 25);
	&movdqa("xmm6", &DWP(592, "esp"));
	&por("xmm1", "xmm4");
	&movdqa(&DWP(544, "esp"), "xmm1");
	&movdqa("xmm0", &DWP(480, "esp"));
	&movdqa("xmm4", &DWP(496, "esp"));
	&movdqa("xmm5", &DWP(560, "esp"));
	&movdqa("xmm1", &DWP(512, "esp"));
	&paddd("xmm0", "xmm5");
	&paddd("xmm4", "xmm1");
	&pxor("xmm7", "xmm0");
	&pshufb("xmm7", &DWP(96, "esp"));
	&pxor("xmm3", "xmm4");
	&pshufb("xmm3", &DWP(96, "esp"));
	&paddd("xmm2", "xmm7");
	&paddd("xmm6", "xmm3");
	&movdqa(&DWP(496, "esp"), "xmm4");
	&pxor("xmm5", "xmm2");
	&pxor("xmm1", "xmm6");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 12);
	&psrld("xmm4", 20);
	&por("xmm5", "xmm4");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&por("xmm1", "xmm4");
	&movdqa("xmm4", &DWP(496, "esp"));
	&paddd("xmm0", "xmm5");
	&paddd("xmm4", "xmm1");
	&pxor("xmm7", "xmm0");
	&pshufb("xmm7", &DWP(112, "esp"));
	&pxor("xmm3", "xmm4");
	&pshufb("xmm3", &DWP(112, "esp"));
	&movdqa(&DWP(480, "esp"), "xmm0");
	&movdqa(&DWP(496, "esp"), "xmm4");
	&paddd("xmm2", "xmm7");
	&paddd("xmm6", "xmm3");
	&movdqa(&DWP(672, "esp"), "xmm3");
	&pxor("xmm5", "xmm2");
	&pxor("xmm1", "xmm6");
	&movdqa("xmm4", "xmm5");
	&pslld("xmm5", 7);
	&psrld("xmm4", 25);
	&movdqa("xmm3", &DWP(640, "esp"));
	&por("xmm5", "xmm4");
	&movdqa(&DWP(560, "esp"), "xmm5");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&psrld("xmm4", 25);
	&por("xmm1", "xmm4");
	&movdqa("xmm0", &DWP(448, "esp"));
	&movdqa("xmm4", &DWP(464, "esp"));
	&movdqa("xmm5", &DWP(528, "esp"));
	&sub("eax", 2);
	&ja(&label("chacha_blocks_ssse3_mainloop1"));
	&movdqa(&DWP(448, "esp"), "xmm0");
	&movdqa(&DWP(512, "esp"), "xmm1");
	&movdqa(&DWP(576, "esp"), "xmm2");
	&movdqa(&DWP(640, "esp"), "xmm3");
	&movdqa(&DWP(464, "esp"), "xmm4");
	&movdqa(&DWP(528, "esp"), "xmm5");
	&movdqa(&DWP(592, "esp"), "xmm6");
	&movdqa(&DWP(656, "esp"), "xmm7");
	&cmp("esi", 0);
	&movdqa("xmm0", &DWP(448, "esp"));
	&movdqa("xmm1", &DWP(464, "esp"));
	&movdqa("xmm2", &DWP(480, "esp"));
	&movdqa("xmm3", &DWP(496, "esp"));
	&paddd("xmm0", &DWP(128, "esp"));
	&paddd("xmm1", &DWP(144, "esp"));
	&paddd("xmm2", &DWP(160, "esp"));
	&paddd("xmm3", &DWP(176, "esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&jbe(&label("chacha_blocks_ssse3_noinput1"));
	&movdqu("xmm2", &DWP(0, "esi"));
	&movdqu("xmm5", &DWP(64, "esi"));
	&movdqu("xmm6", &DWP(128, "esi"));
	&movdqu("xmm7", &DWP(192, "esi"));
	&pxor("xmm4", "xmm2");
	&pxor("xmm1", "xmm5");
	&pxor("xmm0", "xmm6");
	&pxor("xmm3", "xmm7");
	&movdqu(&DWP(0, "edx"), "xmm4");
	&movdqu(&DWP(64, "edx"), "xmm1");
	&movdqu(&DWP(128, "edx"), "xmm0");
	&movdqu(&DWP(192, "edx"), "xmm3");
	&movdqa("xmm0", &DWP(512, "esp"));
	&movdqa("xmm1", &DWP(528, "esp"));
	&movdqa("xmm2", &DWP(544, "esp"));
	&movdqa("xmm3", &DWP(560, "esp"));
	&paddd("xmm0", &DWP(192, "esp"));
	&paddd("xmm1", &DWP(208, "esp"));
	&paddd("xmm2", &DWP(224, "esp"));
	&paddd("xmm3", &DWP(240, "esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu("xmm2", &DWP(16+0, "esi"));
	&movdqu("xmm5", &DWP(16+64, "esi"));
	&movdqu("xmm6", &DWP(16+128, "esi"));
	&movdqu("xmm7", &DWP(16+192, "esi"));
	&pxor("xmm4", "xmm2");
	&pxor("xmm1", "xmm5");
	&pxor("xmm0", "xmm6");
	&pxor("xmm3", "xmm7");
	&movdqu(&DWP(16+0, "edx"), "xmm4");
	&movdqu(&DWP(16+64, "edx"), "xmm1");
	&movdqu(&DWP(16+128, "edx"), "xmm0");
	&movdqu(&DWP(16+192, "edx"), "xmm3");
	&movdqa("xmm0", &DWP(576, "esp"));
	&movdqa("xmm1", &DWP(592, "esp"));
	&movdqa("xmm2", &DWP(608, "esp"));
	&movdqa("xmm3", &DWP(624, "esp"));
	&paddd("xmm0", &DWP(256, "esp"));
	&paddd("xmm1", &DWP(272, "esp"));
	&paddd("xmm2", &DWP(288, "esp"));
	&paddd("xmm3", &DWP(304, "esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu("xmm2", &DWP(32+0, "esi"));
	&movdqu("xmm5", &DWP(32+64, "esi"));
	&movdqu("xmm6", &DWP(32+128, "esi"));
	&movdqu("xmm7", &DWP(32+192, "esi"));
	&pxor("xmm4", "xmm2");
	&pxor("xmm1", "xmm5");
	&pxor("xmm0", "xmm6");
	&pxor("xmm3", "xmm7");
	&movdqu(&DWP(32+0, "edx"), "xmm4");
	&movdqu(&DWP(32+64, "edx"), "xmm1");
	&movdqu(&DWP(32+128, "edx"), "xmm0");
	&movdqu(&DWP(32+192, "edx"), "xmm3");
	&movdqa("xmm0", &DWP(640, "esp"));
	&movdqa("xmm1", &DWP(656, "esp"));
	&movdqa("xmm2", &DWP(672, "esp"));
	&movdqa("xmm3", &DWP(688, "esp"));
	&paddd("xmm0", &DWP(320, "esp"));
	&paddd("xmm1", &DWP(336, "esp"));
	&paddd("xmm2", &DWP(352, "esp"));
	&paddd("xmm3", &DWP(368, "esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu("xmm2", &DWP(48+0, "esi"));
	&movdqu("xmm5", &DWP(48+64, "esi"));
	&movdqu("xmm6", &DWP(48+128, "esi"));
	&movdqu("xmm7", &DWP(48+192, "esi"));
	&pxor("xmm4", "xmm2");
	&pxor("xmm1", "xmm5");
	&pxor("xmm0", "xmm6");
	&pxor("xmm3", "xmm7");
	&movdqu(&DWP(48+0, "edx"), "xmm4");
	&movdqu(&DWP(48+64, "edx"), "xmm1");
	&movdqu(&DWP(48+128, "edx"), "xmm0");
	&movdqu(&DWP(48+192, "edx"), "xmm3");
	&add("esi", 256);
	&jmp(&label("chacha_blocks_ssse3_mainloop1_cont"));
	&set_label("chacha_blocks_ssse3_noinput1");
	&movdqu(&DWP(0, "edx"), "xmm4");
	&movdqu(&DWP(64, "edx"), "xmm1");
	&movdqu(&DWP(128, "edx"), "xmm0");
	&movdqu(&DWP(192, "edx"), "xmm3");
	&movdqa("xmm0", &DWP(512, "esp"));
	&movdqa("xmm1", &DWP(528, "esp"));
	&movdqa("xmm2", &DWP(544, "esp"));
	&movdqa("xmm3", &DWP(560, "esp"));
	&paddd("xmm0", &DWP(192, "esp"));
	&paddd("xmm1", &DWP(208, "esp"));
	&paddd("xmm2", &DWP(224, "esp"));
	&paddd("xmm3", &DWP(240, "esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu(&DWP(16+0, "edx"), "xmm4");
	&movdqu(&DWP(16+64, "edx"), "xmm1");
	&movdqu(&DWP(16+128, "edx"), "xmm0");
	&movdqu(&DWP(16+192, "edx"), "xmm3");
	&movdqa("xmm0", &DWP(576, "esp"));
	&movdqa("xmm1", &DWP(592, "esp"));
	&movdqa("xmm2", &DWP(608, "esp"));
	&movdqa("xmm3", &DWP(624, "esp"));
	&paddd("xmm0", &DWP(256, "esp"));
	&paddd("xmm1", &DWP(272, "esp"));
	&paddd("xmm2", &DWP(288, "esp"));
	&paddd("xmm3", &DWP(304, "esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu(&DWP(32+0, "edx"), "xmm4");
	&movdqu(&DWP(32+64, "edx"), "xmm1");
	&movdqu(&DWP(32+128, "edx"), "xmm0");
	&movdqu(&DWP(32+192, "edx"), "xmm3");
	&movdqa("xmm0", &DWP(640, "esp"));
	&movdqa("xmm1", &DWP(656, "esp"));
	&movdqa("xmm2", &DWP(672, "esp"));
	&movdqa("xmm3", &DWP(688, "esp"));
	&paddd("xmm0", &DWP(320, "esp"));
	&paddd("xmm1", &DWP(336, "esp"));
	&paddd("xmm2", &DWP(352, "esp"));
	&paddd("xmm3", &DWP(368, "esp"));
	&movdqa("xmm4", "xmm0");
	&movdqa("xmm5", "xmm2");
	&punpckldq("xmm4", "xmm1");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm0", "xmm1");
	&punpckhdq("xmm2", "xmm3");
	&movdqa("xmm1", "xmm4");
	&movdqa("xmm3", "xmm0");
	&punpcklqdq("xmm4", "xmm5");
	&punpckhqdq("xmm1", "xmm5");
	&punpcklqdq("xmm0", "xmm2");
	&punpckhqdq("xmm3", "xmm2");
	&movdqu(&DWP(48+0, "edx"), "xmm4");
	&movdqu(&DWP(48+64, "edx"), "xmm1");
	&movdqu(&DWP(48+128, "edx"), "xmm0");
	&movdqu(&DWP(48+192, "edx"), "xmm3");
	&set_label("chacha_blocks_ssse3_mainloop1_cont");
	&mov("eax", &DWP(92, "esp"));
	&sub("eax", 256);
	&add("edx", 256);
	&cmp("eax", 256);
	&jae(&label("chacha_blocks_ssse3_bytesatleast256"));
	&cmp("eax", 0);
	&jbe(&label("chacha_blocks_ssse3_done"));
	&set_label("chacha_blocks_ssse3_bytesbetween1and255");
	&cmp("eax", 64);
	&jae(&label("chacha_blocks_ssse3_nocopy"));
	&mov(&DWP(92, "esp"), "edx");
	&cmp("esi", 0);
	&jbe(&label("chacha_blocks_ssse3_noinput2"));
	&lea("edi", &DWP(128, "esp"));
	&mov("ecx", "eax");
	&rep_movsb();
	&lea("esi", &DWP(128, "esp"));
	&set_label("chacha_blocks_ssse3_noinput2");
	&lea("edx", &DWP(128, "esp"));
	&set_label("chacha_blocks_ssse3_nocopy");
	&mov(&DWP(80, "esp"), "eax");
	&movdqa("xmm0", &DWP(0, "esp"));
	&movdqa("xmm1", &DWP(16, "esp"));
	&movdqa("xmm2", &DWP(32, "esp"));
	&movdqa("xmm3", &DWP(48, "esp"));
	&movdqa("xmm5", &DWP(96, "esp"));
	&movdqa("xmm6", &DWP(112, "esp"));
	&mov("eax", &DWP(88, "esp"));
	&set_label("chacha_blocks_ssse3_mainloop2");
	&paddd("xmm0", "xmm1");
	&pxor("xmm3", "xmm0");
	&pshufb("xmm3", "xmm5");
	&paddd("xmm2", "xmm3");
	&pxor("xmm1", "xmm2");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&pxor("xmm1", "xmm4");
	&paddd("xmm0", "xmm1");
	&pxor("xmm3", "xmm0");
	&pshufb("xmm3", "xmm6");
	&pshufd("xmm0", "xmm0", 0x93);
	&paddd("xmm2", "xmm3");
	&pshufd("xmm3", "xmm3", 0x4e);
	&pxor("xmm1", "xmm2");
	&pshufd("xmm2", "xmm2", 0x39);
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&psrld("xmm4", 25);
	&pxor("xmm1", "xmm4");
	&sub("eax", 2);
	&paddd("xmm0", "xmm1");
	&pxor("xmm3", "xmm0");
	&pshufb("xmm3", "xmm5");
	&paddd("xmm2", "xmm3");
	&pxor("xmm1", "xmm2");
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 12);
	&psrld("xmm4", 20);
	&pxor("xmm1", "xmm4");
	&paddd("xmm0", "xmm1");
	&pxor("xmm3", "xmm0");
	&pshufb("xmm3", "xmm6");
	&pshufd("xmm0", "xmm0", 0x39);
	&paddd("xmm2", "xmm3");
	&pshufd("xmm3", "xmm3", 0x4e);
	&pxor("xmm1", "xmm2");
	&pshufd("xmm2", "xmm2", 0x93);
	&movdqa("xmm4", "xmm1");
	&pslld("xmm1", 7);
	&psrld("xmm4", 25);
	&pxor("xmm1", "xmm4");
	&ja(&label("chacha_blocks_ssse3_mainloop2"));
	&paddd("xmm0", &DWP(0, "esp"));
	&paddd("xmm1", &DWP(16, "esp"));
	&paddd("xmm2", &DWP(32, "esp"));
	&paddd("xmm3", &DWP(48, "esp"));
	&cmp("esi", 0);
	&jbe(&label("chacha_blocks_ssse3_noinput3"));
	&movdqu("xmm4", &DWP(0, "esi"));
	&movdqu("xmm5", &DWP(16, "esi"));
	&movdqu("xmm6", &DWP(32, "esi"));
	&movdqu("xmm7", &DWP(48, "esi"));
	&pxor("xmm0", "xmm4");
	&pxor("xmm1", "xmm5");
	&pxor("xmm2", "xmm6");
	&pxor("xmm3", "xmm7");
	&add("esi", 64);
	&set_label("chacha_blocks_ssse3_noinput3");
	&movdqu(&DWP(0, "edx"), "xmm0");
	&movdqu(&DWP(16, "edx"), "xmm1");
	&movdqu(&DWP(32, "edx"), "xmm2");
	&movdqu(&DWP(48, "edx"), "xmm3");
	&mov("eax", &DWP(80, "esp"));
	&mov("ecx", &DWP(48, "esp"));
	&mov("ebx", &DWP(4+48, "esp"));
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(48, "esp"), "ecx");
	&mov(&DWP(4+48, "esp"), "ebx");
	&cmp("eax", 64);
	&ja(&label("chacha_blocks_ssse3_bytesatleast65"));
	&jae(&label("chacha_blocks_ssse3_bytesatleast64"));
	&mov("esi", "edx");
	&mov("edi", &DWP(92, "esp"));
	&mov("ecx", "eax");
	&rep_movsb();
	&set_label("chacha_blocks_ssse3_bytesatleast64");
	&set_label("chacha_blocks_ssse3_done");
	&mov("eax", &DWP(84, "esp"));
	&movdqa("xmm0", &DWP(48, "esp"));
	&movdqu(&DWP(32, "eax"), "xmm0");
	&mov("eax", &DWP(64, "esp"));
	&mov("ebx", &DWP(68, "esp"));
	&mov("esi", &DWP(72, "esp"));
	&mov("edi", &DWP(76, "esp"));
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
	&set_label("chacha_blocks_ssse3_bytesatleast65");
	&sub("eax", 64);
	&add("edx", 64);
	&jmp(&label("chacha_blocks_ssse3_bytesbetween1and255"));
}

sub chacha_blocks_avx {
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", ~63);
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 1268);
	&mov("ecx", 1);
	&label_to_reg("chacha_constants", "esi");
	&mov("edx", &DWP(16,"ebp"));
	&mov("ebx", &DWP(20,"ebp"));
	&vmovd("xmm2", "ecx");
	&vmovdqu("xmm0", &DWP(16,"esi"));
	&vmovdqu("xmm1", &DWP(32,"esi"));
	&mov("eax", &DWP(12,"ebp"));
	&vmovdqu(&DWP(448,"esp"), "xmm0");
	&vmovdqu(&DWP(432,"esp"), "xmm1");
	&vmovdqu(&DWP(96,"esp"), "xmm2");
	&mov(&DWP(416,"esp"), "edx");
	&test("ebx", "ebx");
	&je(&label("chacha_blocks_avx_36"));
	&set_label("chacha_blocks_avx_2");
	&mov("ecx", &DWP(8,"ebp"));
	&vmovdqu("xmm0", &DWP(0,"esi"));
	&vmovdqu("xmm1", &DWP(0,"ecx"));
	&vmovdqu("xmm2", &DWP(16,"ecx"));
	&mov("edi", &DWP(48,"ecx"));
	&vmovdqu(&DWP(48,"esp"), "xmm0");
	&mov(&DWP(424,"esp"), "edi");
	&vmovdqu(&DWP(64,"esp"), "xmm1");
	&vmovdqu(&DWP(80,"esp"), "xmm2");
	&vmovdqu("xmm0", &DWP(32,"ecx"));
	&cmp("ebx", 256);
	&jb(&label("chacha_blocks_avx_10"));
	&set_label("chacha_blocks_avx_3");
	&vmovdqu("xmm7", &DWP(64,"esp"));
	&vpshufd("xmm1", "xmm7", 255);
	&vmovdqu(&DWP(208,"esp"), "xmm1");
	&vmovdqu("xmm1", &DWP(80,"esp"));
	&vmovdqu("xmm4", &DWP(48,"esp"));
	&vpshufd("xmm2", "xmm7", 0);
	&vmovdqu(&DWP(240,"esp"), "xmm2");
	&vpshufd("xmm3", "xmm7", 85);
	&vpshufd("xmm2", "xmm7", 170);
	&vpshufd("xmm7", "xmm1", 0);
	&vmovdqu(&DWP(192,"esp"), "xmm7");
	&vpshufd("xmm7", "xmm1", 85);
	&vpshufd("xmm5", "xmm4", 85);
	&vmovdqu(&DWP(176,"esp"), "xmm7");
	&vpshufd("xmm7", "xmm1", 170);
	&vpshufd("xmm1", "xmm1", 255);
	&vpshufd("xmm6", "xmm4", 0);
	&vmovdqu(&DWP(224,"esp"), "xmm5");
	&vpshufd("xmm5", "xmm4", 170);
	&vpshufd("xmm4", "xmm4", 255);
	&vmovdqu(&DWP(144,"esp"), "xmm1");
	&vpshufd("xmm1", "xmm0", 170);
	&vpshufd("xmm0", "xmm0", 255);
	&mov("edi", &DWP(32,"ecx"));
	&mov("esi", &DWP(36,"ecx"));
	&vmovdqu(&DWP(160,"esp"), "xmm7");
	&vmovdqu(&DWP(128,"esp"), "xmm1");
	&vmovdqu(&DWP(256,"esp"), "xmm0");
	&vmovdqu(&DWP(336,"esp"), "xmm2");
	&vmovdqu(&DWP(320,"esp"), "xmm3");
	&vmovdqu(&DWP(304,"esp"), "xmm4");
	&vmovdqu(&DWP(288,"esp"), "xmm5");
	&vmovdqu(&DWP(272,"esp"), "xmm6");
	&mov(&DWP(420,"esp"), "ebx");
	&set_label("chacha_blocks_avx_4");
	&mov("ecx", "edi");
	&mov("ebx", "esi");
	&add("ecx", 1);
	&mov(&DWP(36,"esp"), "ecx");
	&mov("ecx", "edi");
	&adc("ebx", 0);
	&add("ecx", 2);
	&mov(&DWP(116,"esp"), "ebx");
	&mov("ebx", "esi");
	&mov(&DWP(40,"esp"), "ecx");
	&mov("ecx", "edi");
	&adc("ebx", 0);
	&add("ecx", 3);
	&mov(&DWP(32,"esp"), "edi");
	&mov(&DWP(44,"esp"), "ecx");
	&vmovdqu("xmm4", &DWP(32,"esp"));
	&vmovdqu("xmm7", &DWP(320,"esp"));
	&vmovdqu("xmm1", &DWP(240,"esp"));
	&mov(&DWP(120,"esp"), "ebx");
	&mov("ebx", "esi");
	&vmovdqu(&DWP(368,"esp"), "xmm4");
	&adc("ebx", 0);
	&vmovdqu(&DWP(464,"esp"), "xmm4");
	&add("edi", 4);
	&vmovdqu("xmm0", &DWP(304,"esp"));
	&vmovdqu("xmm4", &DWP(128,"esp"));
	&vmovdqu(&DWP(560,"esp"), "xmm7");
	&vmovdqu(&DWP(384,"esp"), "xmm1");
	&vmovdqu("xmm7", &DWP(160,"esp"));
	&vmovdqu("xmm1", &DWP(208,"esp"));
	&mov(&DWP(112,"esp"), "esi");
	&adc("esi", 0);
	&mov(&DWP(124,"esp"), "ebx");
	&vmovdqu(&DWP(576,"esp"), "xmm0");
	&vmovdqu(&DWP(480,"esp"), "xmm4");
	&vmovdqu("xmm5", &DWP(272,"esp"));
	&vmovdqu("xmm0", &DWP(336,"esp"));
	&vmovdqu("xmm4", &DWP(256,"esp"));
	&vmovdqu(&DWP(528,"esp"), "xmm7");
	&vmovdqu("xmm6", &DWP(112,"esp"));
	&vmovdqu("xmm3", &DWP(224,"esp"));
	&vmovdqu("xmm2", &DWP(288,"esp"));
	&vmovdqu("xmm7", &DWP(144,"esp"));
	&vmovdqu(&DWP(592,"esp"), "xmm1");
	&vmovdqu(&DWP(544,"esp"), "xmm0");
	&vmovdqu(&DWP(496,"esp"), "xmm4");
	&vmovdqu(&DWP(608,"esp"), "xmm5");
	&vmovdqu(&DWP(352,"esp"), "xmm6");
	&vmovdqu("xmm1", &DWP(192,"esp"));
	&vmovdqu("xmm0", &DWP(176,"esp"));
	&mov("ecx", &DWP(424,"esp"));
	&vmovdqu("xmm5", &DWP(592,"esp"));
	&vmovdqu("xmm4", &DWP(384,"esp"));
	&vmovdqu(&DWP(512,"esp"), "xmm7");
	&vmovdqu(&DWP(640,"esp"), "xmm2");
	&vmovdqu(&DWP(624,"esp"), "xmm3");
	&jmp(&label("chacha_blocks_avx_5"));
	&align(32);
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&set_label("chacha_blocks_avx_5");
	&vpaddd("xmm2", "xmm4", &DWP(608,"esp"));
	&vmovdqu("xmm7", &DWP(560,"esp"));
	&vpxor("xmm3", "xmm2", &DWP(464,"esp"));
	&vmovdqu(&DWP(656,"esp"), "xmm2");
	&vpaddd("xmm7", "xmm7", &DWP(624,"esp"));
	&vmovdqu("xmm2", &DWP(448,"esp"));
	&vpxor("xmm6", "xmm6", "xmm7");
	&vpshufb("xmm3", "xmm3", "xmm2");
	&vpshufb("xmm6", "xmm6", "xmm2");
	&vmovdqu(&DWP(672,"esp"), "xmm7");
	&vpaddd("xmm1", "xmm1", "xmm3");
	&vmovdqu("xmm7", &DWP(544,"esp"));
	&vpaddd("xmm0", "xmm0", "xmm6");
	&vpaddd("xmm7", "xmm7", &DWP(640,"esp"));
	&vpxor("xmm4", "xmm4", "xmm1");
	&vmovdqu(&DWP(592,"esp"), "xmm5");
	&vpaddd("xmm5", "xmm5", &DWP(576,"esp"));
	&vmovdqu(&DWP(688,"esp"), "xmm7");
	&vpxor("xmm7", "xmm7", &DWP(480,"esp"));
	&vmovdqu(&DWP(704,"esp"), "xmm5");
	&vpxor("xmm5", "xmm5", &DWP(496,"esp"));
	&vpshufb("xmm7", "xmm7", "xmm2");
	&vpshufb("xmm5", "xmm5", "xmm2");
	&vmovdqu(&DWP(752,"esp"), "xmm0");
	&vpslld("xmm2", "xmm4", 12);
	&vmovdqu(&DWP(736,"esp"), "xmm1");
	&vpsrld("xmm1", "xmm4", 20);
	&vpxor("xmm0", "xmm0", &DWP(560,"esp"));
	&vpor("xmm4", "xmm1", "xmm2");
	&vpsrld("xmm1", "xmm0", 20);
	&vpslld("xmm0", "xmm0", 12);
	&vpor("xmm1", "xmm1", "xmm0");
	&vpaddd("xmm0", "xmm7", &DWP(528,"esp"));
	&vpaddd("xmm2", "xmm5", &DWP(512,"esp"));
	&vmovdqu(&DWP(720,"esp"), "xmm5");
	&vpxor("xmm5", "xmm0", &DWP(544,"esp"));
	&vmovdqu(&DWP(784,"esp"), "xmm0");
	&vmovdqu(&DWP(800,"esp"), "xmm2");
	&vpxor("xmm0", "xmm2", &DWP(592,"esp"));
	&vpsrld("xmm2", "xmm5", 20);
	&vpslld("xmm5", "xmm5", 12);
	&vpor("xmm2", "xmm2", "xmm5");
	&vpsrld("xmm5", "xmm0", 20);
	&vpslld("xmm0", "xmm0", 12);
	&vpor("xmm0", "xmm5", "xmm0");
	&vpaddd("xmm5", "xmm4", &DWP(656,"esp"));
	&vmovdqu(&DWP(832,"esp"), "xmm5");
	&vpxor("xmm5", "xmm3", "xmm5");
	&vmovdqu("xmm3", &DWP(432,"esp"));
	&vpshufb("xmm5", "xmm5", "xmm3");
	&vmovdqu(&DWP(768,"esp"), "xmm4");
	&vpaddd("xmm4", "xmm1", &DWP(672,"esp"));
	&vmovdqu(&DWP(864,"esp"), "xmm5");
	&vpxor("xmm6", "xmm6", "xmm4");
	&vmovdqu(&DWP(848,"esp"), "xmm4");
	&vpaddd("xmm5", "xmm0", &DWP(704,"esp"));
	&vpaddd("xmm4", "xmm2", &DWP(688,"esp"));
	&vpshufb("xmm6", "xmm6", "xmm3");
	&vmovdqu(&DWP(912,"esp"), "xmm5");
	&vpxor("xmm7", "xmm7", "xmm4");
	&vpxor("xmm5", "xmm5", &DWP(720,"esp"));
	&vmovdqu(&DWP(896,"esp"), "xmm4");
	&vpshufb("xmm4", "xmm7", "xmm3");
	&vpshufb("xmm5", "xmm5", "xmm3");
	&vmovdqu("xmm3", &DWP(864,"esp"));
	&vpaddd("xmm7", "xmm3", &DWP(736,"esp"));
	&vmovdqu(&DWP(816,"esp"), "xmm2");
	&vmovdqu(&DWP(880,"esp"), "xmm6");
	&vpaddd("xmm6", "xmm6", &DWP(752,"esp"));
	&vpxor("xmm2", "xmm7", &DWP(768,"esp"));
	&vmovdqu(&DWP(960,"esp"), "xmm6");
	&vpxor("xmm6", "xmm1", "xmm6");
	&vpsrld("xmm1", "xmm2", 25);
	&vpslld("xmm3", "xmm2", 7);
	&vmovdqu(&DWP(928,"esp"), "xmm4");
	&vmovdqu(&DWP(944,"esp"), "xmm7");
	&vpor("xmm7", "xmm1", "xmm3");
	&vpaddd("xmm4", "xmm4", &DWP(784,"esp"));
	&vpsrld("xmm1", "xmm6", 25);
	&vpslld("xmm6", "xmm6", 7);
	&vpaddd("xmm3", "xmm5", &DWP(800,"esp"));
	&vpor("xmm2", "xmm1", "xmm6");
	&vpxor("xmm1", "xmm4", &DWP(816,"esp"));
	&vpxor("xmm0", "xmm0", "xmm3");
	&vpsrld("xmm6", "xmm1", 25);
	&vpslld("xmm1", "xmm1", 7);
	&vmovdqu(&DWP(1008,"esp"), "xmm3");
	&vpsrld("xmm3", "xmm0", 25);
	&vpslld("xmm0", "xmm0", 7);
	&vpor("xmm1", "xmm6", "xmm1");
	&vpor("xmm6", "xmm3", "xmm0");
	&vpaddd("xmm0", "xmm2", &DWP(832,"esp"));
	&vmovdqu(&DWP(992,"esp"), "xmm2");
	&vpxor("xmm3", "xmm5", "xmm0");
	&vpaddd("xmm2", "xmm1", &DWP(848,"esp"));
	&vmovdqu("xmm5", &DWP(448,"esp"));
	&vmovdqu(&DWP(1056,"esp"), "xmm2");
	&vmovdqu(&DWP(1040,"esp"), "xmm0");
	&vpshufb("xmm0", "xmm3", "xmm5");
	&vpxor("xmm2", "xmm2", &DWP(864,"esp"));
	&vpaddd("xmm4", "xmm4", "xmm0");
	&vpshufb("xmm2", "xmm2", "xmm5");
	&vpaddd("xmm3", "xmm6", &DWP(896,"esp"));
	&vmovdqu(&DWP(1024,"esp"), "xmm6");
	&vpaddd("xmm6", "xmm7", &DWP(912,"esp"));
	&vmovdqu(&DWP(976,"esp"), "xmm7");
	&vpxor("xmm7", "xmm3", &DWP(880,"esp"));
	&vmovdqu(&DWP(1088,"esp"), "xmm6");
	&vpxor("xmm6", "xmm6", &DWP(928,"esp"));
	&vpshufb("xmm7", "xmm7", "xmm5");
	&vmovdqu(&DWP(1072,"esp"), "xmm3");
	&vpshufb("xmm3", "xmm6", "xmm5");
	&vmovdqu(&DWP(1120,"esp"), "xmm4");
	&vpaddd("xmm5", "xmm2", &DWP(1008,"esp"));
	&vpxor("xmm4", "xmm4", &DWP(992,"esp"));
	&vpxor("xmm1", "xmm1", "xmm5");
	&vmovdqu(&DWP(1136,"esp"), "xmm5");
	&vpsrld("xmm6", "xmm4", 20);
	&vpslld("xmm5", "xmm4", 12);
	&vpor("xmm4", "xmm6", "xmm5");
	&vpsrld("xmm6", "xmm1", 20);
	&vpslld("xmm1", "xmm1", 12);
	&vpaddd("xmm5", "xmm7", &DWP(944,"esp"));
	&vpor("xmm1", "xmm6", "xmm1");
	&vpaddd("xmm6", "xmm3", &DWP(960,"esp"));
	&vmovdqu(&DWP(1104,"esp"), "xmm3");
	&vpxor("xmm3", "xmm5", &DWP(1024,"esp"));
	&vmovdqu(&DWP(1168,"esp"), "xmm5");
	&vmovdqu(&DWP(1184,"esp"), "xmm6");
	&vpxor("xmm5", "xmm6", &DWP(976,"esp"));
	&vpsrld("xmm6", "xmm3", 20);
	&vpslld("xmm3", "xmm3", 12);
	&vpor("xmm6", "xmm6", "xmm3");
	&vpsrld("xmm3", "xmm5", 20);
	&vpslld("xmm5", "xmm5", 12);
	&vmovdqu(&DWP(1152,"esp"), "xmm4");
	&vpor("xmm5", "xmm3", "xmm5");
	&vpaddd("xmm4", "xmm4", &DWP(1040,"esp"));
	&vpaddd("xmm3", "xmm1", &DWP(1056,"esp"));
	&vmovdqu(&DWP(608,"esp"), "xmm4");
	&vpxor("xmm4", "xmm0", "xmm4");
	&vmovdqu("xmm0", &DWP(432,"esp"));
	&vpxor("xmm2", "xmm2", "xmm3");
	&vpshufb("xmm4", "xmm4", "xmm0");
	&vpshufb("xmm2", "xmm2", "xmm0");
	&vmovdqu(&DWP(1200,"esp"), "xmm6");
	&vpaddd("xmm6", "xmm6", &DWP(1072,"esp"));
	&vmovdqu(&DWP(624,"esp"), "xmm3");
	&vpxor("xmm7", "xmm7", "xmm6");
	&vpaddd("xmm3", "xmm5", &DWP(1088,"esp"));
	&vmovdqu(&DWP(640,"esp"), "xmm6");
	&vpshufb("xmm6", "xmm7", "xmm0");
	&vpxor("xmm7", "xmm3", &DWP(1104,"esp"));
	&vmovdqu(&DWP(464,"esp"), "xmm2");
	&vmovdqu(&DWP(576,"esp"), "xmm3");
	&vpshufb("xmm0", "xmm7", "xmm0");
	&vpaddd("xmm3", "xmm4", &DWP(1120,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(1136,"esp"));
	&vmovdqu(&DWP(496,"esp"), "xmm4");
	&vpxor("xmm1", "xmm1", "xmm2");
	&vpxor("xmm4", "xmm3", &DWP(1152,"esp"));
	&vmovdqu(&DWP(512,"esp"), "xmm2");
	&vpsrld("xmm7", "xmm4", 25);
	&vpslld("xmm2", "xmm4", 7);
	&vpsrld("xmm4", "xmm1", 25);
	&vpslld("xmm1", "xmm1", 7);
	&vpor("xmm1", "xmm4", "xmm1");
	&vmovdqu(&DWP(480,"esp"), "xmm0");
	&vmovdqu(&DWP(544,"esp"), "xmm1");
	&vpaddd("xmm1", "xmm6", &DWP(1168,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(1184,"esp"));
	&vmovdqu(&DWP(528,"esp"), "xmm3");
	&vpor("xmm3", "xmm7", "xmm2");
	&vpxor("xmm2", "xmm1", &DWP(1200,"esp"));
	&vpxor("xmm7", "xmm5", "xmm0");
	&vmovdqu(&DWP(560,"esp"), "xmm3");
	&vpsrld("xmm5", "xmm2", 25);
	&vpslld("xmm2", "xmm2", 7);
	&vpsrld("xmm3", "xmm7", 25);
	&vpslld("xmm4", "xmm7", 7);
	&vpor("xmm5", "xmm5", "xmm2");
	&vpor("xmm4", "xmm3", "xmm4");
	&add("ecx", -2);
	&jne(&label("chacha_blocks_avx_5"));
	&set_label("chacha_blocks_avx_6");
	&vmovdqu("xmm3", &DWP(624,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(224,"esp"));
	&vmovdqu(&DWP(592,"esp"), "xmm5");
	&vmovdqu(&DWP(384,"esp"), "xmm4");
	&vmovdqu("xmm2", &DWP(640,"esp"));
	&vmovdqu("xmm5", &DWP(608,"esp"));
	&vmovdqu("xmm4", &DWP(576,"esp"));
	&vmovdqu(&DWP(400,"esp"), "xmm3");
	&vpaddd("xmm5", "xmm5", &DWP(272,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(288,"esp"));
	&vpaddd("xmm3", "xmm4", &DWP(304,"esp"));
	&test("eax", "eax");
	&jne(&label("chacha_blocks_avx_7"));
	&vmovdqu(&DWP(0,"esp"), "xmm0");
	&vmovdqu("xmm0", &DWP(400,"esp"));
	&vpunpckldq("xmm7", "xmm5", "xmm0");
	&vpunpckldq("xmm4", "xmm2", "xmm3");
	&vmovdqu(&DWP(16,"esp"), "xmm6");
	&vpunpcklqdq("xmm6", "xmm7", "xmm4");
	&vpunpckhqdq("xmm4", "xmm7", "xmm4");
	&vpunpckhdq("xmm5", "xmm5", "xmm0");
	&vpunpckhdq("xmm2", "xmm2", "xmm3");
	&vpunpcklqdq("xmm3", "xmm5", "xmm2");
	&vpunpckhqdq("xmm5", "xmm5", "xmm2");
	&vmovdqu(&DWP(0,"edx"), "xmm6");
	&vmovdqu(&DWP(64,"edx"), "xmm4");
	&vmovdqu(&DWP(128,"edx"), "xmm3");
	&vmovdqu(&DWP(192,"edx"), "xmm5");
	&vmovdqu("xmm0", &DWP(544,"esp"));
	&vmovdqu("xmm2", &DWP(384,"esp"));
	&vmovdqu("xmm4", &DWP(560,"esp"));
	&vmovdqu("xmm6", &DWP(592,"esp"));
	&vpaddd("xmm7", "xmm0", &DWP(336,"esp"));
	&vpaddd("xmm5", "xmm2", &DWP(240,"esp"));
	&vpaddd("xmm3", "xmm4", &DWP(320,"esp"));
	&vpaddd("xmm0", "xmm6", &DWP(208,"esp"));
	&vpunpckldq("xmm2", "xmm5", "xmm3");
	&vpunpckldq("xmm4", "xmm7", "xmm0");
	&vpunpckhdq("xmm3", "xmm5", "xmm3");
	&vpunpckhdq("xmm7", "xmm7", "xmm0");
	&vpunpcklqdq("xmm6", "xmm2", "xmm4");
	&vpunpckhqdq("xmm4", "xmm2", "xmm4");
	&vpunpcklqdq("xmm5", "xmm3", "xmm7");
	&vpunpckhqdq("xmm3", "xmm3", "xmm7");
	&vpaddd("xmm2", "xmm1", &DWP(192,"esp"));
	&vmovdqu("xmm1", &DWP(0,"esp"));
	&vmovdqu(&DWP(80,"edx"), "xmm4");
	&vmovdqu(&DWP(208,"edx"), "xmm3");
	&vmovdqu(&DWP(144,"edx"), "xmm5");
	&vmovdqu(&DWP(16,"edx"), "xmm6");
	&vmovdqu("xmm3", &DWP(528,"esp"));
	&vmovdqu("xmm4", &DWP(512,"esp"));
	&vpaddd("xmm5", "xmm1", &DWP(176,"esp"));
	&vpaddd("xmm1", "xmm3", &DWP(160,"esp"));
	&vpaddd("xmm7", "xmm4", &DWP(144,"esp"));
	&vpunpckldq("xmm0", "xmm2", "xmm5");
	&vpunpckldq("xmm6", "xmm1", "xmm7");
	&vpunpckhdq("xmm5", "xmm2", "xmm5");
	&vpunpckhdq("xmm1", "xmm1", "xmm7");
	&vpunpcklqdq("xmm3", "xmm0", "xmm6");
	&vpunpcklqdq("xmm2", "xmm5", "xmm1");
	&vmovdqu(&DWP(32,"edx"), "xmm3");
	&vmovdqu(&DWP(160,"edx"), "xmm2");
	&vpunpckhqdq("xmm7", "xmm5", "xmm1");
	&vmovdqu("xmm3", &DWP(464,"esp"));
	&vmovdqu("xmm1", &DWP(16,"esp"));
	&vmovdqu("xmm5", &DWP(480,"esp"));
	&vmovdqu("xmm2", &DWP(496,"esp"));
	&vpaddd("xmm4", "xmm3", &DWP(368,"esp"));
	&vpaddd("xmm5", "xmm5", &DWP(128,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(256,"esp"));
	&vpunpckhqdq("xmm6", "xmm0", "xmm6");
	&vpaddd("xmm0", "xmm1", &DWP(352,"esp"));
	&vmovdqu(&DWP(96,"edx"), "xmm6");
	&vmovdqu(&DWP(224,"edx"), "xmm7");
	&vpunpckldq("xmm6", "xmm4", "xmm0");
	&vpunpckldq("xmm7", "xmm5", "xmm2");
	&vpunpckhdq("xmm0", "xmm4", "xmm0");
	&vpunpckhdq("xmm4", "xmm5", "xmm2");
	&vpunpcklqdq("xmm3", "xmm6", "xmm7");
	&vpunpckhqdq("xmm1", "xmm6", "xmm7");
	&vpunpcklqdq("xmm2", "xmm0", "xmm4");
	&vpunpckhqdq("xmm0", "xmm0", "xmm4");
	&vmovdqu(&DWP(48,"edx"), "xmm3");
	&vmovdqu(&DWP(112,"edx"), "xmm1");
	&vmovdqu(&DWP(176,"edx"), "xmm2");
	&vmovdqu(&DWP(240,"edx"), "xmm0");
	&jmp(&label("chacha_blocks_avx_8"));
	&set_label("chacha_blocks_avx_7");
	&vmovdqu("xmm7", &DWP(400,"esp"));
	&vpunpckldq("xmm4", "xmm5", "xmm7");
	&vpunpckhdq("xmm7", "xmm5", "xmm7");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm3", "xmm2", "xmm3");
	&vpunpcklqdq("xmm2", "xmm4", "xmm5");
	&vpunpckhqdq("xmm4", "xmm4", "xmm5");
	&vpxor("xmm5", "xmm4", &DWP(64,"eax"));
	&vpunpcklqdq("xmm4", "xmm7", "xmm3");
	&vpunpckhqdq("xmm3", "xmm7", "xmm3");
	&vpxor("xmm4", "xmm4", &DWP(128,"eax"));
	&vpxor("xmm7", "xmm3", &DWP(192,"eax"));
	&vpxor("xmm2", "xmm2", &DWP(0,"eax"));
	&vmovdqu(&DWP(128,"edx"), "xmm4");
	&vmovdqu(&DWP(0,"edx"), "xmm2");
	&vmovdqu(&DWP(64,"edx"), "xmm5");
	&vmovdqu(&DWP(192,"edx"), "xmm7");
	&vmovdqu("xmm4", &DWP(384,"esp"));
	&vpaddd("xmm3", "xmm4", &DWP(240,"esp"));
	&vmovdqu("xmm2", &DWP(560,"esp"));
	&vmovdqu("xmm5", &DWP(544,"esp"));
	&vmovdqu("xmm4", &DWP(592,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(320,"esp"));
	&vpaddd("xmm7", "xmm5", &DWP(336,"esp"));
	&vpaddd("xmm4", "xmm4", &DWP(208,"esp"));
	&vpunpckldq("xmm5", "xmm3", "xmm2");
	&vpunpckhdq("xmm2", "xmm3", "xmm2");
	&vpunpckldq("xmm3", "xmm7", "xmm4");
	&vpunpckhdq("xmm4", "xmm7", "xmm4");
	&vpunpcklqdq("xmm7", "xmm5", "xmm3");
	&vpunpckhqdq("xmm5", "xmm5", "xmm3");
	&vpxor("xmm3", "xmm5", &DWP(80,"eax"));
	&vpunpcklqdq("xmm5", "xmm2", "xmm4");
	&vpunpckhqdq("xmm2", "xmm2", "xmm4");
	&vpxor("xmm7", "xmm7", &DWP(16,"eax"));
	&vpxor("xmm5", "xmm5", &DWP(144,"eax"));
	&vpxor("xmm4", "xmm2", &DWP(208,"eax"));
	&vmovdqu(&DWP(16,"edx"), "xmm7");
	&vmovdqu(&DWP(144,"edx"), "xmm5");
	&vmovdqu(&DWP(80,"edx"), "xmm3");
	&vmovdqu(&DWP(208,"edx"), "xmm4");
	&vpaddd("xmm7", "xmm0", &DWP(176,"esp"));
	&vpaddd("xmm5", "xmm1", &DWP(192,"esp"));
	&vmovdqu("xmm0", &DWP(528,"esp"));
	&vmovdqu("xmm1", &DWP(512,"esp"));
	&vpaddd("xmm3", "xmm0", &DWP(160,"esp"));
	&vpaddd("xmm0", "xmm1", &DWP(144,"esp"));
	&vpunpckldq("xmm1", "xmm5", "xmm7");
	&vpunpckldq("xmm2", "xmm3", "xmm0");
	&vpunpckhdq("xmm7", "xmm5", "xmm7");
	&vpunpckhdq("xmm4", "xmm3", "xmm0");
	&vpunpcklqdq("xmm3", "xmm1", "xmm2");
	&vpunpckhqdq("xmm1", "xmm1", "xmm2");
	&vpunpcklqdq("xmm2", "xmm7", "xmm4");
	&vpunpckhqdq("xmm7", "xmm7", "xmm4");
	&vpxor("xmm0", "xmm3", &DWP(32,"eax"));
	&vpxor("xmm1", "xmm1", &DWP(96,"eax"));
	&vpxor("xmm3", "xmm2", &DWP(160,"eax"));
	&vpxor("xmm4", "xmm7", &DWP(224,"eax"));
	&vmovdqu(&DWP(32,"edx"), "xmm0");
	&vmovdqu(&DWP(96,"edx"), "xmm1");
	&vmovdqu(&DWP(160,"edx"), "xmm3");
	&vmovdqu(&DWP(224,"edx"), "xmm4");
	&vpaddd("xmm0", "xmm6", &DWP(352,"esp"));
	&vmovdqu("xmm6", &DWP(480,"esp"));
	&vpaddd("xmm2", "xmm6", &DWP(128,"esp"));
	&vmovdqu("xmm5", &DWP(464,"esp"));
	&vmovdqu("xmm6", &DWP(496,"esp"));
	&vpaddd("xmm1", "xmm5", &DWP(368,"esp"));
	&vpaddd("xmm3", "xmm6", &DWP(256,"esp"));
	&vpunpckldq("xmm7", "xmm1", "xmm0");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm1", "xmm1", "xmm0");
	&vpunpckhdq("xmm0", "xmm2", "xmm3");
	&vpunpcklqdq("xmm4", "xmm7", "xmm5");
	&vpunpckhqdq("xmm7", "xmm7", "xmm5");
	&vpunpcklqdq("xmm2", "xmm1", "xmm0");
	&vpunpckhqdq("xmm0", "xmm1", "xmm0");
	&vpxor("xmm6", "xmm4", &DWP(48,"eax"));
	&vpxor("xmm3", "xmm7", &DWP(112,"eax"));
	&vpxor("xmm4", "xmm2", &DWP(176,"eax"));
	&vpxor("xmm1", "xmm0", &DWP(240,"eax"));
	&add("eax", 256);
	&vmovdqu(&DWP(48,"edx"), "xmm6");
	&vmovdqu(&DWP(112,"edx"), "xmm3");
	&vmovdqu(&DWP(176,"edx"), "xmm4");
	&vmovdqu(&DWP(240,"edx"), "xmm1");
	&set_label("chacha_blocks_avx_8");
	&mov("ecx", &DWP(420,"esp"));
	&add("edx", 256);
	&add("ecx", -256);
	&mov(&DWP(420,"esp"), "ecx");
	&cmp("ecx", 256);
	&jae(&label("chacha_blocks_avx_4"));
	&set_label("chacha_blocks_avx_9");
	&mov("ecx", &DWP(8,"ebp"));
	&mov(&DWP(32,"ecx"), "edi");
	&mov(&DWP(36,"ecx"), "esi");
	&mov("ebx", &DWP(420,"esp"));
	&vmovdqu("xmm0", &DWP(32,"ecx"));
	&set_label("chacha_blocks_avx_10");
	&cmp("ebx", 128);
	&jb(&label("chacha_blocks_avx_16"));
	&set_label("chacha_blocks_avx_11");
	&vmovdqu("xmm3", &DWP(64,"esp"));
	&vmovdqa("xmm1", "xmm0");
	&vmovdqu("xmm4", &DWP(80,"esp"));
	&vmovdqu("xmm6", &DWP(48,"esp"));
	&vpaddq("xmm2", "xmm0", &DWP(96,"esp"));
	&vmovdqu(&DWP(32,"esp"), "xmm3");
	&vmovdqu(&DWP(16,"esp"), "xmm4");
	&vmovdqu(&DWP(128,"esp"), "xmm6");
	&vmovdqu(&DWP(160,"esp"), "xmm6");
	&vmovdqu(&DWP(0,"esp"), "xmm2");
	&vmovdqu(&DWP(144,"esp"), "xmm2");
	&mov("esi", &DWP(424,"esp"));
	&vmovdqu("xmm5", &DWP(16,"esp"));
	&vmovdqu("xmm6", &DWP(32,"esp"));
	&vmovdqu(&DWP(112,"esp"), "xmm0");
	&set_label("chacha_blocks_avx_12");
	&vpaddd("xmm0", "xmm3", &DWP(160,"esp"));
	&vpaddd("xmm2", "xmm6", &DWP(128,"esp"));
	&vpxor("xmm1", "xmm1", "xmm0");
	&vmovdqu("xmm7", &DWP(448,"esp"));
	&vmovdqu(&DWP(176,"esp"), "xmm2");
	&vpshufb("xmm1", "xmm1", "xmm7");
	&vpxor("xmm2", "xmm2", &DWP(144,"esp"));
	&vpaddd("xmm4", "xmm4", "xmm1");
	&vpshufb("xmm2", "xmm2", "xmm7");
	&vpaddd("xmm5", "xmm5", "xmm2");
	&vpxor("xmm3", "xmm3", "xmm4");
	&vpxor("xmm7", "xmm6", "xmm5");
	&vpsrld("xmm6", "xmm3", 20);
	&vpslld("xmm3", "xmm3", 12);
	&vpor("xmm6", "xmm6", "xmm3");
	&vpsrld("xmm3", "xmm7", 20);
	&vpslld("xmm7", "xmm7", 12);
	&vpaddd("xmm0", "xmm0", "xmm6");
	&vpor("xmm7", "xmm3", "xmm7");
	&vpxor("xmm1", "xmm1", "xmm0");
	&vmovdqu(&DWP(192,"esp"), "xmm0");
	&vpaddd("xmm3", "xmm7", &DWP(176,"esp"));
	&vmovdqu("xmm0", &DWP(432,"esp"));
	&vpxor("xmm2", "xmm2", "xmm3");
	&vpshufb("xmm1", "xmm1", "xmm0");
	&vpshufb("xmm0", "xmm2", "xmm0");
	&vpaddd("xmm4", "xmm4", "xmm1");
	&vpaddd("xmm5", "xmm5", "xmm0");
	&vpxor("xmm6", "xmm6", "xmm4");
	&vpxor("xmm7", "xmm7", "xmm5");
	&vpsrld("xmm2", "xmm6", 25);
	&vpslld("xmm6", "xmm6", 7);
	&vpor("xmm2", "xmm2", "xmm6");
	&vpsrld("xmm6", "xmm7", 25);
	&vpslld("xmm7", "xmm7", 7);
	&vpor("xmm6", "xmm6", "xmm7");
	&vpshufd("xmm7", &DWP(192,"esp"), 147);
	&vpshufd("xmm3", "xmm3", 147);
	&vpaddd("xmm7", "xmm7", "xmm2");
	&vpshufd("xmm1", "xmm1", 78);
	&vpaddd("xmm3", "xmm3", "xmm6");
	&vmovdqu(&DWP(208,"esp"), "xmm7");
	&vpxor("xmm7", "xmm1", "xmm7");
	&vmovdqu("xmm1", &DWP(448,"esp"));
	&vpshufd("xmm0", "xmm0", 78);
	&vpshufb("xmm7", "xmm7", "xmm1");
	&vpxor("xmm0", "xmm0", "xmm3");
	&vpshufb("xmm0", "xmm0", "xmm1");
	&vpshufd("xmm4", "xmm4", 57);
	&vpshufd("xmm5", "xmm5", 57);
	&vpaddd("xmm1", "xmm4", "xmm7");
	&vpaddd("xmm5", "xmm5", "xmm0");
	&vpxor("xmm4", "xmm2", "xmm1");
	&vpxor("xmm2", "xmm6", "xmm5");
	&vpsrld("xmm6", "xmm4", 20);
	&vpslld("xmm4", "xmm4", 12);
	&vpor("xmm4", "xmm6", "xmm4");
	&vpsrld("xmm6", "xmm2", 20);
	&vpslld("xmm2", "xmm2", 12);
	&vpor("xmm2", "xmm6", "xmm2");
	&vpaddd("xmm6", "xmm4", &DWP(208,"esp"));
	&vpaddd("xmm3", "xmm3", "xmm2");
	&vmovdqu(&DWP(224,"esp"), "xmm2");
	&vpxor("xmm7", "xmm7", "xmm6");
	&vmovdqu("xmm2", &DWP(432,"esp"));
	&vpxor("xmm0", "xmm0", "xmm3");
	&vpshufb("xmm7", "xmm7", "xmm2");
	&vpshufb("xmm0", "xmm0", "xmm2");
	&vpshufd("xmm3", "xmm3", 57);
	&vpaddd("xmm5", "xmm5", "xmm0");
	&vmovdqu(&DWP(128,"esp"), "xmm3");
	&vpaddd("xmm3", "xmm1", "xmm7");
	&vpshufd("xmm0", "xmm0", 78);
	&vpxor("xmm2", "xmm4", "xmm3");
	&vmovdqu(&DWP(144,"esp"), "xmm0");
	&vpxor("xmm0", "xmm5", &DWP(224,"esp"));
	&vpshufd("xmm6", "xmm6", 57);
	&vmovdqu(&DWP(160,"esp"), "xmm6");
	&vpsrld("xmm6", "xmm2", 25);
	&vpshufd("xmm1", "xmm7", 78);
	&vpslld("xmm7", "xmm2", 7);
	&vpsrld("xmm2", "xmm0", 25);
	&vpslld("xmm0", "xmm0", 7);
	&vpshufd("xmm4", "xmm3", 147);
	&vpor("xmm3", "xmm6", "xmm7");
	&vpshufd("xmm5", "xmm5", 147);
	&vpor("xmm6", "xmm2", "xmm0");
	&add("esi", -2);
	&jne(&label("chacha_blocks_avx_12"));
	&set_label("chacha_blocks_avx_13");
	&vmovdqu("xmm2", &DWP(64,"esp"));
	&vmovdqu(&DWP(16,"esp"), "xmm5");
	&vpaddd("xmm7", "xmm2", "xmm3");
	&vmovdqu(&DWP(32,"esp"), "xmm6");
	&vmovdqu("xmm0", &DWP(112,"esp"));
	&vmovdqu("xmm5", &DWP(48,"esp"));
	&vpaddd("xmm1", "xmm0", "xmm1");
	&vmovdqu("xmm6", &DWP(160,"esp"));
	&vmovdqu("xmm3", &DWP(80,"esp"));
	&vpaddd("xmm6", "xmm5", "xmm6");
	&vpaddd("xmm0", "xmm5", &DWP(128,"esp"));
	&vpaddd("xmm4", "xmm3", "xmm4");
	&vpaddd("xmm5", "xmm2", &DWP(32,"esp"));
	&vpaddd("xmm2", "xmm3", &DWP(16,"esp"));
	&vmovdqu("xmm3", &DWP(0,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(144,"esp"));
	&test("eax", "eax");
	&je(&label("chacha_blocks_avx_15"));
	&set_label("chacha_blocks_avx_14");
	&vpxor("xmm6", "xmm6", &DWP(0,"eax"));
	&vpxor("xmm7", "xmm7", &DWP(16,"eax"));
	&vpxor("xmm4", "xmm4", &DWP(32,"eax"));
	&vpxor("xmm1", "xmm1", &DWP(48,"eax"));
	&vpxor("xmm0", "xmm0", &DWP(64,"eax"));
	&vpxor("xmm5", "xmm5", &DWP(80,"eax"));
	&vpxor("xmm2", "xmm2", &DWP(96,"eax"));
	&vpxor("xmm3", "xmm3", &DWP(112,"eax"));
	&add("eax", 128);
	&set_label("chacha_blocks_avx_15");
	&vmovdqu(&DWP(64,"edx"), "xmm0");
	&vmovdqu(&DWP(0,"edx"), "xmm6");
	&vmovdqu(&DWP(16,"edx"), "xmm7");
	&vmovdqu(&DWP(32,"edx"), "xmm4");
	&vmovdqu(&DWP(48,"edx"), "xmm1");
	&vmovdqu(&DWP(80,"edx"), "xmm5");
	&vmovdqu(&DWP(96,"edx"), "xmm2");
	&vmovdqu(&DWP(112,"edx"), "xmm3");
	&add("ebx", -128);
	&add("edx", 128);
	&vmovdqu("xmm0", &DWP(0,"esp"));
	&vpaddq("xmm0", "xmm0", &DWP(96,"esp"));
	&set_label("chacha_blocks_avx_16");
	&mov("edi", "ebx");
	&test("ebx", "ebx");
	&jne(&label("chacha_blocks_avx_18"));
	&set_label("chacha_blocks_avx_17");
	&vmovdqu(&DWP(32,"ecx"), "xmm0");
	&add("esp", 1268);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
	&set_label("chacha_blocks_avx_18");
	&mov(&DWP(20,"esp"), "edi");
	&xor("esi", "esi");
	&mov(&DWP(420,"esp"), "ebx");
	&vmovdqu("xmm7", &DWP(80,"esp"));
	&vmovdqu("xmm1", &DWP(64,"esp"));
	&vmovdqu("xmm6", &DWP(48,"esp"));
	&jmp(&label("chacha_blocks_avx_19"));
	&set_label("chacha_blocks_avx_35");
	&mov(&DWP(420,"esp"), "ecx");
	&add("edx", 64);
	&set_label("chacha_blocks_avx_19");
	&inc("esi");
	&mov("ecx", "esi");
	&shl("ecx", 6);
	&neg("ecx");
	&add("ecx", &DWP(20,"esp"));
	&lea("ebx", &DWP(64,"ecx"));
	&cmp("ebx", 64);
	&jae(&label("chacha_blocks_avx_30"));
	&set_label("chacha_blocks_avx_20");
	&test("eax", "eax");
	&je(&label("chacha_blocks_avx_29"));
	&set_label("chacha_blocks_avx_21");
	&test("ebx", "ebx");
	&je(&label("chacha_blocks_avx_28"));
	&set_label("chacha_blocks_avx_22");
	&mov("edi", "ebx");
	&shr("edi", 1);
	&mov(&DWP(12,"esp"), "edi");
	&test("edi", "edi");
	&jbe(&label("chacha_blocks_avx_37"));
	&set_label("chacha_blocks_avx_23");
	&mov(&DWP(8,"esp"), "esi");
	&xor("edi", "edi");
	&mov(&DWP(4,"esp"), "ecx");
	&mov(&DWP(16,"esp"), "edx");
	&mov("esi", &DWP(12,"esp"));
	&set_label("chacha_blocks_avx_24");
	&movzb("edx", &DWP(0,"eax","edi",2));
	&mov(&DWP(128,"esp","edi",2), "dl");
	&movzb("ecx", &DWP(1,"eax","edi",2));
	&mov(&DWP(129,"esp","edi",2), "cl");
	&inc("edi");
	&cmp("edi", "esi");
	&jb(&label("chacha_blocks_avx_24"));
	&set_label("chacha_blocks_avx_25");
	&mov("ecx", &DWP(4,"esp"));
	&lea("edi", &DWP(1,"edi","edi"));
	&mov("esi", &DWP(8,"esp"));
	&mov("edx", &DWP(16,"esp"));
	&mov(&DWP(0,"esp"), "edi");
	&set_label("chacha_blocks_avx_26");
	&lea("edi", &DWP(-1,"edi"));
	&cmp("edi", "ebx");
	&jae(&label("chacha_blocks_avx_28"));
	&set_label("chacha_blocks_avx_27");
	&movzb("eax", &DWP(0,"edi","eax"));
	&mov("edi", &DWP(0,"esp"));
	&mov(&DWP(127,"esp","edi"), "al");
	&set_label("chacha_blocks_avx_28");
	&lea("eax", &DWP(128,"esp"));
	&set_label("chacha_blocks_avx_29");
	&mov(&DWP(416,"esp"), "edx");
	&lea("edx", &DWP(128,"esp"));
	&set_label("chacha_blocks_avx_30");
	&vmovdqu(&DWP(112,"esp"), "xmm0");
	&vmovdqa("xmm2", "xmm6");
	&mov(&DWP(16,"esp"), "edx");
	&xor("edi", "edi");
	&mov(&DWP(24,"esp"), "eax");
	&vmovdqa("xmm3", "xmm1");
	&vmovdqu("xmm1", &DWP(448,"esp"));
	&vmovdqa("xmm4", "xmm7");
	&mov("edx", &DWP(424,"esp"));
	&vmovdqa("xmm5", "xmm0");
	&vmovdqu("xmm0", &DWP(432,"esp"));
	&set_label("chacha_blocks_avx_31");
	&vpaddd("xmm6", "xmm2", "xmm3");
	&inc("edi");
	&vpxor("xmm2", "xmm5", "xmm6");
	&vpshufb("xmm5", "xmm2", "xmm1");
	&vpaddd("xmm2", "xmm4", "xmm5");
	&lea("eax", &DWP(0,"edi","edi"));
	&vpxor("xmm4", "xmm3", "xmm2");
	&vpsrld("xmm3", "xmm4", 20);
	&vpslld("xmm7", "xmm4", 12);
	&vpor("xmm7", "xmm3", "xmm7");
	&vpaddd("xmm3", "xmm6", "xmm7");
	&vpxor("xmm5", "xmm5", "xmm3");
	&vpshufb("xmm4", "xmm5", "xmm0");
	&vpaddd("xmm6", "xmm2", "xmm4");
	&vpxor("xmm2", "xmm7", "xmm6");
	&vpsrld("xmm5", "xmm2", 25);
	&vpslld("xmm7", "xmm2", 7);
	&vpshufd("xmm3", "xmm3", 147);
	&vpor("xmm5", "xmm5", "xmm7");
	&vpshufd("xmm4", "xmm4", 78);
	&vpaddd("xmm2", "xmm3", "xmm5");
	&vpxor("xmm4", "xmm4", "xmm2");
	&vpshufb("xmm7", "xmm4", "xmm1");
	&vpshufd("xmm6", "xmm6", 57);
	&vpaddd("xmm3", "xmm6", "xmm7");
	&vpxor("xmm4", "xmm5", "xmm3");
	&vpsrld("xmm5", "xmm4", 20);
	&vpslld("xmm6", "xmm4", 12);
	&vpor("xmm4", "xmm5", "xmm6");
	&vpaddd("xmm5", "xmm2", "xmm4");
	&vpxor("xmm2", "xmm7", "xmm5");
	&vpshufb("xmm7", "xmm2", "xmm0");
	&vpaddd("xmm3", "xmm3", "xmm7");
	&vpxor("xmm6", "xmm4", "xmm3");
	&vpshufd("xmm2", "xmm5", 57);
	&vpshufd("xmm5", "xmm7", 78);
	&vpslld("xmm7", "xmm6", 7);
	&vpshufd("xmm4", "xmm3", 147);
	&vpsrld("xmm3", "xmm6", 25);
	&vpor("xmm3", "xmm3", "xmm7");
	&cmp("eax", "edx");
	&jne(&label("chacha_blocks_avx_31"));
	&set_label("chacha_blocks_avx_32");
	&vmovdqu("xmm7", &DWP(80,"esp"));
	&vmovdqu("xmm1", &DWP(64,"esp"));
	&vpaddd("xmm4", "xmm4", "xmm7");
	&vmovdqu("xmm6", &DWP(48,"esp"));
	&vpaddd("xmm3", "xmm3", "xmm1");
	&vmovdqu("xmm0", &DWP(112,"esp"));
	&vpaddd("xmm2", "xmm2", "xmm6");
	&mov("eax", &DWP(24,"esp"));
	&vpaddd("xmm5", "xmm5", "xmm0");
	&mov("edx", &DWP(16,"esp"));
	&test("eax", "eax");
	&je(&label("chacha_blocks_avx_34"));
	&set_label("chacha_blocks_avx_33");
	&vpxor("xmm2", "xmm2", &DWP(0,"eax"));
	&vpxor("xmm3", "xmm3", &DWP(16,"eax"));
	&vpxor("xmm4", "xmm4", &DWP(32,"eax"));
	&vpxor("xmm5", "xmm5", &DWP(48,"eax"));
	&add("eax", 64);
	&set_label("chacha_blocks_avx_34");
	&vmovdqu(&DWP(0,"edx"), "xmm2");
	&vmovdqu(&DWP(16,"edx"), "xmm3");
	&vmovdqu(&DWP(32,"edx"), "xmm4");
	&vmovdqu(&DWP(48,"edx"), "xmm5");
	&vpaddq("xmm0", "xmm0", &DWP(96,"esp"));
	&cmp("ebx", 64);
	&jbe(&label("chacha_blocks_avx_38"));
	&jmp(&label("chacha_blocks_avx_35"));
	&set_label("chacha_blocks_avx_36");
	&add("esp", 1268);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
	&set_label("chacha_blocks_avx_37");
	&mov("edi", 1);
	&mov(&DWP(0,"esp"), "edi");
	&jmp(&label("chacha_blocks_avx_26"));
	&set_label("chacha_blocks_avx_38");
	&mov("ebx", &DWP(420,"esp"));
	&mov("ecx", &DWP(8,"ebp"));
	&cmp("ebx", 64);
	&jae(&label("chacha_blocks_avx_17"));
	&set_label("chacha_blocks_avx_39");
	&test("ebx", "ebx");
	&jbe(&label("chacha_blocks_avx_17"));
	&set_label("chacha_blocks_avx_40");
	&mov("edi", &DWP(416,"esp"));
	&xor("esi", "esi");
	&set_label("chacha_blocks_avx_41");
	&movzb("eax", &DWP(0,"esi","edx"));
	&mov(&DWP(0,"esi","edi"), "al");
	&inc("esi");
	&cmp("esi", "ebx");
	&jb(&label("chacha_blocks_avx_41"));
	&jmp(&label("chacha_blocks_avx_17"));
}

sub chacha_blocks_xop {
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", ~63);
	&sub("esp", 704);
	&mov(&DWP(68,"esp"), "ebx");
	&mov(&DWP(72,"esp"), "esi");
	&mov(&DWP(76,"esp"), "edi");
	&mov("ecx", &DWP(8,"ebp"));
	&mov(&DWP(84,"esp"), "ecx");
	&mov("esi", &DWP(12,"ebp"));
	&mov("edx", &DWP(16,"ebp"));
	&mov("eax", &DWP(20,"ebp"));
	&label_to_reg("chacha_constants", "ebx");
	&vmovdqa("xmm0", &DWP(0,"ebx"));
	&vmovdqu("xmm1", &DWP(0,"ecx"));
	&vmovdqu("xmm2", &DWP(16,"ecx"));
	&vmovdqu("xmm3", &DWP(32,"ecx"));
	&vmovdqa(&DWP(0,"esp"), "xmm0");
	&vmovdqa(&DWP(16,"esp"), "xmm1");
	&vmovdqa(&DWP(32,"esp"), "xmm2");
	&vmovdqa(&DWP(48,"esp"), "xmm3");
	&mov("ecx", &DWP(48,"ecx"));
	&mov(&DWP(88,"esp"), "ecx");
	&cmp("eax", 128);
	&jb(&label("chacha_blocks_xop_bytesbetween0and127"));
	&cmp("eax", 256);
	&jb(&label("chacha_blocks_xop_bytesatleast128"));
	&vpshufd("xmm4", "xmm0", 0x00);
	&vpshufd("xmm5", "xmm0", 0x55);
	&vpshufd("xmm6", "xmm0", 0xaa);
	&vpshufd("xmm0", "xmm0", 0xff);
	&vmovdqa(&DWP(128,"esp"), "xmm4");
	&vmovdqa(&DWP(192,"esp"), "xmm5");
	&vmovdqa(&DWP(288,"esp"), "xmm6");
	&vmovdqa(&DWP(304,"esp"), "xmm0");
	&vpshufd("xmm0", "xmm1", 0x00);
	&vpshufd("xmm4", "xmm1", 0x55);
	&vpshufd("xmm5", "xmm1", 0xaa);
	&vpshufd("xmm1", "xmm1", 0xff);
	&vmovdqa(&DWP(144,"esp"), "xmm0");
	&vmovdqa(&DWP(208,"esp"), "xmm4");
	&vmovdqa(&DWP(256,"esp"), "xmm5");
	&vmovdqa(&DWP(272,"esp"), "xmm1");
	&vpshufd("xmm0", "xmm2", 0x00);
	&vpshufd("xmm1", "xmm2", 0x55);
	&vpshufd("xmm4", "xmm2", 0xaa);
	&vpshufd("xmm2", "xmm2", 0xff);
	&vmovdqa(&DWP(160,"esp"), "xmm0");
	&vmovdqa(&DWP(224,"esp"), "xmm1");
	&vmovdqa(&DWP(352,"esp"), "xmm4");
	&vmovdqa(&DWP(368,"esp"), "xmm2");
	&vpshufd("xmm0", "xmm3", 0xaa);
	&vpshufd("xmm1", "xmm3", 0xff);
	&vmovdqa(&DWP(320,"esp"), "xmm0");
	&vmovdqa(&DWP(336,"esp"), "xmm1");
	&jmp(&label("chacha_blocks_xop_bytesatleast256"));
	&set_label("chacha_blocks_xop_bytesatleast256",64);
	&mov("ecx", &DWP(48,"esp"));
	&mov("ebx", &DWP(4+48,"esp"));
	&mov(&DWP(176,"esp"), "ecx");
	&mov(&DWP(240,"esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(4+176,"esp"), "ecx");
	&mov(&DWP(4+240,"esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(8+176,"esp"), "ecx");
	&mov(&DWP(8+240,"esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(12+176,"esp"), "ecx");
	&mov(&DWP(12+240,"esp"), "ebx");
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(48,"esp"), "ecx");
	&mov(&DWP(4+48,"esp"), "ebx");
	&mov(&DWP(92,"esp"), "eax");
	&mov("eax", &DWP(88,"esp"));
	&vmovdqa("xmm0", &DWP(256,"esp"));
	&vmovdqa("xmm1", &DWP(272,"esp"));
	&vmovdqa("xmm2", &DWP(288,"esp"));
	&vmovdqa("xmm3", &DWP(304,"esp"));
	&vmovdqa("xmm4", &DWP(320,"esp"));
	&vmovdqa("xmm5", &DWP(336,"esp"));
	&vmovdqa("xmm6", &DWP(352,"esp"));
	&vmovdqa("xmm7", &DWP(368,"esp"));
	&vmovdqa(&DWP(576,"esp"), "xmm0");
	&vmovdqa(&DWP(592,"esp"), "xmm1");
	&vmovdqa(&DWP(608,"esp"), "xmm2");
	&vmovdqa(&DWP(624,"esp"), "xmm3");
	&vmovdqa(&DWP(640,"esp"), "xmm4");
	&vmovdqa(&DWP(656,"esp"), "xmm5");
	&vmovdqa(&DWP(672,"esp"), "xmm6");
	&vmovdqa(&DWP(688,"esp"), "xmm7");
	&vmovdqa("xmm0", &DWP(128,"esp"));
	&vmovdqa("xmm1", &DWP(144,"esp"));
	&vmovdqa("xmm2", &DWP(160,"esp"));
	&vmovdqa("xmm3", &DWP(176,"esp"));
	&vmovdqa("xmm4", &DWP(192,"esp"));
	&vmovdqa("xmm5", &DWP(208,"esp"));
	&vmovdqa("xmm6", &DWP(224,"esp"));
	&vmovdqa("xmm7", &DWP(240,"esp"));
	&set_label("chacha_blocks_xop_mainloop1");
	&vpaddd("xmm0", "xmm0", "xmm1");
	&vpxor("xmm3", "xmm3", "xmm0");
	&vprotd("xmm3", "xmm3", 16);
	&vpaddd("xmm4", "xmm4", "xmm5");
	&vpxor("xmm7", "xmm7", "xmm4");
	&vprotd("xmm7", "xmm7", 16);
	&vpaddd("xmm2", "xmm2", "xmm3");
	&vpaddd("xmm6", "xmm6", "xmm7");
	&vpxor("xmm1", "xmm1", "xmm2");
	&vpxor("xmm5", "xmm5", "xmm6");
	&vprotd("xmm1", "xmm1", 12);
	&vpaddd("xmm0", "xmm0", "xmm1");
	&vprotd("xmm5", "xmm5", 12);
	&vpaddd("xmm4", "xmm4", "xmm5");
	&vpxor("xmm3", "xmm3", "xmm0");
	&vpxor("xmm7", "xmm7", "xmm4");
	&vprotd("xmm3", "xmm3", 8);
	&vprotd("xmm7", "xmm7", 8);
	&vpaddd("xmm2", "xmm2", "xmm3");
	&vpaddd("xmm6", "xmm6", "xmm7");
	&vpxor("xmm1", "xmm1", "xmm2");
	&vpxor("xmm5", "xmm5", "xmm6");
	&vprotd("xmm1", "xmm1", 7);
	&vprotd("xmm5", "xmm5", 7);
	&vmovdqa(&DWP(464,"esp"), "xmm1");
	&vmovdqa("xmm1", &DWP(576,"esp"));
	&vmovdqa(&DWP(528,"esp"), "xmm5");
	&vmovdqa("xmm5", &DWP(592,"esp"));
	&vmovdqa(&DWP(448,"esp"), "xmm0");
	&vmovdqa(&DWP(512,"esp"), "xmm4");
	&vmovdqa(&DWP(480,"esp"), "xmm2");
	&vmovdqa(&DWP(544,"esp"), "xmm6");
	&vmovdqa(&DWP(496,"esp"), "xmm3");
	&vmovdqa(&DWP(560,"esp"), "xmm7");
	&vpaddd("xmm0", "xmm1", &DWP(608,"esp"));
	&vpaddd("xmm4", "xmm5", &DWP(624,"esp"));
	&vpxor("xmm3", "xmm0", &DWP(640,"esp"));
	&vpxor("xmm7", "xmm4", &DWP(656,"esp"));
	&vprotd("xmm3", "xmm3", 16);
	&vpaddd("xmm2", "xmm3", &DWP(672,"esp"));
	&vprotd("xmm7", "xmm7", 16);
	&vpaddd("xmm6", "xmm7", &DWP(688,"esp"));
	&vpxor("xmm1", "xmm1", "xmm2");
	&vpxor("xmm5", "xmm5", "xmm6");
	&vprotd("xmm1", "xmm1", 12);
	&vpaddd("xmm0", "xmm0", "xmm1");
	&vprotd("xmm5", "xmm5", 12);
	&vpaddd("xmm4", "xmm4", "xmm5");
	&vpxor("xmm3", "xmm3", "xmm0");
	&vpxor("xmm7", "xmm7", "xmm4");
	&vprotd("xmm3", "xmm3", 8);
	&vpaddd("xmm2", "xmm2", "xmm3");
	&vprotd("xmm7", "xmm7", 8);
	&vpaddd("xmm6", "xmm6", "xmm7");
	&vpxor("xmm1", "xmm1", "xmm2");
	&vpxor("xmm5", "xmm5", "xmm6");
	&vprotd("xmm1", "xmm1", 7);
	&vprotd("xmm5", "xmm5", 7);
	&vmovdqa(&DWP(640,"esp"), "xmm3");
	&vmovdqa(&DWP(624,"esp"), "xmm4");
	&vmovdqa(&DWP(608,"esp"), "xmm0");
	&vmovdqa(&DWP(592,"esp"), "xmm5");
	&vmovdqa("xmm5", &DWP(528,"esp"));
	&vpaddd("xmm4", "xmm5", &DWP(448,"esp"));
	&vpaddd("xmm0", "xmm1", &DWP(512,"esp"));
	&vpxor("xmm7", "xmm4", "xmm7");
	&vpxor("xmm3", "xmm0", &DWP(496,"esp"));
	&vprotd("xmm7", "xmm7", 16);
	&vpaddd("xmm2", "xmm2", "xmm7");
	&vprotd("xmm3", "xmm3", 16);
	&vpaddd("xmm6", "xmm6", "xmm3");
	&vpxor("xmm5", "xmm5", "xmm2");
	&vpxor("xmm1", "xmm1", "xmm6");
	&vprotd("xmm5", "xmm5", 12);
	&vpaddd("xmm4", "xmm4", "xmm5");
	&vprotd("xmm1", "xmm1", 12);
	&vpaddd("xmm0", "xmm0", "xmm1");
	&vpxor("xmm7", "xmm7", "xmm4");
	&vpxor("xmm3", "xmm3", "xmm0");
	&vprotd("xmm7", "xmm7", 8);
	&vpaddd("xmm2", "xmm2", "xmm7");
	&vprotd("xmm3", "xmm3", 8);
	&vpaddd("xmm6", "xmm6", "xmm3");
	&vpxor("xmm5", "xmm5", "xmm2");
	&vpxor("xmm1", "xmm1", "xmm6");
	&vprotd("xmm5", "xmm5", 7);
	&vprotd("xmm1", "xmm1", 7);
	&vmovdqa(&DWP(448,"esp"), "xmm4");
	&vmovdqa(&DWP(496,"esp"), "xmm3");
	&vmovdqa(&DWP(512,"esp"), "xmm0");
	&vmovdqa(&DWP(528,"esp"), "xmm5");
	&vmovdqa(&DWP(576,"esp"), "xmm1");
	&vmovdqa(&DWP(656,"esp"), "xmm7");
	&vmovdqa(&DWP(672,"esp"), "xmm2");
	&vmovdqa(&DWP(688,"esp"), "xmm6");
	&vmovdqa("xmm1", &DWP(592,"esp"));
	&vmovdqa("xmm5", &DWP(464,"esp"));
	&vpaddd("xmm0", "xmm1", &DWP(608,"esp"));
	&vpaddd("xmm4", "xmm5", &DWP(624,"esp"));
	&vpxor("xmm3", "xmm0", &DWP(560,"esp"));
	&vpxor("xmm7", "xmm4", &DWP(640,"esp"));
	&vprotd("xmm3", "xmm3", 16);
	&vpaddd("xmm2", "xmm3", &DWP(480,"esp"));
	&vprotd("xmm7", "xmm7", 16);
	&vpaddd("xmm6", "xmm7", &DWP(544,"esp"));
	&vpxor("xmm1", "xmm1", "xmm2");
	&vpxor("xmm5", "xmm5", "xmm6");
	&vprotd("xmm1", "xmm1", 12);
	&vpaddd("xmm0", "xmm0", "xmm1");
	&vprotd("xmm5", "xmm5", 12);
	&vpaddd("xmm4", "xmm4", "xmm5");
	&vpxor("xmm3", "xmm3", "xmm0");
	&vpxor("xmm7", "xmm7", "xmm4");
	&vprotd("xmm3", "xmm3", 8);
	&vpaddd("xmm2", "xmm2", "xmm3");
	&vprotd("xmm7", "xmm7", 8);
	&vpaddd("xmm6", "xmm6", "xmm7");
	&vpxor("xmm1", "xmm1", "xmm2");
	&vpxor("xmm5", "xmm5", "xmm6");
	&vprotd("xmm1", "xmm1", 7);
	&vprotd("xmm5", "xmm5", 7);
	&vmovdqa(&DWP(640,"esp"), "xmm7");
	&vmovdqa(&DWP(624,"esp"), "xmm4");
	&vmovdqa(&DWP(608,"esp"), "xmm0");
	&vmovdqa(&DWP(592,"esp"), "xmm1");
	&vmovdqa("xmm7", "xmm3");
	&vmovdqa("xmm1", "xmm5");
	&vmovdqa("xmm0", &DWP(448,"esp"));
	&vmovdqa("xmm3", &DWP(496,"esp"));
	&vmovdqa("xmm4", &DWP(512,"esp"));
	&vmovdqa("xmm5", &DWP(528,"esp"));
	&sub("eax", 2);
	&ja(&label("chacha_blocks_xop_mainloop1"));
	&vmovdqa(&DWP(480,"esp"), "xmm2");
	&vmovdqa(&DWP(544,"esp"), "xmm6");
	&vmovdqa(&DWP(464,"esp"), "xmm1");
	&vmovdqa(&DWP(560,"esp"), "xmm7");
	&cmp("esi", 0);
	&jbe(&label("chacha_blocks_xop_noinput1"));
	&vmovdqa("xmm0", &DWP(448,"esp"));
	&vmovdqa("xmm1", &DWP(512,"esp"));
	&vmovdqa("xmm2", &DWP(608,"esp"));
	&vmovdqa("xmm3", &DWP(624,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(128,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(192,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(288,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(304,"esp"));
	&vpunpckldq("xmm4", "xmm0", "xmm1");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm6", "xmm0", "xmm1");
	&vpunpckhdq("xmm7", "xmm2", "xmm3");
	&vpunpcklqdq("xmm0", "xmm4", "xmm5");
	&vpunpckhqdq("xmm1", "xmm4", "xmm5");
	&vpunpcklqdq("xmm2", "xmm6", "xmm7");
	&vpunpckhqdq("xmm3", "xmm6", "xmm7");
	&vpxor("xmm0", "xmm0", &DWP(0,"esi"));
	&vpxor("xmm1", "xmm1", &DWP(64,"esi"));
	&vpxor("xmm2", "xmm2", &DWP(128,"esi"));
	&vpxor("xmm3", "xmm3", &DWP(192,"esi"));
	&vmovdqu(&DWP(0,"edx"), "xmm0");
	&vmovdqu(&DWP(64,"edx"), "xmm1");
	&vmovdqu(&DWP(128,"edx"), "xmm2");
	&vmovdqu(&DWP(192,"edx"), "xmm3");
	&vmovdqa("xmm0", &DWP(464,"esp"));
	&vmovdqa("xmm1", &DWP(528,"esp"));
	&vmovdqa("xmm2", &DWP(576,"esp"));
	&vmovdqa("xmm3", &DWP(592,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(144,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(208,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(256,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(272,"esp"));
	&vpunpckldq("xmm4", "xmm0", "xmm1");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm6", "xmm0", "xmm1");
	&vpunpckhdq("xmm7", "xmm2", "xmm3");
	&vpunpcklqdq("xmm0", "xmm4", "xmm5");
	&vpunpckhqdq("xmm1", "xmm4", "xmm5");
	&vpunpcklqdq("xmm2", "xmm6", "xmm7");
	&vpunpckhqdq("xmm3", "xmm6", "xmm7");
	&vpxor("xmm0", "xmm0", &DWP(16+0,"esi"));
	&vpxor("xmm1", "xmm1", &DWP(16+64,"esi"));
	&vpxor("xmm2", "xmm2", &DWP(16+128,"esi"));
	&vpxor("xmm3", "xmm3", &DWP(16+192,"esi"));
	&vmovdqu(&DWP(16+0,"edx"), "xmm0");
	&vmovdqu(&DWP(16+64,"edx"), "xmm1");
	&vmovdqu(&DWP(16+128,"edx"), "xmm2");
	&vmovdqu(&DWP(16+192,"edx"), "xmm3");
	&vmovdqa("xmm0", &DWP(480,"esp"));
	&vmovdqa("xmm1", &DWP(544,"esp"));
	&vmovdqa("xmm2", &DWP(672,"esp"));
	&vmovdqa("xmm3", &DWP(688,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(160,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(224,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(352,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(368,"esp"));
	&vpunpckldq("xmm4", "xmm0", "xmm1");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm6", "xmm0", "xmm1");
	&vpunpckhdq("xmm7", "xmm2", "xmm3");
	&vpunpcklqdq("xmm0", "xmm4", "xmm5");
	&vpunpckhqdq("xmm1", "xmm4", "xmm5");
	&vpunpcklqdq("xmm2", "xmm6", "xmm7");
	&vpunpckhqdq("xmm3", "xmm6", "xmm7");
	&vpxor("xmm0", "xmm0", &DWP(32+0,"esi"));
	&vpxor("xmm1", "xmm1", &DWP(32+64,"esi"));
	&vpxor("xmm2", "xmm2", &DWP(32+128,"esi"));
	&vpxor("xmm3", "xmm3", &DWP(32+192,"esi"));
	&vmovdqu(&DWP(32+0,"edx"), "xmm0");
	&vmovdqu(&DWP(32+64,"edx"), "xmm1");
	&vmovdqu(&DWP(32+128,"edx"), "xmm2");
	&vmovdqu(&DWP(32+192,"edx"), "xmm3");
	&vmovdqa("xmm0", &DWP(496,"esp"));
	&vmovdqa("xmm1", &DWP(560,"esp"));
	&vmovdqa("xmm2", &DWP(640,"esp"));
	&vmovdqa("xmm3", &DWP(656,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(176,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(240,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(320,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(336,"esp"));
	&vpunpckldq("xmm4", "xmm0", "xmm1");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm6", "xmm0", "xmm1");
	&vpunpckhdq("xmm7", "xmm2", "xmm3");
	&vpunpcklqdq("xmm0", "xmm4", "xmm5");
	&vpunpckhqdq("xmm1", "xmm4", "xmm5");
	&vpunpcklqdq("xmm2", "xmm6", "xmm7");
	&vpunpckhqdq("xmm3", "xmm6", "xmm7");
	&vpxor("xmm0", "xmm0", &DWP(48+0,"esi"));
	&vpxor("xmm1", "xmm1", &DWP(48+64,"esi"));
	&vpxor("xmm2", "xmm2", &DWP(48+128,"esi"));
	&vpxor("xmm3", "xmm3", &DWP(48+192,"esi"));
	&vmovdqu(&DWP(48+0,"edx"), "xmm0");
	&vmovdqu(&DWP(48+64,"edx"), "xmm1");
	&vmovdqu(&DWP(48+128,"edx"), "xmm2");
	&vmovdqu(&DWP(48+192,"edx"), "xmm3");
	&add("esi", 256);
	&jmp(&label("chacha_blocks_xop_mainloop1_cont"));
	&set_label("chacha_blocks_xop_noinput1");
	&vmovdqa("xmm0", &DWP(448,"esp"));
	&vmovdqa("xmm1", &DWP(512,"esp"));
	&vmovdqa("xmm2", &DWP(608,"esp"));
	&vmovdqa("xmm3", &DWP(624,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(128,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(192,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(288,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(304,"esp"));
	&vpunpckldq("xmm4", "xmm0", "xmm1");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm6", "xmm0", "xmm1");
	&vpunpckhdq("xmm7", "xmm2", "xmm3");
	&vpunpcklqdq("xmm0", "xmm4", "xmm5");
	&vpunpckhqdq("xmm1", "xmm4", "xmm5");
	&vpunpcklqdq("xmm2", "xmm6", "xmm7");
	&vpunpckhqdq("xmm3", "xmm6", "xmm7");
	&vmovdqu(&DWP(0,"edx"), "xmm0");
	&vmovdqu(&DWP(64,"edx"), "xmm1");
	&vmovdqu(&DWP(128,"edx"), "xmm2");
	&vmovdqu(&DWP(192,"edx"), "xmm3");
	&vmovdqa("xmm0", &DWP(464,"esp"));
	&vmovdqa("xmm1", &DWP(528,"esp"));
	&vmovdqa("xmm2", &DWP(576,"esp"));
	&vmovdqa("xmm3", &DWP(592,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(144,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(208,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(256,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(272,"esp"));
	&vpunpckldq("xmm4", "xmm0", "xmm1");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm6", "xmm0", "xmm1");
	&vpunpckhdq("xmm7", "xmm2", "xmm3");
	&vpunpcklqdq("xmm0", "xmm4", "xmm5");
	&vpunpckhqdq("xmm1", "xmm4", "xmm5");
	&vpunpcklqdq("xmm2", "xmm6", "xmm7");
	&vpunpckhqdq("xmm3", "xmm6", "xmm7");
	&vmovdqu(&DWP(16+0,"edx"), "xmm0");
	&vmovdqu(&DWP(16+64,"edx"), "xmm1");
	&vmovdqu(&DWP(16+128,"edx"), "xmm2");
	&vmovdqu(&DWP(16+192,"edx"), "xmm3");
	&vmovdqa("xmm0", &DWP(480,"esp"));
	&vmovdqa("xmm1", &DWP(544,"esp"));
	&vmovdqa("xmm2", &DWP(672,"esp"));
	&vmovdqa("xmm3", &DWP(688,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(160,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(224,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(352,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(368,"esp"));
	&vpunpckldq("xmm4", "xmm0", "xmm1");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm6", "xmm0", "xmm1");
	&vpunpckhdq("xmm7", "xmm2", "xmm3");
	&vpunpcklqdq("xmm0", "xmm4", "xmm5");
	&vpunpckhqdq("xmm1", "xmm4", "xmm5");
	&vpunpcklqdq("xmm2", "xmm6", "xmm7");
	&vpunpckhqdq("xmm3", "xmm6", "xmm7");
	&vmovdqu(&DWP(32+0,"edx"), "xmm0");
	&vmovdqu(&DWP(32+64,"edx"), "xmm1");
	&vmovdqu(&DWP(32+128,"edx"), "xmm2");
	&vmovdqu(&DWP(32+192,"edx"), "xmm3");
	&vmovdqa("xmm0", &DWP(496,"esp"));
	&vmovdqa("xmm1", &DWP(560,"esp"));
	&vmovdqa("xmm2", &DWP(640,"esp"));
	&vmovdqa("xmm3", &DWP(656,"esp"));
	&vpaddd("xmm0", "xmm0", &DWP(176,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(240,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(320,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(336,"esp"));
	&vpunpckldq("xmm4", "xmm0", "xmm1");
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpunpckhdq("xmm6", "xmm0", "xmm1");
	&vpunpckhdq("xmm7", "xmm2", "xmm3");
	&vpunpcklqdq("xmm0", "xmm4", "xmm5");
	&vpunpckhqdq("xmm1", "xmm4", "xmm5");
	&vpunpcklqdq("xmm2", "xmm6", "xmm7");
	&vpunpckhqdq("xmm3", "xmm6", "xmm7");
	&vmovdqu(&DWP(48+0,"edx"), "xmm0");
	&vmovdqu(&DWP(48+64,"edx"), "xmm1");
	&vmovdqu(&DWP(48+128,"edx"), "xmm2");
	&vmovdqu(&DWP(48+192,"edx"), "xmm3");
	&set_label("chacha_blocks_xop_mainloop1_cont");
	&mov("eax", &DWP(92,"esp"));
	&sub("eax", 256);
	&add("edx", 256);
	&cmp("eax", 256);
	&jae(&label("chacha_blocks_xop_bytesatleast256"));
	&cmp("eax", 128);
	&jb(&label("chacha_blocks_xop_bytesbetween0and127"));
	&set_label("chacha_blocks_xop_bytesatleast128");
	&mov("ecx", 1);
	&vmovdqa("xmm0", &DWP(0,"esp"));
	&vmovdqa("xmm1", &DWP(16,"esp"));
	&vmovdqa("xmm2", &DWP(32,"esp"));
	&vmovdqa("xmm3", &DWP(48,"esp"));
	&vmovd("xmm7", "ecx");
	&vmovdqa("xmm4", "xmm0");
	&vmovdqa("xmm5", "xmm1");
	&vmovdqa("xmm6", "xmm2");
	&vpaddq("xmm7", "xmm7", "xmm3");
	&mov(&DWP(92,"esp"), "eax");
	&mov("eax", &DWP(88,"esp"));
	&vmovdqa(&DWP(448,"esp"), "xmm7");
	&set_label("chacha_blocks_xop_mainloop2");
	&vpaddd("xmm0", "xmm1", "xmm0");
	&vpaddd("xmm4", "xmm5", "xmm4");
	&vpxor("xmm3", "xmm0", "xmm3");
	&vpxor("xmm7", "xmm4", "xmm7");
	&vprotd("xmm3", "xmm3", 16);
	&vprotd("xmm7", "xmm7", 16);
	&vpaddd("xmm2", "xmm3", "xmm2");
	&vpaddd("xmm6", "xmm7", "xmm6");
	&vpxor("xmm1", "xmm2", "xmm1");
	&vpxor("xmm5", "xmm6", "xmm5");
	&vprotd("xmm1", "xmm1", 12);
	&vprotd("xmm5", "xmm5", 12);
	&vpaddd("xmm0", "xmm1", "xmm0");
	&vpaddd("xmm4", "xmm5", "xmm4");
	&vpxor("xmm3", "xmm0", "xmm3");
	&vpxor("xmm7", "xmm4", "xmm7");
	&vprotd("xmm3", "xmm3", 8);
	&vprotd("xmm7", "xmm7", 8);
	&vpaddd("xmm2", "xmm3", "xmm2");
	&vpaddd("xmm6", "xmm7", "xmm6");
	&vpshufd("xmm0", "xmm0", 0x93);
	&vpshufd("xmm4", "xmm4", 0x93);
	&vpxor("xmm1", "xmm2", "xmm1");
	&vpxor("xmm5", "xmm6", "xmm5");
	&vprotd("xmm1", "xmm1", 7);
	&vprotd("xmm5", "xmm5", 7);
	&vpshufd("xmm3", "xmm3", 0x4e);
	&vpshufd("xmm7", "xmm7", 0x4e);
	&vpaddd("xmm0", "xmm1", "xmm0");
	&vpaddd("xmm4", "xmm5", "xmm4");
	&vpshufd("xmm2", "xmm2", 0x39);
	&vpshufd("xmm6", "xmm6", 0x39);
	&vpxor("xmm3", "xmm0", "xmm3");
	&vpxor("xmm7", "xmm4", "xmm7");
	&vprotd("xmm3", "xmm3", 16);
	&vprotd("xmm7", "xmm7", 16);
	&vpaddd("xmm2", "xmm3", "xmm2");
	&vpaddd("xmm6", "xmm7", "xmm6");
	&vpxor("xmm1", "xmm2", "xmm1");
	&vpxor("xmm5", "xmm6", "xmm5");
	&vprotd("xmm1", "xmm1", 12);
	&vprotd("xmm5", "xmm5", 12);
	&vpaddd("xmm0", "xmm1", "xmm0");
	&vpaddd("xmm4", "xmm5", "xmm4");
	&vpxor("xmm3", "xmm0", "xmm3");
	&vpxor("xmm7", "xmm4", "xmm7");
	&vprotd("xmm3", "xmm3", 8);
	&vprotd("xmm7", "xmm7", 8);
	&vpaddd("xmm2", "xmm3", "xmm2");
	&vpaddd("xmm6", "xmm7", "xmm6");
	&vpxor("xmm1", "xmm2", "xmm1");
	&vpxor("xmm5", "xmm6", "xmm5");
	&vpshufd("xmm0", "xmm0", 0x39);
	&vpshufd("xmm4", "xmm4", 0x39);
	&vprotd("xmm1", "xmm1", 7);
	&vprotd("xmm5", "xmm5", 7);
	&vpshufd("xmm3", "xmm3", 0x4e);
	&vpshufd("xmm7", "xmm7", 0x4e);
	&vpshufd("xmm2", "xmm2", 0x93);
	&vpshufd("xmm6", "xmm6", 0x93);
	&sub("eax", 2);
	&jnz(&label("chacha_blocks_xop_mainloop2"));
	&vpaddd("xmm0", "xmm0", &DWP(0,"esp"));
	&vpaddd("xmm4", "xmm4", &DWP(0,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(16,"esp"));
	&vpaddd("xmm5", "xmm5", &DWP(16,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(32,"esp"));
	&vpaddd("xmm6", "xmm6", &DWP(32,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(48,"esp"));
	&vpaddd("xmm7", "xmm7", &DWP(448,"esp"));
	&and("esi", "esi");
	&jz(&label("chacha_blocks_xop_noinput2"));
	&vpxor("xmm0", "xmm0", &DWP(0,"esi"));
	&vpxor("xmm1", "xmm1", &DWP(16,"esi"));
	&vpxor("xmm2", "xmm2", &DWP(32,"esi"));
	&vpxor("xmm3", "xmm3", &DWP(48,"esi"));
	&vpxor("xmm4", "xmm4", &DWP(64,"esi"));
	&vpxor("xmm5", "xmm5", &DWP(80,"esi"));
	&vpxor("xmm6", "xmm6", &DWP(96,"esi"));
	&vpxor("xmm7", "xmm7", &DWP(112,"esi"));
	&add("esi", 128);
	&set_label("chacha_blocks_xop_noinput2");
	&vmovdqu(&DWP(0,"edx"), "xmm0");
	&vmovdqu(&DWP(16,"edx"), "xmm1");
	&vmovdqu(&DWP(32,"edx"), "xmm2");
	&vmovdqu(&DWP(48,"edx"), "xmm3");
	&vmovdqu(&DWP(64,"edx"), "xmm4");
	&vmovdqu(&DWP(80,"edx"), "xmm5");
	&vmovdqu(&DWP(96,"edx"), "xmm6");
	&vmovdqu(&DWP(112,"edx"), "xmm7");
	&mov("eax", &DWP(92,"esp"));
	&sub("eax", 128);
	&add("edx", 128);
	&mov("ecx", &DWP(48,"esp"));
	&mov("ebx", &DWP(4+48,"esp"));
	&add("ecx", 2);
	&adc("ebx", 0);
	&mov(&DWP(48,"esp"), "ecx");
	&mov(&DWP(4+48,"esp"), "ebx");
	&set_label("chacha_blocks_xop_bytesbetween0and127");
	&and("eax", "eax");
	&jz(&label("chacha_blocks_xop_done"));
	&cmp("eax", 64);
	&jae(&label("chacha_blocks_xop_nocopy"));
	&mov(&DWP(92,"esp"), "edx");
	&cmp("esi", 0);
	&jbe(&label("chacha_blocks_xop_noinput3"));
	&lea("edi", &DWP(128,"esp"));
	&mov("ecx", "eax");
	&rep_movsb();
	&lea("esi", &DWP(128,"esp"));
	&set_label("chacha_blocks_xop_noinput3");
	&lea("edx", &DWP(128,"esp"));
	&set_label("chacha_blocks_xop_nocopy");
	&mov(&DWP(80,"esp"), "eax");
	&vmovdqa("xmm0", &DWP(0,"esp"));
	&vmovdqa("xmm1", &DWP(16,"esp"));
	&vmovdqa("xmm2", &DWP(32,"esp"));
	&vmovdqa("xmm3", &DWP(48,"esp"));
	&mov("eax", &DWP(88,"esp"));
	&set_label("chacha_blocks_xop_mainloop3");
	&vpaddd("xmm0", "xmm1", "xmm0");
	&vpxor("xmm3", "xmm0", "xmm3");
	&vprotd("xmm3", "xmm3", 16);
	&vpaddd("xmm2", "xmm3", "xmm2");
	&vpxor("xmm1", "xmm2", "xmm1");
	&vprotd("xmm1", "xmm1", 12);
	&vpaddd("xmm0", "xmm1", "xmm0");
	&vpxor("xmm3", "xmm0", "xmm3");
	&vprotd("xmm3", "xmm3", 8);
	&vpaddd("xmm2", "xmm3", "xmm2");
	&vpshufd("xmm0", "xmm0", 0x93);
	&vpxor("xmm1", "xmm2", "xmm1");
	&vprotd("xmm1", "xmm1", 7);
	&vpshufd("xmm3", "xmm3", 0x4e);
	&vpaddd("xmm0", "xmm1", "xmm0");
	&vpshufd("xmm2", "xmm2", 0x39);
	&vpxor("xmm3", "xmm0", "xmm3");
	&vprotd("xmm3", "xmm3", 16);
	&vpaddd("xmm2", "xmm3", "xmm2");
	&vpxor("xmm1", "xmm2", "xmm1");
	&vprotd("xmm1", "xmm1", 12);
	&vpaddd("xmm0", "xmm1", "xmm0");
	&vpxor("xmm3", "xmm0", "xmm3");
	&vprotd("xmm3", "xmm3", 8);
	&vpaddd("xmm2", "xmm3", "xmm2");
	&vpxor("xmm1", "xmm2", "xmm1");
	&vpshufd("xmm0", "xmm0", 0x39);
	&vprotd("xmm1", "xmm1", 7);
	&vpshufd("xmm3", "xmm3", 0x4e);
	&vpshufd("xmm2", "xmm2", 0x93);
	&sub("eax", 2);
	&ja(&label("chacha_blocks_xop_mainloop3"));
	&vpaddd("xmm0", "xmm0", &DWP(0,"esp"));
	&vpaddd("xmm1", "xmm1", &DWP(16,"esp"));
	&vpaddd("xmm2", "xmm2", &DWP(32,"esp"));
	&vpaddd("xmm3", "xmm3", &DWP(48,"esp"));
	&cmp("esi", 0);
	&jbe(&label("chacha_blocks_xop_noinput4"));
	&vpxor("xmm0", "xmm0", &DWP(0,"esi"));
	&vpxor("xmm1", "xmm1", &DWP(16,"esi"));
	&vpxor("xmm2", "xmm2", &DWP(32,"esi"));
	&vpxor("xmm3", "xmm3", &DWP(48,"esi"));
	&add("esi", 64);
	&set_label("chacha_blocks_xop_noinput4");
	&vmovdqu(&DWP(0,"edx"), "xmm0");
	&vmovdqu(&DWP(16,"edx"), "xmm1");
	&vmovdqu(&DWP(32,"edx"), "xmm2");
	&vmovdqu(&DWP(48,"edx"), "xmm3");
	&mov("eax", &DWP(80,"esp"));
	&mov("ecx", &DWP(48,"esp"));
	&mov("ebx", &DWP(4+48,"esp"));
	&add("ecx", 1);
	&adc("ebx", 0);
	&mov(&DWP(48,"esp"), "ecx");
	&mov(&DWP(4+48,"esp"), "ebx");
	&cmp("eax", 64);
	&jb(&label("chacha_blocks_xop_bytesatmost64"));
	&sub("eax", 64);
	&add("edx", 64);
	&jmp(&label("chacha_blocks_xop_bytesbetween0and127"));
	&set_label("chacha_blocks_xop_bytesatmost64");
	&jae(&label("chacha_blocks_xop_bytesatleast64"));
	&mov("esi", "edx");
	&mov("edi", &DWP(92,"esp"));
	&mov("ecx", "eax");
	&rep_movsb();
	&set_label("chacha_blocks_xop_bytesatleast64");
	&set_label("chacha_blocks_xop_done");
	&mov("eax", &DWP(84,"esp"));
	&vmovdqa("xmm0", &DWP(48,"esp"));
	&vmovdqu(&DWP(32,"eax"), "xmm0");
	&mov("eax", &DWP(64,"esp"));
	&mov("ebx", &DWP(68,"esp"));
	&mov("esi", &DWP(72,"esp"));
	&mov("edi", &DWP(76,"esp"));
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
}

sub chacha_blocks_avx2 {
	&set_label("chacha_blocks_avx2_local");
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", ~63);
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 3124);
	&mov("eax", 1);
	&label_to_reg("chacha_constants", "esi");
	&mov("ebx", &DWP(16,"ebp"));
	&mov("edx", &DWP(20,"ebp"));
	&vmovd("xmm0", "eax");
	&vmovdqu("ymm1", &DWP(16,"esi"));
	&vmovdqu("ymm2", &DWP(32,"esi"));
	&vmovdqu(&DWP(592,"esp"), "xmm0");
	&vmovdqu(&DWP(1568,"esp"), "ymm1");
	&vmovdqu(&DWP(1536,"esp"), "ymm2");
	&mov("edi", &DWP(12,"ebp"));
	&mov(&DWP(1508,"esp"), "ebx");
	&test("edx", "edx");
	&je(&label("chacha_blocks_avx2_44"));
	&set_label("chacha_blocks_avx2_2");
	&mov("ecx", &DWP(8,"ebp"));
	&vmovdqu("xmm5", &DWP(0,"esi"));
	&vmovdqu(&DWP(544,"esp"), "xmm5");
	&vmovdqu("xmm0", &DWP(16,"ecx"));
	&vmovdqu("xmm1", &DWP(0,"ecx"));
	&vmovdqu(&DWP(576,"esp"), "xmm0");
	&vmovdqu("xmm0", &DWP(32,"ecx"));
	&mov("eax", &DWP(48,"ecx"));
	&vmovdqu(&DWP(560,"esp"), "xmm1");
	&mov(&DWP(1516,"esp"), "eax");
	&vmovq("xmm5", &DWP(8,"esi"));
	&vmovq("xmm1", &DWP(8,"ecx"));
	&vmovdqu(&DWP(304,"esp"), "xmm0");
	&vpsrldq("xmm2", "xmm0", 4);
	&cmp("edx", 512);
	&jb(&label("chacha_blocks_avx2_10"));
	&set_label("chacha_blocks_avx2_3");
	&vpbroadcastd("ymm7", &DWP(312,"esp"));
	&vmovdqu("xmm4", &DWP(576,"esp"));
	&vmovdqu(&DWP(704,"esp"), "ymm7");
	&vmovdqu(&DWP(288,"esp"), "xmm4");
	&vmovdqu("xmm6", &DWP(560,"esp"));
	&vmovdqu("xmm3", &DWP(544,"esp"));
	&vmovdqu(&DWP(272,"esp"), "xmm6");
	&vpbroadcastd("ymm7", &DWP(316,"esp"));
	&vmovdqu(&DWP(672,"esp"), "ymm7");
	&vmovdqu(&DWP(256,"esp"), "xmm3");
	&vmovd("ecx", "xmm0");
	&vmovd("eax", "xmm2");
	&vpbroadcastd("ymm7", &DWP(292,"esp"));
	&vmovdqu(&DWP(800,"esp"), "ymm7");
	&vpbroadcastd("ymm0", "xmm4");
	&vpbroadcastd("ymm1", "xmm1");
	&vpbroadcastd("ymm7", &DWP(296,"esp"));
	&vmovdqu(&DWP(832,"esp"), "ymm0");
	&vmovdqu(&DWP(864,"esp"), "ymm1");
	&vmovdqu(&DWP(768,"esp"), "ymm7");
	&vpbroadcastd("ymm2", "xmm3");
	&vpbroadcastd("ymm3", "xmm6");
	&vpbroadcastd("ymm0", &DWP(260,"esp"));
	&vpbroadcastd("ymm4", &DWP(276,"esp"));
	&vpbroadcastd("ymm1", &DWP(284,"esp"));
	&vpbroadcastd("ymm7", &DWP(300,"esp"));
	&vpbroadcastd("ymm6", "xmm5");
	&vpbroadcastd("ymm5", &DWP(268,"esp"));
	&vmovdqu(&DWP(736,"esp"), "ymm7");
	&vmovdqu(&DWP(1088,"esp"), "ymm1");
	&vmovdqu(&DWP(1056,"esp"), "ymm4");
	&vmovdqu(&DWP(1024,"esp"), "ymm3");
	&vmovdqu(&DWP(992,"esp"), "ymm5");
	&vmovdqu(&DWP(960,"esp"), "ymm6");
	&vmovdqu(&DWP(928,"esp"), "ymm0");
	&vmovdqu(&DWP(896,"esp"), "ymm2");
	&mov(&DWP(1512,"esp"), "edx");
	&mov(&DWP(1504,"esp"), "esi");
	&set_label("chacha_blocks_avx2_4");
	&mov("edx", "ecx");
	&mov("esi", "eax");
	&add("edx", 1);
	&mov(&DWP(612,"esp"), "edx");
	&mov("edx", "ecx");
	&adc("esi", 0);
	&add("edx", 2);
	&mov(&DWP(644,"esp"), "esi");
	&mov("esi", "eax");
	&mov(&DWP(616,"esp"), "edx");
	&mov("edx", "ecx");
	&adc("esi", 0);
	&add("edx", 3);
	&mov(&DWP(648,"esp"), "esi");
	&mov("esi", "eax");
	&mov(&DWP(620,"esp"), "edx");
	&mov("edx", "ecx");
	&vmovdqu("ymm3", &DWP(864,"esp"));
	&adc("esi", 0);
	&vmovdqu("ymm7", &DWP(832,"esp"));
	&vmovdqu("ymm2", &DWP(896,"esp"));
	&vmovdqu("ymm5", &DWP(960,"esp"));
	&vmovdqu("ymm6", &DWP(1056,"esp"));
	&vmovdqu(&DWP(1760,"esp"), "ymm3");
	&vmovdqu("ymm3", &DWP(800,"esp"));
	&vmovdqu(&DWP(1248,"esp"), "ymm7");
	&vmovdqu("ymm7", &DWP(768,"esp"));
	&vmovdqu("ymm4", &DWP(928,"esp"));
	&vmovdqu(&DWP(1792,"esp"), "ymm6");
	&vmovdqu(&DWP(1216,"esp"), "ymm3");
	&vmovdqu("ymm3", &DWP(736,"esp"));
	&vmovdqu(&DWP(1728,"esp"), "ymm7");
	&vmovdqu("ymm7", &DWP(704,"esp"));
	&vmovdqu(&DWP(1888,"esp"), "ymm5");
	&vmovdqu(&DWP(1824,"esp"), "ymm2");
	&vmovdqu(&DWP(1696,"esp"), "ymm3");
	&vmovdqu("ymm3", &DWP(672,"esp"));
	&vmovdqu("ymm6", &DWP(1088,"esp"));
	&vmovdqu("ymm2", &DWP(1248,"esp"));
	&vmovdqu(&DWP(1632,"esp"), "ymm7");
	&vmovdqu(&DWP(1856,"esp"), "ymm4");
	&vmovdqu(&DWP(1600,"esp"), "ymm3");
	&add("edx", 4);
	&mov(&DWP(652,"esp"), "esi");
	&mov("esi", "eax");
	&mov(&DWP(624,"esp"), "edx");
	&mov("edx", "ecx");
	&adc("esi", 0);
	&add("edx", 5);
	&mov(&DWP(656,"esp"), "esi");
	&mov("esi", "eax");
	&mov(&DWP(628,"esp"), "edx");
	&mov("edx", "ecx");
	&adc("esi", 0);
	&add("edx", 6);
	&mov(&DWP(660,"esp"), "esi");
	&mov("esi", "eax");
	&mov(&DWP(632,"esp"), "edx");
	&mov("edx", "ecx");
	&adc("esi", 0);
	&add("edx", 7);
	&mov(&DWP(664,"esp"), "esi");
	&mov("esi", "eax");
	&adc("esi", 0);
	&mov(&DWP(640,"esp"), "eax");
	&mov(&DWP(668,"esp"), "esi");
	&vmovdqu("ymm1", &DWP(640,"esp"));
	&mov(&DWP(608,"esp"), "ecx");
	&add("ecx", 8);
	&mov(&DWP(636,"esp"), "edx");
	&vmovdqu(&DWP(1120,"esp"), "ymm1");
	&adc("eax", 0);
	&vmovdqu(&DWP(1184,"esp"), "ymm1");
	&vmovdqu("ymm0", &DWP(608,"esp"));
	&vmovdqu("ymm1", &DWP(992,"esp"));
	&vmovdqu("ymm5", &DWP(1184,"esp"));
	&vmovdqu(&DWP(1152,"esp"), "ymm0");
	&vmovdqu(&DWP(1664,"esp"), "ymm0");
	&vmovdqu(&DWP(1920,"esp"), "ymm1");
	&vmovdqu("ymm0", &DWP(1024,"esp"));
	&vmovdqu("ymm1", &DWP(1216,"esp"));
	&mov("edx", &DWP(1516,"esp"));
	&mov(&DWP(304,"esp"), "ecx");
	&mov(&DWP(308,"esp"), "eax");
	&jmp(&label("chacha_blocks_avx2_5"));
	&align(32);
	&nop();
	&nop();
	&nop();
	&nop();
	&nop();
	&set_label("chacha_blocks_avx2_5");
	&vmovdqu("ymm7", &DWP(1792,"esp"));
	&vmovdqu(&DWP(1952,"esp"), "ymm6");
	&vpaddd("ymm3", "ymm0", &DWP(1824,"esp"));
	&vpaddd("ymm6", "ymm6", &DWP(1920,"esp"));
	&vpaddd("ymm7", "ymm7", &DWP(1856,"esp"));
	&vpxor("ymm4", "ymm3", &DWP(1664,"esp"));
	&vmovdqu(&DWP(1984,"esp"), "ymm3");
	&vmovdqu("ymm3", &DWP(1568,"esp"));
	&vmovdqu(&DWP(2016,"esp"), "ymm7");
	&vmovdqu(&DWP(2080,"esp"), "ymm6");
	&vpxor("ymm5", "ymm5", "ymm7");
	&vpxor("ymm6", "ymm6", &DWP(1600,"esp"));
	&vmovdqu("ymm7", &DWP(1760,"esp"));
	&vpshufb("ymm4", "ymm4", "ymm3");
	&vpshufb("ymm5", "ymm5", "ymm3");
	&vpshufb("ymm6", "ymm6", "ymm3");
	&vpaddd("ymm7", "ymm7", &DWP(1888,"esp"));
	&vpaddd("ymm2", "ymm2", "ymm4");
	&vpaddd("ymm1", "ymm1", "ymm5");
	&vmovdqu(&DWP(2112,"esp"), "ymm6");
	&vmovdqu(&DWP(2048,"esp"), "ymm7");
	&vmovdqu(&DWP(2144,"esp"), "ymm2");
	&vmovdqu(&DWP(2176,"esp"), "ymm1");
	&vpxor("ymm7", "ymm7", &DWP(1632,"esp"));
	&vpxor("ymm0", "ymm0", "ymm2");
	&vpxor("ymm1", "ymm1", &DWP(1792,"esp"));
	&vpshufb("ymm7", "ymm7", "ymm3");
	&vpsrld("ymm2", "ymm0", 20);
	&vpslld("ymm3", "ymm0", 12);
	&vpor("ymm0", "ymm2", "ymm3");
	&vpsrld("ymm2", "ymm1", 20);
	&vpslld("ymm1", "ymm1", 12);
	&vpaddd("ymm3", "ymm6", &DWP(1696,"esp"));
	&vmovdqu(&DWP(2208,"esp"), "ymm0");
	&vpor("ymm2", "ymm2", "ymm1");
	&vpaddd("ymm1", "ymm7", &DWP(1728,"esp"));
	&vmovdqu(&DWP(2272,"esp"), "ymm3");
	&vpxor("ymm6", "ymm1", &DWP(1760,"esp"));
	&vmovdqu(&DWP(2240,"esp"), "ymm1");
	&vpxor("ymm1", "ymm3", &DWP(1952,"esp"));
	&vpsrld("ymm3", "ymm6", 20);
	&vpslld("ymm6", "ymm6", 12);
	&vpor("ymm3", "ymm3", "ymm6");
	&vpsrld("ymm6", "ymm1", 20);
	&vpslld("ymm1", "ymm1", 12);
	&vmovdqu(&DWP(2304,"esp"), "ymm3");
	&vpor("ymm1", "ymm6", "ymm1");
	&vpaddd("ymm6", "ymm0", &DWP(1984,"esp"));
	&vpaddd("ymm0", "ymm2", &DWP(2016,"esp"));
	&vmovdqu(&DWP(2336,"esp"), "ymm6");
	&vmovdqu(&DWP(2368,"esp"), "ymm0");
	&vpxor("ymm6", "ymm4", "ymm6");
	&vpxor("ymm5", "ymm5", "ymm0");
	&vpaddd("ymm0", "ymm3", &DWP(2048,"esp"));
	&vmovdqu("ymm4", &DWP(1536,"esp"));
	&vpxor("ymm7", "ymm7", "ymm0");
	&vmovdqu(&DWP(2464,"esp"), "ymm0");
	&vpshufb("ymm6", "ymm6", "ymm4");
	&vpshufb("ymm5", "ymm5", "ymm4");
	&vpshufb("ymm0", "ymm7", "ymm4");
	&vmovdqu(&DWP(2400,"esp"), "ymm6");
	&vmovdqu(&DWP(2432,"esp"), "ymm5");
	&vmovdqu(&DWP(2528,"esp"), "ymm0");
	&vpaddd("ymm6", "ymm1", &DWP(2080,"esp"));
	&vpaddd("ymm5", "ymm5", &DWP(2176,"esp"));
	&vpaddd("ymm0", "ymm0", &DWP(2240,"esp"));
	&vmovdqu(&DWP(2496,"esp"), "ymm6");
	&vmovdqu(&DWP(2592,"esp"), "ymm5");
	&vpxor("ymm6", "ymm6", &DWP(2112,"esp"));
	&vpxor("ymm5", "ymm2", "ymm5");
	&vpshufb("ymm6", "ymm6", "ymm4");
	&vmovdqu("ymm4", &DWP(2400,"esp"));
	&vpaddd("ymm7", "ymm4", &DWP(2144,"esp"));
	&vpxor("ymm3", "ymm7", &DWP(2208,"esp"));
	&vmovdqu(&DWP(2560,"esp"), "ymm7");
	&vpsrld("ymm2", "ymm3", 25);
	&vpslld("ymm4", "ymm3", 7);
	&vpor("ymm7", "ymm2", "ymm4");
	&vpsrld("ymm2", "ymm5", 25);
	&vpslld("ymm5", "ymm5", 7);
	&vpaddd("ymm4", "ymm6", &DWP(2272,"esp"));
	&vmovdqu(&DWP(2624,"esp"), "ymm7");
	&vpor("ymm3", "ymm2", "ymm5");
	&vpxor("ymm2", "ymm0", &DWP(2304,"esp"));
	&vpxor("ymm1", "ymm1", "ymm4");
	&vmovdqu(&DWP(2688,"esp"), "ymm4");
	&vmovdqu(&DWP(2656,"esp"), "ymm3");
	&vpsrld("ymm5", "ymm2", 25);
	&vpslld("ymm2", "ymm2", 7);
	&vpsrld("ymm4", "ymm1", 25);
	&vpslld("ymm1", "ymm1", 7);
	&vpor("ymm2", "ymm5", "ymm2");
	&vpor("ymm5", "ymm4", "ymm1");
	&vpaddd("ymm1", "ymm3", &DWP(2336,"esp"));
	&vpaddd("ymm3", "ymm2", &DWP(2368,"esp"));
	&vmovdqu(&DWP(2720,"esp"), "ymm5");
	&vmovdqu(&DWP(2752,"esp"), "ymm1");
	&vmovdqu(&DWP(2784,"esp"), "ymm3");
	&vpxor("ymm4", "ymm6", "ymm1");
	&vpxor("ymm3", "ymm3", &DWP(2400,"esp"));
	&vmovdqu("ymm6", &DWP(1568,"esp"));
	&vpshufb("ymm1", "ymm4", "ymm6");
	&vpshufb("ymm3", "ymm3", "ymm6");
	&vpaddd("ymm4", "ymm5", &DWP(2464,"esp"));
	&vpaddd("ymm5", "ymm7", &DWP(2496,"esp"));
	&vpaddd("ymm0", "ymm0", "ymm1");
	&vpxor("ymm7", "ymm4", &DWP(2432,"esp"));
	&vmovdqu(&DWP(2848,"esp"), "ymm5");
	&vmovdqu(&DWP(2912,"esp"), "ymm0");
	&vmovdqu(&DWP(2816,"esp"), "ymm4");
	&vpxor("ymm5", "ymm5", &DWP(2528,"esp"));
	&vpxor("ymm0", "ymm0", &DWP(2656,"esp"));
	&vpshufb("ymm7", "ymm7", "ymm6");
	&vpshufb("ymm4", "ymm5", "ymm6");
	&vpaddd("ymm6", "ymm3", &DWP(2688,"esp"));
	&vpsrld("ymm5", "ymm0", 20);
	&vmovdqu(&DWP(2880,"esp"), "ymm4");
	&vpxor("ymm2", "ymm2", "ymm6");
	&vmovdqu(&DWP(2944,"esp"), "ymm6");
	&vpslld("ymm6", "ymm0", 12);
	&vpor("ymm0", "ymm5", "ymm6");
	&vpsrld("ymm5", "ymm2", 20);
	&vpslld("ymm2", "ymm2", 12);
	&vpaddd("ymm6", "ymm7", &DWP(2560,"esp"));
	&vmovdqu(&DWP(2976,"esp"), "ymm0");
	&vpor("ymm2", "ymm5", "ymm2");
	&vpaddd("ymm5", "ymm4", &DWP(2592,"esp"));
	&vpxor("ymm4", "ymm6", &DWP(2720,"esp"));
	&vpaddd("ymm0", "ymm0", &DWP(2752,"esp"));
	&vmovdqu(&DWP(3008,"esp"), "ymm6");
	&vmovdqu(&DWP(3040,"esp"), "ymm5");
	&vpxor("ymm6", "ymm5", &DWP(2624,"esp"));
	&vpsrld("ymm5", "ymm4", 20);
	&vpslld("ymm4", "ymm4", 12);
	&vmovdqu(&DWP(1824,"esp"), "ymm0");
	&vpxor("ymm0", "ymm1", "ymm0");
	&vpor("ymm5", "ymm5", "ymm4");
	&vpsrld("ymm4", "ymm6", 20);
	&vpslld("ymm6", "ymm6", 12);
	&vmovdqu("ymm1", &DWP(1536,"esp"));
	&vmovdqu(&DWP(3072,"esp"), "ymm5");
	&vpor("ymm6", "ymm4", "ymm6");
	&vpaddd("ymm4", "ymm2", &DWP(2784,"esp"));
	&vpaddd("ymm5", "ymm5", &DWP(2816,"esp"));
	&vpshufb("ymm0", "ymm0", "ymm1");
	&vpxor("ymm3", "ymm3", "ymm4");
	&vpxor("ymm7", "ymm7", "ymm5");
	&vmovdqu(&DWP(1856,"esp"), "ymm4");
	&vmovdqu(&DWP(1600,"esp"), "ymm0");
	&vmovdqu(&DWP(1888,"esp"), "ymm5");
	&vpaddd("ymm4", "ymm6", &DWP(2848,"esp"));
	&vpaddd("ymm0", "ymm0", &DWP(2912,"esp"));
	&vpshufb("ymm3", "ymm3", "ymm1");
	&vpshufb("ymm5", "ymm7", "ymm1");
	&vpxor("ymm7", "ymm4", &DWP(2880,"esp"));
	&vmovdqu(&DWP(1664,"esp"), "ymm3");
	&vmovdqu(&DWP(1920,"esp"), "ymm4");
	&vmovdqu(&DWP(1728,"esp"), "ymm0");
	&vpaddd("ymm3", "ymm3", &DWP(2944,"esp"));
	&vpxor("ymm4", "ymm0", &DWP(2976,"esp"));
	&vpshufb("ymm1", "ymm7", "ymm1");
	&vpxor("ymm2", "ymm2", "ymm3");
	&vpsrld("ymm7", "ymm4", 25);
	&vpslld("ymm0", "ymm4", 7);
	&vmovdqu(&DWP(1632,"esp"), "ymm1");
	&vmovdqu(&DWP(1696,"esp"), "ymm3");
	&vpsrld("ymm4", "ymm2", 25);
	&vpslld("ymm2", "ymm2", 7);
	&vpor("ymm3", "ymm7", "ymm0");
	&vpaddd("ymm1", "ymm1", &DWP(3040,"esp"));
	&vpor("ymm2", "ymm4", "ymm2");
	&vmovdqu(&DWP(1792,"esp"), "ymm3");
	&vpxor("ymm7", "ymm6", "ymm1");
	&vmovdqu(&DWP(1760,"esp"), "ymm2");
	&vpaddd("ymm2", "ymm5", &DWP(3008,"esp"));
	&vpsrld("ymm3", "ymm7", 25);
	&vpslld("ymm4", "ymm7", 7);
	&vpxor("ymm0", "ymm2", &DWP(3072,"esp"));
	&vpsrld("ymm6", "ymm0", 25);
	&vpslld("ymm0", "ymm0", 7);
	&vpor("ymm6", "ymm6", "ymm0");
	&vpor("ymm0", "ymm3", "ymm4");
	&add("edx", -2);
	&jne(&label("chacha_blocks_avx2_5"));
	&set_label("chacha_blocks_avx2_6");
	&vmovdqu(&DWP(1248,"esp"), "ymm2");
	&vmovdqu("ymm2", &DWP(1824,"esp"));
	&vmovdqu("ymm4", &DWP(1856,"esp"));
	&vmovdqu(&DWP(1184,"esp"), "ymm5");
	&vmovdqu(&DWP(1216,"esp"), "ymm1");
	&vmovdqu("ymm5", &DWP(1888,"esp"));
	&vmovdqu("ymm1", &DWP(1920,"esp"));
	&vmovdqu("ymm7", &DWP(1792,"esp"));
	&vpaddd("ymm3", "ymm2", &DWP(896,"esp"));
	&vpaddd("ymm2", "ymm4", &DWP(928,"esp"));
	&vpaddd("ymm6", "ymm6", &DWP(1088,"esp"));
	&vpaddd("ymm4", "ymm5", &DWP(960,"esp"));
	&vpaddd("ymm5", "ymm1", &DWP(992,"esp"));
	&vpaddd("ymm1", "ymm7", &DWP(1056,"esp"));
	&vpaddd("ymm0", "ymm0", &DWP(1024,"esp"));
	&vmovdqu(&DWP(1312,"esp"), "ymm6");
	&vmovdqu("ymm7", &DWP(1760,"esp"));
	&vpunpckldq("ymm6", "ymm3", "ymm2");
	&vpunpckhdq("ymm3", "ymm3", "ymm2");
	&vmovdqu("ymm2", &DWP(1312,"esp"));
	&vpaddd("ymm7", "ymm7", &DWP(864,"esp"));
	&vmovdqu(&DWP(1344,"esp"), "ymm6");
	&vpunpckldq("ymm6", "ymm4", "ymm5");
	&vpunpckhdq("ymm4", "ymm4", "ymm5");
	&vpunpckldq("ymm5", "ymm0", "ymm1");
	&vpunpckhdq("ymm0", "ymm0", "ymm1");
	&vpunpckhdq("ymm1", "ymm7", "ymm2");
	&vmovdqu(&DWP(1376,"esp"), "ymm4");
	&vpunpckldq("ymm4", "ymm7", "ymm2");
	&vmovdqu("ymm2", &DWP(1344,"esp"));
	&vpunpcklqdq("ymm7", "ymm2", "ymm6");
	&vpunpckhqdq("ymm6", "ymm2", "ymm6");
	&vmovdqu("ymm2", &DWP(1376,"esp"));
	&vmovdqu(&DWP(1408,"esp"), "ymm7");
	&vpunpcklqdq("ymm7", "ymm5", "ymm4");
	&vpunpckhqdq("ymm4", "ymm5", "ymm4");
	&vpunpcklqdq("ymm5", "ymm3", "ymm2");
	&vpunpckhqdq("ymm3", "ymm3", "ymm2");
	&vmovdqu(&DWP(1440,"esp"), "ymm5");
	&vpunpcklqdq("ymm5", "ymm0", "ymm1");
	&vpunpckhqdq("ymm1", "ymm0", "ymm1");
	&vmovdqu("ymm0", &DWP(1408,"esp"));
	&vperm2i128("ymm2", "ymm0", "ymm7", 32);
	&vmovdqu(&DWP(1280,"esp"), "ymm2");
	&vperm2i128("ymm2", "ymm6", "ymm4", 32);
	&vperm2i128("ymm0", "ymm0", "ymm7", 49);
	&vperm2i128("ymm7", "ymm6", "ymm4", 49);
	&vmovdqu("ymm4", &DWP(1440,"esp"));
	&vperm2i128("ymm6", "ymm4", "ymm5", 32);
	&vmovdqu(&DWP(1472,"esp"), "ymm6");
	&vperm2i128("ymm6", "ymm3", "ymm1", 32);
	&vperm2i128("ymm4", "ymm4", "ymm5", 49);
	&vperm2i128("ymm1", "ymm3", "ymm1", 49);
	&vmovdqu("ymm5", &DWP(1472,"esp"));
	&test("edi", "edi");
	&je(&label("chacha_blocks_avx2_51"));
	&set_label("chacha_blocks_avx2_7");
	&vmovdqu("ymm3", &DWP(1280,"esp"));
	&vpxor("ymm1", "ymm1", &DWP(448,"edi"));
	&vpxor("ymm4", "ymm4", &DWP(384,"edi"));
	&vpxor("ymm7", "ymm7", &DWP(320,"edi"));
	&vpxor("ymm0", "ymm0", &DWP(256,"edi"));
	&vpxor("ymm6", "ymm6", &DWP(192,"edi"));
	&vpxor("ymm5", "ymm5", &DWP(128,"edi"));
	&vpxor("ymm2", "ymm2", &DWP(64,"edi"));
	&vpxor("ymm3", "ymm3", &DWP(0,"edi"));
	&vmovdqu(&DWP(448,"ebx"), "ymm1");
	&vmovdqu("ymm1", &DWP(1184,"esp"));
	&vmovdqu(&DWP(64,"ebx"), "ymm2");
	&vmovdqu(&DWP(0,"ebx"), "ymm3");
	&vmovdqu(&DWP(256,"ebx"), "ymm0");
	&vmovdqu("ymm0", &DWP(1248,"esp"));
	&vmovdqu("ymm2", &DWP(1216,"esp"));
	&vmovdqu(&DWP(128,"ebx"), "ymm5");
	&vmovdqu(&DWP(320,"ebx"), "ymm7");
	&vmovdqu(&DWP(192,"ebx"), "ymm6");
	&vmovdqu("ymm6", &DWP(1728,"esp"));
	&vmovdqu(&DWP(384,"ebx"), "ymm4");
	&vpaddd("ymm3", "ymm1", &DWP(1120,"esp"));
	&vpaddd("ymm5", "ymm0", &DWP(832,"esp"));
	&vpaddd("ymm7", "ymm2", &DWP(800,"esp"));
	&vpaddd("ymm6", "ymm6", &DWP(768,"esp"));
	&vmovdqu("ymm1", &DWP(1600,"esp"));
	&vmovdqu("ymm0", &DWP(1696,"esp"));
	&vmovdqu("ymm2", &DWP(1664,"esp"));
	&vpaddd("ymm1", "ymm1", &DWP(672,"esp"));
	&vpaddd("ymm4", "ymm0", &DWP(736,"esp"));
	&vpaddd("ymm0", "ymm2", &DWP(1152,"esp"));
	&vmovdqu("ymm2", &DWP(1632,"esp"));
	&vmovdqu(&DWP(224,"esp"), "ymm1");
	&vpunpckldq("ymm1", "ymm5", "ymm7");
	&vpunpckhdq("ymm7", "ymm5", "ymm7");
	&vmovdqu("ymm5", &DWP(224,"esp"));
	&vpaddd("ymm2", "ymm2", &DWP(704,"esp"));
	&vmovdqu(&DWP(320,"esp"), "ymm1");
	&vpunpckldq("ymm1", "ymm6", "ymm4");
	&vpunpckhdq("ymm4", "ymm6", "ymm4");
	&vpunpckldq("ymm6", "ymm2", "ymm5");
	&vmovdqu(&DWP(352,"esp"), "ymm4");
	&vpunpckldq("ymm4", "ymm0", "ymm3");
	&vpunpckhdq("ymm0", "ymm0", "ymm3");
	&vpunpckhdq("ymm3", "ymm2", "ymm5");
	&vmovdqu("ymm5", &DWP(320,"esp"));
	&vpunpcklqdq("ymm2", "ymm5", "ymm1");
	&vpunpckhqdq("ymm1", "ymm5", "ymm1");
	&vmovdqu("ymm5", &DWP(352,"esp"));
	&vmovdqu(&DWP(384,"esp"), "ymm2");
	&vpunpcklqdq("ymm2", "ymm4", "ymm6");
	&vpunpckhqdq("ymm4", "ymm4", "ymm6");
	&vpunpcklqdq("ymm6", "ymm7", "ymm5");
	&vpunpckhqdq("ymm7", "ymm7", "ymm5");
	&vpunpckhqdq("ymm5", "ymm0", "ymm3");
	&vmovdqu(&DWP(416,"esp"), "ymm6");
	&vpunpcklqdq("ymm6", "ymm0", "ymm3");
	&vmovdqu("ymm3", &DWP(384,"esp"));
	&vperm2i128("ymm0", "ymm3", "ymm2", 32);
	&vpxor("ymm0", "ymm0", &DWP(32,"edi"));
	&vmovdqu(&DWP(448,"esp"), "ymm0");
	&vperm2i128("ymm0", "ymm1", "ymm4", 32);
	&vpxor("ymm0", "ymm0", &DWP(96,"edi"));
	&vmovdqu(&DWP(480,"esp"), "ymm0");
	&vperm2i128("ymm0", "ymm6", &DWP(416,"esp"), 2);
	&vpxor("ymm0", "ymm0", &DWP(160,"edi"));
	&vmovdqu(&DWP(512,"esp"), "ymm0");
	&vperm2i128("ymm0", "ymm7", "ymm5", 32);
	&vperm2i128("ymm5", "ymm7", "ymm5", 49);
	&vmovdqu("ymm7", &DWP(448,"esp"));
	&vpxor("ymm0", "ymm0", &DWP(224,"edi"));
	&vpxor("ymm5", "ymm5", &DWP(480,"edi"));
	&vperm2i128("ymm3", "ymm3", "ymm2", 49);
	&vperm2i128("ymm4", "ymm1", "ymm4", 49);
	&vperm2i128("ymm1", "ymm6", &DWP(416,"esp"), 19);
	&vpxor("ymm2", "ymm3", &DWP(288,"edi"));
	&vpxor("ymm4", "ymm4", &DWP(352,"edi"));
	&vpxor("ymm1", "ymm1", &DWP(416,"edi"));
	&vmovdqu(&DWP(32,"ebx"), "ymm7");
	&vmovdqu("ymm7", &DWP(480,"esp"));
	&vmovdqu("ymm3", &DWP(512,"esp"));
	&vmovdqu(&DWP(224,"ebx"), "ymm0");
	&vmovdqu(&DWP(288,"ebx"), "ymm2");
	&vmovdqu(&DWP(352,"ebx"), "ymm4");
	&vmovdqu(&DWP(96,"ebx"), "ymm7");
	&vmovdqu(&DWP(160,"ebx"), "ymm3");
	&vmovdqu(&DWP(416,"ebx"), "ymm1");
	&vmovdqu(&DWP(480,"ebx"), "ymm5");
	&add("edi", 512);
	&set_label("chacha_blocks_avx2_8");
	&mov("edx", &DWP(1512,"esp"));
	&add("ebx", 512);
	&add("edx", -512);
	&mov(&DWP(1512,"esp"), "edx");
	&cmp("edx", 512);
	&jae(&label("chacha_blocks_avx2_4"));
	&set_label("chacha_blocks_avx2_9");
	&mov("esi", &DWP(1504,"esp"));
	&vmovdqu("xmm0", &DWP(304,"esp"));
	&set_label("chacha_blocks_avx2_10");
	&cmp("edx", 256);
	&jb(&label("chacha_blocks_avx2_18"));
	&set_label("chacha_blocks_avx2_11");
	&mov("eax", 2);
	&vpxor("ymm3", "ymm3", "ymm3");
	&vinserti128("ymm5", "ymm0", "xmm0", 1);
	&vinserti128("ymm7", "ymm3", &DWP(592,"esp"), 1);
	&vpaddq("ymm0", "ymm5", "ymm7");
	&vmovd("xmm7", "eax");
	&vmovdqu(&DWP(0,"esp"), "ymm0");
	&vbroadcasti128("ymm1", &DWP(560,"esp"));
	&vbroadcasti128("ymm6", &DWP(576,"esp"));
	&vinserti128("ymm5", "ymm7", "xmm7", 1);
	&vmovdqu(&DWP(224,"esp"), "ymm1");
	&vmovdqu(&DWP(192,"esp"), "ymm1");
	&vmovdqu(&DWP(160,"esp"), "ymm6");
	&vmovdqu(&DWP(32,"esp"), "ymm1");
	&vmovdqu(&DWP(128,"esp"), "ymm6");
	&vmovdqu("ymm1", &DWP(224,"esp"));
	&vpaddq("ymm7", "ymm0", "ymm5");
	&vbroadcasti128("ymm2", &DWP(544,"esp"));
	&vmovdqa("ymm4", "ymm2");
	&vmovdqu(&DWP(320,"esp"), "ymm2");
	&vmovdqu(&DWP(64,"esp"), "ymm2");
	&vmovdqu(&DWP(96,"esp"), "ymm7");
	&vmovdqu(&DWP(352,"esp"), "ymm7");
	&vmovdqu(&DWP(384,"esp"), "ymm4");
	&vmovdqu("ymm2", &DWP(192,"esp"));
	&mov("eax", &DWP(1516,"esp"));
	&vmovdqa("ymm3", "ymm0");
	&vmovdqu("ymm0", &DWP(160,"esp"));
	&set_label("chacha_blocks_avx2_12");
	&vpaddd("ymm4", "ymm1", &DWP(384,"esp"));
	&vpaddd("ymm7", "ymm2", &DWP(320,"esp"));
	&vpxor("ymm5", "ymm3", "ymm4");
	&vmovdqu("ymm3", &DWP(1568,"esp"));
	&vmovdqu(&DWP(416,"esp"), "ymm7");
	&vpxor("ymm7", "ymm7", &DWP(352,"esp"));
	&vpshufb("ymm5", "ymm5", "ymm3");
	&vpshufb("ymm3", "ymm7", "ymm3");
	&vpaddd("ymm6", "ymm6", "ymm5");
	&vpaddd("ymm0", "ymm0", "ymm3");
	&vpxor("ymm1", "ymm1", "ymm6");
	&vpxor("ymm7", "ymm2", "ymm0");
	&vpsrld("ymm2", "ymm1", 20);
	&vpslld("ymm1", "ymm1", 12);
	&vpor("ymm1", "ymm2", "ymm1");
	&vpsrld("ymm2", "ymm7", 20);
	&vpslld("ymm7", "ymm7", 12);
	&vpaddd("ymm4", "ymm4", "ymm1");
	&vpor("ymm2", "ymm2", "ymm7");
	&vmovdqu(&DWP(448,"esp"), "ymm4");
	&vpxor("ymm5", "ymm5", "ymm4");
	&vpaddd("ymm7", "ymm2", &DWP(416,"esp"));
	&vmovdqu("ymm4", &DWP(1536,"esp"));
	&vpxor("ymm3", "ymm3", "ymm7");
	&vpshufd("ymm7", "ymm7", 147);
	&vpshufb("ymm5", "ymm5", "ymm4");
	&vpshufb("ymm4", "ymm3", "ymm4");
	&vpaddd("ymm6", "ymm6", "ymm5");
	&vpaddd("ymm3", "ymm0", "ymm4");
	&vpshufd("ymm5", "ymm5", 78);
	&vpshufd("ymm4", "ymm4", 78);
	&vpxor("ymm0", "ymm1", "ymm6");
	&vpxor("ymm2", "ymm2", "ymm3");
	&vpshufd("ymm6", "ymm6", 57);
	&vpshufd("ymm3", "ymm3", 57);
	&vpsrld("ymm1", "ymm0", 25);
	&vpslld("ymm0", "ymm0", 7);
	&vpor("ymm1", "ymm1", "ymm0");
	&vpsrld("ymm0", "ymm2", 25);
	&vpslld("ymm2", "ymm2", 7);
	&vpor("ymm0", "ymm0", "ymm2");
	&vpshufd("ymm2", &DWP(448,"esp"), 147);
	&vpaddd("ymm7", "ymm7", "ymm0");
	&vpaddd("ymm2", "ymm2", "ymm1");
	&vpxor("ymm4", "ymm4", "ymm7");
	&vmovdqu(&DWP(480,"esp"), "ymm2");
	&vpxor("ymm2", "ymm5", "ymm2");
	&vmovdqu("ymm5", &DWP(1568,"esp"));
	&vpshufb("ymm2", "ymm2", "ymm5");
	&vpshufb("ymm4", "ymm4", "ymm5");
	&vpaddd("ymm6", "ymm6", "ymm2");
	&vpaddd("ymm3", "ymm3", "ymm4");
	&vpxor("ymm5", "ymm1", "ymm6");
	&vpxor("ymm1", "ymm0", "ymm3");
	&vpsrld("ymm0", "ymm5", 20);
	&vpslld("ymm5", "ymm5", 12);
	&vpor("ymm0", "ymm0", "ymm5");
	&vpsrld("ymm5", "ymm1", 20);
	&vpslld("ymm1", "ymm1", 12);
	&vpor("ymm1", "ymm5", "ymm1");
	&vpaddd("ymm5", "ymm0", &DWP(480,"esp"));
	&vmovdqu(&DWP(512,"esp"), "ymm1");
	&vpaddd("ymm1", "ymm7", "ymm1");
	&vpxor("ymm7", "ymm2", "ymm5");
	&vmovdqu("ymm2", &DWP(1536,"esp"));
	&vpxor("ymm4", "ymm4", "ymm1");
	&vpshufb("ymm7", "ymm7", "ymm2");
	&vpshufb("ymm4", "ymm4", "ymm2");
	&vpshufd("ymm2", "ymm5", 57);
	&vpaddd("ymm6", "ymm6", "ymm7");
	&vpshufd("ymm5", "ymm1", 57);
	&vpaddd("ymm1", "ymm3", "ymm4");
	&vpshufd("ymm4", "ymm4", 78);
	&vpshufd("ymm3", "ymm7", 78);
	&vpxor("ymm7", "ymm0", "ymm6");
	&vpshufd("ymm0", "ymm1", 147);
	&vpshufd("ymm6", "ymm6", 147);
	&vmovdqu(&DWP(352,"esp"), "ymm4");
	&vmovdqu(&DWP(384,"esp"), "ymm2");
	&vmovdqu(&DWP(320,"esp"), "ymm5");
	&vpxor("ymm4", "ymm1", &DWP(512,"esp"));
	&vpsrld("ymm1", "ymm7", 25);
	&vpslld("ymm2", "ymm7", 7);
	&vpsrld("ymm5", "ymm4", 25);
	&vpor("ymm1", "ymm1", "ymm2");
	&vpslld("ymm2", "ymm4", 7);
	&vpor("ymm2", "ymm5", "ymm2");
	&add("eax", -2);
	&jne(&label("chacha_blocks_avx2_12"));
	&set_label("chacha_blocks_avx2_13");
	&vmovdqu(&DWP(160,"esp"), "ymm0");
	&vmovdqu(&DWP(192,"esp"), "ymm2");
	&vmovdqu(&DWP(224,"esp"), "ymm1");
	&vmovdqu("ymm0", &DWP(0,"esp"));
	&vmovdqu("ymm1", &DWP(32,"esp"));
	&vmovdqu("ymm4", &DWP(384,"esp"));
	&vmovdqu("ymm2", &DWP(64,"esp"));
	&vpaddd("ymm6", "ymm6", &DWP(128,"esp"));
	&vpaddd("ymm5", "ymm1", &DWP(224,"esp"));
	&vpaddd("ymm0", "ymm3", "ymm0");
	&vpaddd("ymm4", "ymm4", "ymm2");
	&vperm2i128("ymm7", "ymm4", "ymm5", 32);
	&vperm2i128("ymm3", "ymm6", "ymm0", 32);
	&vperm2i128("ymm4", "ymm4", "ymm5", 49);
	&vperm2i128("ymm0", "ymm6", "ymm0", 49);
	&test("edi", "edi");
	&je(&label("chacha_blocks_avx2_15"));
	&set_label("chacha_blocks_avx2_14");
	&vpxor("ymm7", "ymm7", &DWP(0,"edi"));
	&vpxor("ymm3", "ymm3", &DWP(32,"edi"));
	&vpxor("ymm4", "ymm4", &DWP(64,"edi"));
	&vpxor("ymm0", "ymm0", &DWP(96,"edi"));
	&set_label("chacha_blocks_avx2_15");
	&vmovdqu(&DWP(32,"ebx"), "ymm3");
	&vmovdqu(&DWP(64,"ebx"), "ymm4");
	&vmovdqu(&DWP(96,"ebx"), "ymm0");
	&vmovdqu(&DWP(0,"ebx"), "ymm7");
	&vpaddd("ymm0", "ymm2", &DWP(320,"esp"));
	&vpaddd("ymm3", "ymm1", &DWP(192,"esp"));
	&vmovdqu("ymm1", &DWP(128,"esp"));
	&vmovdqu("ymm2", &DWP(352,"esp"));
	&vpaddd("ymm4", "ymm1", &DWP(160,"esp"));
	&vpaddd("ymm5", "ymm2", &DWP(96,"esp"));
	&vperm2i128("ymm1", "ymm0", "ymm3", 32);
	&vperm2i128("ymm2", "ymm4", "ymm5", 32);
	&vperm2i128("ymm0", "ymm0", "ymm3", 49);
	&vperm2i128("ymm3", "ymm4", "ymm5", 49);
	&je(&label("chacha_blocks_avx2_17"));
	&set_label("chacha_blocks_avx2_16");
	&vpxor("ymm1", "ymm1", &DWP(128,"edi"));
	&vpxor("ymm2", "ymm2", &DWP(160,"edi"));
	&vpxor("ymm0", "ymm0", &DWP(192,"edi"));
	&vpxor("ymm3", "ymm3", &DWP(224,"edi"));
	&add("edi", 256);
	&set_label("chacha_blocks_avx2_17");
	&add(&DWP(304,"esp"), 4);
	&vmovdqu(&DWP(192,"ebx"), "ymm0");
	&vmovdqu(&DWP(128,"ebx"), "ymm1");
	&vmovdqu(&DWP(160,"ebx"), "ymm2");
	&vmovdqu(&DWP(224,"ebx"), "ymm3");
	&adc(&DWP(308,"esp"), 0);
	&add("edx", -256);
	&vmovdqu("xmm0", &DWP(304,"esp"));
	&add("ebx", 256);
	&set_label("chacha_blocks_avx2_18");
	&cmp("edx", 128);
	&jb(&label("chacha_blocks_avx2_24"));
	&set_label("chacha_blocks_avx2_19");
	&vpxor("ymm7", "ymm7", "ymm7");
	&vinserti128("ymm0", "ymm0", "xmm0", 1);
	&vinserti128("ymm7", "ymm7", &DWP(592,"esp"), 1);
	&vpaddq("ymm7", "ymm0", "ymm7");
	&vbroadcasti128("ymm3", &DWP(560,"esp"));
	&vbroadcasti128("ymm2", &DWP(576,"esp"));
	&vbroadcasti128("ymm6", &DWP(544,"esp"));
	&vmovdqu(&DWP(32,"esp"), "ymm2");
	&vmovdqu(&DWP(64,"esp"), "ymm3");
	&vmovdqu(&DWP(0,"esp"), "ymm7");
	&vmovdqu(&DWP(96,"esp"), "ymm6");
	&vmovdqa("ymm1", "ymm6");
	&vmovdqa("ymm5", "ymm3");
	&vmovdqa("ymm4", "ymm2");
	&vmovdqa("ymm0", "ymm7");
	&mov("eax", &DWP(1516,"esp"));
	&vmovdqu("ymm2", &DWP(1536,"esp"));
	&vmovdqu("ymm3", &DWP(1568,"esp"));
	&set_label("chacha_blocks_avx2_20");
	&vpaddd("ymm6", "ymm1", "ymm5");
	&vpxor("ymm1", "ymm0", "ymm6");
	&vpshufb("ymm0", "ymm1", "ymm3");
	&vpaddd("ymm1", "ymm4", "ymm0");
	&vpxor("ymm4", "ymm5", "ymm1");
	&vpsrld("ymm5", "ymm4", 20);
	&vpslld("ymm7", "ymm4", 12);
	&vpor("ymm7", "ymm5", "ymm7");
	&vpaddd("ymm5", "ymm6", "ymm7");
	&vpxor("ymm0", "ymm0", "ymm5");
	&vpshufd("ymm5", "ymm5", 147);
	&vpshufb("ymm4", "ymm0", "ymm2");
	&vpaddd("ymm6", "ymm1", "ymm4");
	&vpshufd("ymm4", "ymm4", 78);
	&vpxor("ymm1", "ymm7", "ymm6");
	&vpshufd("ymm6", "ymm6", 57);
	&vpsrld("ymm0", "ymm1", 25);
	&vpslld("ymm7", "ymm1", 7);
	&vpor("ymm0", "ymm0", "ymm7");
	&vpaddd("ymm1", "ymm5", "ymm0");
	&vpxor("ymm4", "ymm4", "ymm1");
	&vpshufb("ymm7", "ymm4", "ymm3");
	&vpaddd("ymm5", "ymm6", "ymm7");
	&vpxor("ymm4", "ymm0", "ymm5");
	&vpsrld("ymm0", "ymm4", 20);
	&vpslld("ymm6", "ymm4", 12);
	&vpor("ymm4", "ymm0", "ymm6");
	&vpaddd("ymm0", "ymm1", "ymm4");
	&vpxor("ymm1", "ymm7", "ymm0");
	&vpshufb("ymm7", "ymm1", "ymm2");
	&vpshufd("ymm1", "ymm0", 57);
	&vpaddd("ymm5", "ymm5", "ymm7");
	&vpshufd("ymm0", "ymm7", 78);
	&vpxor("ymm6", "ymm4", "ymm5");
	&vpshufd("ymm4", "ymm5", 147);
	&vpsrld("ymm5", "ymm6", 25);
	&vpslld("ymm7", "ymm6", 7);
	&vpor("ymm5", "ymm5", "ymm7");
	&add("eax", -2);
	&jne(&label("chacha_blocks_avx2_20"));
	&set_label("chacha_blocks_avx2_21");
	&vmovdqu("ymm7", &DWP(0,"esp"));
	&vmovdqu("ymm2", &DWP(32,"esp"));
	&vmovdqu("ymm3", &DWP(64,"esp"));
	&vmovdqu("ymm6", &DWP(96,"esp"));
	&vpaddd("ymm0", "ymm0", "ymm7");
	&vpaddd("ymm2", "ymm4", "ymm2");
	&vpaddd("ymm3", "ymm5", "ymm3");
	&vpaddd("ymm1", "ymm1", "ymm6");
	&vperm2i128("ymm4", "ymm1", "ymm3", 32);
	&vperm2i128("ymm5", "ymm2", "ymm0", 32);
	&vperm2i128("ymm1", "ymm1", "ymm3", 49);
	&vperm2i128("ymm0", "ymm2", "ymm0", 49);
	&test("edi", "edi");
	&je(&label("chacha_blocks_avx2_23"));
	&set_label("chacha_blocks_avx2_22");
	&vpxor("ymm4", "ymm4", &DWP(0,"edi"));
	&vpxor("ymm5", "ymm5", &DWP(32,"edi"));
	&vpxor("ymm1", "ymm1", &DWP(64,"edi"));
	&vpxor("ymm0", "ymm0", &DWP(96,"edi"));
	&add("edi", 128);
	&set_label("chacha_blocks_avx2_23");
	&add(&DWP(304,"esp"), 2);
	&vmovdqu(&DWP(96,"ebx"), "ymm0");
	&vmovdqu(&DWP(0,"ebx"), "ymm4");
	&vmovdqu(&DWP(32,"ebx"), "ymm5");
	&vmovdqu(&DWP(64,"ebx"), "ymm1");
	&adc(&DWP(308,"esp"), 0);
	&add("edx", -128);
	&vmovdqu("xmm0", &DWP(304,"esp"));
	&add("ebx", 128);
	&set_label("chacha_blocks_avx2_24");
	&test("edx", "edx");
	&jne(&label("chacha_blocks_avx2_26"));
	&set_label("chacha_blocks_avx2_25");
	&mov("ecx", &DWP(8,"ebp"));
	&mov("eax", &DWP(304,"esp"));
	&mov("edx", &DWP(308,"esp"));
	&mov(&DWP(32,"ecx"), "eax");
	&mov(&DWP(36,"ecx"), "edx");
	&vzeroupper();
	&add("esp", 3124);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
	&set_label("chacha_blocks_avx2_26");
	&vmovdqu("ymm1", &DWP(1568,"esp"));
	&vmovdqu("ymm2", &DWP(1536,"esp"));
	&vmovdqu("xmm7", &DWP(576,"esp"));
	&vmovdqu("xmm6", &DWP(544,"esp"));
	&mov("esi", &DWP(304,"esp"));
	&mov("ecx", &DWP(308,"esp"));
	&vmovups(&DWP(16,"esp"), "xmm1");
	&vmovups(&DWP(0,"esp"), "xmm2");
	&vmovdqu("xmm1", &DWP(560,"esp"));
	&jmp(&label("chacha_blocks_avx2_27"));
	&set_label("chacha_blocks_avx2_43");
	&add("edx", -64);
	&add("ebx", 64);
	&set_label("chacha_blocks_avx2_27");
	&cmp("edx", 64);
	&jae(&label("chacha_blocks_avx2_38"));
	&set_label("chacha_blocks_avx2_28");
	&test("edi", "edi");
	&je(&label("chacha_blocks_avx2_37"));
	&set_label("chacha_blocks_avx2_29");
	&test("edx", "edx");
	&jbe(&label("chacha_blocks_avx2_36"));
	&set_label("chacha_blocks_avx2_30");
	&mov("eax", "edx");
	&shr("eax", 1);
	&mov(&DWP(60,"esp"), "eax");
	&test("eax", "eax");
	&jbe(&label("chacha_blocks_avx2_45"));
	&set_label("chacha_blocks_avx2_31");
	&mov(&DWP(52,"esp"), "esi");
	&xor("eax", "eax");
	&mov(&DWP(1512,"esp"), "edx");
	&mov(&DWP(56,"esp"), "ebx");
	&mov("esi", &DWP(60,"esp"));
	&set_label("chacha_blocks_avx2_32");
	&movzb("edx", &DWP(0,"edi","eax",2));
	&mov(&DWP(64,"esp","eax",2), "dl");
	&movzb("ebx", &DWP(1,"edi","eax",2));
	&mov(&DWP(65,"esp","eax",2), "bl");
	&inc("eax");
	&cmp("eax", "esi");
	&jb(&label("chacha_blocks_avx2_32"));
	&set_label("chacha_blocks_avx2_33");
	&mov("esi", &DWP(52,"esp"));
	&lea("eax", &DWP(1,"eax","eax"));
	&mov("edx", &DWP(1512,"esp"));
	&mov("ebx", &DWP(56,"esp"));
	&mov(&DWP(48,"esp"), "eax");
	&set_label("chacha_blocks_avx2_34");
	&lea("eax", &DWP(-1,"eax"));
	&cmp("eax", "edx");
	&jae(&label("chacha_blocks_avx2_36"));
	&set_label("chacha_blocks_avx2_35");
	&movzb("eax", &DWP(0,"eax","edi"));
	&mov("edi", &DWP(48,"esp"));
	&mov(&DWP(63,"esp","edi"), "al");
	&set_label("chacha_blocks_avx2_36");
	&lea("edi", &DWP(64,"esp"));
	&set_label("chacha_blocks_avx2_37");
	&mov(&DWP(1508,"esp"), "ebx");
	&lea("ebx", &DWP(64,"esp"));
	&set_label("chacha_blocks_avx2_38");
	&vmovdqu(&DWP(32,"esp"), "xmm0");
	&vmovdqa("xmm2", "xmm6");
	&mov("eax", &DWP(1516,"esp"));
	&vmovdqa("xmm3", "xmm1");
	&vmovups("xmm1", &DWP(16,"esp"));
	&vmovdqa("xmm4", "xmm7");
	&vmovdqa("xmm5", "xmm0");
	&vmovups("xmm0", &DWP(0,"esp"));
	&set_label("chacha_blocks_avx2_39");
	&vpaddd("xmm6", "xmm2", "xmm3");
	&vpxor("xmm2", "xmm5", "xmm6");
	&vpshufb("xmm5", "xmm2", "xmm1");
	&vpaddd("xmm2", "xmm4", "xmm5");
	&vpxor("xmm4", "xmm3", "xmm2");
	&vpsrld("xmm3", "xmm4", 20);
	&vpslld("xmm7", "xmm4", 12);
	&vpxor("xmm7", "xmm3", "xmm7");
	&vpaddd("xmm3", "xmm6", "xmm7");
	&vpxor("xmm5", "xmm5", "xmm3");
	&vpshufb("xmm4", "xmm5", "xmm0");
	&vpaddd("xmm6", "xmm2", "xmm4");
	&vpxor("xmm2", "xmm7", "xmm6");
	&vpsrld("xmm5", "xmm2", 25);
	&vpslld("xmm7", "xmm2", 7);
	&vpshufd("xmm3", "xmm3", 147);
	&vpxor("xmm5", "xmm5", "xmm7");
	&vpshufd("xmm4", "xmm4", 78);
	&vpaddd("xmm2", "xmm3", "xmm5");
	&vpxor("xmm4", "xmm4", "xmm2");
	&vpshufb("xmm7", "xmm4", "xmm1");
	&vpshufd("xmm6", "xmm6", 57);
	&vpaddd("xmm3", "xmm6", "xmm7");
	&vpxor("xmm4", "xmm5", "xmm3");
	&vpsrld("xmm5", "xmm4", 20);
	&vpslld("xmm6", "xmm4", 12);
	&vpxor("xmm4", "xmm5", "xmm6");
	&vpaddd("xmm5", "xmm2", "xmm4");
	&vpxor("xmm2", "xmm7", "xmm5");
	&vpshufb("xmm7", "xmm2", "xmm0");
	&vpaddd("xmm3", "xmm3", "xmm7");
	&vpxor("xmm6", "xmm4", "xmm3");
	&vpshufd("xmm2", "xmm5", 57);
	&vpshufd("xmm5", "xmm7", 78);
	&vpslld("xmm7", "xmm6", 7);
	&vpshufd("xmm4", "xmm3", 147);
	&vpsrld("xmm3", "xmm6", 25);
	&vpxor("xmm3", "xmm3", "xmm7");
	&add("eax", -2);
	&jne(&label("chacha_blocks_avx2_39"));
	&set_label("chacha_blocks_avx2_40");
	&vmovdqu("xmm7", &DWP(576,"esp"));
	&vmovdqu("xmm1", &DWP(560,"esp"));
	&vpaddd("xmm4", "xmm4", "xmm7");
	&vmovdqu("xmm6", &DWP(544,"esp"));
	&vpaddd("xmm3", "xmm3", "xmm1");
	&vmovdqu("xmm0", &DWP(32,"esp"));
	&vpaddd("xmm2", "xmm2", "xmm6");
	&vpaddd("xmm5", "xmm5", "xmm0");
	&test("edi", "edi");
	&je(&label("chacha_blocks_avx2_42"));
	&set_label("chacha_blocks_avx2_41");
	&vpxor("xmm2", "xmm2", &DWP(0,"edi"));
	&vpxor("xmm3", "xmm3", &DWP(16,"edi"));
	&vpxor("xmm4", "xmm4", &DWP(32,"edi"));
	&vpxor("xmm5", "xmm5", &DWP(48,"edi"));
	&add("edi", 64);
	&set_label("chacha_blocks_avx2_42");
	&add("esi", 1);
	&vmovdqu(&DWP(0,"ebx"), "xmm2");
	&vmovdqu(&DWP(16,"ebx"), "xmm3");
	&vmovdqu(&DWP(32,"ebx"), "xmm4");
	&vmovdqu(&DWP(48,"ebx"), "xmm5");
	&vpaddq("xmm0", "xmm0", &DWP(592,"esp"));
	&adc("ecx", 0);
	&cmp("edx", 64);
	&jbe(&label("chacha_blocks_avx2_46"));
	&jmp(&label("chacha_blocks_avx2_43"));
	&set_label("chacha_blocks_avx2_44");
	&vzeroupper();
	&add("esp", 3124);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
	&set_label("chacha_blocks_avx2_45");
	&mov("eax", 1);
	&mov(&DWP(48,"esp"), "eax");
	&jmp(&label("chacha_blocks_avx2_34"));
	&set_label("chacha_blocks_avx2_46");
	&mov(&DWP(304,"esp"), "esi");
	&mov(&DWP(308,"esp"), "ecx");
	&jae(&label("chacha_blocks_avx2_25"));
	&set_label("chacha_blocks_avx2_47");
	&test("edx", "edx");
	&jbe(&label("chacha_blocks_avx2_25"));
	&set_label("chacha_blocks_avx2_48");
	&mov("esi", &DWP(1508,"esp"));
	&xor("ecx", "ecx");
	&set_label("chacha_blocks_avx2_49");
	&movzb("eax", &DWP(0,"ecx","ebx"));
	&mov(&DWP(0,"ecx","esi"), "al");
	&inc("ecx");
	&cmp("ecx", "edx");
	&jb(&label("chacha_blocks_avx2_49"));
	&jmp(&label("chacha_blocks_avx2_25"));
	&set_label("chacha_blocks_avx2_51");
	&vmovdqu(&DWP(256,"ebx"), "ymm0");
	&vmovdqu("ymm0", &DWP(1248,"esp"));
	&vmovdqu(&DWP(192,"ebx"), "ymm6");
	&vmovdqu(&DWP(320,"ebx"), "ymm7");
	&vmovdqu("ymm3", &DWP(1280,"esp"));
	&vmovdqu("ymm6", &DWP(1216,"esp"));
	&vmovdqu(&DWP(384,"ebx"), "ymm4");
	&vmovdqu(&DWP(448,"ebx"), "ymm1");
	&vmovdqu("ymm4", &DWP(1696,"esp"));
	&vmovdqu(&DWP(0,"ebx"), "ymm3");
	&vmovdqu(&DWP(64,"ebx"), "ymm2");
	&vmovdqu(&DWP(128,"ebx"), "ymm5");
	&vmovdqu("ymm5", &DWP(1728,"esp"));
	&vpaddd("ymm7", "ymm0", &DWP(832,"esp"));
	&vpaddd("ymm1", "ymm6", &DWP(800,"esp"));
	&vpaddd("ymm2", "ymm4", &DWP(736,"esp"));
	&vpaddd("ymm5", "ymm5", &DWP(768,"esp"));
	&vmovdqu("ymm0", &DWP(1664,"esp"));
	&vmovdqu("ymm6", &DWP(1184,"esp"));
	&vpaddd("ymm3", "ymm0", &DWP(1152,"esp"));
	&vpaddd("ymm4", "ymm6", &DWP(1120,"esp"));
	&vmovdqu("ymm0", &DWP(1632,"esp"));
	&vpaddd("ymm6", "ymm0", &DWP(704,"esp"));
	&vmovdqu("ymm0", &DWP(1600,"esp"));
	&vpaddd("ymm0", "ymm0", &DWP(672,"esp"));
	&vmovdqu(&DWP(0,"esp"), "ymm0");
	&vpunpckldq("ymm0", "ymm7", "ymm1");
	&vpunpckhdq("ymm1", "ymm7", "ymm1");
	&vmovdqu("ymm7", &DWP(0,"esp"));
	&vmovdqu(&DWP(32,"esp"), "ymm0");
	&vpunpckldq("ymm0", "ymm5", "ymm2");
	&vpunpckhdq("ymm2", "ymm5", "ymm2");
	&vpunpckldq("ymm5", "ymm6", "ymm7");
	&vpunpckhdq("ymm6", "ymm6", "ymm7");
	&vmovdqu(&DWP(64,"esp"), "ymm2");
	&vpunpckldq("ymm2", "ymm3", "ymm4");
	&vpunpckhdq("ymm4", "ymm3", "ymm4");
	&vmovdqu("ymm3", &DWP(32,"esp"));
	&vpunpcklqdq("ymm7", "ymm3", "ymm0");
	&vpunpckhqdq("ymm3", "ymm3", "ymm0");
	&vpunpckhqdq("ymm0", "ymm2", "ymm5");
	&vmovdqu(&DWP(96,"esp"), "ymm7");
	&vpunpcklqdq("ymm7", "ymm2", "ymm5");
	&vmovdqu("ymm5", &DWP(64,"esp"));
	&vpunpcklqdq("ymm2", "ymm1", "ymm5");
	&vpunpckhqdq("ymm1", "ymm1", "ymm5");
	&vmovdqu("ymm5", &DWP(96,"esp"));
	&vmovdqu(&DWP(128,"esp"), "ymm2");
	&vpunpcklqdq("ymm2", "ymm4", "ymm6");
	&vpunpckhqdq("ymm4", "ymm4", "ymm6");
	&vperm2i128("ymm6", "ymm5", "ymm7", 32);
	&vmovdqu(&DWP(160,"esp"), "ymm6");
	&vperm2i128("ymm6", "ymm3", "ymm0", 32);
	&vperm2i128("ymm3", "ymm3", "ymm0", 49);
	&vmovdqu("ymm0", &DWP(128,"esp"));
	&vmovdqu(&DWP(192,"esp"), "ymm3");
	&vmovdqu(&DWP(96,"ebx"), "ymm6");
	&vperm2i128("ymm7", "ymm5", "ymm7", 49);
	&vperm2i128("ymm3", "ymm0", "ymm2", 32);
	&vperm2i128("ymm5", "ymm1", "ymm4", 32);
	&vperm2i128("ymm0", "ymm0", "ymm2", 49);
	&vperm2i128("ymm2", "ymm1", "ymm4", 49);
	&vmovdqu("ymm1", &DWP(160,"esp"));
	&vmovdqu(&DWP(160,"ebx"), "ymm3");
	&vmovdqu(&DWP(224,"ebx"), "ymm5");
	&vmovdqu(&DWP(288,"ebx"), "ymm7");
	&vmovdqu(&DWP(416,"ebx"), "ymm0");
	&vmovdqu(&DWP(32,"ebx"), "ymm1");
	&vmovdqu("ymm1", &DWP(192,"esp"));
	&vmovdqu(&DWP(480,"ebx"), "ymm2");
	&vmovdqu(&DWP(352,"ebx"), "ymm1");
	&jmp(&label("chacha_blocks_avx2_8"));
}

add_function("chacha_blocks", "x86", \&chacha_blocks_x86);
add_function("chacha_blocks", "sse2", \&chacha_blocks_sse2);
add_function("chacha_blocks", "ssse3", \&chacha_blocks_ssse3);
add_function("chacha_blocks", "avx", \&chacha_blocks_avx);
add_function("chacha_blocks", "xop", \&chacha_blocks_xop);
add_function("chacha_blocks", "avx2", \&chacha_blocks_avx2);

} # $ext{'x86'}

&asm_finish();
