#!/usr/bin/env perl

#==============================================================================
#
#  Optimized Poly1305 routines for x86 by Andrew Moon
#  except for
#  x86 implementation is modified from the public domain implementation by 
#  D. J. Bernstein
#
#  x86: 16 byte block size
#  sse2: 32 byte block size, but can process 64 byte blocks
#  avx: 32 byte block size, but can process 64 byte blocks
#  avx2: 64 byte block size
#
#  License: Public Domain or MIT
#
#==============================================================================
#
# struct POLY1305_CTX {
#     unsigned char opaque_state[192];
# }
#
# void poly1305_init_ext(
#     POLY1305_CTX *ctx,
#     const unsigned char *key,
#     size_t bytes_hint
# )
#
# 'bytes_hint' should be 0, or the exact number of bytes that will be processed
#
#
# void poly1305_blocks_ext(
#     POLY1305_CTX *ctx,
#     const unsigned char *in,
#     size_t inl
# )
#
#
# void poly1305_finish_ext(
#     POLY1305_CTX *ctx,
#     const unsigned char *in,
#     size_t remaining,
#     unsigned char *mac
# )
#
# 'remaining' must be between 0 and (block size - 1), inclusive
#
#
# void poly1305_ext(
#     unsigned char *mac,
#     const unsigned char *in,
#     size_t inl,
#     const unsigned char *key
# )
#
#==============================================================================


$0 =~ m/(.*[\/\\])[^\/\\]+$/; 
my $dir = $1;
push(@INC,"${dir}","${dir}../../perlasm");
require "x86asm.pl";

&asm_init($ARGV[0],"poly1305_x86.pl",$ARGV[$#ARGV] eq "386");

# why isn't this already handled for implementors?
my %ext = (
	x86 => 0,
	sse2 => 0,
	sse3 => 0,
	ssse3 => 0,
	sse4_1 => 0,
	sse4_2 => 0,
	avx => 0,
	xop => 0,
	avx2 => 0,
	avx512 => 0,
);

if (exists $INC{'x86gas.pl'}) {
	if (`$ENV{CC} -Wa,-v -c -o /dev/null -x assembler /dev/null 2>&1` =~ /GNU assembler version ([2-9]\.[0-9]+)/) {
		$ext{'x86'} = 1;
		$ext{'sse2'} = ($1 >= 2.16);
		$ext{'sse3'} = ($1 >= 2.17);
		$ext{'ssse3'} = ($1 >= 2.18);
		$ext{'sse4_1'} = ($1 >= 2.18);
		$ext{'sse4_2'} = ($1 >= 2.18);
		$ext{'avx'} = ($1 >= 2.19);
		$ext{'xop'} = ($1 >= 2.21);
		$ext{'avx2'} = ($1 >= 2.22);
		$ext{'avx512'} = ($1 >= 2.24);
	}

	if ((exists $INC{'x86gas.pl'}) && `$ENV{CC} -v 2>&1` =~ /(^clang version|based on LLVM) ([3-9])\.([0-9]+)/) {
		my $ver = $2 + ($3 / 100.0); # 3.1->3.01, 3.10->3.10
		$ext{'x86'} = 1;
		$ext{'sse2'} = ($1 >= 2.01); # ?
		$ext{'sse3'} = ($1 >= 2.01); # ?
		$ext{'ssse3'} = ($1 >= 2.01);
		$ext{'sse4_1'} = ($1 >= 2.01); # ?
		$ext{'sse4_2'} = ($1 >= 2.01); # ?
		$ext{'avx'} = ($ver >= 3.0);
		$ext{'xop'} = ($ver >= 3.01);
		$ext{'avx2'} = ($ver >= 3.01);
		$ext{'avx512'} = ($ver >= 3.05); # ?
	}

	*{x86gas::opsize} = sub () {
		my $reg = shift;
    		if    ($reg =~ m/^%e/o)          { "l"; }
    		elsif ($reg =~ m/^%[a-d][hl]$/o) { "b"; }
    		elsif ($reg =~ m/^%[yxm]/o)      { undef; }
    		else                             { "w"; }
	}
}

if (exists $INC{'x86nasm.pl'}) {
	if (`nasm -v 2>&1` =~ /NASM version ([2-9]\.[0-9]+)/) {
		$ext{'x86'} = 1;
		$ext{'sse2'} = ($1 >= 2.00);
		$ext{'sse3'} = ($1 >= 2.01);
		$ext{'ssse3'} = ($1 >= 2.01);
		$ext{'sse4_1'} = ($1 >= 2.01);
		$ext{'sss4_2'} = ($1 >= 2.01);
		$ext{'avx'} = ($1 >= 2.09);
		$ext{'xop'} = ($1 >= 2.09); # ?
		$ext{'avx2'} = ($1 >= 2.10);
		$ext{'avx512'} = ($1 >= 2.12); # ?
	}
}

if (exists $INC{'x86masm.pl'}) {
	if (`ml64 2>&1` =~ /Version ([0-9]+)\./) {
		$ext{'x86'} = 1;
		$ext{'sse2'} = ($1 >= 7); # ?
		$ext{'sse3'} = ($1 >= 9); # ?
		$ext{'ssse3'} = ($1 >= 9);
		$ext{'sse4_1'} = ($1 >= 9);
		$ext{'sse4_2'} = ($1 >= 9);
		$ext{'avx'} = ($1 >= 10);
		$ext{'xop'} = ($1 >= 10);
		$ext{'avx2'} = ($1 >= 11);
		$ext{'avx512'} = ($1 >= 14); # ?
	}
}

my %last_valid_fallback = ();

sub add_function {
	my ($name, $flag, $fn) = @_;

	my $fullname = "$name\_$flag";
	&function_begin_B($fullname);
	if ($ext{$flag}) {
		$fn->();
		$last_valid_fallback{$name} = "$fullname\_local";
	} else {
		&jmp($last_valid_fallback{$name});
	}
	&function_end_B($fullname);
}

if ($ext{'x86'}) {

sub rep_movsb {
	&data_byte(0xf3,0xa4);
}

sub label_to_reg {
	my ($label, $reg) = @_;
	&call(&label("get_pic"));
	&set_label("get_pic");
	&pop($reg);
	&lea($reg,&DWP(&label($label)."-".&label("get_pic"),$reg));
}

&align(16);
&static_label("poly1305_constants_x86");
&set_label("poly1305_constants_x86");
&data_word(0x00000000,0x37f40000); # scale
&data_word(0x00000000,0x41f00000); # two32
&data_word(0x00000000,0x43f00000); # two64
&data_word(0x00000000,0x45f00000); # two96
&data_word(0x00000000,0x45e80000); # alpha32
&data_word(0x00000000,0x47e80000); # alpha64
&data_word(0x00000000,0x49e80000); # alpha96
&data_word(0x00000000,0x4c080000); # alpha130
&data_word(0x00000000,0x43300000); # doffset0
&data_word(0x00000000,0x45300000); # doffset1
&data_word(0x00000000,0x47300000); # doffset2
&data_word(0x00000000,0x49300000); # doffset3
&data_word(0x00000000,0x492ffffe); # doffset3minustwo128
&data_word(0xfffffffb,0x43300001); # hoffset0
&data_word(0xfffffffe,0x45300001); # doffset1
&data_word(0xfffffffe,0x47300001); # doffset2
&data_word(0xfffffffe,0x49300003); # doffset3

sub poly1305_init_ex_x86 {
	&static_label("poly1305_init_ex_x86_local");
	&set_label("poly1305_init_ex_x86_local");
	&push("ebx");
	&push("esi");
	&push("edi");
	&push("eax");
	&push("eax");
	&mov("eax", &DWP(24,"esp"));
	&mov("edx", &DWP(28,"esp"));
	&mov("ecx", 0x137f);
	&mov(&DWP(0,"esp"), "ecx");
	&fstcw(&DWP(4,"esp"));
	&fldcw(&DWP(0,"esp"));
	&mov("ecx", &DWP(16,"edx"));
	&mov("ebx", &DWP(20,"edx"));
	&mov("esi", &DWP(24,"edx"));
	&mov("edi", &DWP(28,"edx"));
	&mov(&DWP(104,"eax"), "ecx");
	&mov(&DWP(108,"eax"), "ebx");
	&mov(&DWP(112,"eax"), "esi");
	&mov(&DWP(116,"eax"), "edi");
	&mov("ecx", &DWP(0,"edx"));
	&mov("ebx", &DWP(4,"edx"));
	&mov("esi", &DWP(8,"edx"));
	&mov("edi", &DWP(12,"edx"));
	&and("ecx", 0x0fffffff);
	&and("ebx", 0x0ffffffc);
	&and("esi", 0x0ffffffc);
	&and("edi", 0x0ffffffc);
	&mov(&DWP(0,"eax"), "ecx");
	&mov(&DWP(4,"eax"), 0x43300000);
	&mov(&DWP(8,"eax"), "ebx");
	&mov(&DWP(12,"eax"), 0x45300000);
	&mov(&DWP(16,"eax"), "esi");
	&mov(&DWP(20,"eax"), 0x47300000);
	&mov(&DWP(24,"eax"), "edi");
	&mov(&DWP(28,"eax"), 0x49300000);
	&label_to_reg("poly1305_constants_x86", "edx");
	&fld(&FQWP(0,"eax"));
	&fsub(&FQWP(64,"edx"));
	&fld(&FQWP(8,"eax"));
	&fsub(&FQWP(72,"edx"));
	&fld(&FQWP(16,"eax"));
	&fsub(&FQWP(80,"edx"));
	&fld(&FQWP(24,"eax"));
	&fsub(&FQWP(88,"edx"));
	&fxch("st(3)");
	&fstp(&FQWP(0,"eax"));
	&fxch("st(1)");
	&fst(&FQWP(8,"eax"));
	&fmul(&FQWP(0,"edx"));
	&fstp(&FQWP(16,"eax"));
	&fst(&FQWP(24,"eax"));
	&fmul(&FQWP(0,"edx"));
	&fstp(&FQWP(32,"eax"));
	&fst(&FQWP(40,"eax"));
	&fmul(&FQWP(0,"edx"));
	&fstp(&FQWP(48,"eax"));
	&fldz();
	&fst(&FQWP(56,"eax"));
	&fst(&FQWP(64,"eax"));
	&fst(&FQWP(72,"eax"));
	&fst(&FQWP(80,"eax"));
	&fst(&FQWP(88,"eax"));
	&fstp(&FQWP(96,"eax"));
	&fldcw(&DWP(4,"esp"));
	&pop("eax");
	&pop("eax");
	&pop("edi");
	&pop("esi");
	&pop("ebx");
	&ret();
}


sub poly1305_blocks_x86 {
	&static_label("poly1305_blocks_x86_local");
	&set_label("poly1305_blocks_x86_local");
	&mov("eax", "esp");
	&and("eax", 63);
	&add("eax", 192);
	&sub("esp", "eax");
	&mov(&DWP(0,"esp"), "eax");
	&mov(&DWP(4,"esp"), "ebx");
	&mov(&DWP(8,"esp"), "esi");
	&mov(&DWP(12,"esp"), "edi");
	&mov(&DWP(16,"esp"), "ebp");
	&mov("ecx", 0x137f);
	&mov(&DWP(188,"esp"), "ecx");
	&fstcw(&DWP(184,"esp"));
	&fldcw(&DWP(188,"esp"));
	&mov(&DWP(100,"esp"), 0x43300000);
	&mov(&DWP(108,"esp"), 0x45300000);
	&mov(&DWP(116,"esp"), 0x47300000);
	&mov(&DWP(124,"esp"), 0x49300000);
	&mov("ebp", &DWP(4,"esp","eax"));
	&mov("esi", &DWP(8,"esp","eax"));
	&mov("ecx", &DWP(12,"esp","eax"));
	&label_to_reg("poly1305_constants_x86", "edx");
	&cmp("ecx", 16);
	&jb(&label("poly1305_blocks_x86_nomorebytes"));
	&fld(&FTWP(92,"ebp"));
	&fld(&FTWP(80,"ebp"));
	&fld(&FTWP(68,"ebp"));
	&fld(&FTWP(56,"ebp"));
	&add("esi", 16);
	&sub("ecx", 16);
	&mov(&DWP(20,"esp"), "ecx");
	&mov("eax", &DWP(-4,"esi"));
	&mov("ecx", &DWP(-8,"esi"));
	&mov("ebx", &DWP(-12,"esi"));
	&mov("edi", &DWP(-16,"esi"));
	&mov(&DWP(120,"esp"), "eax");
	&mov(&DWP(112,"esp"), "ecx");
	&mov(&DWP(104,"esp"), "ebx");
	&mov(&DWP(96,"esp"), "edi");
	&fxch("st(3)");
	&fadd(&FQWP(120,"esp"));
	&fsub(&FQWP(96,"edx"));
	&fxch("st(1)");
	&fadd(&FQWP(104,"esp"));
	&fsub(&FQWP(72,"edx"));
	&fxch("st(2)");
	&fadd(&FQWP(112,"esp"));
	&fsub(&FQWP(80,"edx"));
	&fxch("st(3)");
	&fadd(&FQWP(96,"esp"));
	&fsub(&FQWP(64,"edx"));
	&mov("ecx", &DWP(20,"esp"));
	&cmp("ecx", 16);
	&jb(&label("poly1305_blocks_x86_lastmultiply"));
	&set_label("poly1305_blocks_x86_multiplyaddatleast16bytes");
	&add("esi", 16);
	&sub("ecx", 16);
	&mov(&DWP(20,"esp"), "ecx");
	&mov("eax", &DWP(-4,"esi"));
	&mov("ecx", &DWP(-8,"esi"));
	&mov("ebx", &DWP(-12,"esi"));
	&mov("edi", &DWP(-16,"esi"));
	&mov(&DWP(120,"esp"), "eax");
	&mov(&DWP(112,"esp"), "ecx");
	&mov(&DWP(104,"esp"), "ebx");
	&mov(&DWP(96,"esp"), "edi");
	&fld(&FQWP(56,"edx"));
	&fadd("st(0)", "st(2)");
	&fsub(&FQWP(56,"edx"));
	&fsubr("st(2)", "st(0)");
	&fmul(&FQWP(0,"edx"));
	&fld(&FQWP(32,"edx"));
	&fadd("st(0)", "st(2)");
	&fsub(&FQWP(32,"edx"));
	&fsubr("st(2)", "st(0)");
	&fxch("st(2)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(40,"edx"));
	&fadd("st(0)", "st(4)");
	&fsub(&FQWP(40,"edx"));
	&fsubr("st(4)", "st(0)");
	&fld(&FQWP(48,"edx"));
	&fadd("st(0)", "st(6)");
	&fsub(&FQWP(48,"edx"));
	&fsubr("st(6)", "st(0)");
	&fxch("st(6)");
	&faddp("st(1)", "st(0)");
	&fxch("st(3)");
	&faddp("st(5)", "st(0)");
	&fxch("st(3)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(40,"ebp"));
	&fmul("st(0)", "st(3)");
	&fld(&FQWP(24,"ebp"));
	&fmul("st(0)", "st(4)");
	&fld(&FQWP(8,"ebp"));
	&fmul("st(0)", "st(5)");
	&fld(&FQWP(0,"ebp"));
	&fmulp("st(6)", "st(0)");
	&fld(&FQWP(24,"ebp"));
	&fmul("st(0)", "st(4)");
	&faddp("st(3)", "st(0)");
	&fld(&FQWP(8,"ebp"));
	&fmul("st(0)", "st(4)");
	&faddp("st(2)", "st(0)");
	&fld(&FQWP(0,"ebp"));
	&fmul("st(0)", "st(4)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(48,"ebp"));
	&fmulp("st(4)", "st(0)");
	&fxch("st(3)");
	&faddp("st(5)", "st(0)");
	&fld(&FQWP(8,"ebp"));
	&fmul("st(0)", "st(4)");
	&faddp("st(2)", "st(0)");
	&fld(&FQWP(0,"ebp"));
	&fmul("st(0)", "st(4)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(48,"ebp"));
	&fmul("st(0)", "st(4)");
	&faddp("st(3)", "st(0)");
	&fld(&FQWP(32,"ebp"));
	&fmulp("st(4)", "st(0)");
	&fxch("st(3)");
	&faddp("st(4)", "st(0)");
	&fld(&FQWP(0,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(1)", "st(0)");
	&fxch("st(3)");
	&fld(&FQWP(48,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(3)", "st(0)");
	&fxch("st(1)");
	&fld(&FQWP(32,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(16,"ebp"));
	&fmulp("st(5)", "st(0)");
	&fxch("st(4)");
	&faddp("st(1)", "st(0)");
	&mov("ecx", &DWP(20,"esp"));
	&fxch("st(2)");
	&fld(&FQWP(120,"esp"));
	&fsub(&FQWP(96,"edx"));
	&faddp("st(1)", "st(0)");
	&fxch("st(1)");
	&fld(&FQWP(112,"esp"));
	&fsub(&FQWP(80,"edx"));
	&cmp("ecx", 16);
	&faddp("st(1)", "st(0)");
	&fxch("st(3)");
	&fld(&FQWP(104,"esp"));
	&fsub(&FQWP(72,"edx"));
	&faddp("st(1)", "st(0)");
	&fxch("st(2)");
	&fld(&FQWP(96,"esp"));
	&fsub(&FQWP(64,"edx"));
	&faddp("st(1)", "st(0)");
	&jae(&label("poly1305_blocks_x86_multiplyaddatleast16bytes"));
	&set_label("poly1305_blocks_x86_lastmultiply");
	&fld(&FQWP(56,"edx"));
	&fadd("st(0)", "st(2)");
	&fsub(&FQWP(56,"edx"));
	&fsubr("st(2)", "st(0)");
	&fmul(&FQWP(0,"edx"));
	&fld(&FQWP(32,"edx"));
	&fadd("st(0)", "st(2)");
	&fsub(&FQWP(32,"edx"));
	&fsubr("st(2)", "st(0)");
	&fld(&FQWP(40,"edx"));
	&fadd("st(0)", "st(5)");
	&fsub(&FQWP(40,"edx"));
	&fsubr("st(5)", "st(0)");
	&fld(&FQWP(48,"edx"));
	&fadd("st(0)", "st(7)");
	&fsub(&FQWP(48,"edx"));
	&fsubr("st(7)", "st(0)");
	&fxch("st(7)");
	&faddp("st(1)", "st(0)");
	&fxch("st(5)");
	&faddp("st(1)", "st(0)");
	&fxch("st(3)");
	&faddp("st(5)", "st(0)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(40,"ebp"));
	&fmul("st(0)", "st(1)");
	&fld(&FQWP(24,"ebp"));
	&fmul("st(0)", "st(2)");
	&fld(&FQWP(8,"ebp"));
	&fmul("st(0)", "st(3)");
	&fld(&FQWP(0,"ebp"));
	&fmulp("st(4)", "st(0)");
	&fld(&FQWP(24,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(3)", "st(0)");
	&fld(&FQWP(8,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(2)", "st(0)");
	&fld(&FQWP(0,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(48,"ebp"));
	&fmulp("st(5)", "st(0)");
	&fxch("st(4)");
	&faddp("st(3)", "st(0)");
	&fld(&FQWP(8,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(2)", "st(0)");
	&fld(&FQWP(0,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(48,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(4)", "st(0)");
	&fld(&FQWP(32,"ebp"));
	&fmulp("st(5)", "st(0)");
	&fxch("st(4)");
	&faddp("st(2)", "st(0)");
	&fld(&FQWP(0,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(48,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(4)", "st(0)");
	&fld(&FQWP(32,"ebp"));
	&fmul("st(0)", "st(5)");
	&faddp("st(3)", "st(0)");
	&fld(&FQWP(16,"ebp"));
	&fmulp("st(5)", "st(0)");
	&fxch("st(4)");
	&faddp("st(1)", "st(0)");
	&fstp(&FTWP(56,"ebp"));
	&fstp(&FTWP(68,"ebp"));
	&fstp(&FTWP(80,"ebp"));
	&fstp(&FTWP(92,"ebp"));
	&set_label("poly1305_blocks_x86_nomorebytes");
	&fldcw(&DWP(184,"esp"));
	&mov("eax", &DWP(0,"esp"));
	&mov("ebx", &DWP(4,"esp"));
	&mov("esi", &DWP(8,"esp"));
	&mov("edi", &DWP(12,"esp"));
	&mov("ebp", &DWP(16,"esp"));
	&add("esp", "eax");
	&ret();
}

sub poly1305_finish_ex_x86 {
	&static_label("poly1305_finish_ex_x86_local");
	&set_label("poly1305_finish_ex_x86_local");
	&push("ebx");
	&push("esi");
	&push("edi");
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", ~63);
	&sub("esp", 256);
	&mov("ecx", 0x137f);
	&mov(&DWP(0,"esp"), "ecx");
	&fstcw(&DWP(4,"esp"));
	&fldcw(&DWP(0,"esp"));
	&mov("ebx", &DWP(20,"ebp"));
	&mov("esi", &DWP(24,"ebp"));
	&mov("ecx", &DWP(28,"ebp"));
	&mov("eax", &DWP(32,"ebp"));
	&mov(&DWP(0,"esp"), "ebp");
	&mov(&DWP(4,"esp"), "ebx");
	&mov(&DWP(8,"esp"), "eax");
	&fld(&FTWP(92,"ebx"));
	&fld(&FTWP(80,"ebx"));
	&fld(&FTWP(68,"ebx"));
	&fld(&FTWP(56,"ebx"));
	&label_to_reg("poly1305_constants_x86", "edx");
	&mov(&DWP(128,"esp"), "edx");
	&and("ecx", "ecx");
	&jz(&label("poly1305_finish_x86_nomorebytes"));
	&mov(&DWP(100,"esp"), 0x43300000);
	&mov(&DWP(108,"esp"), 0x45300000);
	&mov(&DWP(116,"esp"), 0x47300000);
	&mov(&DWP(124,"esp"), 0x49300000);
	&mov(&DWP(64,"esp"), 0);
	&mov(&DWP(4+64,"esp"), 0);
	&mov(&DWP(8+64,"esp"), 0);
	&mov(&DWP(12+64,"esp"), 0);
	&lea("edi", &DWP(64,"esp"));
	&rep_movsb();
	&mov(&DWP(0,"edi"), 1);
	&mov("eax", &DWP(12+64,"esp"));
	&mov("ecx", &DWP(8+64,"esp"));
	&mov("edx", &DWP(4+64,"esp"));
	&mov("esi", &DWP(64,"esp"));
	&mov(&DWP(120,"esp"), "eax");
	&mov(&DWP(112,"esp"), "ecx");
	&mov(&DWP(104,"esp"), "edx");
	&mov(&DWP(96,"esp"), "esi");
	&mov("edx", &DWP(128,"esp"));
	&fxch("st(3)");
	&fadd(&FQWP(120,"esp"));
	&fsub(&FQWP(88,"edx"));
	&fxch("st(2)");
	&fadd(&FQWP(112,"esp"));
	&fsub(&FQWP(80,"edx"));
	&fxch("st(1)");
	&fadd(&FQWP(104,"esp"));
	&fsub(&FQWP(72,"edx"));
	&fxch("st(3)");
	&fadd(&FQWP(96,"esp"));
	&fsub(&FQWP(64,"edx"));
	&fld(&FQWP(56,"edx"));
	&fadd("st(0)", "st(3)");
	&fsub(&FQWP(56,"edx"));
	&fsubr("st(3)", "st(0)");
	&fmul(&FQWP(0,"edx"));
	&fld(&FQWP(32,"edx"));
	&fadd("st(0)", "st(2)");
	&fsub(&FQWP(32,"edx"));
	&fsubr("st(2)", "st(0)");
	&fld(&FQWP(40,"edx"));
	&fadd("st(0)", "st(6)");
	&fsub(&FQWP(40,"edx"));
	&fsubr("st(6)", "st(0)");
	&fld(&FQWP(48,"edx"));
	&fadd("st(0)", "st(5)");
	&fsub(&FQWP(48,"edx"));
	&fsubr("st(5)", "st(0)");
	&fxch("st(4)");
	&faddp("st(3)", "st(0)");
	&fxch("st(6)");
	&faddp("st(1)", "st(0)");
	&fxch("st(3)");
	&faddp("st(5)", "st(0)");
	&fxch("st(3)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(40,"ebx"));
	&fmul("st(0)", "st(3)");
	&fld(&FQWP(24,"ebx"));
	&fmul("st(0)", "st(4)");
	&fld(&FQWP(8,"ebx"));
	&fmul("st(0)", "st(5)");
	&fld(&FQWP(0,"ebx"));
	&fmulp("st(6)", "st(0)");
	&fld(&FQWP(24,"ebx"));
	&fmul("st(0)", "st(5)");
	&faddp("st(3)", "st(0)");
	&fld(&FQWP(8,"ebx"));
	&fmul("st(0)", "st(5)");
	&faddp("st(2)", "st(0)");
	&fld(&FQWP(0,"ebx"));
	&fmul("st(0)", "st(5)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(48,"ebx"));
	&fmulp("st(5)", "st(0)");
	&fxch("st(4)");
	&faddp("st(5)", "st(0)");
	&fld(&FQWP(8,"ebx"));
	&fmul("st(0)", "st(6)");
	&faddp("st(2)", "st(0)");
	&fld(&FQWP(0,"ebx"));
	&fmul("st(0)", "st(6)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(48,"ebx"));
	&fmul("st(0)", "st(6)");
	&faddp("st(4)", "st(0)");
	&fld(&FQWP(32,"ebx"));
	&fmulp("st(6)", "st(0)");
	&fxch("st(5)");
	&faddp("st(4)", "st(0)");
	&fld(&FQWP(0,"ebx"));
	&fmul("st(0)", "st(2)");
	&faddp("st(1)", "st(0)");
	&fld(&FQWP(48,"ebx"));
	&fmul("st(0)", "st(2)");
	&faddp("st(5)", "st(0)");
	&fld(&FQWP(32,"ebx"));
	&fmul("st(0)", "st(2)");
	&faddp("st(3)", "st(0)");
	&fld(&FQWP(16,"ebx"));
	&fmulp("st(2)", "st(0)");
	&fxch("st(1)");
	&faddp("st(3)", "st(0)");
	&fxch("st(3)");
	&fxch("st(2)");
	&set_label("poly1305_finish_x86_nomorebytes");
	&mov("edx", &DWP(128,"esp"));
	&fld(&FQWP(56,"edx"));
	&fadd("st(0)", "st(4)");
	&fsub(&FQWP(56,"edx"));
	&fsubr("st(4)", "st(0)");
	&fmul(&FQWP(0,"edx"));
	&fld(&FQWP(32,"edx"));
	&fadd("st(0)", "st(2)");
	&fsub(&FQWP(32,"edx"));
	&fsubr("st(2)", "st(0)");
	&fld(&FQWP(40,"edx"));
	&fadd("st(0)", "st(4)");
	&fsub(&FQWP(40,"edx"));
	&fsubr("st(4)", "st(0)");
	&fld(&FQWP(48,"edx"));
	&fadd("st(0)", "st(6)");
	&fsub(&FQWP(48,"edx"));
	&fxch("st(6)");
	&fsub("st(0)", "st(6)");
	&fxch("st(4)");
	&faddp("st(3)", "st(0)");
	&fxch("st(4)");
	&faddp("st(1)", "st(0)");
	&fxch("st(2)");
	&faddp("st(3)", "st(0)");
	&fxch("st(4)");
	&faddp("st(3)", "st(0)");
	&fxch("st(3)");
	&fadd(&FQWP(104,"edx"));
	&fxch("st(3)");
	&fadd(&FQWP(112,"edx"));
	&fxch("st(1)");
	&fadd(&FQWP(120,"edx"));
	&fxch("st(2)");
	&fadd(&FQWP(128,"edx"));
	&fxch("st(3)");
	&fstp(&FQWP(96,"esp"));
	&fstp(&FQWP(104,"esp"));
	&fstp(&FQWP(112,"esp"));
	&fstp(&FQWP(120,"esp"));
	&mov("eax", &DWP(100,"esp"));
	&and("eax", 63);
	&mov("ecx", &DWP(108,"esp"));
	&and("ecx", 63);
	&mov("edx", &DWP(116,"esp"));
	&and("edx", 63);
	&mov("ebx", &DWP(124,"esp"));
	&and("ebx", 63);
	&mov("esi", &DWP(104,"esp"));
	&add("esi", "eax");
	&mov(&DWP(28,"esp"), "esi");
	&mov("eax", &DWP(112,"esp"));
	&adc("eax", "ecx");
	&mov(&DWP(32,"esp"), "eax");
	&mov("eax", &DWP(120,"esp"));
	&adc("eax", "edx");
	&mov(&DWP(36,"esp"), "eax");
	&mov("eax", 0);
	&adc("eax", "ebx");
	&mov(&DWP(40,"esp"), "eax");
	&mov("eax", 5);
	&mov("ecx", &DWP(96,"esp"));
	&add("eax", "ecx");
	&mov(&DWP(44,"esp"), "eax");
	&mov("eax", 0);
	&mov("edx", &DWP(28,"esp"));
	&adc("eax", "edx");
	&mov(&DWP(28,"esp"), "eax");
	&mov("eax", 0);
	&mov("ebx", &DWP(32,"esp"));
	&adc("eax", "ebx");
	&mov(&DWP(32,"esp"), "eax");
	&mov("eax", 0);
	&mov("esi", &DWP(36,"esp"));
	&adc("eax", "esi");
	&mov(&DWP(36,"esp"), "eax");
	&mov("eax", 0xfffffffc);
	&mov("edi", &DWP(40,"esp"));
	&adc("eax", "edi");
	&sar("eax", 16);
	&mov("edi", "eax");
	&xor("edi", 0xffffffff);
	&and("ecx", "eax");
	&mov("ebp", &DWP(44,"esp"));
	&and("ebp", "edi");
	&or("ecx", "ebp");
	&and("edx", "eax");
	&mov("ebp", &DWP(28,"esp"));
	&and("ebp", "edi");
	&or("edx", "ebp");
	&and("ebx", "eax");
	&mov("ebp", &DWP(32,"esp"));
	&and("ebp", "edi");
	&or("ebx", "ebp");
	&and("esi", "eax");
	&mov("eax", &DWP(36,"esp"));
	&and("eax", "edi");
	&or("esi", "eax");
	&mov("eax", &DWP(4,"esp"));
	&add("ecx", &DWP(104,"eax"));
	&adc("edx", &DWP(108,"eax"));
	&adc("ebx", &DWP(112,"eax"));
	&adc("esi", &DWP(116,"eax"));
	&mov("eax", &DWP(8,"esp"));
	&mov(&DWP(0,"eax"), "ecx");
	&mov(&DWP(4,"eax"), "edx");
	&mov(&DWP(8,"eax"), "ebx");
	&mov(&DWP(12,"eax"), "esi");
	&xor("eax", "eax");
	&mov("edi", &DWP(4,"esp"));
	&fldz();
	&fst(&FQWP(0,"edi"));
	&fst(&FQWP(8,"edi"));
	&fst(&FQWP(16,"edi"));
	&fst(&FQWP(24,"edi"));
	&fst(&FQWP(32,"edi"));
	&fst(&FQWP(40,"edi"));
	&fst(&FQWP(48,"edi"));
	&fst(&FQWP(56,"edi"));
	&fst(&FQWP(64,"edi"));
	&fst(&FQWP(72,"edi"));
	&fst(&FQWP(80,"edi"));
	&fst(&FQWP(88,"edi"));
	&fst(&FQWP(96,"edi"));
	&fst(&FQWP(104,"edi"));
	&fstp(&FQWP(112,"edi"));
	&mov("esp", &DWP(0,"esp"));
	&pop("ebp");
	&pop("edi");
	&pop("esi");
	&pop("ebx");
	&ret();
}

sub poly1305_x86 {
	&push("ebp");
	&push("edi");
	&mov("ebp", "esp");
	&sub("esp", 128);
	&and("esp", ~63);
	&mov("edi", "esp");
	&push(&DWP(24,"ebp"));
	&push("edi");
	&call(&LABEL("poly1305_init_ex_x86_local"));
	&mov("ecx", &DWP(20,"ebp"));
	&and("ecx", ~15);
	&jz(&label("poly1305_x86_no_data"));
	&push("ecx");
	&push(&DWP(16,"ebp"));
	&add(&DWP(16,"ebp"), "ecx");
	&push("edi");
	&call(&LABEL("poly1305_blocks_x86_local"));
	&set_label("poly1305_x86_no_data");
	&push(&DWP(12,"ebp"));
	&mov("ecx", &DWP(20,"ebp"));
	&and("ecx", 15);
	&push("ecx");
	&push(&DWP(16,"ebp"));
	&push("edi");
	&call(&LABEL("poly1305_finish_ex_x86_local"));
	&mov("esp", "ebp");
	&pop("edi");
	&pop("ebp");
	&ret();
}

sub poly1305_finish_ex_sse2 {
	&static_label("poly1305_finish_ex_sse2_local");
	&set_label("poly1305_finish_ex_sse2_local");
	&push("esi");
	&push("edi");
	&push("ebx");
	&push("ebp");
	&sub("esp", 60);
	&mov("ebp", &DWP(88,"esp"));
	&test("ebp", "ebp");
	&mov("ebx", &DWP(80,"esp"));
	&je(&label("poly1305_finish_ex_sse2_18"));
	&set_label("poly1305_finish_ex_sse2_2");
	&pxor("xmm0", "xmm0");
	&movaps(&DWP(16,"esp"), "xmm0");
	&movaps(&DWP(32,"esp"), "xmm0");
	&set_label("poly1305_finish_ex_sse2_3");
	&mov("ecx", &DWP(84,"esp"));
	&lea("edx", &DWP(16,"esp"));
	&sub("ecx", "edx");
	&test("ebp", 16);
	&je(&label("poly1305_finish_ex_sse2_5"));
	&set_label("poly1305_finish_ex_sse2_4");
	&lea("edx", &DWP(32,"esp"));
	&movdqu("xmm0", &DWP(16,"esp","ecx"));
	&movdqa(&DWP(16,"esp"), "xmm0");
	&set_label("poly1305_finish_ex_sse2_5");
	&test("ebp", 8);
	&je(&label("poly1305_finish_ex_sse2_7"));
	&set_label("poly1305_finish_ex_sse2_6");
	&mov("esi", &DWP(0,"edx","ecx"));
	&mov("edi", &DWP(4,"edx","ecx"));
	&mov(&DWP(0,"edx"), "esi");
	&mov(&DWP(4,"edx"), "edi");
	&add("edx", 8);
	&set_label("poly1305_finish_ex_sse2_7");
	&test("ebp", 4);
	&je(&label("poly1305_finish_ex_sse2_9"));
	&set_label("poly1305_finish_ex_sse2_8");
	&mov("esi", &DWP(0,"edx","ecx"));
	&mov(&DWP(0,"edx"), "esi");
	&add("edx", 4);
	&set_label("poly1305_finish_ex_sse2_9");
	&test("ebp", 2);
	&je(&label("poly1305_finish_ex_sse2_11"));
	&set_label("poly1305_finish_ex_sse2_10");
	&movzw("esi", &DWP(0,"edx","ecx"));
	&mov(&DWP(0,"edx"), "si");
	&add("edx", 2);
	&set_label("poly1305_finish_ex_sse2_11");
	&test("ebp", 1);
	&je(&label("poly1305_finish_ex_sse2_13"));
	&set_label("poly1305_finish_ex_sse2_12");
	&movzb("ecx", &DWP(0,"edx","ecx"));
	&mov(&DWP(0,"edx"), "cl");
	&set_label("poly1305_finish_ex_sse2_13");
	&cmp("ebp", 16);
	&je(&label("poly1305_finish_ex_sse2_16"));
	&set_label("poly1305_finish_ex_sse2_14");
	&mov("dl", 1);
	&mov(&BP(16,"esp","ebp"), "dl");
	&jae(&label("poly1305_finish_ex_sse2_16"));
	&set_label("poly1305_finish_ex_sse2_15");
	&mov("edx", 8);
	&jmp(&label("poly1305_finish_ex_sse2_17"));
	&set_label("poly1305_finish_ex_sse2_16");
	&mov("edx", 4);
	&set_label("poly1305_finish_ex_sse2_17");
	&or(&DWP(116,"ebx"), "edx");
	&mov("eax", "ebx");
	&mov("ecx", 32);
	&lea("edx", &DWP(16,"esp"));
	&call(&LABEL("poly1305_blocks_sse2_local"));
	&set_label("poly1305_finish_ex_sse2_18");
	&mov("edx", &DWP(116,"ebx"));
	&test("dl", 1);
	&je(&label("poly1305_finish_ex_sse2_24"));
	&set_label("poly1305_finish_ex_sse2_19");
	&test("ebp", "ebp");
	&je(&label("poly1305_finish_ex_sse2_21"));
	&set_label("poly1305_finish_ex_sse2_20");
	&cmp("ebp", 16);
	&jbe(&label("poly1305_finish_ex_sse2_22"));
	&set_label("poly1305_finish_ex_sse2_21");
	&or("edx", 16);
	&mov(&DWP(116,"ebx"), "edx");
	&jmp(&label("poly1305_finish_ex_sse2_23"));
	&set_label("poly1305_finish_ex_sse2_22");
	&or("edx", 32);
	&mov(&DWP(116,"ebx"), "edx");
	&set_label("poly1305_finish_ex_sse2_23");
	&mov("eax", "ebx");
	&xor("edx", "edx");
	&mov("ecx", 32);
	&call(&LABEL("poly1305_blocks_sse2_local"));
	&set_label("poly1305_finish_ex_sse2_24");
	&mov("edx", &DWP(8,"ebx"));
	&mov("eax", "edx");
	&mov("ecx", &DWP(4,"ebx"));
	&mov("esi", "ecx");
	&shr("ecx", 6);
	&shl("eax", 20);
	&pxor("xmm0", "xmm0");
	&mov("ebp", &DWP(12,"ebx"));
	&or("ecx", "eax");
	&mov("eax", "ebp");
	&shr("edx", 12);
	&shl("eax", 14);
	&or("edx", "eax");
	&mov("eax", &DWP(16,"ebx"));
	&shl("esi", 26);
	&shr("ebp", 18);
	&shl("eax", 8);
	&mov("edi", &DWP(92,"esp"));
	&or("ebp", "eax");
	&or("esi", &DWP(0,"ebx"));
	&add("esi", &DWP(100,"ebx"));
	&adc("ecx", &DWP(104,"ebx"));
	&adc("edx", &DWP(108,"ebx"));
	&adc("ebp", &DWP(112,"ebx"));
	&movdqu(&DWP(0,"ebx"), "xmm0");
	&movdqu(&DWP(16,"ebx"), "xmm0");
	&movdqu(&DWP(32,"ebx"), "xmm0");
	&movdqu(&DWP(48,"ebx"), "xmm0");
	&movdqu(&DWP(64,"ebx"), "xmm0");
	&movdqu(&DWP(80,"ebx"), "xmm0");
	&movdqu(&DWP(96,"ebx"), "xmm0");
	&movdqu(&DWP(112,"ebx"), "xmm0");
	&mov(&DWP(0,"edi"), "esi");
	&mov(&DWP(4,"edi"), "ecx");
	&mov(&DWP(8,"edi"), "edx");
	&mov(&DWP(12,"edi"), "ebp");
	&add("esp", 60);
	&pop("ebp");
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
}


sub poly1305_blocks_sse2 {
	&mov("eax", &DWP(4,"esp"));
	&mov("edx", &DWP(8,"esp"));
	&mov("ecx", &DWP(12,"esp"));
	&static_label("poly1305_blocks_sse2_local");
	&set_label("poly1305_blocks_sse2_local");
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 544);
	&mov("ebx", 16777216);
	&mov("esi", 67108863);
	&mov("edi", 5);
	&movd("xmm0", "ebx");
	&movd("xmm2", "esi");
	&movd("xmm4", "edi");
	&mov("ebx", &DWP(116,"eax"));
	&test("bl", 4);
	&pshufd("xmm1", "xmm0", 68);
	&pshufd("xmm3", "xmm2", 68);
	&pshufd("xmm5", "xmm4", 68);
	&movdqa(&DWP(272,"esp"), "xmm1");
	&movdqa(&DWP(256,"esp"), "xmm3");
	&movdqa(&DWP(160,"esp"), "xmm5");
	&je(&label("poly1305_blocks_sse2_3"));
	&set_label("poly1305_blocks_sse2_2");
	&movdqa("xmm0", &DWP(272,"esp"));
	&psrldq("xmm0", 8);
	&movdqa(&DWP(272,"esp"), "xmm0");
	&set_label("poly1305_blocks_sse2_3");
	&test("bl", 8);
	&je(&label("poly1305_blocks_sse2_5"));
	&set_label("poly1305_blocks_sse2_4");
	&pxor("xmm0", "xmm0");
	&movdqa(&DWP(272,"esp"), "xmm0");
	&set_label("poly1305_blocks_sse2_5");
	&test("bl", 1);
	&jne(&label("poly1305_blocks_sse2_7"));
	&set_label("poly1305_blocks_sse2_6");
	&movq("xmm0", &DWP(8,"edx"));
	&or("ebx", 1);
	&movq("xmm1", &DWP(0,"edx"));
	&add("ecx", -32);
	&movhpd("xmm0", &DWP(24,"edx"));
	&movdqa("xmm4", &DWP(256,"esp"));
	&movaps("xmm2", "xmm0");
	&movhpd("xmm1", &DWP(16,"edx"));
	&movdqa("xmm7", "xmm4");
	&pand("xmm7", "xmm1");
	&movaps("xmm6", "xmm1");
	&psrlq("xmm1", 52);
	&psllq("xmm2", 12);
	&por("xmm1", "xmm2");
	&movdqa("xmm3", "xmm4");
	&psrlq("xmm6", 26);
	&pand("xmm3", "xmm1");
	&psrlq("xmm1", 26);
	&psrlq("xmm0", 40);
	&movdqa(&DWP(32,"esp"), "xmm3");
	&pand("xmm6", "xmm4");
	&por("xmm0", &DWP(272,"esp"));
	&pand("xmm1", "xmm4");
	&mov(&DWP(116,"eax"), "ebx");
	&add("edx", 32);
	&jmp(&label("poly1305_blocks_sse2_8"));
	&set_label("poly1305_blocks_sse2_7");
	&movdqu("xmm0", &DWP(16,"eax"));
	&movdqu("xmm6", &DWP(0,"eax"));
	&movdqu("xmm2", &DWP(32,"eax"));
	&pshufd("xmm1", "xmm0", 80);
	&pshufd("xmm7", "xmm6", 80);
	&pshufd("xmm6", "xmm6", 250);
	&movdqa(&DWP(32,"esp"), "xmm1");
	&pshufd("xmm1", "xmm0", 250);
	&pshufd("xmm0", "xmm2", 80);
	&set_label("poly1305_blocks_sse2_8");
	&test("bl", 48);
	&je(&label("poly1305_blocks_sse2_13"));
	&set_label("poly1305_blocks_sse2_9");
	&movdqu("xmm2", &DWP(40,"eax"));
	&mov("esi", &DWP(56,"eax"));
	&test("bl", 16);
	&je(&label("poly1305_blocks_sse2_11"));
	&set_label("poly1305_blocks_sse2_10");
	&movdqu("xmm4", &DWP(60,"eax"));
	&movdqa("xmm5", "xmm4");
	&movd("xmm3", &DWP(76,"eax"));
	&punpckldq("xmm5", "xmm2");
	&punpckhdq("xmm4", "xmm2");
	&movd("xmm2", "esi");
	&punpcklqdq("xmm3", "xmm2");
	&movdqa(&DWP(240,"esp"), "xmm3");
	&jmp(&label("poly1305_blocks_sse2_12"));
	&set_label("poly1305_blocks_sse2_11");
	&mov("ebx", 1);
	&movdqa("xmm5", "xmm2");
	&movdqa("xmm4", "xmm2");
	&movd("xmm2", "esi");
	&movdqa(&DWP(240,"esp"), "xmm2");
	&movd("xmm3", "ebx");
	&punpckldq("xmm5", "xmm3");
	&punpckhdq("xmm4", "xmm3");
	&set_label("poly1305_blocks_sse2_12");
	&pshufd("xmm2", "xmm5", 80);
	&pshufd("xmm3", "xmm5", 250);
	&pshufd("xmm5", "xmm4", 80);
	&pshufd("xmm4", "xmm4", 250);
	&movdqa(&DWP(176,"esp"), "xmm2");
	&movdqa(&DWP(224,"esp"), "xmm3");
	&movdqa(&DWP(208,"esp"), "xmm5");
	&movdqa(&DWP(192,"esp"), "xmm4");
	&jmp(&label("poly1305_blocks_sse2_14"));
	&set_label("poly1305_blocks_sse2_13");
	&movdqu("xmm3", &DWP(60,"eax"));
	&movd("xmm2", &DWP(76,"eax"));
	&pshufd("xmm4", "xmm3", 0);
	&movdqa(&DWP(176,"esp"), "xmm4");
	&pshufd("xmm5", "xmm3", 85);
	&pshufd("xmm4", "xmm3", 170);
	&pshufd("xmm3", "xmm3", 255);
	&pshufd("xmm2", "xmm2", 0);
	&movdqa(&DWP(224,"esp"), "xmm5");
	&movdqa(&DWP(208,"esp"), "xmm4");
	&movdqa(&DWP(192,"esp"), "xmm3");
	&movdqa(&DWP(240,"esp"), "xmm2");
	&set_label("poly1305_blocks_sse2_14");
	&movdqa("xmm2", &DWP(160,"esp"));
	&cmp("ecx", 64);
	&movdqa("xmm5", &DWP(192,"esp"));
	&pmuludq("xmm5", "xmm2");
	&movdqa(&DWP(320,"esp"), "xmm5");
	&movdqa("xmm4", &DWP(224,"esp"));
	&movdqa("xmm3", &DWP(208,"esp"));
	&movdqa("xmm5", &DWP(240,"esp"));
	&pmuludq("xmm4", "xmm2");
	&pmuludq("xmm3", "xmm2");
	&pmuludq("xmm5", "xmm2");
	&movdqa("xmm2", &DWP(320,"esp"));
	&jb(&label("poly1305_blocks_sse2_18"));
	&set_label("poly1305_blocks_sse2_15");
	&movdqa(&DWP(304,"esp"), "xmm3");
	&movdqu("xmm3", &DWP(80,"eax"));
	&movaps(&DWP(400,"esp"), "xmm0");
	&movd("xmm0", &DWP(96,"eax"));
	&movdqa(&DWP(320,"esp"), "xmm2");
	&pshufd("xmm2", "xmm3", 85);
	&movdqa(&DWP(128,"esp"), "xmm2");
	&pshufd("xmm2", "xmm0", 0);
	&movdqa("xmm0", &DWP(160,"esp"));
	&movdqa(&DWP(336,"esp"), "xmm5");
	&pshufd("xmm5", "xmm3", 0);
	&movdqa(&DWP(0,"esp"), "xmm5");
	&movdqa("xmm5", "xmm0");
	&pmuludq("xmm5", &DWP(128,"esp"));
	&movdqa(&DWP(288,"esp"), "xmm4");
	&pshufd("xmm4", "xmm3", 170);
	&movdqa(&DWP(80,"esp"), "xmm5");
	&movdqa("xmm5", "xmm0");
	&movdqa(&DWP(112,"esp"), "xmm4");
	&pshufd("xmm3", "xmm3", 255);
	&pmuludq("xmm5", "xmm4");
	&movdqa("xmm4", "xmm0");
	&pmuludq("xmm4", "xmm3");
	&pmuludq("xmm0", "xmm2");
	&movdqa(&DWP(144,"esp"), "xmm3");
	&movdqa(&DWP(96,"esp"), "xmm2");
	&movdqa(&DWP(64,"esp"), "xmm5");
	&movdqa(&DWP(48,"esp"), "xmm4");
	&movdqa(&DWP(16,"esp"), "xmm0");
	&movaps("xmm0", &DWP(400,"esp"));
	&movaps(&DWP(384,"esp"), "xmm1");
	&movdqa(&DWP(368,"esp"), "xmm6");
	&movdqa(&DWP(352,"esp"), "xmm7");
	&set_label("poly1305_blocks_sse2_16");
	&movq("xmm6", &DWP(8,"edx"));
	&add("ecx", -64);
	&movq("xmm2", &DWP(0,"edx"));
	&movhpd("xmm6", &DWP(24,"edx"));
	&movdqa("xmm1", &DWP(256,"esp"));
	&movaps("xmm7", "xmm6");
	&movhpd("xmm2", &DWP(16,"edx"));
	&movdqa("xmm4", "xmm1");
	&pand("xmm4", "xmm2");
	&movaps("xmm3", "xmm2");
	&movaps("xmm5", "xmm6");
	&psrlq("xmm2", 52);
	&psllq("xmm7", 12);
	&psrlq("xmm3", 26);
	&psrlq("xmm5", 14);
	&por("xmm2", "xmm7");
	&pand("xmm3", "xmm1");
	&pand("xmm5", "xmm1");
	&pand("xmm2", "xmm1");
	&psrlq("xmm6", 40);
	&movdqu("xmm1", &DWP(32,"edx"));
	&movdqu("xmm7", &DWP(48,"edx"));
	&movdqa(&DWP(416,"esp"), "xmm4");
	&movdqa("xmm4", "xmm1");
	&punpckldq("xmm4", "xmm7");
	&add("edx", 64);
	&punpckhdq("xmm1", "xmm7");
	&cmp("ecx", 64);
	&movaps("xmm7", &DWP(384,"esp"));
	&movaps(&DWP(400,"esp"), "xmm0");
	&pmuludq("xmm0", &DWP(80,"esp"));
	&pmuludq("xmm7", &DWP(64,"esp"));
	&movdqa(&DWP(480,"esp"), "xmm1");
	&movdqa("xmm1", &DWP(32,"esp"));
	&pmuludq("xmm1", &DWP(48,"esp"));
	&paddq("xmm0", "xmm7");
	&movdqa("xmm7", &DWP(368,"esp"));
	&pmuludq("xmm7", &DWP(16,"esp"));
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", &DWP(352,"esp"));
	&paddq("xmm0", "xmm7");
	&movdqa("xmm7", &DWP(0,"esp"));
	&pmuludq("xmm1", "xmm7");
	&por("xmm6", &DWP(272,"esp"));
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", &DWP(288,"esp"));
	&pmuludq("xmm1", "xmm6");
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", &DWP(304,"esp"));
	&pmuludq("xmm1", "xmm5");
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", &DWP(320,"esp"));
	&pmuludq("xmm1", "xmm2");
	&movaps(&DWP(448,"esp"), "xmm2");
	&movdqa("xmm2", &DWP(336,"esp"));
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", "xmm2");
	&pmuludq("xmm1", "xmm3");
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", &DWP(176,"esp"));
	&movdqa(&DWP(432,"esp"), "xmm3");
	&movdqa("xmm3", "xmm1");
	&pmuludq("xmm3", &DWP(416,"esp"));
	&pmuludq("xmm1", "xmm5");
	&paddq("xmm0", "xmm3");
	&pxor("xmm3", "xmm3");
	&movdqa(&DWP(464,"esp"), "xmm4");
	&punpckldq("xmm4", "xmm3");
	&paddq("xmm0", "xmm4");
	&movdqa(&DWP(496,"esp"), "xmm0");
	&movaps("xmm4", &DWP(400,"esp"));
	&movaps("xmm0", &DWP(384,"esp"));
	&pmuludq("xmm4", &DWP(16,"esp"));
	&pmuludq("xmm0", "xmm7");
	&movdqa("xmm7", &DWP(32,"esp"));
	&pmuludq("xmm7", &DWP(128,"esp"));
	&paddq("xmm4", "xmm0");
	&movdqa("xmm0", &DWP(368,"esp"));
	&pmuludq("xmm0", &DWP(112,"esp"));
	&paddq("xmm4", "xmm7");
	&movdqa("xmm7", &DWP(352,"esp"));
	&pmuludq("xmm7", &DWP(144,"esp"));
	&paddq("xmm4", "xmm0");
	&movdqa("xmm0", "xmm2");
	&pmuludq("xmm0", "xmm6");
	&paddq("xmm4", "xmm7");
	&movdqa("xmm7", &DWP(224,"esp"));
	&paddq("xmm4", "xmm0");
	&movaps("xmm0", &DWP(448,"esp"));
	&pmuludq("xmm7", "xmm0");
	&pmuludq("xmm2", "xmm0");
	&paddq("xmm4", "xmm1");
	&movdqa("xmm1", &DWP(208,"esp"));
	&pmuludq("xmm1", &DWP(432,"esp"));
	&paddq("xmm4", "xmm7");
	&movdqa("xmm7", &DWP(192,"esp"));
	&paddq("xmm4", "xmm1");
	&movdqa("xmm1", &DWP(416,"esp"));
	&pmuludq("xmm7", "xmm1");
	&paddq("xmm4", "xmm7");
	&movdqa("xmm7", &DWP(480,"esp"));
	&punpckhdq("xmm7", "xmm3");
	&psllq("xmm7", 18);
	&paddq("xmm4", "xmm7");
	&movdqa(&DWP(512,"esp"), "xmm4");
	&movaps("xmm4", &DWP(400,"esp"));
	&movaps("xmm3", &DWP(384,"esp"));
	&pmuludq("xmm4", &DWP(64,"esp"));
	&pmuludq("xmm3", &DWP(48,"esp"));
	&movdqa("xmm7", &DWP(32,"esp"));
	&pmuludq("xmm7", &DWP(16,"esp"));
	&paddq("xmm4", "xmm3");
	&movdqa("xmm3", &DWP(368,"esp"));
	&pmuludq("xmm3", &DWP(0,"esp"));
	&paddq("xmm4", "xmm7");
	&movdqa("xmm7", &DWP(352,"esp"));
	&paddq("xmm4", "xmm3");
	&movdqa("xmm3", &DWP(128,"esp"));
	&pmuludq("xmm7", "xmm3");
	&paddq("xmm4", "xmm7");
	&movdqa("xmm7", &DWP(304,"esp"));
	&pmuludq("xmm7", "xmm6");
	&paddq("xmm4", "xmm7");
	&movdqa("xmm7", &DWP(320,"esp"));
	&pmuludq("xmm7", "xmm5");
	&paddq("xmm4", "xmm7");
	&movdqa("xmm7", &DWP(176,"esp"));
	&movdqa("xmm0", "xmm7");
	&pmuludq("xmm0", &DWP(432,"esp"));
	&paddq("xmm4", "xmm2");
	&movdqa("xmm2", &DWP(224,"esp"));
	&paddq("xmm4", "xmm0");
	&movdqa("xmm0", "xmm2");
	&pmuludq("xmm0", "xmm1");
	&pmuludq("xmm2", "xmm5");
	&pmuludq("xmm5", &DWP(336,"esp"));
	&paddq("xmm4", "xmm0");
	&movdqa("xmm0", &DWP(464,"esp"));
	&pxor("xmm1", "xmm1");
	&punpckhdq("xmm0", "xmm1");
	&psllq("xmm0", 6);
	&movdqa("xmm1", &DWP(496,"esp"));
	&paddq("xmm4", "xmm0");
	&psrlq("xmm1", 26);
	&paddq("xmm4", "xmm1");
	&movdqa(&DWP(528,"esp"), "xmm4");
	&movaps("xmm4", &DWP(400,"esp"));
	&movaps("xmm1", &DWP(384,"esp"));
	&movaps("xmm0", "xmm4");
	&pmuludq("xmm0", &DWP(0,"esp"));
	&pmuludq("xmm1", "xmm3");
	&pmuludq("xmm4", &DWP(48,"esp"));
	&paddq("xmm0", "xmm1");
	&movdqa("xmm3", &DWP(32,"esp"));
	&pmuludq("xmm3", &DWP(112,"esp"));
	&paddq("xmm0", "xmm3");
	&movdqa("xmm3", &DWP(368,"esp"));
	&movdqa("xmm1", "xmm3");
	&pmuludq("xmm1", &DWP(144,"esp"));
	&pmuludq("xmm3", &DWP(128,"esp"));
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", &DWP(352,"esp"));
	&pmuludq("xmm1", &DWP(96,"esp"));
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", "xmm7");
	&pmuludq("xmm1", "xmm6");
	&pmuludq("xmm6", &DWP(320,"esp"));
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", &DWP(208,"esp"));
	&pmuludq("xmm1", &DWP(448,"esp"));
	&paddq("xmm0", "xmm2");
	&movdqa("xmm2", &DWP(192,"esp"));
	&pmuludq("xmm2", &DWP(432,"esp"));
	&paddq("xmm0", "xmm1");
	&movdqa("xmm1", &DWP(416,"esp"));
	&paddq("xmm0", "xmm2");
	&movdqa("xmm2", &DWP(240,"esp"));
	&pmuludq("xmm2", "xmm1");
	&pmuludq("xmm1", &DWP(208,"esp"));
	&paddq("xmm0", "xmm2");
	&movdqa("xmm2", &DWP(512,"esp"));
	&paddq("xmm0", &DWP(272,"esp"));
	&psrlq("xmm2", 26);
	&paddq("xmm0", "xmm2");
	&movaps("xmm2", &DWP(384,"esp"));
	&pmuludq("xmm2", &DWP(16,"esp"));
	&paddq("xmm4", "xmm2");
	&movdqa("xmm2", &DWP(32,"esp"));
	&pmuludq("xmm2", &DWP(0,"esp"));
	&paddq("xmm4", "xmm2");
	&paddq("xmm4", "xmm3");
	&movdqa("xmm3", &DWP(352,"esp"));
	&pmuludq("xmm3", &DWP(112,"esp"));
	&paddq("xmm4", "xmm3");
	&paddq("xmm4", "xmm6");
	&movaps("xmm6", &DWP(448,"esp"));
	&movdqa("xmm3", "xmm0");
	&pmuludq("xmm6", "xmm7");
	&psrlq("xmm3", 26);
	&pmuludq("xmm3", &DWP(160,"esp"));
	&paddq("xmm4", "xmm5");
	&movdqa("xmm5", &DWP(432,"esp"));
	&pmuludq("xmm5", &DWP(224,"esp"));
	&paddq("xmm4", "xmm6");
	&paddq("xmm4", "xmm5");
	&movdqa("xmm7", &DWP(480,"esp"));
	&pxor("xmm5", "xmm5");
	&punpckldq("xmm7", "xmm5");
	&paddq("xmm4", "xmm1");
	&movdqa("xmm2", &DWP(528,"esp"));
	&psllq("xmm7", 12);
	&movdqa("xmm1", "xmm2");
	&paddq("xmm4", "xmm7");
	&psrlq("xmm1", 26);
	&paddq("xmm4", "xmm1");
	&movdqa("xmm5", &DWP(256,"esp"));
	&movaps("xmm1", "xmm4");
	&movdqa("xmm6", &DWP(496,"esp"));
	&psrlq("xmm1", 26);
	&movdqa("xmm7", &DWP(512,"esp"));
	&pand("xmm6", "xmm5");
	&pand("xmm7", "xmm5");
	&pand("xmm2", "xmm5");
	&paddq("xmm6", "xmm3");
	&paddq("xmm7", "xmm1");
	&movdqa("xmm3", "xmm5");
	&movdqa("xmm1", "xmm5");
	&pand("xmm3", "xmm6");
	&psrlq("xmm6", 26);
	&pand("xmm1", "xmm4");
	&movdqa("xmm4", "xmm5");
	&paddq("xmm2", "xmm6");
	&pand("xmm4", "xmm7");
	&pand("xmm0", "xmm5");
	&psrlq("xmm7", 26);
	&movdqa(&DWP(352,"esp"), "xmm3");
	&movdqa(&DWP(368,"esp"), "xmm2");
	&movdqa(&DWP(32,"esp"), "xmm1");
	&movaps(&DWP(384,"esp"), "xmm4");
	&paddq("xmm0", "xmm7");
	&jae(&label("poly1305_blocks_sse2_16"));
	&set_label("poly1305_blocks_sse2_17");
	&movdqa("xmm5", &DWP(336,"esp"));
	&movdqa("xmm2", &DWP(320,"esp"));
	&movdqa("xmm3", &DWP(304,"esp"));
	&movdqa("xmm4", &DWP(288,"esp"));
	&movaps("xmm1", &DWP(384,"esp"));
	&movdqa("xmm6", &DWP(368,"esp"));
	&movdqa("xmm7", &DWP(352,"esp"));
	&set_label("poly1305_blocks_sse2_18");
	&cmp("ecx", 32);
	&jb(&label("poly1305_blocks_sse2_22"));
	&set_label("poly1305_blocks_sse2_19");
	&movaps(&DWP(384,"esp"), "xmm1");
	&test("edx", "edx");
	&pmuludq("xmm4", "xmm0");
	&pmuludq("xmm1", "xmm3");
	&pmuludq("xmm3", "xmm0");
	&paddq("xmm4", "xmm1");
	&movdqa("xmm1", &DWP(32,"esp"));
	&pmuludq("xmm1", "xmm2");
	&paddq("xmm4", "xmm1");
	&movdqa("xmm1", "xmm6");
	&pmuludq("xmm1", "xmm5");
	&paddq("xmm4", "xmm1");
	&movdqa("xmm1", &DWP(176,"esp"));
	&movdqa(&DWP(352,"esp"), "xmm7");
	&pmuludq("xmm7", "xmm1");
	&paddq("xmm4", "xmm7");
	&movdqa(&DWP(288,"esp"), "xmm4");
	&movaps("xmm4", &DWP(384,"esp"));
	&movaps("xmm7", "xmm4");
	&pmuludq("xmm7", "xmm2");
	&pmuludq("xmm2", "xmm0");
	&paddq("xmm3", "xmm7");
	&movdqa("xmm7", &DWP(32,"esp"));
	&pmuludq("xmm7", "xmm5");
	&movdqa(&DWP(368,"esp"), "xmm6");
	&pmuludq("xmm6", "xmm1");
	&paddq("xmm3", "xmm7");
	&paddq("xmm3", "xmm6");
	&movdqa("xmm7", &DWP(352,"esp"));
	&movdqa("xmm6", &DWP(224,"esp"));
	&pmuludq("xmm7", "xmm6");
	&paddq("xmm3", "xmm7");
	&movdqa(&DWP(304,"esp"), "xmm3");
	&movaps("xmm3", "xmm4");
	&pmuludq("xmm3", "xmm5");
	&pmuludq("xmm5", "xmm0");
	&pmuludq("xmm0", "xmm1");
	&paddq("xmm2", "xmm3");
	&movdqa("xmm3", &DWP(32,"esp"));
	&movdqa("xmm7", "xmm3");
	&pmuludq("xmm7", "xmm1");
	&paddq("xmm2", "xmm7");
	&movdqa("xmm7", &DWP(368,"esp"));
	&pmuludq("xmm7", "xmm6");
	&movdqa("xmm6", &DWP(352,"esp"));
	&paddq("xmm2", "xmm7");
	&movdqa("xmm7", &DWP(208,"esp"));
	&pmuludq("xmm6", "xmm7");
	&paddq("xmm2", "xmm6");
	&movdqa(&DWP(320,"esp"), "xmm2");
	&movaps("xmm2", "xmm4");
	&pmuludq("xmm2", "xmm1");
	&paddq("xmm5", "xmm2");
	&movdqa("xmm6", &DWP(224,"esp"));
	&movdqa("xmm2", "xmm3");
	&pmuludq("xmm2", "xmm6");
	&pmuludq("xmm4", "xmm6");
	&pmuludq("xmm3", &DWP(208,"esp"));
	&paddq("xmm5", "xmm2");
	&paddq("xmm0", "xmm4");
	&movdqa("xmm2", &DWP(368,"esp"));
	&pmuludq("xmm2", "xmm7");
	&paddq("xmm0", "xmm3");
	&paddq("xmm5", "xmm2");
	&movdqa("xmm2", &DWP(192,"esp"));
	&movdqa("xmm4", &DWP(368,"esp"));
	&pmuludq("xmm4", "xmm2");
	&movdqa("xmm7", &DWP(352,"esp"));
	&movdqa("xmm1", &DWP(352,"esp"));
	&pmuludq("xmm1", &DWP(240,"esp"));
	&pmuludq("xmm7", "xmm2");
	&paddq("xmm0", "xmm4");
	&paddq("xmm5", "xmm7");
	&paddq("xmm0", "xmm1");
	&movdqa("xmm4", &DWP(288,"esp"));
	&movdqa("xmm3", &DWP(304,"esp"));
	&movdqa("xmm2", &DWP(320,"esp"));
	&je(&label("poly1305_blocks_sse2_21"));
	&set_label("poly1305_blocks_sse2_20");
	&movdqu("xmm1", &DWP(0,"edx"));
	&movdqu("xmm7", &DWP(16,"edx"));
	&movdqa("xmm6", "xmm1");
	&movaps(&DWP(400,"esp"), "xmm0");
	&punpckldq("xmm6", "xmm7");
	&pxor("xmm0", "xmm0");
	&punpckhdq("xmm1", "xmm7");
	&movdqa("xmm7", "xmm6");
	&punpckhdq("xmm6", "xmm0");
	&psllq("xmm6", 6);
	&paddq("xmm3", "xmm6");
	&movdqa("xmm6", "xmm1");
	&punpckldq("xmm6", "xmm0");
	&punpckhdq("xmm1", "xmm0");
	&psllq("xmm6", 12);
	&punpckldq("xmm7", "xmm0");
	&psllq("xmm1", 18);
	&movaps("xmm0", &DWP(400,"esp"));
	&paddq("xmm0", &DWP(272,"esp"));
	&paddq("xmm4", "xmm7");
	&paddq("xmm2", "xmm6");
	&paddq("xmm5", "xmm1");
	&set_label("poly1305_blocks_sse2_21");
	&movdqa("xmm6", "xmm5");
	&movdqa("xmm7", "xmm4");
	&psrlq("xmm6", 26);
	&psrlq("xmm7", 26);
	&paddq("xmm0", "xmm6");
	&paddq("xmm3", "xmm7");
	&movaps("xmm7", "xmm0");
	&movdqa("xmm1", "xmm3");
	&psrlq("xmm7", 26);
	&psrlq("xmm1", 26);
	&pmuludq("xmm7", &DWP(160,"esp"));
	&paddq("xmm2", "xmm1");
	&movdqa("xmm1", &DWP(256,"esp"));
	&movdqa("xmm6", "xmm2");
	&pand("xmm5", &DWP(256,"esp"));
	&psrlq("xmm6", 26);
	&pand("xmm3", "xmm1");
	&pand("xmm4", &DWP(256,"esp"));
	&paddq("xmm5", "xmm6");
	&paddq("xmm4", "xmm7");
	&movdqa("xmm6", "xmm3");
	&movdqa("xmm3", "xmm1");
	&movdqa("xmm7", "xmm1");
	&pand("xmm3", "xmm2");
	&movdqa("xmm2", "xmm1");
	&pand("xmm7", "xmm4");
	&psrlq("xmm4", 26);
	&pand("xmm1", "xmm5");
	&pand("xmm0", "xmm2");
	&psrlq("xmm5", 26);
	&paddq("xmm6", "xmm4");
	&paddq("xmm0", "xmm5");
	&movdqa(&DWP(32,"esp"), "xmm3");
	&set_label("poly1305_blocks_sse2_22");
	&test("edx", "edx");
	&je(&label("poly1305_blocks_sse2_24"));
	&set_label("poly1305_blocks_sse2_23");
	&pshufd("xmm2", "xmm7", 8);
	&pshufd("xmm6", "xmm6", 8);
	&pshufd("xmm3", &DWP(32,"esp"), 8);
	&pshufd("xmm1", "xmm1", 8);
	&punpcklqdq("xmm2", "xmm6");
	&punpcklqdq("xmm3", "xmm1");
	&pshufd("xmm0", "xmm0", 8);
	&movdqu(&DWP(0,"eax"), "xmm2");
	&movdqu(&DWP(16,"eax"), "xmm3");
	&movq(&DWP(32,"eax"), "xmm0");
	&add("esp", 544);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
	&set_label("poly1305_blocks_sse2_24");
	&movdqa("xmm2", "xmm7");
	&movdqa("xmm3", "xmm6");
	&psrldq("xmm2", 8);
	&paddq("xmm7", "xmm2");
	&psrldq("xmm3", 8);
	&movd("ecx", "xmm7");
	&paddq("xmm6", "xmm3");
	&movdqa("xmm5", &DWP(32,"esp"));
	&mov("esi", "ecx");
	&movdqa("xmm4", "xmm5");
	&and("ecx", 67108863);
	&movd("ebx", "xmm6");
	&movaps("xmm6", "xmm1");
	&psrldq("xmm4", 8);
	&paddq("xmm5", "xmm4");
	&shr("esi", 26);
	&add("ebx", "esi");
	&psrldq("xmm6", 8);
	&movd("edi", "xmm5");
	&paddq("xmm1", "xmm6");
	&mov(&DWP(0,"esp"), "eax");
	&mov("eax", "ebx");
	&shr("eax", 26);
	&and("ebx", 67108863);
	&add("edi", "eax");
	&movd("eax", "xmm1");
	&movaps("xmm1", "xmm0");
	&psrldq("xmm1", 8);
	&paddq("xmm0", "xmm1");
	&mov("edx", "edi");
	&and("edi", 67108863);
	&shr("edx", 26);
	&add("eax", "edx");
	&movd("edx", "xmm0");
	&mov("esi", "eax");
	&shr("esi", 26);
	&and("eax", 67108863);
	&add("edx", "esi");
	&mov("esi", "edx");
	&and("edx", 67108863);
	&shr("esi", 26);
	&lea("esi", &DWP(0,"esi","esi",4));
	&add("ecx", "esi");
	&mov("esi", "ecx");
	&and("ecx", 67108863);
	&shr("esi", 26);
	&add("ebx", "esi");
	&mov("esi", "ebx");
	&and("ebx", 67108863);
	&shr("esi", 26);
	&add("edi", "esi");
	&mov("esi", "edi");
	&shr("edi", 26);
	&and("esi", 67108863);
	&add("eax", "edi");
	&mov("edi", "eax");
	&shr("eax", 26);
	&and("edi", 67108863);
	&add("edx", "eax");
	&mov("eax", "edx");
	&shr("edx", 26);
	&and("eax", 67108863);
	&mov(&DWP(8,"esp"), "eax");
	&mov(&DWP(4,"esp"), "edi");
	&lea("edx", &DWP(0,"edx","edx",4));
	&add("ecx", "edx");
	&mov("edx", "ecx");
	&and("edx", 67108863);
	&shr("ecx", 26);
	&add("ebx", "ecx");
	&lea("ecx", &DWP(5,"edx"));
	&mov(&DWP(12,"esp"), "ecx");
	&shr("ecx", 26);
	&add("ecx", "ebx");
	&mov(&DWP(16,"esp"), "ecx");
	&shr("ecx", 26);
	&add("ecx", "esi");
	&mov(&DWP(20,"esp"), "ecx");
	&shr("ecx", 26);
	&add("ecx", "edi");
	&mov(&DWP(24,"esp"), "ecx");
	&shr("ecx", 26);
	&mov("edi", &DWP(12,"esp"));
	&and("edi", 67108863);
	&lea("eax", &DWP(-67108864,"ecx","eax"));
	&mov(&DWP(28,"esp"), "eax");
	&shr("eax", 31);
	&dec("eax");
	&mov("ecx", "eax");
	&and("edi", "eax");
	&not("ecx");
	&and("edx", "ecx");
	&and("ebx", "ecx");
	&or("edx", "edi");
	&and("esi", "ecx");
	&mov("edi", &DWP(0,"esp"));
	&mov(&DWP(0,"edi"), "edx");
	&mov("edx", &DWP(16,"esp"));
	&and("edx", 67108863);
	&and("edx", "eax");
	&or("ebx", "edx");
	&mov(&DWP(4,"edi"), "ebx");
	&mov("ebx", &DWP(20,"esp"));
	&and("ebx", 67108863);
	&and("ebx", "eax");
	&mov("edx", &DWP(24,"esp"));
	&or("esi", "ebx");
	&and("edx", 67108863);
	&mov(&DWP(8,"edi"), "esi");
	&and("edx", "eax");
	&mov("esi", &DWP(4,"esp"));
	&and("esi", "ecx");
	&or("esi", "edx");
	&mov("edx", &DWP(28,"esp"));
	&and("ecx", &DWP(8,"esp"));
	&and("edx", "eax");
	&or("ecx", "edx");
	&mov(&DWP(12,"edi"), "esi");
	&mov(&DWP(16,"edi"), "ecx");
	&set_label("poly1305_blocks_sse2_25");
	&add("esp", 544);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
}

sub poly1305_init_ex_sse2 {
	&mov("eax", &DWP(4,"esp"));
	&mov("edx", &DWP(8,"esp"));
	&mov("ecx", &DWP(12,"esp"));
	&static_label("poly1305_init_ex_sse2_local");
	&set_label("poly1305_init_ex_sse2_local");
	&push("esi");
	&push("edi");
	&push("ebx");
	&push("ebp");
	&sub("esp", 76);
	&mov("ebx", "edx");
	&mov("edx", -1);
	&test("ecx", "ecx");
	&pxor("xmm0", "xmm0");
	&movdqu(&DWP(0,"eax"), "xmm0");
	&movdqu(&DWP(16,"eax"), "xmm0");
	&movdqu(&DWP(32,"eax"), "xmm0");
	&cmove("ecx", "edx");
	&mov("edx", &DWP(4,"ebx"));
	&mov("ebp", "edx");
	&mov("edi", &DWP(0,"ebx"));
	&mov("esi", "edi");
	&shr("edi", 26);
	&and("esi", 67108863);
	&shl("ebp", 6);
	&mov(&DWP(12,"esp"), "ecx");
	&or("edi", "ebp");
	&mov("ecx", &DWP(8,"ebx"));
	&mov("ebp", "ecx");
	&shr("edx", 20);
	&and("edi", 67108611);
	&shl("ebp", 12);
	&mov(&DWP(0,"esp"), "ebx");
	&or("edx", "ebp");
	&mov("ebx", &DWP(12,"ebx"));
	&mov("ebp", "ebx");
	&shr("ecx", 14);
	&and("edx", 67092735);
	&shl("ebp", 18);
	&or("ecx", "ebp");
	&mov("ebp", &DWP(0,"esp"));
	&and("ecx", 66076671);
	&shr("ebx", 8);
	&and("ebx", 1048575);
	&mov(&DWP(40,"eax"), "esi");
	&mov(&DWP(44,"eax"), "edi");
	&mov(&DWP(48,"eax"), "edx");
	&mov(&DWP(52,"eax"), "ecx");
	&mov(&DWP(56,"eax"), "ebx");
	&mov(&DWP(20,"esp"), "esi");
	&mov("esi", &DWP(16,"ebp"));
	&mov(&DWP(100,"eax"), "esi");
	&mov(&DWP(24,"esp"), "edi");
	&mov("edi", &DWP(20,"ebp"));
	&mov(&DWP(104,"eax"), "edi");
	&mov("esi", &DWP(24,"ebp"));
	&mov(&DWP(108,"eax"), "esi");
	&lea("esi", &DWP(80,"eax"));
	&mov("ebp", &DWP(28,"ebp"));
	&mov("edi", &DWP(12,"esp"));
	&cmp("edi", 16);
	&mov(&DWP(112,"eax"), "ebp");
	&lea("ebp", &DWP(60,"eax"));
	&mov(&DWP(28,"esp"), 0);
	&mov(&DWP(16,"esp"), "ebp");
	&jbe(&label("poly1305_init_ex_sse2_9"));
	&set_label("poly1305_init_ex_sse2_2");
	&mov(&DWP(8,"esp"), "ebp");
	&mov(&DWP(0,"esp"), "esi");
	&mov(&DWP(32,"esp"), "ebx");
	&mov(&DWP(40,"esp"), "ecx");
	&mov(&DWP(36,"esp"), "edx");
	&mov(&DWP(12,"esp"), "edi");
	&mov(&DWP(4,"esp"), "eax");
	&set_label("poly1305_init_ex_sse2_3");
	&mov("ebp", &DWP(40,"esp"));
	&mov("eax", &DWP(20,"esp"));
	&mul("eax");
	&mov("ebx", &DWP(32,"esp"));
	&lea("esi", &DWP(0,"ebp","ebp",4));
	&mov("ebp", &DWP(24,"esp"));
	&mov("ecx", "eax");
	&mov(&DWP(48,"esp"), "esi");
	&lea("edi", &DWP(0,"ebx","ebx"));
	&mov("ebx", "edx");
	&mov(&DWP(44,"esp"), "edi");
	&lea("eax", &DWP(0,"ebp","ebp",4));
	&mul("edi");
	&mov("edi", &DWP(36,"esp"));
	&add("ecx", "eax");
	&mov(&DWP(60,"esp"), "edi");
	&adc("ebx", "edx");
	&lea("eax", &DWP(0,"edi","edi"));
	&mul("esi");
	&add("ecx", "eax");
	&lea("esi", &DWP(0,"ebp","ebp"));
	&mov("eax", &DWP(20,"esp"));
	&adc("ebx", "edx");
	&mov("ebp", &DWP(40,"esp"));
	&mov(&DWP(56,"esp"), "esi");
	&lea("edx", &DWP(0,"eax","eax"));
	&mov(&DWP(68,"esp"), "edx");
	&mul("esi");
	&mov("esi", "edx");
	&add("ebp", "ebp");
	&mov(&DWP(64,"esp"), "ebp");
	&mov("ebp", "eax");
	&mov(&DWP(52,"esp"), "ecx");
	&lea("eax", &DWP(0,"edi","edi",4));
	&mul(&DWP(44,"esp"));
	&add("ebp", "eax");
	&mov("eax", &DWP(40,"esp"));
	&mov("edi", &DWP(48,"esp"));
	&adc("esi", "edx");
	&mul("edi");
	&shl("ebx", 6);
	&shr("ecx", 26);
	&or("ebx", "ecx");
	&add("eax", "ebx");
	&adc("edx", 0);
	&add("ebp", "eax");
	&mov("eax", &DWP(24,"esp"));
	&adc("esi", "edx");
	&mul("eax");
	&mov("ecx", "eax");
	&mov("ebx", "edx");
	&mov("eax", &DWP(60,"esp"));
	&mul(&DWP(68,"esp"));
	&add("ecx", "eax");
	&mov("eax", &DWP(44,"esp"));
	&adc("ebx", "edx");
	&mul("edi");
	&mov(&DWP(72,"esp"), "ebp");
	&shl("esi", 6);
	&shr("ebp", 26);
	&or("esi", "ebp");
	&add("eax", "esi");
	&adc("edx", 0);
	&add("ecx", "eax");
	&mov("edi", "ecx");
	&adc("ebx", "edx");
	&and("edi", 67108863);
	&shl("ebx", 6);
	&shr("ecx", 26);
	&mov("eax", &DWP(56,"esp"));
	&or("ebx", "ecx");
	&mov("ecx", &DWP(60,"esp"));
	&mul("ecx");
	&mov(&DWP(36,"esp"), "edi");
	&mov("esi", "eax");
	&mov("eax", &DWP(20,"esp"));
	&mov("ebp", "edx");
	&mov("edi", &DWP(64,"esp"));
	&mul("edi");
	&add("esi", "eax");
	&mov("eax", &DWP(32,"esp"));
	&adc("ebp", "edx");
	&lea("edx", &DWP(0,"eax","eax",4));
	&mul("edx");
	&add("esi", "eax");
	&mov("eax", "ecx");
	&adc("ebp", "edx");
	&add("ebx", "esi");
	&mov("esi", "ebx");
	&adc("ebp", 0);
	&and("esi", 67108863);
	&mul("ecx");
	&shl("ebp", 6);
	&mov("ecx", "eax");
	&shr("ebx", 26);
	&mov("eax", "edi");
	&or("ebp", "ebx");
	&mov("ebx", "edx");
	&mul(&DWP(24,"esp"));
	&add("ecx", "eax");
	&mov("eax", &DWP(68,"esp"));
	&adc("ebx", "edx");
	&mul(&DWP(32,"esp"));
	&add("ecx", "eax");
	&mov("edi", &DWP(52,"esp"));
	&adc("ebx", "edx");
	&add("ebp", "ecx");
	&mov("ecx", "ebp");
	&adc("ebx", 0);
	&and("edi", 67108863);
	&shl("ebx", 6);
	&and("ecx", 67108863);
	&shr("ebp", 26);
	&or("ebx", "ebp");
	&mov("eax", &DWP(72,"esp"));
	&and("eax", 67108863);
	&mov(&DWP(32,"esp"), "ecx");
	&mov("ecx", &DWP(28,"esp"));
	&lea("ebx", &DWP(0,"ebx","ebx",4));
	&add("edi", "ebx");
	&inc("ecx");
	&mov("ebp", "edi");
	&shr("edi", 26);
	&and("ebp", 67108863);
	&mov(&DWP(40,"esp"), "esi");
	&cmp("ecx", 2);
	&mov(&DWP(20,"esp"), "ebp");
	&mov(&DWP(28,"esp"), "ecx");
	&lea("edx", &DWP(0,"eax","edi"));
	&mov(&DWP(24,"esp"), "edx");
	&jae(&label("poly1305_init_ex_sse2_8"));
	&set_label("poly1305_init_ex_sse2_4");
	&cmp(&DWP(28,"esp"), 0);
	&jne(&label("poly1305_init_ex_sse2_6"));
	&set_label("poly1305_init_ex_sse2_5");
	&mov("esi", &DWP(8,"esp"));
	&mov("eax", &DWP(32,"esp"));
	&mov("edx", &DWP(40,"esp"));
	&mov("ecx", &DWP(36,"esp"));
	&mov("ebx", &DWP(24,"esp"));
	&mov("edi", &DWP(16,"esp"));
	&mov(&DWP(16,"esi"), "eax");
	&mov(&DWP(12,"esi"), "edx");
	&mov(&DWP(8,"esi"), "ecx");
	&mov(&DWP(4,"esi"), "ebx");
	&mov(&DWP(0,"esi"), "ebp");
	&mov(&DWP(8,"esp"), "edi");
	&jmp(&label("poly1305_init_ex_sse2_3"));
	&set_label("poly1305_init_ex_sse2_6");
	&cmp(&DWP(28,"esp"), 1);
	&jne(&label("poly1305_init_ex_sse2_3"));
	&set_label("poly1305_init_ex_sse2_7");
	&mov("esi", &DWP(8,"esp"));
	&mov("eax", &DWP(32,"esp"));
	&mov("edx", &DWP(40,"esp"));
	&mov("ecx", &DWP(36,"esp"));
	&mov("ebx", &DWP(24,"esp"));
	&mov("edi", &DWP(0,"esp"));
	&mov(&DWP(16,"esi"), "eax");
	&mov(&DWP(12,"esi"), "edx");
	&mov(&DWP(8,"esi"), "ecx");
	&mov(&DWP(4,"esi"), "ebx");
	&mov(&DWP(0,"esi"), "ebp");
	&mov(&DWP(8,"esp"), "edi");
	&cmp(&DWP(12,"esp"), 96);
	&jae(&label("poly1305_init_ex_sse2_3"));
	&jmp(&label("poly1305_init_ex_sse2_10"));
	&set_label("poly1305_init_ex_sse2_8");
	&mov("ebp", &DWP(8,"esp"));
	&mov("ecx", "esi");
	&mov("edx", &DWP(36,"esp"));
	&mov(&DWP(12,"ebp"), "ecx");
	&mov(&DWP(8,"ebp"), "edx");
	&mov("ebx", &DWP(32,"esp"));
	&mov("edx", &DWP(24,"esp"));
	&mov("ecx", &DWP(20,"esp"));
	&mov("eax", &DWP(4,"esp"));
	&mov(&DWP(16,"ebp"), "ebx");
	&mov(&DWP(4,"ebp"), "edx");
	&mov(&DWP(0,"ebp"), "ecx");
	&set_label("poly1305_init_ex_sse2_9");
	&mov(&DWP(116,"eax"), 0);
	&add("esp", 76);
	&pop("ebp");
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
	&set_label("poly1305_init_ex_sse2_10");
	&mov("eax", &DWP(4,"esp"));
	&jmp(&label("poly1305_init_ex_sse2_9"));
}

sub poly1305_sse2 {
	#&cmp(&DWP(12,"esp"), 64);
	#&jb(&label("poly1305_x86_local"));
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", -64);
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 244);
	&mov("esi", &DWP(16,"ebp"));
	&lea("eax", &DWP(64,"esp"));
	&mov("ecx", "esi");
	&mov("edx", &DWP(20,"ebp"));
	&mov("edi", &DWP(12,"ebp"));
	&call(&LABEL("poly1305_init_ex_sse2_local"));
	&set_label("poly1305_sse2_2");
	&mov("ebx", "esi");
	&and("ebx", -32);
	&je(&label("poly1305_sse2_5"));
	&set_label("poly1305_sse2_3");
	&mov("edx", "edi");
	&lea("eax", &DWP(64,"esp"));
	&mov("ecx", "ebx");
	&call(&LABEL("poly1305_blocks_sse2_local"));
	&set_label("poly1305_sse2_4");
	&add("edi", "ebx");
	&sub("esi", "ebx");
	&set_label("poly1305_sse2_5");
	&push(&DWP(8,"ebp"));
	&push("esi");
	&push("edi");
	&lea("eax", &DWP(76,"esp"));
	&push("eax");
	&call(&LABEL("poly1305_finish_ex_sse2_local"));
	&set_label("poly1305_sse2_6");
	&add("esp", 260);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
}



sub poly1305_blocks_avx {
	&static_label("poly1305_blocks_avx_local");
	&set_label("poly1305_blocks_avx_local");
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 560);
	&mov("ebx", 16777216);
	&mov("esi", 67108863);
	&mov("edi", 5);
	&mov("ecx", &DWP(576,"esp"));
	&mov("eax", &DWP(580,"esp"));
	&vmovd("xmm0", "ebx");
	&vmovd("xmm2", "esi");
	&vmovd("xmm4", "edi");
	&mov("esi", &DWP(116,"ecx"));
	&vpshufd("xmm1", "xmm0", 68);
	&vpshufd("xmm3", "xmm2", 68);
	&vpshufd("xmm5", "xmm4", 68);
	&mov("edx", &DWP(584,"esp"));
	&vmovdqu(&DWP(96,"esp"), "xmm1");
	&vmovdqu(&DWP(80,"esp"), "xmm3");
	&vmovdqu(&DWP(48,"esp"), "xmm5");
	&test("esi", 4);
	&je(&label("poly1305_blocks_avx_3"));
	&set_label("poly1305_blocks_avx_2");
	&vmovdqu("xmm0", &DWP(96,"esp"));
	&vpsrldq("xmm1", "xmm0", 8);
	&vmovdqu(&DWP(96,"esp"), "xmm1");
	&set_label("poly1305_blocks_avx_3");
	&test("esi", 8);
	&je(&label("poly1305_blocks_avx_5"));
	&set_label("poly1305_blocks_avx_4");
	&vpxor("xmm0", "xmm0", "xmm0");
	&vmovdqu(&DWP(96,"esp"), "xmm0");
	&set_label("poly1305_blocks_avx_5");
	&mov("ebx", "esi");
	&bts("ebx", 0);
	&jc(&label("poly1305_blocks_avx_7"));
	&set_label("poly1305_blocks_avx_6");
	&vmovq("xmm0", &DWP(0,"eax"));
	&mov("esi", "ebx");
	&vmovq("xmm1", &DWP(8,"eax"));
	&add("edx", -32);
	&vmovhpd("xmm3", "xmm0", &DWP(16,"eax"));
	&vmovhpd("xmm2", "xmm1", &DWP(24,"eax"));
	&vpsrlq("xmm4", "xmm3", 52);
	&vmovdqu("xmm5", &DWP(80,"esp"));
	&vpsllq("xmm7", "xmm2", 12);
	&vpand("xmm0", "xmm5", "xmm3");
	&vpsrlq("xmm6", "xmm3", 26);
	&vpor("xmm3", "xmm4", "xmm7");
	&vpsrlq("xmm2", "xmm2", 40);
	&vpsrlq("xmm4", "xmm3", 26);
	&vpand("xmm1", "xmm5", "xmm6");
	&mov(&DWP(116,"ecx"), "esi");
	&add("eax", 32);
	&vpand("xmm6", "xmm5", "xmm3");
	&vpand("xmm5", "xmm5", "xmm4");
	&vpor("xmm3", "xmm2", &DWP(96,"esp"));
	&jmp(&label("poly1305_blocks_avx_8"));
	&set_label("poly1305_blocks_avx_7");
	&vmovdqu("xmm1", &DWP(0,"ecx"));
	&vmovdqu("xmm2", &DWP(16,"ecx"));
	&vpshufd("xmm3", &DWP(32,"ecx"), 80);
	&vpshufd("xmm0", "xmm1", 80);
	&vpshufd("xmm1", "xmm1", 250);
	&vpshufd("xmm6", "xmm2", 80);
	&vpshufd("xmm5", "xmm2", 250);
	&set_label("poly1305_blocks_avx_8");
	&test("esi", 48);
	&je(&label("poly1305_blocks_avx_13"));
	&set_label("poly1305_blocks_avx_9");
	&vmovdqu("xmm2", &DWP(40,"ecx"));
	&mov("edi", &DWP(56,"ecx"));
	&test("esi", 16);
	&je(&label("poly1305_blocks_avx_11"));
	&set_label("poly1305_blocks_avx_10");
	&vmovdqu("xmm7", &DWP(60,"ecx"));
	&vpunpckldq("xmm4", "xmm7", "xmm2");
	&vpunpckhdq("xmm2", "xmm7", "xmm2");
	&vmovd("xmm7", "edi");
	&vmovdqu(&DWP(0,"esp"), "xmm2");
	&vmovd("xmm2", &DWP(76,"ecx"));
	&vpunpcklqdq("xmm2", "xmm2", "xmm7");
	&vmovdqu(&DWP(64,"esp"), "xmm2");
	&vmovdqu("xmm2", &DWP(0,"esp"));
	&jmp(&label("poly1305_blocks_avx_12"));
	&set_label("poly1305_blocks_avx_11");
	&mov("ebx", 1);
	&vmovd("xmm7", "ebx");
	&vpunpckldq("xmm4", "xmm2", "xmm7");
	&vpunpckhdq("xmm2", "xmm2", "xmm7");
	&vmovd("xmm7", "edi");
	&vmovdqu(&DWP(64,"esp"), "xmm7");
	&set_label("poly1305_blocks_avx_12");
	&vpshufd("xmm7", "xmm4", 80);
	&vpshufd("xmm4", "xmm4", 250);
	&vmovdqu(&DWP(176,"esp"), "xmm4");
	&vpshufd("xmm4", "xmm2", 80);
	&vpshufd("xmm2", "xmm2", 250);
	&vmovdqu(&DWP(112,"esp"), "xmm7");
	&vmovdqu(&DWP(144,"esp"), "xmm4");
	&vmovdqu(&DWP(128,"esp"), "xmm2");
	&jmp(&label("poly1305_blocks_avx_14"));
	&set_label("poly1305_blocks_avx_13");
	&vpshufd("xmm4", &DWP(60,"ecx"), 0);
	&vpshufd("xmm7", &DWP(60,"ecx"), 85);
	&vmovd("xmm2", &DWP(76,"ecx"));
	&vmovdqu(&DWP(112,"esp"), "xmm4");
	&vmovdqu(&DWP(176,"esp"), "xmm7");
	&vpshufd("xmm4", &DWP(60,"ecx"), 170);
	&vpshufd("xmm7", &DWP(60,"ecx"), 255);
	&vpshufd("xmm2", "xmm2", 0);
	&vmovdqu(&DWP(144,"esp"), "xmm4");
	&vmovdqu(&DWP(128,"esp"), "xmm7");
	&vmovdqu(&DWP(64,"esp"), "xmm2");
	&set_label("poly1305_blocks_avx_14");
	&vmovdqu("xmm4", &DWP(48,"esp"));
	&vpmuludq("xmm2", "xmm4", &DWP(176,"esp"));
	&vpmuludq("xmm7", "xmm4", &DWP(128,"esp"));
	&vmovdqu(&DWP(208,"esp"), "xmm2");
	&vpmuludq("xmm2", "xmm4", &DWP(144,"esp"));
	&vpmuludq("xmm4", "xmm4", &DWP(64,"esp"));
	&vmovdqu(&DWP(192,"esp"), "xmm7");
	&vmovdqu(&DWP(160,"esp"), "xmm4");
	&cmp("edx", 64);
	&jb(&label("poly1305_blocks_avx_18"));
	&set_label("poly1305_blocks_avx_15");
	&vmovdqu(&DWP(240,"esp"), "xmm6");
	&vpshufd("xmm6", &DWP(80,"ecx"), 0);
	&vmovd("xmm7", &DWP(96,"ecx"));
	&vmovdqu(&DWP(0,"esp"), "xmm5");
	&vmovdqu(&DWP(400,"esp"), "xmm3");
	&vmovdqu(&DWP(224,"esp"), "xmm1");
	&vmovdqu(&DWP(304,"esp"), "xmm6");
	&vpshufd("xmm5", &DWP(80,"ecx"), 170);
	&vpshufd("xmm4", &DWP(80,"ecx"), 255);
	&vpshufd("xmm1", "xmm7", 0);
	&vmovdqu("xmm3", &DWP(48,"esp"));
	&vpshufd("xmm6", &DWP(80,"ecx"), 85);
	&vmovdqu(&DWP(352,"esp"), "xmm5");
	&vmovdqu(&DWP(336,"esp"), "xmm4");
	&vmovdqu(&DWP(320,"esp"), "xmm1");
	&vpmuludq("xmm5", "xmm5", "xmm3");
	&vpmuludq("xmm4", "xmm4", "xmm3");
	&vpmuludq("xmm1", "xmm1", "xmm3");
	&vpmuludq("xmm7", "xmm6", "xmm3");
	&vmovdqu(&DWP(384,"esp"), "xmm6");
	&vmovdqu(&DWP(272,"esp"), "xmm5");
	&vmovdqu(&DWP(32,"esp"), "xmm4");
	&vmovdqu(&DWP(16,"esp"), "xmm1");
	&vmovdqu(&DWP(288,"esp"), "xmm7");
	&vmovdqu("xmm3", &DWP(400,"esp"));
	&vmovdqu("xmm5", &DWP(0,"esp"));
	&vmovdqu("xmm6", &DWP(240,"esp"));
	&vmovdqu("xmm1", &DWP(224,"esp"));
	&vmovdqu(&DWP(368,"esp"), "xmm2");
	&set_label("poly1305_blocks_avx_16");
	&vmovq("xmm2", &DWP(0,"eax"));
	&add("edx", -64);
	&vmovdqu(&DWP(256,"esp"), "xmm0");
	&vmovhpd("xmm0", "xmm2", &DWP(16,"eax"));
	&vmovq("xmm4", &DWP(8,"eax"));
	&vmovdqu("xmm2", &DWP(80,"esp"));
	&vmovhpd("xmm7", "xmm4", &DWP(24,"eax"));
	&vpand("xmm4", "xmm2", "xmm0");
	&vmovdqu(&DWP(0,"esp"), "xmm4");
	&vpsrlq("xmm4", "xmm0", 26);
	&vpand("xmm4", "xmm2", "xmm4");
	&vpsrlq("xmm0", "xmm0", 52);
	&vmovdqu(&DWP(416,"esp"), "xmm4");
	&vpsrlq("xmm4", "xmm7", 14);
	&vpand("xmm4", "xmm2", "xmm4");
	&vmovdqu(&DWP(432,"esp"), "xmm4");
	&vpsllq("xmm4", "xmm7", 12);
	&vpor("xmm0", "xmm0", "xmm4");
	&vpsrlq("xmm7", "xmm7", 40);
	&vpand("xmm2", "xmm2", "xmm0");
	&vpor("xmm0", "xmm7", &DWP(96,"esp"));
	&vmovdqu("xmm7", &DWP(32,"eax"));
	&vpunpckldq("xmm4", "xmm7", &DWP(48,"eax"));
	&vpunpckhdq("xmm7", "xmm7", &DWP(48,"eax"));
	&add("eax", 64);
	&vmovdqu(&DWP(400,"esp"), "xmm3");
	&vmovdqu(&DWP(496,"esp"), "xmm7");
	&vpmuludq("xmm3", "xmm3", &DWP(288,"esp"));
	&vpmuludq("xmm7", "xmm5", &DWP(272,"esp"));
	&vmovdqu(&DWP(240,"esp"), "xmm6");
	&vpmuludq("xmm6", "xmm6", &DWP(32,"esp"));
	&vpaddq("xmm3", "xmm3", "xmm7");
	&vpaddq("xmm3", "xmm3", "xmm6");
	&vmovdqu("xmm6", &DWP(16,"esp"));
	&vmovdqu(&DWP(224,"esp"), "xmm1");
	&vpmuludq("xmm1", "xmm1", "xmm6");
	&vmovdqu("xmm7", &DWP(304,"esp"));
	&vpaddq("xmm1", "xmm3", "xmm1");
	&vpmuludq("xmm3", "xmm7", &DWP(256,"esp"));
	&vpmuludq("xmm7", "xmm5", "xmm7");
	&vpaddq("xmm1", "xmm1", "xmm3");
	&vmovdqu(&DWP(464,"esp"), "xmm0");
	&vpmuludq("xmm0", "xmm0", &DWP(208,"esp"));
	&vpaddq("xmm1", "xmm1", "xmm0");
	&vmovdqu("xmm0", &DWP(432,"esp"));
	&vpmuludq("xmm3", "xmm0", &DWP(368,"esp"));
	&vmovdqu(&DWP(448,"esp"), "xmm2");
	&vpmuludq("xmm2", "xmm2", &DWP(192,"esp"));
	&vpaddq("xmm0", "xmm1", "xmm3");
	&vmovdqu("xmm3", &DWP(160,"esp"));
	&vpmuludq("xmm1", "xmm3", &DWP(416,"esp"));
	&vpaddq("xmm0", "xmm0", "xmm2");
	&vpaddq("xmm2", "xmm0", "xmm1");
	&vmovdqu("xmm1", &DWP(112,"esp"));
	&vpmuludq("xmm0", "xmm1", &DWP(0,"esp"));
	&vpmuludq("xmm1", "xmm1", &DWP(432,"esp"));
	&vpaddq("xmm0", "xmm2", "xmm0");
	&vpxor("xmm2", "xmm2", "xmm2");
	&vmovdqu(&DWP(480,"esp"), "xmm4");
	&vpunpckldq("xmm4", "xmm4", "xmm2");
	&vpaddq("xmm0", "xmm0", "xmm4");
	&vmovdqu(&DWP(512,"esp"), "xmm0");
	&vmovdqu("xmm0", &DWP(400,"esp"));
	&vpmuludq("xmm6", "xmm0", "xmm6");
	&vpmuludq("xmm0", "xmm0", &DWP(272,"esp"));
	&vpaddq("xmm4", "xmm6", "xmm7");
	&vmovdqu("xmm6", &DWP(240,"esp"));
	&vpmuludq("xmm7", "xmm6", &DWP(384,"esp"));
	&vmovdqu("xmm6", &DWP(224,"esp"));
	&vpaddq("xmm4", "xmm4", "xmm7");
	&vpmuludq("xmm7", "xmm6", &DWP(352,"esp"));
	&vmovdqu("xmm6", &DWP(256,"esp"));
	&vpaddq("xmm4", "xmm4", "xmm7");
	&vpmuludq("xmm7", "xmm6", &DWP(336,"esp"));
	&vpaddq("xmm4", "xmm4", "xmm7");
	&vmovdqu("xmm7", &DWP(464,"esp"));
	&vpmuludq("xmm3", "xmm7", "xmm3");
	&vpmuludq("xmm7", "xmm7", &DWP(368,"esp"));
	&vpaddq("xmm3", "xmm4", "xmm3");
	&vpaddq("xmm3", "xmm3", "xmm1");
	&vmovdqu("xmm1", &DWP(448,"esp"));
	&vpmuludq("xmm4", "xmm1", &DWP(176,"esp"));
	&vmovdqu("xmm1", &DWP(416,"esp"));
	&vpaddq("xmm3", "xmm3", "xmm4");
	&vpmuludq("xmm4", "xmm1", &DWP(144,"esp"));
	&vmovdqu("xmm1", &DWP(0,"esp"));
	&vpaddq("xmm3", "xmm3", "xmm4");
	&vpmuludq("xmm4", "xmm1", &DWP(128,"esp"));
	&vpaddq("xmm4", "xmm3", "xmm4");
	&vmovdqu("xmm3", &DWP(496,"esp"));
	&vpunpckhdq("xmm3", "xmm3", "xmm2");
	&vpsllq("xmm3", "xmm3", 18);
	&vpaddq("xmm4", "xmm4", "xmm3");
	&vpmuludq("xmm3", "xmm5", &DWP(32,"esp"));
	&vmovdqu(&DWP(528,"esp"), "xmm4");
	&vpaddq("xmm4", "xmm0", "xmm3");
	&vmovdqu("xmm0", &DWP(240,"esp"));
	&vpmuludq("xmm0", "xmm0", &DWP(16,"esp"));
	&vpaddq("xmm3", "xmm4", "xmm0");
	&vmovdqu("xmm4", &DWP(304,"esp"));
	&vpmuludq("xmm0", "xmm4", &DWP(224,"esp"));
	&vpaddq("xmm0", "xmm3", "xmm0");
	&vmovdqu("xmm3", &DWP(384,"esp"));
	&vpmuludq("xmm6", "xmm6", "xmm3");
	&vpmuludq("xmm3", "xmm5", "xmm3");
	&vpmuludq("xmm5", "xmm5", &DWP(16,"esp"));
	&vpaddq("xmm6", "xmm0", "xmm6");
	&vpaddq("xmm0", "xmm6", "xmm7");
	&vmovdqu("xmm6", &DWP(432,"esp"));
	&vpmuludq("xmm7", "xmm6", &DWP(192,"esp"));
	&vmovdqu("xmm6", &DWP(448,"esp"));
	&vpaddq("xmm0", "xmm0", "xmm7");
	&vpmuludq("xmm7", "xmm6", &DWP(160,"esp"));
	&vpaddq("xmm6", "xmm0", "xmm7");
	&vmovdqu("xmm7", &DWP(112,"esp"));
	&vpmuludq("xmm0", "xmm7", &DWP(416,"esp"));
	&vpaddq("xmm0", "xmm6", "xmm0");
	&vmovdqu("xmm6", &DWP(176,"esp"));
	&vpmuludq("xmm1", "xmm1", "xmm6");
	&vpmuludq("xmm6", "xmm6", &DWP(432,"esp"));
	&vpaddq("xmm0", "xmm0", "xmm1");
	&vmovdqu("xmm1", &DWP(480,"esp"));
	&vpunpckhdq("xmm2", "xmm1", "xmm2");
	&vpsllq("xmm1", "xmm2", 6);
	&vpaddq("xmm2", "xmm0", "xmm1");
	&vmovdqu("xmm0", &DWP(512,"esp"));
	&vpsrlq("xmm0", "xmm0", 26);
	&vpaddq("xmm1", "xmm2", "xmm0");
	&vmovdqu("xmm0", &DWP(400,"esp"));
	&vpmuludq("xmm2", "xmm0", "xmm4");
	&vpmuludq("xmm0", "xmm0", &DWP(32,"esp"));
	&vpaddq("xmm2", "xmm2", "xmm3");
	&vpaddq("xmm0", "xmm0", "xmm5");
	&vmovdqu(&DWP(544,"esp"), "xmm1");
	&vmovdqu("xmm1", &DWP(240,"esp"));
	&vpmuludq("xmm3", "xmm1", &DWP(352,"esp"));
	&vpmuludq("xmm5", "xmm1", "xmm4");
	&vpaddq("xmm2", "xmm2", "xmm3");
	&vpaddq("xmm4", "xmm0", "xmm5");
	&vmovdqu("xmm3", &DWP(224,"esp"));
	&vpmuludq("xmm3", "xmm3", &DWP(336,"esp"));
	&vpaddq("xmm2", "xmm2", "xmm3");
	&vmovdqu("xmm3", &DWP(256,"esp"));
	&vpmuludq("xmm3", "xmm3", &DWP(320,"esp"));
	&vpaddq("xmm3", "xmm2", "xmm3");
	&vmovdqu("xmm2", &DWP(464,"esp"));
	&vpmuludq("xmm7", "xmm2", "xmm7");
	&vpmuludq("xmm2", "xmm2", &DWP(192,"esp"));
	&vpaddq("xmm3", "xmm3", "xmm7");
	&vpaddq("xmm7", "xmm3", "xmm6");
	&vmovdqu("xmm6", &DWP(448,"esp"));
	&vpmuludq("xmm3", "xmm6", &DWP(144,"esp"));
	&vpmuludq("xmm6", "xmm6", &DWP(112,"esp"));
	&vpaddq("xmm3", "xmm7", "xmm3");
	&vmovdqu("xmm7", &DWP(416,"esp"));
	&vpmuludq("xmm7", "xmm7", &DWP(128,"esp"));
	&vpaddq("xmm7", "xmm3", "xmm7");
	&vmovdqu("xmm3", &DWP(0,"esp"));
	&vpmuludq("xmm3", "xmm3", &DWP(64,"esp"));
	&vpaddq("xmm7", "xmm7", "xmm3");
	&vpaddq("xmm3", "xmm7", &DWP(96,"esp"));
	&vmovdqu("xmm7", &DWP(528,"esp"));
	&vmovdqu("xmm1", &DWP(224,"esp"));
	&vpsrlq("xmm7", "xmm7", 26);
	&vpaddq("xmm3", "xmm3", "xmm7");
	&vpmuludq("xmm7", "xmm1", &DWP(384,"esp"));
	&vmovdqu("xmm0", &DWP(256,"esp"));
	&vpmuludq("xmm1", "xmm0", &DWP(352,"esp"));
	&vpaddq("xmm5", "xmm4", "xmm7");
	&vmovdqu("xmm0", &DWP(432,"esp"));
	&vpaddq("xmm4", "xmm5", "xmm1");
	&vpmuludq("xmm1", "xmm0", &DWP(160,"esp"));
	&vpaddq("xmm5", "xmm4", "xmm2");
	&vpaddq("xmm2", "xmm5", "xmm1");
	&vpaddq("xmm0", "xmm2", "xmm6");
	&vmovdqu("xmm6", &DWP(416,"esp"));
	&vpmuludq("xmm5", "xmm6", &DWP(176,"esp"));
	&vmovdqu("xmm1", &DWP(0,"esp"));
	&vpxor("xmm6", "xmm6", "xmm6");
	&vpmuludq("xmm4", "xmm1", &DWP(144,"esp"));
	&vpaddq("xmm2", "xmm0", "xmm5");
	&vmovdqu("xmm7", &DWP(496,"esp"));
	&vpunpckldq("xmm0", "xmm7", "xmm6");
	&vpsrlq("xmm7", "xmm3", 26);
	&vpaddq("xmm5", "xmm2", "xmm4");
	&vpsllq("xmm1", "xmm0", 12);
	&vpaddq("xmm2", "xmm5", "xmm1");
	&vpmuludq("xmm5", "xmm7", &DWP(48,"esp"));
	&vmovdqu("xmm1", &DWP(544,"esp"));
	&vpsrlq("xmm4", "xmm1", 26);
	&vpaddq("xmm6", "xmm2", "xmm4");
	&vmovdqu("xmm4", &DWP(80,"esp"));
	&vpsrlq("xmm7", "xmm6", 26);
	&vpand("xmm0", "xmm4", &DWP(512,"esp"));
	&vpand("xmm1", "xmm1", "xmm4");
	&vpand("xmm2", "xmm4", &DWP(528,"esp"));
	&vpand("xmm3", "xmm3", "xmm4");
	&vpaddq("xmm5", "xmm0", "xmm5");
	&vpaddq("xmm2", "xmm2", "xmm7");
	&vpand("xmm0", "xmm5", "xmm4");
	&vpsrlq("xmm5", "xmm5", 26);
	&vpaddq("xmm1", "xmm1", "xmm5");
	&vpand("xmm5", "xmm2", "xmm4");
	&vpsrlq("xmm2", "xmm2", 26);
	&vpaddq("xmm3", "xmm3", "xmm2");
	&vpand("xmm6", "xmm6", "xmm4");
	&cmp("edx", 64);
	&jae(&label("poly1305_blocks_avx_16"));
	&set_label("poly1305_blocks_avx_17");
	&vmovdqu("xmm2", &DWP(368,"esp"));
	&set_label("poly1305_blocks_avx_18");
	&cmp("edx", 32);
	&jb(&label("poly1305_blocks_avx_22"));
	&set_label("poly1305_blocks_avx_19");
	&vpmuludq("xmm4", "xmm3", &DWP(208,"esp"));
	&vpmuludq("xmm7", "xmm5", "xmm2");
	&vpaddq("xmm7", "xmm4", "xmm7");
	&vmovdqu("xmm4", &DWP(192,"esp"));
	&vmovdqu(&DWP(240,"esp"), "xmm6");
	&vpmuludq("xmm6", "xmm6", "xmm4");
	&vpaddq("xmm7", "xmm7", "xmm6");
	&vmovdqu("xmm6", &DWP(160,"esp"));
	&vmovdqu(&DWP(224,"esp"), "xmm1");
	&vpmuludq("xmm1", "xmm1", "xmm6");
	&vpaddq("xmm7", "xmm7", "xmm1");
	&vmovdqu("xmm1", &DWP(112,"esp"));
	&vmovdqu(&DWP(256,"esp"), "xmm0");
	&vpmuludq("xmm0", "xmm0", "xmm1");
	&vpaddq("xmm0", "xmm7", "xmm0");
	&vpmuludq("xmm7", "xmm3", "xmm2");
	&vpmuludq("xmm2", "xmm5", "xmm4");
	&vpmuludq("xmm4", "xmm3", "xmm4");
	&vmovdqu(&DWP(32,"esp"), "xmm0");
	&vpaddq("xmm0", "xmm7", "xmm2");
	&vpmuludq("xmm7", "xmm6", &DWP(240,"esp"));
	&vpaddq("xmm2", "xmm0", "xmm7");
	&vpmuludq("xmm0", "xmm1", &DWP(224,"esp"));
	&vpaddq("xmm7", "xmm2", "xmm0");
	&vmovdqu("xmm2", &DWP(176,"esp"));
	&vpmuludq("xmm0", "xmm2", &DWP(256,"esp"));
	&vpaddq("xmm7", "xmm7", "xmm0");
	&vpmuludq("xmm0", "xmm5", "xmm6");
	&vpmuludq("xmm6", "xmm3", "xmm6");
	&vpmuludq("xmm3", "xmm3", "xmm1");
	&vmovdqu(&DWP(0,"esp"), "xmm7");
	&vpaddq("xmm7", "xmm4", "xmm0");
	&vpmuludq("xmm4", "xmm1", &DWP(240,"esp"));
	&vpaddq("xmm0", "xmm7", "xmm4");
	&vpmuludq("xmm7", "xmm2", &DWP(224,"esp"));
	&vmovdqu("xmm4", &DWP(144,"esp"));
	&vpaddq("xmm0", "xmm0", "xmm7");
	&vpmuludq("xmm7", "xmm4", &DWP(256,"esp"));
	&vpaddq("xmm0", "xmm0", "xmm7");
	&vmovdqu(&DWP(16,"esp"), "xmm0");
	&vpmuludq("xmm0", "xmm5", "xmm1");
	&vpmuludq("xmm5", "xmm5", "xmm2");
	&vpaddq("xmm7", "xmm6", "xmm0");
	&vpaddq("xmm1", "xmm3", "xmm5");
	&vpmuludq("xmm6", "xmm2", &DWP(240,"esp"));
	&vpmuludq("xmm2", "xmm4", &DWP(240,"esp"));
	&vpaddq("xmm0", "xmm7", "xmm6");
	&vpaddq("xmm3", "xmm1", "xmm2");
	&vpmuludq("xmm7", "xmm4", &DWP(224,"esp"));
	&vpaddq("xmm6", "xmm0", "xmm7");
	&vmovdqu("xmm7", &DWP(128,"esp"));
	&vpmuludq("xmm0", "xmm7", &DWP(256,"esp"));
	&vpmuludq("xmm4", "xmm7", &DWP(224,"esp"));
	&vpaddq("xmm0", "xmm6", "xmm0");
	&vpaddq("xmm6", "xmm3", "xmm4");
	&vmovdqu("xmm5", &DWP(256,"esp"));
	&vpmuludq("xmm7", "xmm5", &DWP(64,"esp"));
	&vmovdqu("xmm1", &DWP(32,"esp"));
	&vpaddq("xmm4", "xmm6", "xmm7");
	&test("eax", "eax");
	&je(&label("poly1305_blocks_avx_21"));
	&set_label("poly1305_blocks_avx_20");
	&vmovdqu("xmm5", &DWP(0,"eax"));
	&vpaddq("xmm4", "xmm4", &DWP(96,"esp"));
	&vpunpckldq("xmm7", "xmm5", &DWP(16,"eax"));
	&vpxor("xmm3", "xmm3", "xmm3");
	&vpunpckhdq("xmm2", "xmm5", &DWP(16,"eax"));
	&vpunpckldq("xmm6", "xmm7", "xmm3");
	&vpunpckhdq("xmm5", "xmm7", "xmm3");
	&vpaddq("xmm1", "xmm1", "xmm6");
	&vpsllq("xmm6", "xmm5", 6);
	&vpunpckldq("xmm5", "xmm2", "xmm3");
	&vpaddq("xmm7", "xmm6", &DWP(0,"esp"));
	&vpunpckhdq("xmm2", "xmm2", "xmm3");
	&vpsllq("xmm6", "xmm5", 12);
	&vmovdqu(&DWP(0,"esp"), "xmm7");
	&vpsllq("xmm3", "xmm2", 18);
	&vpaddq("xmm7", "xmm6", &DWP(16,"esp"));
	&vpaddq("xmm0", "xmm0", "xmm3");
	&vmovdqu(&DWP(16,"esp"), "xmm7");
	&set_label("poly1305_blocks_avx_21");
	&vpsrlq("xmm5", "xmm0", 26);
	&vpsrlq("xmm7", "xmm1", 26);
	&vpaddq("xmm2", "xmm4", "xmm5");
	&vpaddq("xmm6", "xmm7", &DWP(0,"esp"));
	&vmovdqu("xmm3", &DWP(80,"esp"));
	&vpsrlq("xmm4", "xmm6", 26);
	&vpand("xmm7", "xmm1", "xmm3");
	&vpsrlq("xmm1", "xmm2", 26);
	&vpmuludq("xmm1", "xmm1", &DWP(48,"esp"));
	&vpand("xmm0", "xmm0", "xmm3");
	&vpaddq("xmm5", "xmm4", &DWP(16,"esp"));
	&vpaddq("xmm1", "xmm7", "xmm1");
	&vpsrlq("xmm4", "xmm5", 26);
	&vpand("xmm6", "xmm6", "xmm3");
	&vpaddq("xmm4", "xmm0", "xmm4");
	&vpand("xmm0", "xmm1", "xmm3");
	&vpsrlq("xmm1", "xmm1", 26);
	&vpaddq("xmm1", "xmm6", "xmm1");
	&vpand("xmm6", "xmm5", "xmm3");
	&vpand("xmm5", "xmm4", "xmm3");
	&vpand("xmm2", "xmm2", "xmm3");
	&vpsrlq("xmm3", "xmm4", 26);
	&vpaddq("xmm3", "xmm2", "xmm3");
	&set_label("poly1305_blocks_avx_22");
	&test("eax", "eax");
	&je(&label("poly1305_blocks_avx_24"));
	&set_label("poly1305_blocks_avx_23");
	&vpshufd("xmm0", "xmm0", 8);
	&vpshufd("xmm1", "xmm1", 8);
	&vpshufd("xmm6", "xmm6", 8);
	&vpshufd("xmm5", "xmm5", 8);
	&vpunpcklqdq("xmm2", "xmm0", "xmm1");
	&vpunpcklqdq("xmm4", "xmm6", "xmm5");
	&vpshufd("xmm3", "xmm3", 8);
	&vmovdqu(&DWP(0,"ecx"), "xmm2");
	&vmovdqu(&DWP(16,"ecx"), "xmm4");
	&vmovq(&DWP(32,"ecx"), "xmm3");
	&add("esp", 560);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
	&set_label("poly1305_blocks_avx_24");
	&vpsrldq("xmm2", "xmm0", 8);
	&vpaddq("xmm0", "xmm0", "xmm2");
	&vpsrldq("xmm4", "xmm1", 8);
	&vmovd("ecx", "xmm0");
	&vpaddq("xmm1", "xmm1", "xmm4");
	&vmovd("ebx", "xmm1");
	&mov("esi", "ecx");
	&vpsrldq("xmm7", "xmm6", 8);
	&vpaddq("xmm6", "xmm6", "xmm7");
	&shr("esi", 26);
	&and("ecx", 67108863);
	&add("ebx", "esi");
	&vmovd("edi", "xmm6");
	&mov("eax", "ebx");
	&vpsrldq("xmm0", "xmm5", 8);
	&vpaddq("xmm5", "xmm5", "xmm0");
	&shr("eax", 26);
	&and("ebx", 67108863);
	&add("edi", "eax");
	&vmovd("eax", "xmm5");
	&mov("edx", "edi");
	&vpsrldq("xmm0", "xmm3", 8);
	&vpaddq("xmm3", "xmm3", "xmm0");
	&shr("edx", 26);
	&and("edi", 67108863);
	&add("eax", "edx");
	&vmovd("edx", "xmm3");
	&mov("esi", "eax");
	&shr("esi", 26);
	&and("eax", 67108863);
	&add("edx", "esi");
	&mov("esi", "edx");
	&and("edx", 67108863);
	&shr("esi", 26);
	&lea("esi", &DWP(0,"esi","esi",4));
	&add("ecx", "esi");
	&mov("esi", "ecx");
	&and("ecx", 67108863);
	&shr("esi", 26);
	&add("ebx", "esi");
	&mov("esi", "ebx");
	&and("ebx", 67108863);
	&shr("esi", 26);
	&add("edi", "esi");
	&mov("esi", "edi");
	&shr("edi", 26);
	&and("esi", 67108863);
	&add("eax", "edi");
	&mov("edi", "eax");
	&shr("eax", 26);
	&and("edi", 67108863);
	&add("edx", "eax");
	&mov("eax", "edx");
	&shr("edx", 26);
	&and("eax", 67108863);
	&mov(&DWP(4,"esp"), "eax");
	&mov(&DWP(0,"esp"), "edi");
	&lea("edx", &DWP(0,"edx","edx",4));
	&add("ecx", "edx");
	&mov("edx", "ecx");
	&and("edx", 67108863);
	&shr("ecx", 26);
	&add("ebx", "ecx");
	&lea("ecx", &DWP(5,"edx"));
	&mov(&DWP(8,"esp"), "ecx");
	&shr("ecx", 26);
	&add("ecx", "ebx");
	&mov(&DWP(12,"esp"), "ecx");
	&shr("ecx", 26);
	&add("ecx", "esi");
	&mov(&DWP(16,"esp"), "ecx");
	&shr("ecx", 26);
	&add("ecx", "edi");
	&mov(&DWP(20,"esp"), "ecx");
	&shr("ecx", 26);
	&mov("edi", &DWP(8,"esp"));
	&and("edi", 67108863);
	&lea("eax", &DWP(-67108864,"ecx","eax"));
	&mov(&DWP(24,"esp"), "eax");
	&shr("eax", 31);
	&dec("eax");
	&mov("ecx", "eax");
	&and("edi", "eax");
	&not("ecx");
	&and("edx", "ecx");
	&and("ebx", "ecx");
	&or("edx", "edi");
	&and("esi", "ecx");
	&mov("edi", &DWP(576,"esp"));
	&mov(&DWP(0,"edi"), "edx");
	&mov("edx", &DWP(12,"esp"));
	&and("edx", 67108863);
	&and("edx", "eax");
	&or("ebx", "edx");
	&mov(&DWP(4,"edi"), "ebx");
	&mov("ebx", &DWP(16,"esp"));
	&and("ebx", 67108863);
	&and("ebx", "eax");
	&mov("edx", &DWP(20,"esp"));
	&or("esi", "ebx");
	&and("edx", 67108863);
	&mov(&DWP(8,"edi"), "esi");
	&and("edx", "eax");
	&mov("esi", &DWP(0,"esp"));
	&and("esi", "ecx");
	&or("esi", "edx");
	&mov("edx", &DWP(24,"esp"));
	&and("ecx", &DWP(4,"esp"));
	&and("edx", "eax");
	&or("ecx", "edx");
	&mov(&DWP(12,"edi"), "esi");
	&mov(&DWP(16,"edi"), "ecx");
	&set_label("poly1305_blocks_avx_25");
	&add("esp", 560);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
}

sub poly1305_finish_ex_avx {
	&static_label("poly1305_finish_ex_avx_local");
	&set_label("poly1305_finish_ex_avx_local");
	&push("esi");
	&push("edi");
	&push("ebx");
	&push("ebp");
	&sub("esp", 60);
	&mov("ebp", &DWP(88,"esp"));
	&mov("esi", &DWP(80,"esp"));
	&test("ebp", "ebp");
	&je(&label("poly1305_finish_ex_avx_18"));
	&set_label("poly1305_finish_ex_avx_2");
	&vpxor("xmm0", "xmm0", "xmm0");
	&vmovups(&DWP(32,"esp"), "xmm0");
	&vmovups(&DWP(16,"esp"), "xmm0");
	&set_label("poly1305_finish_ex_avx_3");
	&mov("eax", &DWP(84,"esp"));
	&lea("edx", &DWP(16,"esp"));
	&sub("eax", "edx");
	&test("ebp", 16);
	&je(&label("poly1305_finish_ex_avx_5"));
	&set_label("poly1305_finish_ex_avx_4");
	&lea("edx", &DWP(32,"esp"));
	&vmovdqu("xmm0", &DWP(16,"esp","eax"));
	&vmovdqu(&DWP(16,"esp"), "xmm0");
	&set_label("poly1305_finish_ex_avx_5");
	&test("ebp", 8);
	&je(&label("poly1305_finish_ex_avx_7"));
	&set_label("poly1305_finish_ex_avx_6");
	&mov("ecx", &DWP(0,"edx","eax"));
	&mov("edi", &DWP(4,"edx","eax"));
	&mov(&DWP(0,"edx"), "ecx");
	&mov(&DWP(4,"edx"), "edi");
	&add("edx", 8);
	&set_label("poly1305_finish_ex_avx_7");
	&test("ebp", 4);
	&je(&label("poly1305_finish_ex_avx_9"));
	&set_label("poly1305_finish_ex_avx_8");
	&mov("ecx", &DWP(0,"edx","eax"));
	&mov(&DWP(0,"edx"), "ecx");
	&add("edx", 4);
	&set_label("poly1305_finish_ex_avx_9");
	&test("ebp", 2);
	&je(&label("poly1305_finish_ex_avx_11"));
	&set_label("poly1305_finish_ex_avx_10");
	&movzw("ecx", &DWP(0,"edx","eax"));
	&mov(&DWP(0,"edx"), "cx");
	&add("edx", 2);
	&set_label("poly1305_finish_ex_avx_11");
	&test("ebp", 1);
	&je(&label("poly1305_finish_ex_avx_13"));
	&set_label("poly1305_finish_ex_avx_12");
	&movzb("eax", &DWP(0,"edx","eax"));
	&mov(&DWP(0,"edx"), "al");
	&set_label("poly1305_finish_ex_avx_13");
	&cmp("ebp", 16);
	&je(&label("poly1305_finish_ex_avx_16"));
	&set_label("poly1305_finish_ex_avx_14");
	&mov("al", 1);
	&mov(&DWP(16,"esp","ebp"), "al");
	&jae(&label("poly1305_finish_ex_avx_16"));
	&set_label("poly1305_finish_ex_avx_15");
	&mov("eax", 8);
	&jmp(&label("poly1305_finish_ex_avx_17"));
	&set_label("poly1305_finish_ex_avx_16");
	&mov("eax", 4);
	&set_label("poly1305_finish_ex_avx_17");
	&add("esp", 12);
	&push(32);
	&or(&DWP(116,"esi"), "eax");
	&lea("eax", &DWP(8,"esp"));
	&push("eax");
	&push("esi");
	&call(&LABEL("poly1305_blocks_avx_local"));
	&set_label("poly1305_finish_ex_avx_18");
	&mov("eax", &DWP(116,"esi"));
	&test("al", 1);
	&je(&label("poly1305_finish_ex_avx_24"));
	&set_label("poly1305_finish_ex_avx_19");
	&test("ebp", "ebp");
	&je(&label("poly1305_finish_ex_avx_21"));
	&set_label("poly1305_finish_ex_avx_20");
	&cmp("ebp", 16);
	&jbe(&label("poly1305_finish_ex_avx_22"));
	&set_label("poly1305_finish_ex_avx_21");
	&or("eax", 16);
	&mov(&DWP(116,"esi"), "eax");
	&jmp(&label("poly1305_finish_ex_avx_23"));
	&set_label("poly1305_finish_ex_avx_22");
	&or("eax", 32);
	&mov(&DWP(116,"esi"), "eax");
	&set_label("poly1305_finish_ex_avx_23");
	&add("esp", 12);
	&push(32);
	&push(0);
	&push("esi");
	&call(&LABEL("poly1305_blocks_avx_local"));
	&set_label("poly1305_finish_ex_avx_24");
	&mov("eax", &DWP(8,"esi"));
	&mov("edi", "eax");
	&mov("edx", &DWP(4,"esi"));
	&mov("ebx", "edx");
	&shr("edx", 6);
	&shl("edi", 20);
	&vpxor("xmm0", "xmm0", "xmm0");
	&mov("ecx", &DWP(12,"esi"));
	&or("edx", "edi");
	&mov("edi", "ecx");
	&shr("eax", 12);
	&shl("edi", 14);
	&or("eax", "edi");
	&mov("edi", &DWP(16,"esi"));
	&shl("ebx", 26);
	&shr("ecx", 18);
	&shl("edi", 8);
	&mov("ebp", &DWP(92,"esp"));
	&or("ecx", "edi");
	&or("ebx", &DWP(0,"esi"));
	&add("ebx", &DWP(100,"esi"));
	&adc("edx", &DWP(104,"esi"));
	&adc("eax", &DWP(108,"esi"));
	&adc("ecx", &DWP(112,"esi"));
	&vmovdqu(&DWP(0,"esi"), "xmm0");
	&vmovdqu(&DWP(16,"esi"), "xmm0");
	&vmovdqu(&DWP(32,"esi"), "xmm0");
	&vmovdqu(&DWP(48,"esi"), "xmm0");
	&vmovdqu(&DWP(64,"esi"), "xmm0");
	&vmovdqu(&DWP(80,"esi"), "xmm0");
	&vmovdqu(&DWP(96,"esi"), "xmm0");
	&vmovdqu(&DWP(112,"esi"), "xmm0");
	&mov(&DWP(0,"ebp"), "ebx");
	&mov(&DWP(4,"ebp"), "edx");
	&mov(&DWP(8,"ebp"), "eax");
	&mov(&DWP(12,"ebp"), "ecx");
	&add("esp", 60);
	&pop("ebp");
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
}

sub poly1305_init_ex_avx {
	&static_label("poly1305_init_ex_avx_local");
	&set_label("poly1305_init_ex_avx_local");
	&push("esi");
	&push("edi");
	&push("ebx");
	&push("ebp");
	&sub("esp", 76);
	&mov("ecx", -1);
	&mov("eax", &DWP(96,"esp"));
	&vpxor("xmm0", "xmm0", "xmm0");
	&mov("ebp", &DWP(100,"esp"));
	&mov("edx", &DWP(104,"esp"));
	&test("edx", "edx");
	&vmovdqu(&DWP(0,"eax"), "xmm0");
	&vmovdqu(&DWP(16,"eax"), "xmm0");
	&vmovdqu(&DWP(32,"eax"), "xmm0");
	&cmove("edx", "ecx");
	&mov("ecx", &DWP(4,"ebp"));
	&mov("esi", "ecx");
	&mov(&DWP(12,"esp"), "edx");
	&mov("edx", &DWP(0,"ebp"));
	&mov("edi", "edx");
	&shr("edx", 26);
	&and("edi", 67108863);
	&shl("esi", 6);
	&mov("ebx", &DWP(8,"ebp"));
	&or("edx", "esi");
	&mov("esi", "ebx");
	&and("edx", 67108611);
	&shr("ecx", 20);
	&shl("esi", 12);
	&mov("ebp", &DWP(12,"ebp"));
	&or("ecx", "esi");
	&mov("esi", "ebp");
	&and("ecx", 67092735);
	&shr("ebx", 14);
	&shl("esi", 18);
	&or("ebx", "esi");
	&mov("esi", &DWP(100,"esp"));
	&and("ebx", 66076671);
	&shr("ebp", 8);
	&and("ebp", 1048575);
	&mov(&DWP(40,"eax"), "edi");
	&mov(&DWP(44,"eax"), "edx");
	&mov(&DWP(48,"eax"), "ecx");
	&mov(&DWP(52,"eax"), "ebx");
	&mov(&DWP(56,"eax"), "ebp");
	&mov(&DWP(24,"esp"), "edi");
	&mov("edi", &DWP(16,"esi"));
	&mov(&DWP(100,"eax"), "edi");
	&mov("edi", &DWP(20,"esi"));
	&mov(&DWP(104,"eax"), "edi");
	&mov("edi", &DWP(24,"esi"));
	&mov(&DWP(108,"eax"), "edi");
	&mov("esi", &DWP(28,"esi"));
	&mov(&DWP(112,"eax"), "esi");
	&lea("esi", &DWP(60,"eax"));
	&mov(&DWP(36,"esp"), 0);
	&lea("eax", &DWP(80,"eax"));
	&mov(&DWP(8,"esp"), "esi");
	&cmp(&DWP(12,"esp"), 16);
	&jbe(&label("poly1305_init_ex_avx_9"));
	&set_label("poly1305_init_ex_avx_2");
	&mov(&DWP(4,"esp"), "esi");
	&mov(&DWP(0,"esp"), "eax");
	&mov(&DWP(16,"esp"), "ebp");
	&mov(&DWP(32,"esp"), "ebx");
	&mov(&DWP(20,"esp"), "ecx");
	&mov(&DWP(28,"esp"), "edx");
	&set_label("poly1305_init_ex_avx_3");
	&mov("ebp", &DWP(32,"esp"));
	&mov("eax", &DWP(24,"esp"));
	&mul("eax");
	&mov("ebx", &DWP(16,"esp"));
	&lea("esi", &DWP(0,"ebp","ebp",4));
	&mov("ebp", &DWP(28,"esp"));
	&mov("ecx", "eax");
	&mov(&DWP(44,"esp"), "esi");
	&lea("edi", &DWP(0,"ebx","ebx"));
	&mov("ebx", "edx");
	&mov(&DWP(40,"esp"), "edi");
	&lea("eax", &DWP(0,"ebp","ebp",4));
	&mul("edi");
	&mov("edi", &DWP(20,"esp"));
	&add("ecx", "eax");
	&mov(&DWP(56,"esp"), "edi");
	&adc("ebx", "edx");
	&lea("eax", &DWP(0,"edi","edi"));
	&mul("esi");
	&add("ecx", "eax");
	&lea("esi", &DWP(0,"ebp","ebp"));
	&mov("eax", &DWP(24,"esp"));
	&adc("ebx", "edx");
	&mov("ebp", &DWP(32,"esp"));
	&mov(&DWP(52,"esp"), "esi");
	&lea("edx", &DWP(0,"eax","eax"));
	&mov(&DWP(64,"esp"), "edx");
	&mul("esi");
	&mov("esi", "edx");
	&add("ebp", "ebp");
	&mov(&DWP(60,"esp"), "ebp");
	&mov("ebp", "eax");
	&mov(&DWP(48,"esp"), "ecx");
	&lea("eax", &DWP(0,"edi","edi",4));
	&mul(&DWP(40,"esp"));
	&add("ebp", "eax");
	&mov("eax", &DWP(32,"esp"));
	&mov("edi", &DWP(44,"esp"));
	&adc("esi", "edx");
	&mul("edi");
	&shl("ebx", 6);
	&shr("ecx", 26);
	&or("ebx", "ecx");
	&add("eax", "ebx");
	&adc("edx", 0);
	&add("ebp", "eax");
	&mov("eax", &DWP(28,"esp"));
	&adc("esi", "edx");
	&mul("eax");
	&mov("ecx", "eax");
	&mov("ebx", "edx");
	&mov("eax", &DWP(56,"esp"));
	&mul(&DWP(64,"esp"));
	&add("ecx", "eax");
	&mov("eax", &DWP(40,"esp"));
	&adc("ebx", "edx");
	&mul("edi");
	&mov(&DWP(68,"esp"), "ebp");
	&shl("esi", 6);
	&shr("ebp", 26);
	&or("esi", "ebp");
	&add("eax", "esi");
	&adc("edx", 0);
	&add("ecx", "eax");
	&mov("edi", "ecx");
	&adc("ebx", "edx");
	&and("edi", 67108863);
	&shl("ebx", 6);
	&shr("ecx", 26);
	&mov("eax", &DWP(52,"esp"));
	&or("ebx", "ecx");
	&mov("ecx", &DWP(56,"esp"));
	&mul("ecx");
	&mov(&DWP(20,"esp"), "edi");
	&mov("esi", "eax");
	&mov("eax", &DWP(24,"esp"));
	&mov("ebp", "edx");
	&mov("edi", &DWP(60,"esp"));
	&mul("edi");
	&add("esi", "eax");
	&mov("eax", &DWP(16,"esp"));
	&adc("ebp", "edx");
	&lea("edx", &DWP(0,"eax","eax",4));
	&mul("edx");
	&add("esi", "eax");
	&mov("eax", "ecx");
	&adc("ebp", "edx");
	&add("ebx", "esi");
	&mov("esi", "ebx");
	&adc("ebp", 0);
	&and("esi", 67108863);
	&mul("ecx");
	&shl("ebp", 6);
	&mov("ecx", "eax");
	&shr("ebx", 26);
	&mov("eax", "edi");
	&or("ebp", "ebx");
	&mov("ebx", "edx");
	&mul(&DWP(28,"esp"));
	&add("ecx", "eax");
	&mov("eax", &DWP(64,"esp"));
	&adc("ebx", "edx");
	&mul(&DWP(16,"esp"));
	&add("ecx", "eax");
	&mov("edi", &DWP(48,"esp"));
	&adc("ebx", "edx");
	&add("ebp", "ecx");
	&mov("ecx", "ebp");
	&adc("ebx", 0);
	&and("edi", 67108863);
	&shl("ebx", 6);
	&and("ecx", 67108863);
	&shr("ebp", 26);
	&or("ebx", "ebp");
	&mov("eax", &DWP(68,"esp"));
	&and("eax", 67108863);
	&mov(&DWP(16,"esp"), "ecx");
	&mov("ecx", &DWP(36,"esp"));
	&lea("ebx", &DWP(0,"ebx","ebx",4));
	&add("edi", "ebx");
	&inc("ecx");
	&mov("ebp", "edi");
	&shr("edi", 26);
	&and("ebp", 67108863);
	&mov(&DWP(32,"esp"), "esi");
	&mov(&DWP(24,"esp"), "ebp");
	&mov(&DWP(36,"esp"), "ecx");
	&lea("edx", &DWP(0,"eax","edi"));
	&mov(&DWP(28,"esp"), "edx");
	&cmp("ecx", 2);
	&jae(&label("poly1305_init_ex_avx_8"));
	&set_label("poly1305_init_ex_avx_4");
	&cmp(&DWP(36,"esp"), 0);
	&jne(&label("poly1305_init_ex_avx_6"));
	&set_label("poly1305_init_ex_avx_5");
	&mov("esi", &DWP(4,"esp"));
	&mov("eax", &DWP(16,"esp"));
	&mov("edx", &DWP(32,"esp"));
	&mov("ecx", &DWP(20,"esp"));
	&mov("ebx", &DWP(28,"esp"));
	&mov("edi", &DWP(8,"esp"));
	&mov(&DWP(16,"esi"), "eax");
	&mov(&DWP(12,"esi"), "edx");
	&mov(&DWP(8,"esi"), "ecx");
	&mov(&DWP(4,"esi"), "ebx");
	&mov(&DWP(0,"esi"), "ebp");
	&mov(&DWP(4,"esp"), "edi");
	&jmp(&label("poly1305_init_ex_avx_3"));
	&set_label("poly1305_init_ex_avx_6");
	&cmp(&DWP(36,"esp"), 1);
	&jne(&label("poly1305_init_ex_avx_3"));
	&set_label("poly1305_init_ex_avx_7");
	&mov("esi", &DWP(4,"esp"));
	&mov("eax", &DWP(16,"esp"));
	&mov("edx", &DWP(32,"esp"));
	&mov("ecx", &DWP(20,"esp"));
	&mov("ebx", &DWP(28,"esp"));
	&mov("edi", &DWP(0,"esp"));
	&mov(&DWP(16,"esi"), "eax");
	&mov(&DWP(12,"esi"), "edx");
	&mov(&DWP(8,"esi"), "ecx");
	&mov(&DWP(4,"esi"), "ebx");
	&mov(&DWP(0,"esi"), "ebp");
	&mov(&DWP(4,"esp"), "edi");
	&cmp(&DWP(12,"esp"), 96);
	&jae(&label("poly1305_init_ex_avx_3"));
	&jmp(&label("poly1305_init_ex_avx_9"));
	&set_label("poly1305_init_ex_avx_8");
	&mov("esi", &DWP(4,"esp"));
	&mov("ebp", &DWP(16,"esp"));
	&mov("ebx", &DWP(32,"esp"));
	&mov("ecx", &DWP(20,"esp"));
	&mov("eax", &DWP(24,"esp"));
	&mov(&DWP(16,"esi"), "ebp");
	&mov(&DWP(12,"esi"), "ebx");
	&mov(&DWP(8,"esi"), "ecx");
	&mov(&DWP(4,"esi"), "edx");
	&mov(&DWP(0,"esi"), "eax");
	&set_label("poly1305_init_ex_avx_9");
	&mov("eax", &DWP(96,"esp"));
	&mov(&DWP(116,"eax"), 0);
	&add("esp", 76);
	&pop("ebp");
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
}

sub poly1305_avx {
	#&cmp(&DWP(12,"esp"), 64);
	#&jbe(&label("poly1305_x86_local"));
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", -64);
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 232);
	&mov("edi", &DWP(16,"ebp"));
	&lea("eax", &DWP(52,"esp"));
	&push("edi");
	&push(&DWP(20,"ebp"));
	&push("eax");
	&mov("esi", &DWP(12,"ebp"));
	&call(&LABEL("poly1305_init_ex_avx_local"));
	&set_label("poly1305_avx_2");
	&mov("edx", "edi");
	&and("edx", -32);
	&je(&label("poly1305_avx_5"));
	&set_label("poly1305_avx_3");
	&add("esp", 12);
	&push("edx");
	&push("esi");
	&lea("eax", &DWP(60,"esp"));
	&push("eax");
	&mov(&DWP(192,"esp"), "edx");
	&call(&LABEL("poly1305_blocks_avx_local"));
	&set_label("poly1305_avx_4");
	&mov("edx", &DWP(192,"esp"));
	&add("esi", "edx");
	&sub("edi", "edx");
	&set_label("poly1305_avx_5");
	&push(&DWP(8,"ebp"));
	&push("edi");
	&push("esi");
	&lea("eax", &DWP(76,"esp"));
	&push("eax");
	&call(&LABEL("poly1305_finish_ex_avx_local"));
	&set_label("poly1305_avx_6");
	&add("esp", 260);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
}




sub poly1305_blocks_avx2 {
	&mov("eax", &DWP(4,"esp"));
	&mov("edx", &DWP(8,"esp"));
	&mov("ecx", &DWP(12,"esp"));
	&static_label("poly1305_blocks_avx2_local");
	&set_label("poly1305_blocks_avx2_local");
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", -64);
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 628);
	&mov("ebx", 16777216);
	&mov("esi", 67108863);
	&mov("edi", 5);
	&vmovd("xmm0", "ebx");
	&vmovd("xmm2", "esi");
	&vmovd("xmm4", "edi");
	&vpbroadcastq("ymm1", "xmm0");
	&vpbroadcastq("ymm3", "xmm2");
	&vpbroadcastq("ymm5", "xmm4");
	&mov("ebx", &DWP(176,"eax"));
	&vmovdqu(&DWP(256,"esp"), "ymm1");
	&vmovdqu(&DWP(224,"esp"), "ymm3");
	&vmovdqu(&DWP(192,"esp"), "ymm5");
	&test("bl", 60);
	&je(&label("poly1305_blocks_avx2_12"));
	&set_label("poly1305_blocks_avx2_2");
	&vmovdqu("ymm0", &DWP(256,"esp"));
	&vpsrldq("ymm5", "ymm0", 8);
	&test("bl", 4);
	&je(&label("poly1305_blocks_avx2_4"));
	&set_label("poly1305_blocks_avx2_3");
	&vpermq("ymm5", "ymm5", 192);
	&set_label("poly1305_blocks_avx2_4");
	&test("bl", 8);
	&je(&label("poly1305_blocks_avx2_6"));
	&set_label("poly1305_blocks_avx2_5");
	&vpermq("ymm5", "ymm5", 240);
	&set_label("poly1305_blocks_avx2_6");
	&test("bl", 16);
	&je(&label("poly1305_blocks_avx2_8"));
	&set_label("poly1305_blocks_avx2_7");
	&vpermq("ymm5", "ymm5", 252);
	&set_label("poly1305_blocks_avx2_8");
	&test("bl", 32);
	&je(&label("poly1305_blocks_avx2_10"));
	&set_label("poly1305_blocks_avx2_9");
	&vpxor("ymm5", "ymm5", "ymm5");
	&set_label("poly1305_blocks_avx2_10");
	&vmovdqu(&DWP(256,"esp"), "ymm5");
	&set_label("poly1305_blocks_avx2_12");
	&mov("esi", "ebx");
	&bts("esi", 0);
	&jc(&label("poly1305_blocks_avx2_14"));
	&set_label("poly1305_blocks_avx2_13");
	&vmovdqu("ymm3", &DWP(0,"edx"));
	&mov("ebx", "esi");
	&vmovdqu("ymm2", &DWP(224,"esp"));
	&vpunpcklqdq("ymm4", "ymm3", &DWP(32,"edx"));
	&vpunpckhqdq("ymm6", "ymm3", &DWP(32,"edx"));
	&vpermq("ymm1", "ymm4", 216);
	&add("edx", 64);
	&vpermq("ymm0", "ymm6", 216);
	&add("ecx", -64);
	&vpand("ymm4", "ymm2", "ymm1");
	&vpsrlq("ymm7", "ymm1", 26);
	&vpsrlq("ymm1", "ymm1", 52);
	&vpsllq("ymm6", "ymm0", 12);
	&vpsrlq("ymm0", "ymm0", 40);
	&vpand("ymm3", "ymm2", "ymm7");
	&vpor("ymm7", "ymm1", "ymm6");
	&vpor("ymm0", "ymm0", &DWP(256,"esp"));
	&vpsrlq("ymm6", "ymm7", 26);
	&vpand("ymm1", "ymm2", "ymm7");
	&vmovdqu(&DWP(128,"esp"), "ymm7");
	&vpand("ymm2", "ymm2", "ymm6");
	&vmovdqu(&DWP(160,"esp"), "ymm6");
	&mov(&DWP(176,"eax"), "ebx");
	&jmp(&label("poly1305_blocks_avx2_15"));
	&set_label("poly1305_blocks_avx2_14");
	&vpermq("ymm5", &DWP(0,"eax"), 216);
	&vpxor("ymm6", "ymm6", "ymm6");
	&vpermq("ymm2", &DWP(32,"eax"), 216);
	&vpermq("ymm0", &DWP(64,"eax"), 216);
	&vmovdqu(&DWP(128,"esp"), "ymm2");
	&vmovdqu(&DWP(160,"esp"), "ymm0");
	&vpunpckldq("ymm4", "ymm5", "ymm6");
	&vpunpckhdq("ymm3", "ymm5", "ymm6");
	&vpunpckldq("ymm1", "ymm2", "ymm6");
	&vpunpckhdq("ymm2", "ymm2", "ymm6");
	&vpunpckldq("ymm0", "ymm0", "ymm6");
	&set_label("poly1305_blocks_avx2_15");
	&cmp("ecx", 64);
	&jb(&label("poly1305_blocks_avx2_35"));
	&set_label("poly1305_blocks_avx2_16");
	&vmovdqu("ymm6", &DWP(140,"eax"));
	&test("ebx", 8064);
	&je(&label("poly1305_blocks_avx2_31"));
	&set_label("poly1305_blocks_avx2_17");
	&vpermq("ymm7", &DWP(80,"eax"), 216);
	&vmovdqu(&DWP(32,"esp"), "ymm7");
	&vpermq("ymm7", &DWP(100,"eax"), 216);
	&vmovdqu(&DWP(64,"esp"), "ymm7");
	&vpermq("ymm7", &DWP(120,"eax"), 216);
	&vmovdqu(&DWP(0,"esp"), "ymm7");
	&vpermq("ymm6", "ymm6", 216);
	&test("ebx", 128);
	&je(&label("poly1305_blocks_avx2_19"));
	&set_label("poly1305_blocks_avx2_18");
	&vmovdqa("ymm5", "ymm6");
	&vmovdqu(&DWP(128,"esp"), "ymm6");
	&vmovdqu(&DWP(160,"esp"), "ymm6");
	&jmp(&label("poly1305_blocks_avx2_30"));
	&set_label("poly1305_blocks_avx2_19");
	&test("ebx", 256);
	&je(&label("poly1305_blocks_avx2_21"));
	&set_label("poly1305_blocks_avx2_20");
	&vmovdqu(&DWP(128,"esp"), "ymm6");
	&vmovdqu("ymm7", &DWP(64,"esp"));
	&vmovdqa("ymm5", "ymm6");
	&vmovdqu("ymm6", &DWP(0,"esp"));
	&vmovdqu(&DWP(0,"esp"), "ymm7");
	&vmovdqu(&DWP(160,"esp"), "ymm6");
	&jmp(&label("poly1305_blocks_avx2_30"));
	&set_label("poly1305_blocks_avx2_21");
	&test("ebx", 512);
	&je(&label("poly1305_blocks_avx2_23"));
	&set_label("poly1305_blocks_avx2_22");
	&vmovdqa("ymm5", "ymm6");
	&vmovdqu("ymm6", &DWP(0,"esp"));
	&vmovdqu("ymm7", &DWP(64,"esp"));
	&vmovdqu(&DWP(128,"esp"), "ymm6");
	&vmovdqu("ymm6", &DWP(32,"esp"));
	&vmovdqu(&DWP(160,"esp"), "ymm7");
	&vmovdqu(&DWP(0,"esp"), "ymm6");
	&jmp(&label("poly1305_blocks_avx2_30"));
	&set_label("poly1305_blocks_avx2_23");
	&test("ebx", 1024);
	&je(&label("poly1305_blocks_avx2_25"));
	&set_label("poly1305_blocks_avx2_24");
	&mov("edi", 1);
	&vmovdqu("ymm6", &DWP(64,"esp"));
	&vmovdqu("ymm7", &DWP(32,"esp"));
	&vmovdqu("ymm5", &DWP(0,"esp"));
	&vmovdqu(&DWP(128,"esp"), "ymm6");
	&vpxor("ymm6", "ymm6", "ymm6");
	&vmovd("xmm6", "edi");
	&vmovdqu(&DWP(160,"esp"), "ymm7");
	&vmovdqu(&DWP(0,"esp"), "ymm6");
	&jmp(&label("poly1305_blocks_avx2_30"));
	&set_label("poly1305_blocks_avx2_25");
	&test("ebx", 2048);
	&je(&label("poly1305_blocks_avx2_27"));
	&set_label("poly1305_blocks_avx2_26");
	&mov("edi", 1);
	&vmovdqu("ymm6", &DWP(32,"esp"));
	&vpxor("ymm7", "ymm7", "ymm7");
	&vmovd("xmm7", "edi");
	&vmovdqu("ymm5", &DWP(64,"esp"));
	&vmovdqu(&DWP(128,"esp"), "ymm6");
	&vmovdqu(&DWP(160,"esp"), "ymm7");
	&vmovdqa("ymm6", "ymm7");
	&vmovdqu(&DWP(0,"esp"), "ymm6");
	&jmp(&label("poly1305_blocks_avx2_30"));
	&set_label("poly1305_blocks_avx2_27");
	&test("ebx", 4096);
	&je(&label("poly1305_blocks_avx2_29"));
	&set_label("poly1305_blocks_avx2_28");
	&mov("edi", 1);
	&vpxor("ymm6", "ymm6", "ymm6");
	&vmovd("xmm6", "edi");
	&vmovdqu("ymm5", &DWP(32,"esp"));
	&vmovdqu(&DWP(128,"esp"), "ymm6");
	&vmovdqu(&DWP(0,"esp"), "ymm6");
	&vmovdqa("ymm7", "ymm6");
	&vmovdqu(&DWP(160,"esp"), "ymm7");
	&jmp(&label("poly1305_blocks_avx2_30"));
	&set_label("poly1305_blocks_avx2_29");
	&vmovdqu(&DWP(0,"esp"), "ymm6");
	&set_label("poly1305_blocks_avx2_30");
	&vmovdqu(&DWP(416,"esp"), "ymm3");
	&vmovdqu(&DWP(32,"esp"), "ymm1");
	&vmovdqu(&DWP(64,"esp"), "ymm2");
	&vmovdqu("ymm1", &DWP(128,"esp"));
	&vmovdqu("ymm2", &DWP(0,"esp"));
	&vmovdqu("ymm3", &DWP(160,"esp"));
	&vmovdqu(&DWP(96,"esp"), "ymm0");
	&vpunpcklqdq("ymm6", "ymm5", "ymm1");
	&vpunpckhqdq("ymm5", "ymm5", "ymm1");
	&vpunpcklqdq("ymm7", "ymm3", "ymm2");
	&vmovdqu("ymm1", &DWP(32,"esp"));
	&vperm2i128("ymm0", "ymm6", "ymm7", 32);
	&vmovdqu(&DWP(288,"esp"), "ymm0");
	&vpsrlq("ymm0", "ymm0", 32);
	&vmovdqu(&DWP(320,"esp"), "ymm0");
	&vpunpckhqdq("ymm0", "ymm3", "ymm2");
	&vmovdqu("ymm2", &DWP(64,"esp"));
	&vmovdqu("ymm3", &DWP(416,"esp"));
	&vperm2i128("ymm6", "ymm6", "ymm7", 49);
	&vpsrlq("ymm7", "ymm6", 32);
	&vmovdqu(&DWP(352,"esp"), "ymm6");
	&vmovdqu(&DWP(384,"esp"), "ymm7");
	&vperm2i128("ymm5", "ymm5", "ymm0", 32);
	&vmovdqu("ymm0", &DWP(96,"esp"));
	&jmp(&label("poly1305_blocks_avx2_32"));
	&set_label("poly1305_blocks_avx2_31");
	&vpsrlq("ymm5", "ymm6", 32);
	&vpermq("ymm7", "ymm6", 0);
	&vmovdqu(&DWP(288,"esp"), "ymm7");
	&vpermq("ymm7", "ymm5", 0);
	&vmovdqu(&DWP(320,"esp"), "ymm7");
	&vpermq("ymm7", "ymm6", 85);
	&vpermq("ymm5", "ymm5", 85);
	&vmovdqu(&DWP(352,"esp"), "ymm7");
	&vmovdqu(&DWP(384,"esp"), "ymm5");
	&vpermq("ymm5", "ymm6", 170);
	&set_label("poly1305_blocks_avx2_32");
	&vmovdqu("ymm6", &DWP(192,"esp"));
	&vmovdqu(&DWP(0,"esp"), "ymm5");
	&vpmuludq("ymm7", "ymm6", &DWP(320,"esp"));
	&vmovdqu(&DWP(128,"esp"), "ymm7");
	&vpmuludq("ymm7", "ymm6", &DWP(352,"esp"));
	&vmovdqu(&DWP(96,"esp"), "ymm7");
	&vpmuludq("ymm7", "ymm6", &DWP(384,"esp"));
	&vpmuludq("ymm6", "ymm5", "ymm6");
	&vmovdqu(&DWP(64,"esp"), "ymm7");
	&vmovdqu(&DWP(32,"esp"), "ymm6");
	&jmp(&label("poly1305_blocks_avx2_33"));
	&set_label("poly1305_blocks_avx2_33",64);
	&vmovdqu("ymm6", &DWP(0,"edx"));
	&add("ecx", -64);
	&vmovdqu(&DWP(416,"esp"), "ymm3");
	&vmovdqu(&DWP(160,"esp"), "ymm4");
	&vperm2i128("ymm7", "ymm6", &DWP(32,"edx"), 32);
	&vperm2i128("ymm5", "ymm6", &DWP(32,"edx"), 49);
	&add("edx", 64);
	&vpunpckldq("ymm6", "ymm7", "ymm5");
	&vpunpckhdq("ymm7", "ymm7", "ymm5");
	&vmovdqu(&DWP(448,"esp"), "ymm6");
	&vmovdqu(&DWP(480,"esp"), "ymm7");
	&vpmuludq("ymm7", "ymm0", &DWP(128,"esp"));
	&vpmuludq("ymm5", "ymm2", &DWP(96,"esp"));
	&vpaddq("ymm7", "ymm7", "ymm5");
	&vpmuludq("ymm5", "ymm1", &DWP(64,"esp"));
	&vpaddq("ymm5", "ymm7", "ymm5");
	&vmovdqu("ymm7", &DWP(32,"esp"));
	&vpmuludq("ymm3", "ymm3", "ymm7");
	&vpaddq("ymm5", "ymm5", "ymm3");
	&vmovdqu("ymm3", &DWP(288,"esp"));
	&vpmuludq("ymm4", "ymm4", "ymm3");
	&vpaddq("ymm4", "ymm5", "ymm4");
	&vpxor("ymm5", "ymm5", "ymm5");
	&vpunpckldq("ymm6", "ymm6", "ymm5");
	&vpaddq("ymm4", "ymm4", "ymm6");
	&vpmuludq("ymm6", "ymm0", "ymm7");
	&vpmuludq("ymm3", "ymm2", "ymm3");
	&vmovdqu(&DWP(512,"esp"), "ymm4");
	&vpaddq("ymm4", "ymm6", "ymm3");
	&vmovdqu("ymm6", &DWP(416,"esp"));
	&vpmuludq("ymm3", "ymm1", &DWP(320,"esp"));
	&vpaddq("ymm4", "ymm4", "ymm3");
	&vpmuludq("ymm3", "ymm6", &DWP(352,"esp"));
	&vmovdqu("ymm6", &DWP(160,"esp"));
	&vpaddq("ymm4", "ymm4", "ymm3");
	&vpmuludq("ymm3", "ymm6", &DWP(384,"esp"));
	&vpaddq("ymm4", "ymm4", "ymm3");
	&vmovdqu("ymm3", &DWP(480,"esp"));
	&vpunpckhdq("ymm3", "ymm3", "ymm5");
	&vpsllq("ymm3", "ymm3", 18);
	&vpaddq("ymm4", "ymm4", "ymm3");
	&vmovdqu(&DWP(544,"esp"), "ymm4");
	&vpmuludq("ymm4", "ymm0", &DWP(96,"esp"));
	&vpmuludq("ymm3", "ymm2", &DWP(64,"esp"));
	&vpaddq("ymm4", "ymm4", "ymm3");
	&vmovdqu("ymm3", &DWP(288,"esp"));
	&vpmuludq("ymm7", "ymm1", "ymm7");
	&vpaddq("ymm4", "ymm4", "ymm7");
	&vpmuludq("ymm7", "ymm3", &DWP(416,"esp"));
	&vpaddq("ymm7", "ymm4", "ymm7");
	&vmovdqu("ymm4", &DWP(320,"esp"));
	&vpmuludq("ymm6", "ymm6", "ymm4");
	&vpaddq("ymm7", "ymm7", "ymm6");
	&vmovdqu("ymm6", &DWP(448,"esp"));
	&vpunpckhdq("ymm5", "ymm6", "ymm5");
	&vpsllq("ymm5", "ymm5", 6);
	&vpaddq("ymm6", "ymm7", "ymm5");
	&vmovdqu("ymm7", &DWP(512,"esp"));
	&vpsrlq("ymm7", "ymm7", 26);
	&vpaddq("ymm5", "ymm6", "ymm7");
	&vpmuludq("ymm6", "ymm0", "ymm3");
	&vpmuludq("ymm7", "ymm2", "ymm4");
	&vmovdqu(&DWP(576,"esp"), "ymm5");
	&vpaddq("ymm5", "ymm6", "ymm7");
	&vmovdqu("ymm7", &DWP(416,"esp"));
	&vpmuludq("ymm6", "ymm1", &DWP(352,"esp"));
	&vpaddq("ymm5", "ymm5", "ymm6");
	&vpmuludq("ymm6", "ymm7", &DWP(384,"esp"));
	&vpaddq("ymm6", "ymm5", "ymm6");
	&vmovdqu("ymm5", &DWP(160,"esp"));
	&vpmuludq("ymm5", "ymm5", &DWP(0,"esp"));
	&vpaddq("ymm6", "ymm6", "ymm5");
	&vpaddq("ymm5", "ymm6", &DWP(256,"esp"));
	&vmovdqu("ymm6", &DWP(544,"esp"));
	&vpmuludq("ymm0", "ymm0", &DWP(64,"esp"));
	&vpmuludq("ymm2", "ymm2", &DWP(32,"esp"));
	&vpsrlq("ymm6", "ymm6", 26);
	&vpaddq("ymm2", "ymm0", "ymm2");
	&vmovdqu("ymm0", &DWP(480,"esp"));
	&vpaddq("ymm5", "ymm5", "ymm6");
	&vpxor("ymm6", "ymm6", "ymm6");
	&vpmuludq("ymm1", "ymm1", "ymm3");
	&vpaddq("ymm2", "ymm2", "ymm1");
	&vmovdqu("ymm1", &DWP(160,"esp"));
	&vpmuludq("ymm4", "ymm7", "ymm4");
	&vpaddq("ymm3", "ymm2", "ymm4");
	&vpunpckldq("ymm2", "ymm0", "ymm6");
	&vpsrlq("ymm6", "ymm5", 26);
	&vpmuludq("ymm7", "ymm1", &DWP(352,"esp"));
	&vpaddq("ymm4", "ymm3", "ymm7");
	&vpsllq("ymm1", "ymm2", 12);
	&vmovdqu("ymm3", &DWP(576,"esp"));
	&vpaddq("ymm7", "ymm4", "ymm1");
	&vpsrlq("ymm0", "ymm3", 26);
	&vpaddq("ymm2", "ymm7", "ymm0");
	&vmovdqu("ymm0", &DWP(224,"esp"));
	&vpsrlq("ymm7", "ymm2", 26);
	&vpand("ymm4", "ymm0", &DWP(512,"esp"));
	&vpand("ymm3", "ymm3", "ymm0");
	&vpand("ymm5", "ymm5", "ymm0");
	&vpmuludq("ymm1", "ymm6", &DWP(192,"esp"));
	&vpaddq("ymm1", "ymm4", "ymm1");
	&vpand("ymm4", "ymm0", &DWP(544,"esp"));
	&vpaddq("ymm7", "ymm4", "ymm7");
	&vpand("ymm4", "ymm1", "ymm0");
	&vpsrlq("ymm1", "ymm1", 26);
	&vpaddq("ymm3", "ymm3", "ymm1");
	&vpand("ymm1", "ymm2", "ymm0");
	&vpand("ymm2", "ymm7", "ymm0");
	&vpsrlq("ymm0", "ymm7", 26);
	&vpaddq("ymm0", "ymm5", "ymm0");
	&cmp("ecx", 64);
	&jae(&label("poly1305_blocks_avx2_33"));
	&set_label("poly1305_blocks_avx2_35");
	&test("bl", 64);
	&jne(&label("poly1305_blocks_avx2_37"));
	&set_label("poly1305_blocks_avx2_36");
	&vpshufd("ymm4", "ymm4", 8);
	&vpshufd("ymm5", "ymm3", 8);
	&vpshufd("ymm1", "ymm1", 8);
	&vpshufd("ymm2", "ymm2", 8);
	&vpshufd("ymm0", "ymm0", 8);
	&vpermq("ymm6", "ymm4", 8);
	&vpermq("ymm7", "ymm5", 8);
	&vpermq("ymm1", "ymm1", 8);
	&vpermq("ymm2", "ymm2", 8);
	&vpermq("ymm0", "ymm0", 8);
	&vperm2i128("ymm3", "ymm6", "ymm7", 32);
	&vperm2i128("ymm4", "ymm1", "ymm2", 32);
	&vmovdqu(&DWP(0,"eax"), "ymm3");
	&vmovdqu(&DWP(32,"eax"), "ymm4");
	&vmovdqu(&DWP(64,"eax"), "xmm0");
	&jmp(&label("poly1305_blocks_avx2_38"));
	&set_label("poly1305_blocks_avx2_37");
	&vpermq("ymm5", "ymm4", 245);
	&vpaddq("ymm4", "ymm4", "ymm5");
	&vpermq("ymm6", "ymm3", 245);
	&vpaddq("ymm3", "ymm3", "ymm6");
	&vpermq("ymm7", "ymm1", 245);
	&vpaddq("ymm1", "ymm1", "ymm7");
	&vpermq("ymm7", "ymm4", 170);
	&vpermq("ymm5", "ymm2", 245);
	&vpaddq("ymm4", "ymm4", "ymm7");
	&vpaddq("ymm2", "ymm2", "ymm5");
	&vpermq("ymm5", "ymm3", 170);
	&vpaddq("ymm3", "ymm3", "ymm5");
	&mov(&DWP(0,"esp"), "eax");
	&vpermq("ymm6", "ymm0", 245);
	&vpaddq("ymm0", "ymm0", "ymm6");
	&vmovd("eax", "xmm4");
	&vpermq("ymm4", "ymm1", 170);
	&mov("edx", "eax");
	&vpaddq("ymm1", "ymm1", "ymm4");
	&and("eax", 67108863);
	&shr("edx", 26);
	&vmovd("ebx", "xmm3");
	&vpermq("ymm3", "ymm2", 170);
	&add("ebx", "edx");
	&vpaddq("ymm2", "ymm2", "ymm3");
	&mov("edi", "ebx");
	&shr("edi", 26);
	&and("ebx", 67108863);
	&vmovd("ecx", "xmm1");
	&vpermq("ymm1", "ymm0", 170);
	&add("ecx", "edi");
	&vpaddq("ymm0", "ymm0", "ymm1");
	&mov("edx", "ecx");
	&shr("edx", 26);
	&and("ecx", 67108863);
	&vmovd("esi", "xmm2");
	&add("esi", "edx");
	&mov("edx", "esi");
	&and("esi", 67108863);
	&shr("edx", 26);
	&vmovd("edi", "xmm0");
	&add("edi", "edx");
	&mov("edx", "edi");
	&and("edi", 67108863);
	&shr("edx", 26);
	&lea("edx", &DWP(0,"edx","edx",4));
	&add("eax", "edx");
	&mov("edx", "eax");
	&and("eax", 67108863);
	&shr("edx", 26);
	&add("ebx", "edx");
	&mov("edx", "ebx");
	&and("ebx", 67108863);
	&shr("edx", 26);
	&add("ecx", "edx");
	&mov("edx", "ecx");
	&shr("ecx", 26);
	&and("edx", 67108863);
	&add("esi", "ecx");
	&mov("ecx", "esi");
	&shr("esi", 26);
	&and("ecx", 67108863);
	&add("edi", "esi");
	&mov("esi", "edi");
	&shr("edi", 26);
	&and("esi", 67108863);
	&mov(&DWP(4,"esp"), "esi");
	&lea("edi", &DWP(0,"edi","edi",4));
	&add("eax", "edi");
	&mov("edi", "eax");
	&and("edi", 67108863);
	&shr("eax", 26);
	&add("ebx", "eax");
	&lea("eax", &DWP(5,"edi"));
	&mov(&DWP(8,"esp"), "eax");
	&shr("eax", 26);
	&add("eax", "ebx");
	&mov(&DWP(12,"esp"), "eax");
	&shr("eax", 26);
	&add("eax", "edx");
	&mov(&DWP(16,"esp"), "eax");
	&shr("eax", 26);
	&add("eax", "ecx");
	&mov(&DWP(20,"esp"), "eax");
	&shr("eax", 26);
	&lea("eax", &DWP(-67108864,"eax","esi"));
	&mov(&DWP(24,"esp"), "eax");
	&shr("eax", 31);
	&dec("eax");
	&andn("esi", "eax", "edi");
	&mov("edi", &DWP(8,"esp"));
	&and("edi", 67108863);
	&and("edi", "eax");
	&or("esi", "edi");
	&mov("edi", &DWP(0,"esp"));
	&mov(&DWP(0,"edi"), "esi");
	&andn("esi", "eax", "ebx");
	&mov("ebx", &DWP(12,"esp"));
	&and("ebx", 67108863);
	&and("ebx", "eax");
	&or("esi", "ebx");
	&andn("ebx", "eax", "edx");
	&mov("edx", &DWP(16,"esp"));
	&and("edx", 67108863);
	&and("edx", "eax");
	&or("ebx", "edx");
	&andn("edx", "eax", "ecx");
	&mov("ecx", &DWP(20,"esp"));
	&and("ecx", 67108863);
	&mov(&DWP(8,"edi"), "ebx");
	&and("ecx", "eax");
	&mov(&DWP(4,"edi"), "esi");
	&or("edx", "ecx");
	&mov("ebx", &DWP(24,"esp"));
	&andn("esi", "eax", &DWP(4,"esp"));
	&and("ebx", "eax");
	&or("esi", "ebx");
	&mov(&DWP(12,"edi"), "edx");
	&mov(&DWP(16,"edi"), "esi");
	&set_label("poly1305_blocks_avx2_38");
	&vzeroupper();
	&add("esp", 628);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
}

sub poly1305_finish_ex_avx2 {
	&static_label("poly1305_finish_ex_avx2_local");
	&set_label("poly1305_finish_ex_avx2_local");
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", -64);
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 180);
	&mov("ebx", &DWP(16,"ebp"));
	&test("ebx", "ebx");
	&je(&label("poly1305_finish_ex_avx2_29"));
	&set_label("poly1305_finish_ex_avx2_2");
	&mov("edx", &DWP(12,"ebp"));
	&lea("eax", &DWP(64,"esp"));
	&vpxor("ymm0", "ymm0", "ymm0");
	&sub("edx", "eax");
	&vmovdqu(&DWP(64,"esp"), "ymm0");
	&vmovdqu(&DWP(96,"esp"), "ymm0");
	&test("bl", 32);
	&je(&label("poly1305_finish_ex_avx2_4"));
	&set_label("poly1305_finish_ex_avx2_3");
	&vmovdqu("ymm0", &DWP(64,"esp","edx"));
	&lea("eax", &DWP(96,"esp"));
	&vmovdqu(&DWP(64,"esp"), "ymm0");
	&set_label("poly1305_finish_ex_avx2_4");
	&test("bl", 16);
	&je(&label("poly1305_finish_ex_avx2_6"));
	&set_label("poly1305_finish_ex_avx2_5");
	&vmovdqu("xmm0", &DWP(0,"eax","edx"));
	&vmovdqu(&DWP(0,"eax"), "xmm0");
	&add("eax", 16);
	&set_label("poly1305_finish_ex_avx2_6");
	&test("bl", 8);
	&je(&label("poly1305_finish_ex_avx2_8"));
	&set_label("poly1305_finish_ex_avx2_7");
	&mov("ecx", &DWP(0,"eax","edx"));
	&mov("esi", &DWP(4,"eax","edx"));
	&mov(&DWP(0,"eax"), "ecx");
	&mov(&DWP(4,"eax"), "esi");
	&add("eax", 8);
	&set_label("poly1305_finish_ex_avx2_8");
	&test("bl", 4);
	&je(&label("poly1305_finish_ex_avx2_10"));
	&set_label("poly1305_finish_ex_avx2_9");
	&mov("ecx", &DWP(0,"eax","edx"));
	&mov(&DWP(0,"eax"), "ecx");
	&add("eax", 4);
	&set_label("poly1305_finish_ex_avx2_10");
	&test("bl", 2);
	&je(&label("poly1305_finish_ex_avx2_12"));
	&set_label("poly1305_finish_ex_avx2_11");
	&movzw("ecx", &DWP(0,"eax","edx"));
	&mov(&DWP(0,"eax"), "cx");
	&add("eax", 2);
	&set_label("poly1305_finish_ex_avx2_12");
	&test("bl", 1);
	&je(&label("poly1305_finish_ex_avx2_14"));
	&set_label("poly1305_finish_ex_avx2_13");
	&movzb("edx", &DWP(0,"eax","edx"));
	&mov(&DWP(0,"eax"), "dl");
	&set_label("poly1305_finish_ex_avx2_14");
	&test("bl", 15);
	&je(&label("poly1305_finish_ex_avx2_16"));
	&set_label("poly1305_finish_ex_avx2_15");
	&mov("dl", 1);
	&mov(&DWP(64,"esp","ebx"), "dl");
	&set_label("poly1305_finish_ex_avx2_16");
	&mov("eax", &DWP(8,"ebp"));
	&mov("eax", &DWP(176,"eax"));
	&and("eax", -8125);
	&cmp("ebx", 48);
	&jb(&label("poly1305_finish_ex_avx2_18"));
	&set_label("poly1305_finish_ex_avx2_17");
	&or("eax", 4);
	&jmp(&label("poly1305_finish_ex_avx2_21"));
	&set_label("poly1305_finish_ex_avx2_18");
	&cmp("ebx", 32);
	&jb(&label("poly1305_finish_ex_avx2_20"));
	&set_label("poly1305_finish_ex_avx2_19");
	&or("eax", 8);
	&jmp(&label("poly1305_finish_ex_avx2_21"));
	&set_label("poly1305_finish_ex_avx2_20");
	&mov("edx", "eax");
	&or("eax", 32);
	&or("edx", 16);
	&cmp("ebx", 16);
	&cmovae("eax", "edx");
	&set_label("poly1305_finish_ex_avx2_21");
	&test("al", 1);
	&je(&label("poly1305_finish_ex_avx2_27"));
	&set_label("poly1305_finish_ex_avx2_22");
	&cmp("ebx", 16);
	&ja(&label("poly1305_finish_ex_avx2_24"));
	&set_label("poly1305_finish_ex_avx2_23");
	&mov("edx", &DWP(8,"ebp"));
	&or("eax", 256);
	&mov(&DWP(176,"edx"), "eax");
	&jmp(&label("poly1305_finish_ex_avx2_28"));
	&set_label("poly1305_finish_ex_avx2_24");
	&cmp("ebx", 32);
	&ja(&label("poly1305_finish_ex_avx2_27"));
	&set_label("poly1305_finish_ex_avx2_25");
	&mov("edx", &DWP(8,"ebp"));
	&or("eax", 128);
	&mov(&DWP(176,"edx"), "eax");
	&jmp(&label("poly1305_finish_ex_avx2_28"));
	&set_label("poly1305_finish_ex_avx2_27");
	&mov("edx", &DWP(8,"ebp"));
	&mov(&DWP(176,"edx"), "eax");
	&set_label("poly1305_finish_ex_avx2_28");
	&mov("ecx", 64);
	&lea("edx", &DWP(64,"esp"));
	&mov("eax", &DWP(8,"ebp"));
	&vzeroupper();
	&call(&LABEL("poly1305_blocks_avx2_local"));
	&set_label("poly1305_finish_ex_avx2_29");
	&mov("eax", &DWP(8,"ebp"));
	&mov("edx", &DWP(176,"eax"));
	&test("dl", 1);
	&je(&label("poly1305_finish_ex_avx2_37"));
	&set_label("poly1305_finish_ex_avx2_30");
	&and("edx", -8125);
	&test("ebx", "ebx");
	&je(&label("poly1305_finish_ex_avx2_32"));
	&set_label("poly1305_finish_ex_avx2_31");
	&cmp("ebx", 48);
	&jbe(&label("poly1305_finish_ex_avx2_33"));
	&set_label("poly1305_finish_ex_avx2_32");
	&or("edx", 512);
	&jmp(&label("poly1305_finish_ex_avx2_36"));
	&set_label("poly1305_finish_ex_avx2_33");
	&cmp("ebx", 32);
	&jbe(&label("poly1305_finish_ex_avx2_35"));
	&set_label("poly1305_finish_ex_avx2_34");
	&or("edx", 1024);
	&jmp(&label("poly1305_finish_ex_avx2_36"));
	&set_label("poly1305_finish_ex_avx2_35");
	&mov("eax", "edx");
	&or("edx", 4096);
	&or("eax", 2048);
	&cmp("ebx", 16);
	&cmova("edx", "eax");
	&set_label("poly1305_finish_ex_avx2_36");
	&mov("eax", &DWP(8,"ebp"));
	&or("edx", 96);
	&vpxor("ymm0", "ymm0", "ymm0");
	&mov("ecx", 64);
	&mov(&DWP(176,"eax"), "edx");
	&lea("edx", &DWP(64,"esp"));
	&vmovdqu(&DWP(64,"esp"), "ymm0");
	&vmovdqu(&DWP(96,"esp"), "ymm0");
	&vzeroupper();
	&call(&LABEL("poly1305_blocks_avx2_local"));
	&set_label("poly1305_finish_ex_avx2_37");
	&mov("esi", &DWP(8,"ebp"));
	&mov("edi", &DWP(20,"ebp"));
	&mov(&DWP(128,"esp"), "edi");
	&mov("eax", &DWP(8,"esi"));
	&mov("edi", "eax");
	&mov("edx", &DWP(4,"esi"));
	&mov("ebx", "edx");
	&shr("edx", 6);
	&shl("edi", 20);
	&mov("ecx", &DWP(12,"esi"));
	&or("edx", "edi");
	&mov("edi", "ecx");
	&shr("eax", 12);
	&shl("edi", 14);
	&vpxor("ymm0", "ymm0", "ymm0");
	&or("eax", "edi");
	&mov("edi", &DWP(16,"esi"));
	&shl("ebx", 26);
	&shr("ecx", 18);
	&shl("edi", 8);
	&or("ebx", &DWP(0,"esi"));
	&or("ecx", "edi");
	&add("ebx", &DWP(160,"esi"));
	&adc("edx", &DWP(164,"esi"));
	&adc("eax", &DWP(168,"esi"));
	&adc("ecx", &DWP(172,"esi"));
	&vmovdqu(&DWP(0,"esi"), "ymm0");
	&vmovdqu(&DWP(32,"esi"), "ymm0");
	&vmovdqu(&DWP(64,"esi"), "ymm0");
	&vmovdqu(&DWP(96,"esi"), "ymm0");
	&vmovdqu(&DWP(128,"esi"), "ymm0");
	&vmovdqu(&DWP(160,"esi"), "ymm0");
	&mov("esi", &DWP(128,"esp"));
	&vzeroupper();
	&mov(&DWP(0,"esi"), "ebx");
	&mov(&DWP(4,"esi"), "edx");
	&mov(&DWP(8,"esi"), "eax");
	&mov(&DWP(12,"esi"), "ecx");
	&add("esp", 180);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
}


sub poly1305_init_ex_avx2 {
	&mov("eax", &DWP(4,"esp"));
	&mov("edx", &DWP(8,"esp"));
	&mov("ecx", &DWP(12,"esp"));
	&static_label("poly1305_init_ex_avx2_local");
	&set_label("poly1305_init_ex_avx2_local");
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 272);
	&mov("edi", "edx");
	&vpxor("ymm0", "ymm0", "ymm0");
	&mov("edx", "eax");
	&vmovdqu(&DWP(0,"edx"), "ymm0");
	&mov("eax", -1);
	&test("ecx", "ecx");
	&vmovdqu(&DWP(32,"edx"), "ymm0");
	&vpxor("xmm0", "xmm0", "xmm0");
	&vmovdqu(&DWP(64,"edx"), "xmm0");
	&cmove("ecx", "eax");
	&mov(&DWP(4,"esp"), "ecx");
	&mov("esi", &DWP(0,"edi"));
	&mov("ecx", &DWP(4,"edi"));
	&mov("ebx", &DWP(8,"edi"));
	&mov("eax", &DWP(12,"edi"));
	&mov(&DWP(0,"esp"), "edi");
	&mov("edi", "esi");
	&and("edi", 67108863);
	&mov(&DWP(80,"edx"), "edi");
	&mov("edi", "ecx");
	&shr("esi", 26);
	&shl("edi", 6);
	&or("esi", "edi");
	&and("esi", 67108611);
	&mov(&DWP(84,"edx"), "esi");
	&mov("esi", "ebx");
	&shr("ecx", 20);
	&shl("esi", 12);
	&or("ecx", "esi");
	&and("ecx", 67092735);
	&mov(&DWP(88,"edx"), "ecx");
	&mov("ecx", "eax");
	&shr("ebx", 14);
	&shl("ecx", 18);
	&shr("eax", 8);
	&or("ebx", "ecx");
	&and("ebx", 66076671);
	&and("eax", 1048575);
	&mov(&DWP(92,"edx"), "ebx");
	&xor("ecx", "ecx");
	&mov(&DWP(96,"edx"), "eax");
	&mov("edi", &DWP(4,"esp"));
	&mov("esi", &DWP(0,"esp"));
	&set_label("poly1305_init_ex_avx2_2");
	&mov("ebx", &DWP(16,"esi","ecx",4));
	&mov(&DWP(160,"edx","ecx",4), "ebx");
	&inc("ecx");
	&cmp("ecx", 4);
	&jl(&label("poly1305_init_ex_avx2_2"));
	&set_label("poly1305_init_ex_avx2_3");
	&cmp("edi", 16);
	&jbe(&label("poly1305_init_ex_avx2_6"));
	&set_label("poly1305_init_ex_avx2_4");
	&mov("ecx", 5);
	&vmovd("xmm5", "eax");
	&vmovdqu("xmm4", &DWP(80,"edx"));
	&vmovd("xmm3", "ecx");
	&mov("ecx", 67108863);
	&vmovdqu(&DWP(224,"esp"), "xmm4");
	&vpshufd("xmm3", "xmm3", 68);
	&vpunpcklqdq("xmm1", "xmm5", "xmm0");
	&vmovd("xmm2", "ecx");
	&vpshufd("xmm6", "xmm2", 68);
	&vpunpcklqdq("xmm2", "xmm4", "xmm0");
	&vpunpckhqdq("xmm4", "xmm4", "xmm0");
	&vpsrlq("xmm7", "xmm2", 32);
	&vpsrlq("xmm0", "xmm4", 32);
	&vmovdqu(&DWP(208,"esp"), "xmm5");
	&vmovdqu(&DWP(32,"esp"), "xmm0");
	&vpmuludq("xmm5", "xmm0", "xmm3");
	&vpaddd("xmm0", "xmm0", "xmm0");
	&vmovdqu(&DWP(48,"esp"), "xmm0");
	&vpaddd("xmm0", "xmm2", "xmm2");
	&vmovdqu(&DWP(16,"esp"), "xmm7");
	&vmovdqu(&DWP(64,"esp"), "xmm0");
	&vpaddd("xmm0", "xmm7", "xmm7");
	&vpmuludq("xmm7", "xmm7", "xmm3");
	&vmovdqu(&DWP(256,"esp"), "xmm6");
	&vpaddd("xmm6", "xmm1", "xmm1");
	&vmovdqu(&DWP(80,"esp"), "xmm0");
	&vpaddd("xmm0", "xmm4", "xmm4");
	&vpmuludq("xmm0", "xmm0", "xmm5");
	&vpmuludq("xmm7", "xmm6", "xmm7");
	&vpaddq("xmm0", "xmm0", "xmm7");
	&vpmuludq("xmm7", "xmm2", "xmm2");
	&vpaddq("xmm0", "xmm0", "xmm7");
	&vmovdqu(&DWP(96,"esp"), "xmm0");
	&vpmuludq("xmm0", "xmm1", "xmm3");
	&vmovdqu(&DWP(0,"esp"), "xmm1");
	&vpmuludq("xmm1", "xmm1", "xmm0");
	&vpmuludq("xmm0", "xmm2", &DWP(48,"esp"));
	&vmovdqu("xmm7", &DWP(80,"esp"));
	&vmovdqu(&DWP(240,"esp"), "xmm3");
	&vpaddq("xmm1", "xmm1", "xmm0");
	&vpmuludq("xmm0", "xmm7", "xmm4");
	&vpmuludq("xmm3", "xmm4", "xmm3");
	&vpmuludq("xmm2", "xmm7", "xmm2");
	&vpmuludq("xmm3", "xmm6", "xmm3");
	&vpmuludq("xmm6", "xmm6", "xmm5");
	&vpaddq("xmm1", "xmm1", "xmm0");
	&vpmuludq("xmm0", "xmm5", &DWP(32,"esp"));
	&vpaddq("xmm0", "xmm0", "xmm3");
	&vmovdqu("xmm7", &DWP(96,"esp"));
	&vpaddq("xmm2", "xmm0", "xmm2");
	&vpsrlq("xmm0", "xmm7", 26);
	&vpaddq("xmm7", "xmm2", "xmm0");
	&vmovdqu(&DWP(112,"esp"), "xmm7");
	&vmovdqu("xmm7", &DWP(16,"esp"));
	&vmovdqu("xmm2", &DWP(64,"esp"));
	&vpmuludq("xmm3", "xmm7", &DWP(48,"esp"));
	&vpmuludq("xmm0", "xmm2", &DWP(0,"esp"));
	&vpaddq("xmm3", "xmm3", "xmm0");
	&vpmuludq("xmm0", "xmm4", "xmm4");
	&vpmuludq("xmm4", "xmm2", "xmm4");
	&vpmuludq("xmm2", "xmm7", "xmm7");
	&vpaddq("xmm3", "xmm3", "xmm0");
	&vpaddq("xmm5", "xmm6", "xmm4");
	&vpsrlq("xmm0", "xmm1", 26);
	&vpaddq("xmm0", "xmm3", "xmm0");
	&vpaddq("xmm6", "xmm5", "xmm2");
	&vmovdqu("xmm7", &DWP(112,"esp"));
	&vpsrlq("xmm5", "xmm0", 26);
	&vpsrlq("xmm4", "xmm7", 26);
	&vpaddq("xmm4", "xmm6", "xmm4");
	&vpmuludq("xmm6", "xmm5", &DWP(240,"esp"));
	&vmovdqu("xmm2", &DWP(256,"esp"));
	&vpand("xmm3", "xmm2", &DWP(96,"esp"));
	&vpand("xmm1", "xmm1", "xmm2");
	&vpaddq("xmm5", "xmm3", "xmm6");
	&vpsrlq("xmm6", "xmm4", 26);
	&vpand("xmm0", "xmm0", "xmm2");
	&vpaddq("xmm6", "xmm1", "xmm6");
	&vpsrlq("xmm1", "xmm6", 26);
	&vpand("xmm7", "xmm7", "xmm2");
	&vpsrlq("xmm3", "xmm5", 26);
	&vpand("xmm4", "xmm4", "xmm2");
	&vpaddq("xmm0", "xmm0", "xmm1");
	&vpand("xmm1", "xmm5", "xmm2");
	&vpand("xmm6", "xmm6", "xmm2");
	&vpaddq("xmm5", "xmm7", "xmm3");
	&vpunpcklqdq("xmm1", "xmm1", "xmm5");
	&vpunpcklqdq("xmm3", "xmm4", "xmm6");
	&vshufps("xmm4", "%xmm1", "xmm3", 136);
	&vmovups(&DWP(100,"edx"), "xmm4");
	&vmovd(&DWP(116,"edx"), "xmm0");
	&cmp("edi", 32);
	&jbe(&label("poly1305_init_ex_avx2_6"));
	&set_label("poly1305_init_ex_avx2_5");
	&vmovdqu("xmm3", &DWP(224,"esp"));
	&vmovdqu(&DWP(256,"esp"), "xmm2");
	&vpunpcklqdq("xmm1", "xmm3", "xmm4");
	&vpunpckhqdq("xmm2", "xmm3", "xmm4");
	&vpsrlq("xmm5", "xmm1", 32);
	&vmovdqu("xmm4", &DWP(208,"esp"));
	&vpsrlq("xmm3", "xmm2", 32);
	&vpunpcklqdq("xmm4", "xmm4", "xmm0");
	&vpshufd("xmm7", "xmm1", 238);
	&vmovdqu(&DWP(32,"esp"), "xmm7");
	&vmovdqu(&DWP(0,"esp"), "xmm1");
	&vmovdqu(&DWP(16,"esp"), "xmm5");
	&vpshufd("xmm0", "xmm5", 238);
	&vpshufd("xmm5", "xmm2", 238);
	&vpshufd("xmm6", "xmm3", 238);
	&vpshufd("xmm1", "xmm4", 238);
	&vmovdqu("xmm7", &DWP(240,"esp"));
	&vmovdqu(&DWP(64,"esp"), "xmm5");
	&vmovdqu(&DWP(80,"esp"), "xmm6");
	&vmovdqu(&DWP(96,"esp"), "xmm1");
	&vpmuludq("xmm5", "xmm5", "xmm7");
	&vpmuludq("xmm6", "xmm6", "xmm7");
	&vpmuludq("xmm1", "xmm1", "xmm7");
	&vpmuludq("xmm7", "xmm0", "xmm7");
	&vmovdqu(&DWP(112,"esp"), "xmm5");
	&vpmuludq("xmm7", "xmm4", "xmm7");
	&vpmuludq("xmm5", "xmm3", "xmm5");
	&vmovdqu(&DWP(128,"esp"), "xmm6");
	&vpmuludq("xmm6", "xmm2", "xmm6");
	&vpaddq("xmm7", "xmm7", "xmm5");
	&vpmuludq("xmm5", "xmm1", &DWP(16,"esp"));
	&vpaddq("xmm7", "xmm7", "xmm6");
	&vmovdqu("xmm6", &DWP(32,"esp"));
	&vpaddq("xmm7", "xmm7", "xmm5");
	&vpmuludq("xmm5", "xmm6", &DWP(0,"esp"));
	&vpaddq("xmm7", "xmm7", "xmm5");
	&vpmuludq("xmm5", "xmm4", "xmm1");
	&vmovdqu(&DWP(160,"esp"), "xmm7");
	&vpmuludq("xmm7", "xmm3", "xmm6");
	&vmovdqu(&DWP(48,"esp"), "xmm0");
	&vpmuludq("xmm0", "xmm2", "xmm0");
	&vpaddq("xmm5", "xmm5", "xmm7");
	&vmovdqu("xmm7", &DWP(16,"esp"));
	&vpaddq("xmm5", "xmm5", "xmm0");
	&vpmuludq("xmm0", "xmm7", &DWP(64,"esp"));
	&vpaddq("xmm7", "xmm5", "xmm0");
	&vmovdqu("xmm0", &DWP(0,"esp"));
	&vpmuludq("xmm5", "xmm0", &DWP(80,"esp"));
	&vpaddq("xmm7", "xmm7", "xmm5");
	&vpmuludq("xmm5", "xmm4", &DWP(112,"esp"));
	&vmovdqu(&DWP(176,"esp"), "xmm7");
	&vpmuludq("xmm7", "xmm3", &DWP(128,"esp"));
	&vmovdqu(&DWP(144,"esp"), "xmm1");
	&vpmuludq("xmm1", "xmm2", "xmm1");
	&vpaddq("xmm5", "xmm5", "xmm7");
	&vpaddq("xmm7", "xmm5", "xmm1");
	&vmovdqu("xmm1", &DWP(16,"esp"));
	&vpmuludq("xmm5", "xmm1", "xmm6");
	&vpaddq("xmm5", "xmm7", "xmm5");
	&vmovdqu("xmm7", &DWP(48,"esp"));
	&vpmuludq("xmm0", "xmm0", "xmm7");
	&vpaddq("xmm5", "xmm5", "xmm0");
	&vmovdqu("xmm0", &DWP(160,"esp"));
	&vpsrlq("xmm0", "xmm0", 26);
	&vpaddq("xmm5", "xmm5", "xmm0");
	&vpmuludq("xmm0", "xmm3", "xmm7");
	&vpmuludq("xmm3", "xmm3", &DWP(144,"esp"));
	&vpmuludq("xmm7", "xmm1", "xmm7");
	&vmovdqu(&DWP(192,"esp"), "xmm5");
	&vpmuludq("xmm5", "xmm4", "xmm6");
	&vpmuludq("xmm4", "xmm4", &DWP(128,"esp"));
	&vpaddq("xmm5", "xmm5", "xmm0");
	&vpaddq("xmm3", "xmm4", "xmm3");
	&vpmuludq("xmm0", "xmm2", &DWP(64,"esp"));
	&vpmuludq("xmm2", "xmm2", "xmm6");
	&vpaddq("xmm5", "xmm5", "xmm0");
	&vpaddq("xmm3", "xmm3", "xmm2");
	&vpmuludq("xmm0", "xmm1", &DWP(80,"esp"));
	&vpaddq("xmm2", "xmm3", "xmm7");
	&vpaddq("xmm0", "xmm5", "xmm0");
	&vmovdqu("xmm5", &DWP(0,"esp"));
	&vpmuludq("xmm5", "xmm5", &DWP(96,"esp"));
	&vmovdqu("xmm4", &DWP(0,"esp"));
	&vpmuludq("xmm1", "xmm4", &DWP(64,"esp"));
	&vpaddq("xmm5", "xmm0", "xmm5");
	&vpaddq("xmm6", "xmm2", "xmm1");
	&vmovdqu("xmm0", &DWP(176,"esp"));
	&vpsrlq("xmm0", "xmm0", 26);
	&vpaddq("xmm5", "xmm5", "xmm0");
	&vmovdqu("xmm0", &DWP(192,"esp"));
	&vpsrlq("xmm4", "xmm5", 26);
	&vpsrlq("xmm7", "xmm0", 26);
	&vpaddq("xmm2", "xmm6", "xmm7");
	&vpmuludq("xmm6", "xmm4", &DWP(240,"esp"));
	&vmovdqu("xmm3", &DWP(256,"esp"));
	&vpsrlq("xmm4", "xmm2", 26);
	&vpand("xmm7", "xmm3", &DWP(176,"esp"));
	&vpand("xmm5", "xmm5", "xmm3");
	&vpand("xmm1", "xmm3", &DWP(160,"esp"));
	&vpand("xmm0", "xmm0", "xmm3");
	&vpaddq("xmm4", "xmm7", "xmm4");
	&vpaddq("xmm6", "xmm1", "xmm6");
	&vpsrlq("xmm7", "xmm4", 26);
	&vpaddq("xmm7", "xmm5", "xmm7");
	&vpand("xmm5", "xmm6", "xmm3");
	&vpshufd("xmm1", "xmm5", 8);
	&vpsrlq("xmm5", "xmm6", 26);
	&vpaddq("xmm0", "xmm0", "xmm5");
	&vpshufd("xmm6", "xmm0", 8);
	&vpunpckldq("xmm5", "xmm1", "xmm6");
	&vpand("xmm1", "xmm2", "xmm3");
	&vpand("xmm3", "xmm4", "xmm3");
	&vpshufd("xmm2", "xmm1", 8);
	&vpshufd("xmm4", "xmm3", 8);
	&vpunpckldq("xmm0", "xmm2", "xmm4");
	&vpunpcklqdq("xmm1", "xmm5", "xmm0");
	&vpunpckhqdq("xmm3", "xmm5", "xmm0");
	&vpshufd("xmm2", "xmm7", 8);
	&vpshufd("xmm7", "xmm7", 2);
	&vmovdqu(&DWP(120,"edx"), "xmm1");
	&vmovdqu(&DWP(140,"edx"), "xmm3");
	&vmovd(&DWP(136,"edx"), "xmm2");
	&vmovd(&DWP(156,"edx"), "xmm7");
	&set_label("poly1305_init_ex_avx2_6");
	&mov(&DWP(176,"edx"), 0);
	&vzeroupper();
	&add("esp", 272);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&ret();
}

sub poly1305_avx2 {
	#&cmp(&DWP(12,"esp"), 64);
	#&jbe(&label("poly1305_x86_local"));
	&push("ebp");
	&mov("ebp", "esp");
	&and("esp", -64);
	&push("esi");
	&push("edi");
	&push("ebx");
	&sub("esp", 308);
	&mov("esi", &DWP(16,"ebp"));
	&lea("eax", &DWP(64,"esp"));
	&mov("ecx", "esi");
	&mov("edx", &DWP(20,"ebp"));
	&mov("edi", &DWP(12,"ebp"));
	&call(&LABEL("poly1305_init_ex_avx2_local"));
	&mov("ebx", "esi");
	&and("ebx", -64);
	&je(&label("poly1305_avx2_5"));
	&mov("edx", "edi");
	&lea("eax", &DWP(64,"esp"));
	&mov("ecx", "ebx");
	&call(&LABEL("poly1305_blocks_avx2_local"));
	&add("edi", "ebx");
	&sub("esi", "ebx");
	&set_label("poly1305_avx2_5");
	&push(&DWP(8,"ebp"));
	&push("esi");
	&push("edi");
	&lea("eax", &DWP(76,"esp"));
	&push("eax");
	&call(&LABEL("poly1305_finish_ex_avx2_local"));
	&add("esp", 324);
	&pop("ebx");
	&pop("edi");
	&pop("esi");
	&mov("esp", "ebp");
	&pop("ebp");
	&ret();
}


add_function("poly1305_init_ex", "x86", \&poly1305_init_ex_x86);
add_function("poly1305_blocks", "x86", \&poly1305_blocks_x86);
add_function("poly1305_finish_ex", "x86", \&poly1305_finish_ex_x86);
add_function("poly1305", "x86", \&poly1305_x86);

add_function("poly1305_init_ex", "sse2", \&poly1305_init_ex_sse2);
add_function("poly1305_blocks", "sse2", \&poly1305_blocks_sse2);
add_function("poly1305_finish_ex", "sse2", \&poly1305_finish_ex_sse2);
add_function("poly1305", "sse2", \&poly1305_sse2);

add_function("poly1305_init_ex", "avx", \&poly1305_init_ex_avx);
add_function("poly1305_blocks", "avx", \&poly1305_blocks_avx);
add_function("poly1305_finish_ex", "avx", \&poly1305_finish_ex_avx);
add_function("poly1305", "avx", \&poly1305_avx);

add_function("poly1305_init_ex", "avx2", \&poly1305_init_ex_avx2);
add_function("poly1305_blocks", "avx2", \&poly1305_blocks_avx2);
add_function("poly1305_finish_ex", "avx2", \&poly1305_finish_ex_avx2);
add_function("poly1305", "avx2", \&poly1305_avx2);

} # $ext{'x86'}

&asm_finish();
